// ----------------------------------------------------------------------------
// ErrorStream.java
// A stream for printing errors.
// $Id: ErrorStream.java,v 1.23 2012/10/18 16:12:58 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import java.io.*;
import fmrisc.ProgramExplorer.SWT.*;
import fmrisc.ProgramExplorer.Syntax.*;

/**
 *
 * A stream for printing errors.
 *
 */
public class ErrorStream 
{
  private PrintWriter error; // the stream on which to write error messages
  private String prefix;     // the prefix for each error message
  private String prefix2;    // the prefix for each explanation
  private String prefix3;    // the prefix for each normal message
  private int errnum;        // the number of errors
  
  private TopWindow top;      // the top-level window (may be null)
  private SourcePosition pos; // the current source code position (may be null)
  
  /****************************************************************************
   * Construct an error stream.
   * @param prefix the string which shall prefix each error message.
   * @param prefix2 the string which shall prefix each explanation of an error.
   * @param prefix3 the string which shall prefix each normal message.
   * @param error the stream on which to write error messages.
   * @param top the top window for signaling errors (may be null).
   ***************************************************************************/
  public ErrorStream(String prefix, String prefix2, String prefix3,
    PrintWriter error, TopWindow top)
  {
    this.error = error;
    this.top = top;
    this.pos = null;
    this.prefix = prefix;
    this.prefix2 = prefix2;
    this.prefix3 = prefix3;
    this.errnum = 0;
  }
  
  /****************************************************************************
   * Return the number of errors generated.
   * @return the number of errros.
   ***************************************************************************/
  public int getErrorNumber()
  {
    return errnum;
  }
  
  /****************************************************************************
   * Set current source position.
   * @param pos the position.
   ***************************************************************************/
  public void setPosition(SourcePosition pos)
  {
    this.pos = pos;
  }
  
  /****************************************************************************
   * Get current source position.
   * @return the position.
   ***************************************************************************/
  public SourcePosition getPosition()
  {
    return pos;
  }
  
  /****************************************************************************
   * Clear errors for denoted source.
   * @param source the source to be cleared.
   ***************************************************************************/
  public void clearErrors(Source source)
  {
    if (top != null) top.clearAnnotations(source);
  }
  
  /****************************************************************************
   * Generate a new error.
   * @param pos the source position to which the error refers
   * @param message the message indicating the error.
   ***************************************************************************/
  public void newError(SourcePosition pos, String message)
  {
    // System.out.println(message);
    // new Exception().printStackTrace(System.out);
    message = message.trim();
    if (pos == null)
      error.println(prefix + ": " + message);
    else if (pos instanceof SourceAnnotationPosition) 
    {
      SourceAnnotationPosition pos0 = (SourceAnnotationPosition)pos;
      String text = "in annotation line\n" +
        pos0.getAnnotation().getLine(pos0.getAnnotationLine()) + "\n";
      for (int i=0; i<pos0.getAnnotationColumn(); i++) text = text + " ";
      text = text + "^\n";
      message = text + message;
      error.println(prefix + " (" + pos0 + "): " + message);
    }
    else
      error.println(prefix + " (" + pos + "): " + message);
    errnum = errnum + 1;
    if (top != null && pos != null) top.addAnnotation(pos, message);
  } 
  
  /****************************************************************************
   * Generate a new error for currently set source position.
   * @param message the message indicating the error.
   ***************************************************************************/
  public void newError(String message)
  {
    newError(pos, message);
  }
  
  /****************************************************************************
   * Generate a new error for source position denoted by abstract syntax tree.
   * @param ast the abstract syntax tree referring to a source position.
   * @param message the message indicating the error.
   ***************************************************************************/
  public void newError(ASTSource ast, String message)
  {
    newError(ast.getPosition(), message);
  }
  
  /****************************************************************************
   * Add an explanation to the error.
   * @param message the message explaining the error.
   ***************************************************************************/
  public void explainError(String message)
  {
    boolean off = true; if (off) return;
    error.println(prefix2 + message);
    if (top != null && pos != null) top.markSource(pos.getSource(), true);
  }   
  
 
  /****************************************************************************
   * Add an explanation to the error.
   * @param message the message explaining the error.
   ***************************************************************************/
  public void explainError2(String message)
  {
    error.println(prefix2 + message);
    if (top != null && pos != null) top.markSource(pos.getSource(), true);
  }  
  
  /****************************************************************************
   * Add a message not counting as an error.
   * @param message the message.
   ***************************************************************************/
  public void newMessage(String message)
  {
    error.println(prefix3 + message);
  }  
  
  /****************************************************************************
   * Start a message not counting as an error.
   * @param message the message.
   ***************************************************************************/
  public void startMessage(String message)
  {
    error.print(prefix3 + message);
    error.flush();
  } 
  
  /****************************************************************************
   * End a message not counting as an error.
   * @param message the message.
   ***************************************************************************/
  public void endMessage(String message)
  {
    error.println(message);
  } 
  
  /****************************************************************************
   * Reset number of errors to zero.
   ***************************************************************************/
  public void reset()
  {
    errnum = 0;
  }
}
