// ----------------------------------------------------------------------------
// Presenter.java
// Interface to visual presentation of mathematical information.
// $Id: Presenter.java,v 1.27 2019/11/28 08:29:17 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import java.io.*;
import java.util.*;
import org.w3c.dom.*;

import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Judgements.*;

/**
 * 
 * Interface to visual presentation of declarations and proof states.
 *
 */
public final class Presenter
{
  // common extension
  public static final String fileExt = ".xhtml";
  
  // frame identifiers
  public static final String methodId = "method";
  public static final String detailId = "detail";
  
  // names of method file, detail files, prefixes of statement file
  // extension "-2" indicates file with the simplified version of the formula
  public static final String methodName = methodId + fileExt;
  public static final String detailName = detailId + fileExt;
  public static final String detailName2 = detailId + "-2" + fileExt;
  public static final String statPrefix = "stat";
  public static final String statPrefix2 = statPrefix + "-2";
  
  // names of form action and buttons
  public static final String nameFormStatAction = "formstat";
  public static final String nameMethod = "method";
  public static final String nameStatement = "statement";
  public static final String nameFormula = "formula";
  public static final String namePrePost = "prepost";
  
  // current method (may be null)
  private ParamSymbol method;
  
  // directory where presentation files are stored
  private VirtualDirectory directory;

  // size of fonts to be used in documents
  private int fontSize;

  // number of characters to be used per line
  private int width;
  
  // number of statements written so far
  private int statCounter;
  
  // port number of command server (-1, if unknown)
  private int commandPort;
  
  // maps of statements to pre-state conditions (may be null).
  private Map<Statement,Formula> preMap;
  private Map<Statement,Formula> postMap;
  
  // number of presenters created so far
  private static int presenterCounter = 0;
  
  /***************************************************************************
   * Create instance of presenter with files in denoted directory.
   * @param method the current method (may be null)
   * @param directory for storing the presentation files.
   * @param fontSize the base size of the fonts to be used.
   * @param width the number of characters per line to be used.
   * @param port port number of command server (-1, if unknown).
   * @param preMap a map of statements to pre-state conditions (may be null).
   * @param postMap a map of statements to post-state conditions (may be null).
   **************************************************************************/
  private Presenter(ParamSymbol method, VirtualDirectory directory, 
    int fontSize, int width, int port,
    Map<Statement,Formula> preMap, Map<Statement,Formula> postMap)
  {
    this.method = method;
    this.directory = directory;
    this.fontSize = fontSize;
    this.width = width;
    this.statCounter = 0;
    this.commandPort = port;
    this.preMap = preMap;
    this.postMap = postMap;
    presenterCounter = presenterCounter+1;
  }
  
  /***************************************************************************
   * Present method in a certain font size.
   * @param symbol the symbol of the method.
   * @param fontSize the font size.
   * @param width the number of characters per line to be used
   *        for the detail view.
   * @param port the port number of the command server.
   * @return a directory whose main file contains the output
   *         (the file may contain links to other files in the directory).
   **************************************************************************/
  public static VirtualDirectory present(ParamSymbol symbol, 
    int fontSize, int width, int port)
  {
    VirtualDirectory directory = new VirtualDirectory();
    Presenter presenter = new Presenter(symbol, 
        directory, fontSize, width, port, null, null);
    presenter.print(symbol);
    return directory;
  }
  
  /***************************************************************************
   * Present formula in a certain font size.
   * @param formula the formula.
   * @param fontSize the font size.
   * @param width the number of characters per line to be used
   * @return a directory whose main file contains the output
   *         (the file may contain links to other files in the directory).
   **************************************************************************/
  public static VirtualDirectory present(Formula formula, 
    int fontSize, int width)
  {
    VirtualDirectory directory = new VirtualDirectory();
    PrintWriter out = directory.create(VirtualDirectory.MAIN);
    Presenter presenter = new Presenter(null, 
        directory, fontSize, width, -1, null, null);
    presenter.print(formula, out);
    out.close();
    return directory;
  }

  /***************************************************************************
   * Present method in a certain font size.
   * @param symbol the symbol of the method.
   * @param statement the statement.
   * @param formula the formula.
   * @param pre true iff formula is precondition.
   * @param preMap the precondition map.
   * @param postMap the postcondition map.
   * @param fontSize the font size.
   * @param width the number of characters per line to be used
   *        for the detail view.
   * @param port the port number of the command server.
   * @return a directory whose main file contains the output
   *         (the file may contain links to other files in the directory).
   **************************************************************************/
  public static VirtualDirectory present(ParamSymbol symbol, 
    Statement statement, Formula formula, boolean pre,
    Map<Statement,Formula> preMap, Map<Statement,Formula> postMap,
    int fontSize, int width, int port)
  {
    VirtualDirectory directory = new VirtualDirectory();
    Presenter presenter = new Presenter(symbol, 
        directory, fontSize, width, port, preMap, postMap);
    presenter.print(symbol, statement, formula, pre);
    return directory;
  }
  
  // --------------------------------------------------------------------------
  //
  // Construction of presentation documents
  //
  // --------------------------------------------------------------------------
  
  /***************************************************************************
   * Pretty-print method.
   * @param symbol the symbol of the method.
   **************************************************************************/
  private void print(ParamSymbol symbol)
  {
    // create index file
    PrintWriter indexOut = directory.create(VirtualDirectory.MAIN);
    Document indexDoc = createFrameset("Method " + getLongName(symbol));
    Element indexNode = getRoot(indexDoc);
    indexNode.setAttribute("cols", "40%, 60%");
    Element methodFrame = indexDoc.createElement("frame");
    indexNode.appendChild(methodFrame);
    methodFrame.setAttribute("name", methodId);
    methodFrame.setAttribute("id", methodId);
    methodFrame.setAttribute("src", methodName);
    Element detailFrame = indexDoc.createElement("frame");
    indexNode.appendChild(detailFrame);
    detailFrame.setAttribute("name", detailId);
    detailFrame.setAttribute("id", detailId);
    detailFrame.setAttribute("src", detailName2);
    writeDocument(indexDoc, indexOut);
    indexOut.close();
    
    // write other documents
    writeMethodFile(symbol);
    writeDetailFileBodySpec(symbol, false, detailName, detailName2);
    writeDetailFileBodySpec(symbol, true, detailName2, detailName);
  }
  
  // detail frame created in last call of print() below
  private Element detailFrame;
  
  /***************************************************************************
   * Pretty-print method where condition for a body statement has been fixed.
   * @param symbol the symbol of the method.
   * @param statement the body statement.
   * @param formula the condition attached to the body statement.
   * @param pre true if condition was attached to pre-state.
   **************************************************************************/
  private void print(ParamSymbol symbol,
    Statement statement, Formula formula, boolean pre)
  {
    // create index file, set detail frame
    PrintWriter indexOut = directory.create(VirtualDirectory.MAIN);
    Document indexDoc = createFrameset("Method " + getLongName(symbol));
    Element indexNode = getRoot(indexDoc);
    indexNode.setAttribute("cols", "40%, 60%");
    Element methodFrame = indexDoc.createElement("frame");
    indexNode.appendChild(methodFrame);
    methodFrame.setAttribute("name", methodId);
    methodFrame.setAttribute("id", methodId);
    methodFrame.setAttribute("src", methodName);
    detailFrame = indexDoc.createElement("frame");
    indexNode.appendChild(detailFrame);
    detailFrame.setAttribute("name", detailId);
    detailFrame.setAttribute("id", detailId);
    detailFrame.setAttribute("src", detailName2);
    
    // write other documents
    writeMethodFile(symbol, statement, formula, pre);
    writeDetailFileBodySpec(symbol, false, detailName, detailName2);
    writeDetailFileBodySpec(symbol, true, detailName2, detailName);
    
    // write index file with potentially modified detail frame
    writeDocument(indexDoc, indexOut);
    indexOut.close();
  }
  
  /***************************************************************************
   * Set detail frame of current frame to denoted file.
   * @param file the denoted file.
   **************************************************************************/
  public void setDetailFrame(String file)
  {
    detailFrame.setAttribute("src", file);
  }
  
  /***************************************************************************
   * Get long name of method.
   * @param symbol the method symbol.
   * @return the long name of the method.
   **************************************************************************/
  private static String getLongName(ParamSymbol symbol)
  {
    return
      symbol.getClassSymbol().getName().getString() + "." +
      symbol.getName().getString();
  }
  
  /***************************************************************************
   * Write method file for method.
   * @param symbol the symbol denoting the method.
   **************************************************************************/
  private void writeMethodFile(ParamSymbol symbol)
  {
    // prepare document
    Document document = createDocument("Method", true);
    Element top = createTopNode(document, fontSize);
    
    // fill document
    Element header = document.createElement("h1");
    top.appendChild(header);
    Node title = document.createTextNode(getLongName(symbol));
    header.appendChild(title);
        
    // write source
    ParamDeclaration decl = (ParamDeclaration)(symbol.getDeclaration());
    if (decl != null) 
    {
      // the method specification
      writeMethodSpecification(top, (MethodSpec)decl.getAnnotation());
      
      // create form
      Element form = document.createElement("form");
      top.appendChild(form);
      form.setAttribute("action", 
          "http://localhost:" + commandPort + "/" + nameFormStatAction);
      form.setAttribute("method", "post");
      form.setAttribute("target", "output");
            
      // the method declaration
      Element div = document.createElement("div");
      form.appendChild(div);
      div.setAttribute("class", "statblock");
      Element table = document.createElement("table");
      div.appendChild(table);
      table.setAttribute("class", "statblock");
      new ASTPrinterHTML(table, this, true, null).append(decl);
      
      // the input area
      form.appendChild(document.createTextNode("Select a statement and define a condition for its pre/poststate:"));
      form.appendChild(document.createElement("br"));
      Element textarea = document.createElement("textarea");
      form.appendChild(textarea);
      textarea.setAttribute("name", nameFormula);
      textarea.setAttribute("cols", "45");
      textarea.setAttribute("rows", "2");
      textarea.setAttribute("style", "font-size: " + fontSize + "pt");
      form.appendChild(document.createElement("p"));
      
      // the buttons
      Element submit = document.createElement("input");
      form.appendChild(submit);
      submit.setAttribute("type", "submit");
      submit.setAttribute("value", "Submit");
      Element reset = document.createElement("input");
      form.appendChild(reset);
      reset.setAttribute("type", "reset");
      reset.setAttribute("value", "Reset");
      
      // the radio buttons
      Element pre = document.createElement("input");
      form.appendChild(pre);
      form.appendChild(document.createTextNode("Prestate"));
      pre.setAttribute("type", "radio");
      pre.setAttribute("name", namePrePost);
      pre.setAttribute("value", "pre");
      pre.setAttribute("checked", "");
      Element post = document.createElement("input");
      form.appendChild(post);
      form.appendChild(document.createTextNode("Poststate"));
      post.setAttribute("type", "radio");
      post.setAttribute("name", namePrePost);
      post.setAttribute("value", "post");
      
      // a hidden button holding the name + offset of the method
      Element method = document.createElement("input");
      form.appendChild(method);
      method.setAttribute("type", "hidden");
      method.setAttribute("name", nameMethod);
      method.setAttribute("value", 
          symbol.getLongName().toString() + "." +
          symbol.getDeclaration().getPosition().getOffset());
      
      // create frame for form output
      // top.appendChild(document.createElement("p"));
      Element output = document.createElement("iframe");
      top.appendChild(output);
      output.setAttribute("id", "output");
      output.setAttribute("name", "output");
      output.setAttribute("frameborder", "0");
      output.setAttribute("width", "100%");
      output.setAttribute("height", "150");
    }
    
    // write document to file
    PrintWriter out = directory.create(methodName);
    writeDocument(document, out);
    out.close();
  }
  
  /***************************************************************************
   * Write method file for method.
   * @param symbol the symbol denoting the method.
   **************************************************************************/
  private void writeMethodFile(ParamSymbol symbol,
    Statement statement, Formula formula, boolean prestate)
  {
    // prepare document
    Document document = createDocument("Method", true);
    Element top = createTopNode(document, fontSize);
    
    // fill document
    Element header = document.createElement("h1");
    top.appendChild(header);
    Node title = document.createTextNode(getLongName(symbol));
    header.appendChild(title);
        
    // write source
    ParamDeclaration decl = (ParamDeclaration)(symbol.getDeclaration());
    if (decl != null) 
    {
      // the method specification
      writeMethodSpecification(top, (MethodSpec)decl.getAnnotation());
            
      // the method declaration
      Element div = document.createElement("div");
      top.appendChild(div);
      div.setAttribute("class", "statblock");
      Element table = document.createElement("table");
      div.appendChild(table);
      table.setAttribute("class", "statblock");
      new ASTPrinterHTML(table, this, false, statement).append(decl);
      
      // the input area
      top.appendChild(document.createTextNode("The " +
          (prestate ? "prestate" : "poststate") +
          " condition assumed for the selected statement:"));
      Element p = document.createElement("p");
      top.appendChild(p);
      new PrettyMathML(document, width).appendFormula(p, formula);
    }
    
    // write document to file
    PrintWriter out = directory.create(methodName);
    writeDocument(document, out);
    out.close();
  }
  
  /***************************************************************************
   * Write method specification to given top element.
   * @param top the top element.
   * @param spec the specification (may be null).
   **************************************************************************/
  private void writeMethodSpecification(Element top, MethodSpec spec)
  {
    if (spec == null) return;
    Document document = top.getOwnerDocument();
    Element div = document.createElement("div");
    top.appendChild(div);
    boolean written = false;
    if (spec.getPreCondition() != null)
    {
      writeClause(div, spec.getPreCondition(), MethodSpec.preKeyword);
      written = true;
    }
    if (spec.getVariables() != null)
    {
      if (written) addBreak(div);
      writeClause(div, spec.getVariables(), MethodSpec.variablesKeyword);
      written = true;
    }
    if (spec.getExceptions() != null)
    {
      if (written) addBreak(div);
      writeClause(div, spec.getExceptions(), MethodSpec.exceptionsKeyword);
      written = true;
    }
    if (spec.getPostCondition() != null)
    {
      if (written) addBreak(div);
      writeClause(div, spec.getPostCondition(), MethodSpec.postKeyword);
      written = true;
    }
    if (spec.getDivergesCondition() != null)
    {
      if (written) addBreak(div);
      writeClause(div, spec.getDivergesCondition(), MethodSpec.divergesKeyword);
      written = true;
    }
    if (spec.getTerminationTerm() != null)
    {
      if (written) addBreak(div);
      writeClause(div, spec.getTerminationTerm(), MethodSpec.terminationKeyword);
      written = true;
    }
  }
    
  /***************************************************************************
   * Append <br> to parent.
   * @param parent the parent node.
   **************************************************************************/
  private void addBreak(Element parent)
  {
    Document document = parent.getOwnerDocument();
    Element br = document.createElement("br");
    parent.appendChild(br);
  }
  
  /***************************************************************************
   * Write specification clause.
   * @param parent the parent of the clause.
   * @param formula the formula to be written (may be null).
   * @param keyword the keyword to be used.
   **************************************************************************/
  private void writeClause(Element parent, Formula formula, String keyword)
  {
    if (formula == null) return;
    appendKeyword(parent, keyword);
    new PrettyMathML(parent.getOwnerDocument(), width).appendFormula(parent, formula);
  }
  
  /***************************************************************************
   * Write specification clause.
   * @param parent the parent of the clause.
   * @param term the term to be written (may be null).
   * @param keyword the keyword to be used.
   **************************************************************************/
  private void writeClause(Element parent, Term[] term, String keyword)
  {
    if (term == null) return;
    appendKeyword(parent, keyword);
    Document document = parent.getOwnerDocument();
    PrettyMathML printer = new PrettyMathML(document, width);
    int n = term.length;
    for (int i=0; i<n; i++)
    {
      printer.appendTerm(parent, term[i]);
      if (i+1<n) parent.appendChild(document.createTextNode(", "));
    }
    return;
  }
  
  /***************************************************************************
   * Write specification clause.
   * @param parent the parent of the clause.
   * @param names the names to be written (may be null).
   * @param keyword the keyword to be used.
   **************************************************************************/
  private void writeClause(Element parent, Name[] names, String keyword)
  {
    if (names == null) return;
    appendKeyword(parent, keyword);
    appendNames(parent, names);
  }
  
  /***************************************************************************
   * Write names.
   * @param parent the parent of the clause.
   * @param names the names to be written.
   **************************************************************************/
  private void appendNames(Element parent, Name[] names)
  {
    int n = names.length;
    Document document = parent.getOwnerDocument();
    PrettyMathML printer = new PrettyMathML(document, width);
    for (int i=0; i<n; i++)
    {
      printer.appendName(parent, names[i]);
      if (i<n-1) parent.appendChild(document.createTextNode(", "));
    }
  }
  
  /***************************************************************************
   * Write keyword.
   * @param parent the parent of the keyword.
   * @param keyword the keyword to be written.
   **************************************************************************/
  private void appendKeyword(Element parent, String keyword)
  {
    Document document = parent.getOwnerDocument();
    Element span = document.createElement("span");
    parent.appendChild(span);
    span.setAttribute("class", "keyword");
    Node text = document.createTextNode(keyword);
    span.appendChild(text);
  }
  
  /***************************************************************************
   * Write detail file for method with body specification.
   * @param symbol the symbol denoting the method.
   * @param simplify true if the specification formula is to be simplified.
   * @param file name of the file to write to.
   * @param link name of the file to link to.
   **************************************************************************/
  private void writeDetailFileBodySpec(ParamSymbol symbol, 
    boolean simplify, String file, String link)
  {
    // prepare document
    Document document = createDocument("Detail", true);
    Element top = createTopNode(document, fontSize);
    
    // fill document
    Element header = document.createElement("h1");
    top.appendChild(header);
    Node title = document.createTextNode("Body Knowledge");
    header.appendChild(title);
    
    // link to formulas before simplification
    Element node = document.createElement("p");
    top.appendChild(node);
    if (simplify)
      writeLink(node, link, "Show Original Formulas");
    else
      writeLink(node, link, "Show Simplified Formulas");
    
    // precondition header
    Element preHeader = document.createElement("h2");
    top.appendChild(preHeader);
    Node preTitle = document.createTextNode("Pre-State Knowledge");
    preHeader.appendChild(preTitle);
    
    // write precondition
    Element p = document.createElement("p");
    top.appendChild(p);
    Formula formula = Normalization.preCondition(symbol, true);
    if (formula == null)
    {
      p.appendChild(document.createTextNode("(none)"));
    }
    else
    {
      if (simplify) formula = Simplification.perform(formula);
      new PrettyMathML(document, width).appendFormula(p, formula);
    }
    
    // write judgement
    writeJudgement(top, symbol.getBodyJudgement(), null, simplify, link);
    
    // write document to file
    PrintWriter out = directory.create(file);
    writeDocument(document, out);
    out.close();
  }
  
  /***************************************************************************
   * Write judgement.
   * @param parent the parent to which to append the output.
   * @param judgement the judgement (may be null)
   * @param pre alternative precondition (if null, use judgement precondition)
   * @param simplify true if the judgement formula is to be simplified.
   * @param link name of the file to link to
   **************************************************************************/
  private void writeJudgement(Element parent, StatJudgement judgement,
    Formula pre, boolean simplify, String link)
  {
    if (judgement == null) return;
    Document document = parent.getOwnerDocument();
    PrettyMathML printer = new PrettyMathML(document, width);
    
    // precondition
    if (pre == null) pre = judgement.getPrecondition();
    if (pre != null)
    {
      Element preHeader = document.createElement("h2");
      parent.appendChild(preHeader);
      Node preTitle = document.createTextNode("Precondition");
      preHeader.appendChild(preTitle);
      if (simplify) pre = Simplification.perform(pre);
      printer.appendFormula(parent, pre);
    }
    
    // statement frame
    Element frameHeader = document.createElement("h2");
    parent.appendChild(frameHeader);
    Node frameTitle = document.createTextNode("Effects");
    frameHeader.appendChild(frameTitle);
    writeBoolean(parent, "executes", judgement.executes);
    parent.appendChild(document.createTextNode(", "));
    writeBoolean(parent, "continues", judgement.continues);
    parent.appendChild(document.createTextNode(", "));
    writeBoolean(parent, "breaks", judgement.breaks);
    parent.appendChild(document.createTextNode(", "));
    writeBoolean(parent, "returns", judgement.returns);
    parent.appendChild(document.createElement("br"));
    writeVariableSymbols(parent, "variables", judgement.variables);
    parent.appendChild(document.createTextNode("; "));
    writeClassSymbols(parent, "exceptions", judgement.exceptions);
    parent.appendChild(document.createElement("br"));
    
    // transition relation
    Element transHeader = document.createElement("h2");
    parent.appendChild(transHeader);
    Node transTitle = document.createTextNode("Transition Relation");
    transHeader.appendChild(transTitle);
    Formula formula;
    if (simplify)
      formula = judgement.sformula;
    else
      formula = judgement.formula;
    printer.appendFormula(parent, formula);
    
    // termination condition
    Formula termination = judgement.getTermination();
    if (termination != null)
    {
      if (simplify) termination = Simplification.perform(termination);
      if (!(termination instanceof True))
      {
        parent.appendChild(document.createElement("br"));
        Element termHeader = document.createElement("h2");
        parent.appendChild(termHeader);
        Node termTitle = document.createTextNode("Termination Condition");
        termHeader.appendChild(termTitle);
        if (simplify) termination = Simplification.perform(termination);
        printer.appendFormula(parent, termination);
      }
    }
  }
  
  /***************************************************************************
   * Write boolean flag.
   * @param parent the parent to which to append the output.
   * @param label the label to be written.
   * @param flag the flag to be written.
   **************************************************************************/
  private void writeBoolean(Element parent, String label, boolean flag)
  {
    appendKeyword(parent, label + ": ");
    Document document = parent.getOwnerDocument();
    String ftext = flag ? "true" : "false";
    parent.appendChild(document.createTextNode(ftext));
  }
  
  /***************************************************************************
   * Write link.
   * @param parent the parent to which to append the output.
   * @param link the file to link to.
   * @param text the link text
   **************************************************************************/
  private void writeLink(Element parent, String link, String text)
  {
    Document document = parent.getOwnerDocument();
    parent.appendChild(document.createTextNode("["));
    Element a = document.createElement("a");
    parent.appendChild(a);
    a.setAttribute("href", link);
    a.appendChild(document.createTextNode(text));
    parent.appendChild(document.createTextNode("]"));
  }
  
  /***************************************************************************
   * Write variable symbols.
   * @param parent the parent to which to append the output.
   * @param label the label to be written.
   * @param symbols the symbols to be written
   **************************************************************************/
  private void writeVariableSymbols(Element parent, 
    String label, final Collection<VariableSymbol> symbols)
  {
    appendKeyword(parent, label + ": ");
    int n = symbols.size();
    if (n == 0)
    {
      Document document = parent.getOwnerDocument();
      parent.appendChild(document.createTextNode("-"));
      return;
    }
    Name[] names = new Name[n];
    int i=0;
    for (VariableSymbol symbol : symbols)
    {
      names[i] = NameUtils.uniqueName(symbol);
      i = i+1;
    }
    appendNames(parent, names);
  }
  
  /***************************************************************************
   * Write symbols.
   * @param parent the parent to which to append the output.
   * @param label the label to be written.
   * @param symbols the symbols to be written
   **************************************************************************/
  private void writeClassSymbols(Element parent, 
    String label, final Collection<ClassSymbol> symbols)
  {
    appendKeyword(parent, label + ":");
    int n = symbols.size();
    if (n == 0)
    {
      Document document = parent.getOwnerDocument();
      parent.appendChild(document.createTextNode("-"));
      return;
    }
    Name[] names = new Name[n];
    int i=0;
    for (ClassSymbol symbol : symbols)
    {
      names[i] = symbol.getLongName();
      i = i+1;
    }
    appendNames(parent, names);
  }
  
  /***************************************************************************
   * Write file for statement and statement formula.
   * @param stat the statement to be written
   * @return the name of the file
   **************************************************************************/
  public String writeStatement(Statement stat)
  {
    // determine file names
    String ext = presenterCounter + "." + statCounter + fileExt;
    statCounter = statCounter+1;
    String fileName = statPrefix + ext;
    String fileName2 = statPrefix2 + ext;
    
    if (preMap == null)
    {
      // write cross-linked files with statement specification
      writeStatement(stat, false, fileName, fileName2);
      writeStatement(stat, true, fileName2, fileName);
    }
    else
    {
      // write cross-linked files with pre/post-state conditions
      writeStatementCond(stat, false, fileName, fileName2);
      writeStatementCond(stat, true, fileName2, fileName);
    }
    
    // return name of one of the files
    return fileName2;
  }
    
  /***************************************************************************
   * Write file for statement and statement formula.
   * @param stat the statement to be written
   * @param simplify true if the formula is to be simplified.
   * @param file the file to write to.
   * @param link the file to link to.
   **************************************************************************/
  private void writeStatement(Statement stat, boolean simplify,
    String file, String link)
  {
    // prepare document
    Document document = createDocument("Statement", true);
    Element top = createTopNode(document, fontSize);
    
    // document header
    Element header = document.createElement("h1");
    top.appendChild(header);
    Node title = document.createTextNode("Statement Knowledge");
    header.appendChild(title);
    
    // link to formulas before simplification
    Element node = document.createElement("p");
    top.appendChild(node);
    if (simplify)
      writeLink(node, link, "Show Original Formulas");
    else
      writeLink(node, link, "Show Simplified Formulas");
    
    // the pre-state knowledge
    Element preHeader = document.createElement("h2");
    top.appendChild(preHeader);
    Node preTitle = document.createTextNode("Pre-State Knowledge");
    preHeader.appendChild(preTitle);
    
    // the pre-state formula
    Element preNode = document.createElement("p");
    top.appendChild(preNode);
    Map<Statement, Formula> preMap = method.getPreMap();
    Formula formula = preMap == null ? null : preMap.get(stat);
    if (formula == null)
    {
      top.appendChild(document.createTextNode("(none)"));
    }
    else
    {
      if (simplify) formula = Simplification.perform(formula);
      new PrettyMathML(document, width).appendFormula(preNode, formula);
    }
    
    // fill document with simplified specification formula
    Formula precondition = null;
    if (stat instanceof ForLoopStatement)
      precondition = ((ForLoopStatement)stat).getPrecondition();
    writeJudgement(top, stat.getJudgement(), precondition, simplify, link);
    
    // write statement itself
    Element pre = document.createElement("pre");
    top.appendChild(pre);
    pre.setAttribute("class", "statblock");
    pre.appendChild(document.createTextNode(stat.toString()));
    
    // write document to file
    PrintWriter out = directory.create(file);
    writeDocument(document, out);
    out.close();
  }
  
  /***************************************************************************
   * Write file for statement and pre/post-state condition.
   * @param stat the statement to be written
   * @param simplify true if the formula is to be simplified.
   * @param file the file to write to.
   * @param link the file to link to.
   **************************************************************************/
  private void writeStatementCond(Statement stat, boolean simplify,
    String file, String link)
  {
    // prepare document
    Document document = createDocument("Statement", true);
    Element top = createTopNode(document, fontSize);
    
    // document header
    Element header = document.createElement("h1");
    top.appendChild(header);
    Node title = document.createTextNode("State Conditions");
    header.appendChild(title);
    
    // link to formulas before simplification
    Element node = document.createElement("p");
    top.appendChild(node);
    if (simplify)
      writeLink(node, link, "Show Original Formulas");
    else
      writeLink(node, link, "Show Simplified Formulas");
    
    // the pre-state knowledge
    Element preHeader = document.createElement("h2");
    top.appendChild(preHeader);
    Node preTitle = document.createTextNode("Pre-State Condition");
    preHeader.appendChild(preTitle);
    Element preNode = document.createElement("p");
    top.appendChild(preNode);
    Formula preFormula = preMap.get(stat);
    if (preFormula == null)
    {
      top.appendChild(document.createTextNode("(none)"));
    }
    else
    {
      // show variable references in post-state form
      preFormula = (Formula)PostStateCloner.clone(preFormula);
      if (simplify) preFormula = Simplification.perform(preFormula);
      new PrettyMathML(document, width).appendFormula(preNode, preFormula);
    }
    
    // the post-state knowledge
    Element postHeader = document.createElement("h2");
    top.appendChild(postHeader);
    Node postTitle = document.createTextNode("Post-State Condition");
    postHeader.appendChild(postTitle);
    Element postNode = document.createElement("p");
    top.appendChild(postNode);
    Formula postFormula = postMap.get(stat);
    if (postFormula == null)
    {
      top.appendChild(document.createTextNode("(none)"));
    }
    else
    {
      // show variable references in post-state form
      postFormula = (Formula)PostStateCloner.clone(postFormula);
      if (simplify) postFormula = Simplification.perform(postFormula);
      new PrettyMathML(document, width).appendFormula(postNode, postFormula);
    }
    
    // write statement itself
    Element pre = document.createElement("pre");
    top.appendChild(pre);
    pre.setAttribute("class", "statblock");
    pre.appendChild(document.createTextNode(stat.toString()));
    
    // write document to file
    PrintWriter out = directory.create(file);
    writeDocument(document, out);
    out.close();
  }
  
  /***************************************************************************
   * Pretty-print formula to denoted output stream.
   * @param formula the formula.
   * @param out the output stream.
   **************************************************************************/
  private void print(Formula formula, PrintWriter out)
  {
    Document document = createDocument("Formula", true);
    Element top = createTopNode(document, fontSize);
    PrettyMathML printer = new PrettyMathML(document, width);
    printer.appendFormula(top, formula);
    // scrollToBottom(document);
    writeDocument(document, out);
  }
  
  // --------------------------------------------------------------------------
  //
  // XHTML/MathML handling
  //
  // --------------------------------------------------------------------------
  
  /***************************************************************************
   * Create new XHTML/MathML document.
   * @param title the document title
   * @param fontSize the font size
   * @return the new document
   **************************************************************************/
  public static Document createDocument(String title, int fontSize)
  {
    return createDocument(title, 0, false, fontSize);
  }
  
  /***************************************************************************
   * Create new XHTML/MathML document.
   * @param title the document title
   * @param bottom true if extra space is added at bottom
   * @return the new document
   **************************************************************************/
  private Document createDocument(String title, boolean bottom)
  {
    return createDocument(title, 0, bottom, fontSize);
  }

  /***************************************************************************
   * Create new XHTML frameset document.
   * @param title the document title
   * @return the new document
   **************************************************************************/
  private Document createFrameset(String title)
  {
    return createDocument(title, 1, false, fontSize);
  }
  
  /***************************************************************************
   * Create new XHTMl document.
   * @param title the document title
   * @param type (0 = MathML, 1 = Frameset)
   * @param bottom if true some extra space is added at bottom for type 0
   * @param fontSize the size of the font
   * @return the new document
   **************************************************************************/
  private static Document createDocument(String title, int type, boolean bottom,
    int fontSize)
  {
    try
    {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(true);
      DocumentBuilder builder = factory.newDocumentBuilder();
      DOMImplementation impl = builder.getDOMImplementation();
      DocumentType doctype = null;
      if (type == 0) 
        doctype = impl.createDocumentType("html", 
            "-//W3C//DTD XHTML 1.1 plus MathML 2.0//EN",
        "http://www.w3.org/TR/MathML2/dtd/xhtml-math11-f.dtd"); 
      else if (type == 1)
        doctype = impl.createDocumentType("html", 
            "-//W3C//DTD XHTML 1.0 Frameset//EN",
        "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd");
      Document document = 
        impl.createDocument("http://www.w3.org/1999/xhtml", "html", doctype);
      Element root = document.getDocumentElement();
      Element head = document.createElement("head");
      root.appendChild(head);
      Element tnode = document.createElement("title");
      head.appendChild(tnode);
      tnode.appendChild(document.createTextNode(title));
      Element style = document.createElement("style");
      style.setAttribute("type", "text/css");
      head.appendChild(style);
      style.appendChild(document.createTextNode(
          "h1 { font-size: 140%; margin-top: 0px; } " +
          "h2 { font-size: 120%; } " +
          "span.keyword { font-weight: bold; color: blue; } " +
          "dd.option { margin-left: 12px; } " +
          "math { font-size: 110%; } " +
          "a { text-decoration: none; color: blue; } " +
          "a:visited { color: blue; } " +
          "pre.statblock { padding-left: 6px; padding-top: 6px; padding-bottom: 6px; " +
            "margin-top: 14px; margin-bottom: 14px; " +
            "background-color: #F5F5F5; " +
            "border: 1px solid; border-color: black; }" +
          "div.statblock { padding-top: 6px; padding-bottom: 6px; " +
              "margin-top: 14px; margin-bottom: 14px; " +
              "background-color: #F5F5F5; " +
              "border: 1px solid; border-color: black; }" +
          "table.statblock { border-spacing: 0px; padding-left: 1px; }" +
          "tr.statblock { }" +
          "div.statlabel { position: relative; } " +
          "td.statlabel { border: 1px outset; padding-right: 15px; } " +
          "td.statlabel:hover { background: blue; color: white; } " +
          "td.statcode { padding: 0px; padding-left: 6px; font-family: mono; } " +
          "div.lblock { position: relative; } " +
          "td.label { border: 1px outset; padding: 1px; } " +
          "td.label:hover { background: blue; color: white; } " +
          "table.menu { display: none; } " +
          "td:hover table.menu { display: block; position: absolute; " +
          "z-index: 1000; width: 300px; top: 0px; left: 12px; " +
          "border: 2px outset; background: #eeeeee; color: black; " +
          "border-collapse: collapse; } " +
          ".menu caption { width: 100%; color: red;" + 
            " border-bottom: 1px solid; border-color: black; font-weight: bold; } " +
          ".menu td { width: 1%; } " +
          "td.entry:hover { background: blue; color: white; } " +
          ".entry a { text-decoration: none; color: black; } " +
          ".entry a:hover { color: white; }"
      ));
      if (type == 0)
      {
        Element body = document.createElement("body");
        root.appendChild(body);
        String b = "0px";
        if (bottom) b = "60px";
        body.setAttribute("style",
            "margin:0 0 " + b + " 0; font-family: Helvetica, sans; font-size: " + 
            fontSize + "pt;");
      }
      else if (type == 1)
      {
        root.appendChild(document.createElement("frameset"));
      }
      return document;
    }
    catch(ParserConfigurationException e)
    {
      Main.internalError("error in creating document: exception " + e.getMessage());
      return null;
    } 
  }

  /***************************************************************************
   * Modify document such that on loading it is automatically scrolled
   * to bottom. Must be called after all content has been added to document.
   * @param document a document.
   **************************************************************************/
  protected void scrollToBottom(Document document)
  {
    Element root = getRoot(document);
    root.setAttribute("onload", 
    "document.getElementById('bottom').scrollIntoView(false);");
    Element bottom = document.createElement("a");
    bottom.setAttribute("id", "bottom");
    root.appendChild(bottom);
  }

  /***************************************************************************
   * Get root node of document to which new nodes should be appended.
   * @param document the document.
   * @return its root node.
   **************************************************************************/
  private static Element getRoot(Document document)
  {
    return (Element)document.getDocumentElement().getChildNodes().item(1);
  }
  
  /***************************************************************************
   * Create top node for document content.
   * @param document the document.
   * @param fontSize the font size
   * @return the node to which to append content.
   **************************************************************************/
  public static Element createTopNode(Document document, int fontSize)
  {
    Element root = getRoot(document);
    Element top = document.createElement("div");
    root.appendChild(top);
    top.setAttribute("style", 
        "padding: 6px; " +
        "font-size: " + fontSize + "pt; " +
        "position: relative; ");
    return top;
  }
  
  /***************************************************************************
   * Write document to writer.
   * @param document the document.
   * @param out the writer.
   **************************************************************************/
  public static void writeDocument(Document document, PrintWriter out)
  {
    try
    {
      TransformerFactory tfactory = TransformerFactory.newInstance();
      Transformer transformer = tfactory.newTransformer();
      transformer.setOutputProperty(OutputKeys.METHOD, "xml");
      transformer.setOutputProperty(OutputKeys.INDENT, "yes");
      transformer.setOutputProperty(OutputKeys.DOCTYPE_PUBLIC,  
          document.getDoctype().getPublicId());
      transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM,
          document.getDoctype().getSystemId());
      DOMSource source = new DOMSource(document);
      StreamResult result = new StreamResult(out);
      transformer.transform(source, result);
      // transformer.transform(source, new StreamResult(System.out));
    }
    catch(TransformerException e)
    {
      Main.internalError("error in transforming document: exception " + e.getMessage());
    }
  }
}
