// ----------------------------------------------------------------------------
// PrettyPrinter.java
// Printing abstract syntax trees with appropriate line breaks (old version).
// $Id: PrettyPrinter.java,v 1.12 2012/10/18 16:12:58 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import java.io.*;
import java.util.*;

import fmrisc.ProgramExplorer.InOut.Breaks.Info;
import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Judgements.*;

/**
 * 
 * Printing abstract syntax trees with appropriate line breaks (old version).
 *
 */
public final class PrettyPrinter extends ASTVisitorBase
{
  /****************************************************************************
   * The pretty printer depends on the current printing mode (LINE or BLOCK) 
   * and the current indentation level LEVEL.
   * 
   * Printing a tree starts in BLOCK mode, when a subtree is encountered 
   * that demands switching to LINE mode, the whole subtree is printed in 
   * LINE mode (no switching back to BLOCK mode within the subtree).
   * 
   * When printed in LINE mode, a syntax tree fills the current line as much 
   * as possible; if a line break has to be inserted, the remainder of the
   * tree is printed at increased level LEVEL+1 (not further increased).
   * 
   *  token token token token token ...      LEVEL 
   *    token token token token token token  LEVEL+1
   *    token token token token token token  LEVEL+1
   *    
   * The only trees that make use of the BLOCK mode are the ones listed below. 
   * If such a tree does not fit into the current line, parts of it are printed
   * at indentation level LEVEL+1 (which may be further increased within 
   * the parts).
   *   
   * QUANTIFIER(VARIABLES): LEVEL
   *   EXPRESSION           LEVEL+1 
   *   
   *   FORMULA          LEVEL+1
   * BINARY-CONNECTIVE  LEVEL
   *   FORMULA          LEVEL+1
   * BINARY-CONNECTIVE  LEVEL    
   *   FORMULA          LEVEL+1
   * 
   * IF FORMULA THEN     LEVEL    
   *   EXPRESSION        LEVEL+1
   * ELSIF FORMULA THEN  LEVEL
   *   EXPRESSION        LEVEL+1
   * ELSE                LEVEL
   *   EXPRESSION        LEVEL+1
   * ENDIF               LEVEL
   * 
   * LET             LEVEL
   *   DECLARATION,  LEVEL+1 
   *   DECLARATION,  LEVEL+1
   *   DECLARATION   LEVEL+1
   * IN              LEVEL
   *   EXPRESSION    LEVEL+1
   *   
   * The indentation level is not increased beyond MAXLEVEL; if this bound
   * is reached, the corresponding subtree is printed in LINE mode. Likewise,
   * if the number of character positions in a line drops below MINFREE, the
   * remainder of the tree is printed in LINE mode.
   ***************************************************************************/
  
  private final int MAXLEVEL = 7; // maximum indentation level
  private final int MINFREE = 20; // no further breaks if this is reached
  private final int INDENT = 2;   // amount of indentation per level
  
  private int width;       // the suggested line width
  private PrintWriter out; // the output stream to be used
  
  private boolean lineMode = false;  // we are in line mode
  private int level = 0;             // the current indentation level
  private int pos = 0;               // the current column position
  
  // the information for the currently visited node
  private Info current;
  
  /****************************************************************************
   * Create pretty printer to print abstract syntax trees.
   * @param width the line width (suggestion, not a hard bound).
   * @param out the stream to print on.
   ***************************************************************************/
  private PrettyPrinter(int width, PrintWriter out)
  {
    this.width = width;
    this.out = out;
  }
  
  /****************************************************************************
   * Print formula.
   * @param formula the formula to be printed.
   * @param width the line width (suggestion, not a hard bound).
   * @param out the stream to print on.
   ***************************************************************************/
  public static void print(Formula formula, int width, PrintWriter out)
  {
    new PrettyPrinter(width, out).print(formula);
  }
  
  /***************************************************************************
   * Print formula.
   * @param formula a formula.
   **************************************************************************/
  private void print(Formula formula)
  {
    print(Breaks.compute(formula, false));
  }

  /****************************************************************************
   * Print statement judgement.
   * @param spec the judgement to be printed.
   * @param width the line width (suggestion, not a hard bound).
   * @param out the stream to print on.
   ***************************************************************************/
  public static void print(StatJudgement spec, int width, PrintWriter out)
  {
    new PrettyPrinter(width, out).print(spec);
  }
  
  /***************************************************************************
   * Print statement judgement.
   * @param spec a statement judgement (may be null).
   **************************************************************************/
  private void print(StatJudgement spec)
  {
    if (spec == null) return;
    print("[\n");
    print(spec.sformula);
    print("\n]^");
    print(spec.continues);
    print(",");
    print(spec.breaks);
    print(",");
    print(spec.returns);
    print("_");
    printClasses(spec.exceptions);
    print(",");
    printVariables(spec.variables);
    Formula termination = spec.getTermination();
    if (termination != null)
    {
      print("(");
      print(termination);
      print(")");
    }
  }
  
  /****************************************************************************
   * Print boolean.
   * @param b the boolean.
   ***************************************************************************/
  private void print(boolean b)
  {
    if (b) print("TRUE"); else print("FALSE");
  }

  
  /****************************************************************************
   * Print name.
   * @param name the name to be printed
   ***************************************************************************/
  private void print(Name name)
  {
    // quick and dirty
    print(name.toString());
  }
  
  /****************************************************************************
   * Print collection of class symbols.
   * @param c the collection.
   ***************************************************************************/
  private void printClasses(Collection<ClassSymbol> c)
  {
    print("{");
    if (c != null)
    {
      Iterator<ClassSymbol> iter = c.iterator();
      while (iter.hasNext())
      {
        ClassSymbol s = iter.next();
        print(s.getLongName());
        if (iter.hasNext()) print(",");
      }
    }
    print("}");
  }
  
  /****************************************************************************
   * Print collection of variable symbols
   * @param c the collection.
   ***************************************************************************/
  private void printVariables(Collection<VariableSymbol> c)
  {
    print("{");
    if (c != null)
    {
      Iterator<VariableSymbol> iter = c.iterator();
      while (iter.hasNext())
      {
        VariableSymbol s = iter.next();
        print(NameUtils.uniqueName(s));
        if (iter.hasNext()) print(",");
      }
    }
    print("}");
  }
  
  // --------------------------------------------------------------------------
  //
  // auxiliary methods
  //
  // --------------------------------------------------------------------------
  
  /***************************************************************************
   * Print text.
   * @param text the text to be printed.
   **************************************************************************/
  private void print(String text)
  {
    if (width-pos < text.length()) 
    {
      resetLine();
      indent(true);
    }
    out.print(text);
    pos = pos+text.length();
  }
  
  /****************************************************************************
   * Change indentation level.
   * @param increase true if level is increased (decreased otherwise)
   ***************************************************************************/
  private void changeLevel(boolean increase)
  {
    if (lineMode) return;
    if (increase) 
    {
      level++;
      indent(false);
    }
    else
      level--;
  }
  
  /***************************************************************************
   * Start new (non-overflow)line and indent it appropriately.
   ***************************************************************************/
  private void newLine()
  {
    if (lineMode) 
      print(" ");
    else
    {
      resetLine();
      indent(false);
    }
  }
  
  /***************************************************************************
   * Insert a space (or a newline at the end of the line).
   **************************************************************************/
  private void space()
  {
    if (pos == width)
      newLine();
    else
      print(" ");
  }
  
  /***************************************************************************
   * Reset to begin of line.
   *************************************************************************/
  private void resetLine()
  {
    out.println();
    pos = 0;
  }
  
  /***************************************************************************
   * If in block mode, indent to the apropriate position.
   * @param overflow true iff line is an overflow line
   **************************************************************************/
  private void indent(boolean overflow)
  {
    int indent = level*INDENT;
    if (overflow) indent = indent+2*INDENT;
    for (int i=pos; i<indent; i++)
      out.print(' ');
    pos = indent;
  }
  
  /***************************************************************************
   * Print expression occurring as a left subexpression of an outer expression,
   * if necessary with parentheses.
   * @param info the info for the expression to be printed.
   * @param outer the outer expression.
   *************************************************************************/
  private void printLeftPar(Info info, Expression outer)
  {
    Expression exp = (Expression)info.tree;
    boolean parentheses =
      exp.getPrecedence() < outer.getPrecedence() &&
      !(exp.getClass() == outer.getClass() && exp.isLeftAssociative());
    if (parentheses) print("(");
    print(info);
    if (parentheses) print(")");
  }
  
  /***************************************************************************
   * Print expression occurring as a right subexpression of an outer expression,
   * if necessary with parentheses.
   * @param info the info for the expression to be printed.
   * @param outer the outer expression.
   *************************************************************************/
  private void printRightPar(Info info, Expression outer)
  {
    Expression exp = (Expression)info.tree;
    boolean parentheses =
      exp.getPrecedence() < outer.getPrecedence() &&
      !(exp.getClass() == outer.getClass() && exp.isRightAssociative());
    if (parentheses) print("(");
    print(info);
    if (parentheses) print(")");
  }
  
  /****************************************************************************
   * Print sequence of trees.
   * @param info the sequence to be printed.
   * @param from the index of the first element to be printed.
   * @param to the index of the last element to be printed.
   * @param separator the separator printed between elements.
   * @param nl true iff a new line is started before each element
   ***************************************************************************/
  private void print(Info[] info, int from, int to,
    String separator, boolean nl)
  {
    int n = to-from;
    
    // check whether sequence fits in single line
    int l = Breaks.length(info);
    if (n>1) l = l+(n-1)*(separator.length()+1);
    boolean single = (pos+l <= width);
    boolean mode = lineMode;
    if (single) 
    {
      // newLine();
      lineMode = true;
    }
    
    for (int i=from; i<to; i++)
    {
      if (!single && nl) newLine();
      print(info[i]);
      if (i<n-1) { print(separator); space(); }
    }
    
    // restore mode
    lineMode = mode;
  }
  
  /****************************************************************************
   * Print tree based on previously computed breaks information.
   * @param info the information on which the printing is based.
   ***************************************************************************/
  private void print(Info info)
  {
    AST tree = info.tree;
    int len = info.getLength();
    int free = width-pos;
    
    // enough space left, we can print the simple way
    if (len <= free)
    {
      tree.print(out);
      pos = pos+len;
      return;
    }
    
    // we may switch to line mode
    boolean oldMode = lineMode;
    if (level > MAXLEVEL || free < MINFREE || tree instanceof Type)
      lineMode = true;

    // the core printing method
    current = info;
    tree.accept(this);
    
    // restore original mode
    lineMode = oldMode;
  }
  
  // --------------------------------------------------------------------------
  //
  // the visitor methods.
  //
  // --------------------------------------------------------------------------
  
  public void visit(Identifier exp)
  {
    print(exp.getString());
    // return true;
  }
 
  public void visit(Name exp)
  {
    print(exp);
    // return true;
  }
  
  public void visit(ApplicationExpression exp)
  {
    // information for current node
    Info info = current;
    
    // we print applications in line mode
    boolean mode = lineMode;
    lineMode = true;
    
    // print the application
    Info[] infos = info.getChildren();
    print(infos[0]);
    print("(");
    print(infos, 1, infos.length, ",", false);
    print(")");
    
    // restore the line mode
    lineMode = mode;
    // return true;
  }
  
  public void visit(BinaryTerm exp)
  { 
    // information for this node
    Info info = current;
    
    // terms printed in line mode
    boolean mode = lineMode;
    lineMode = true;
    
    // print expression
    Info[] children = info.getChildren();
    printLeftPar(children[0], exp);
    print(exp.getOperator());
    printRightPar(children[0], exp);
    
    // restore the line mode
    lineMode = mode;
    // return true;
  }
  
  public void visit(BinaryAtomicFormula exp)
  { 
    // information for this node
    Info info = current;
    
    // terms printed in line mode
    boolean mode = lineMode;
    lineMode = true;
    
    // print expression
    Info[] children = info.getChildren();
    printLeftPar(children[0], exp);
    space(); print(exp.getOperator()); space();
    printRightPar(children[0], exp);
    
    // restore the line mode
    lineMode = mode;
    // return true;
  }
  
  public void visit(AndFormula exp)
  { 
    printAndOrFormula(exp);
    // return true;
  }
  
  public void visit(OrFormula exp)
  { 
    printAndOrFormula(exp);
    // return true;
  }
  
  private void printAndOrFormula(BinaryFormula exp)
  {
    // information for this node
    Info info = current;
    BinaryFormula outer = exp;
    Info[] infos = info.getChildren();
    
    // fix the operator
    String operator = exp.getOperator();
    int olen = operator.length();
    
    // iterate over subformulas of same type
    changeLevel(true);
    while (true)
    {
      while(true)
      {
        printLeftPar(infos[0], outer);
        Formula second = outer.getSecond();
        if (second.getClass() != exp.getClass()) break;
        Info[] infos0 = infos[1].getChildren();
        if (2+olen+infos0[0].getLength()+INDENT > width-pos) break;
        space(); print(operator); space();
        info = infos[1];
        outer = (BinaryFormula)second;
        infos = info.getChildren();
      }
      changeLevel(false);
      newLine();
      print(operator);
      Formula second = outer.getSecond();
      if (second.getClass() != exp.getClass()) break;
      if (infos[1].getLength()+INDENT <= width-pos) break;
      changeLevel(true);
      newLine();
      info = infos[1];
      outer = (BinaryFormula)second;
      infos = info.getChildren();
    }
    
    // print last subformula
    changeLevel(true);
    newLine();
    printRightPar(infos[1], outer);
    changeLevel(false);
  }
  
  public void visit(EquivalentFormula exp)
  { 
    printEquivFormula(exp);
    // return true;
  }
  
  public void visit(NotEquivalentFormula exp)
  { 
    printEquivFormula(exp);
    // return true;
  }
  
  private void printEquivFormula(BinaryFormula exp)
  {
    // information for this node
    Info info = current;
    
    // print expression
    Info[] children = info.getChildren();
    changeLevel(true);
    printLeftPar(children[0], exp);
    changeLevel(false);
    newLine();
    print(exp.getOperator());
    changeLevel(true);
    newLine();
    printRightPar(children[0], exp);
    changeLevel(false);
  }
  
  public void visit(ImpliesFormula exp)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // iterate as long as second formula is of same type
    changeLevel(true);
    Formula outer = (Formula)info.tree;
    while (true)
    {
      printLeftPar(infos[0], outer);
      changeLevel(false);
      newLine();
      print(exp.getOperator());
      if (!(infos[1].tree instanceof ImpliesFormula)) break;
      if (infos[1].getLength()+INDENT <= width-pos) break;
      changeLevel(true);
      newLine();
      info = infos[1];
      infos = info.getChildren();
      outer = (Formula)info.tree;
    }
    changeLevel(true);
    newLine();
    printRightPar(infos[1], outer);
    changeLevel(false);
    
    // we are done
    // return true;
  }
  
  public void visit(UnaryTerm exp)
  {
    // information for this node
    Info info = current;
    
    // terms printed in line mode
    boolean mode = lineMode;
    lineMode = true;
    
    // print expression
    Info[] children = info.getChildren();
    print(exp.getOperator());
    printRightPar(children[0], exp);
    
    // restore the line mode
    lineMode = mode;
    // return true;
  }
  
  public void visit(UnaryFormula exp)
  {
    // information for this node
    Info info = current;
    
    // print expression
    Info[] children = info.getChildren();
    print(exp.getOperator());
    space();
    printRightPar(children[0], exp);
    
    // we are done
    // return true;
  }
  
  public void visit(QuantifiedExpression exp)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    int n = infos.length;
    
    // print the expression
    print(exp.getQuantifier());
    print("(");
    boolean mode = lineMode;
    lineMode = true;
    print(infos, 0, n-1, ",", false);
    lineMode = mode;
    print("):");
    changeLevel(true);
    newLine();
    printRightPar(infos[n-1], exp);
    changeLevel(false);
    
    // we are done
    // return true;
  }
  
  public void visit(ArrayType exp)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    print(ArrayType.arrayKeyword); space();
    print(infos[0]);
    space(); print(ArrayType.baseKeyword); space();
    printRightPar(infos[1], exp);
    
    // we are done
    // return true;
  }
  
  public void visit(AtomicType exp)
  {
    if (!Main.toLogical())
      print(exp.getName());
    else if (exp instanceof BooleanType)
      print("LOGICAL");
    else if (exp instanceof BitType)
      print("BOOLEAN");
    else
      print(exp.getName());
    // return true;
  }
 
  public void visit(Constant exp)
  {
    if (!Main.toLogical())
      print(exp.getName());
    else if (exp instanceof True)
      print("LOGICALTRUE");
    else if (exp instanceof False)
      print("LOGICALFALSE");
    else if (exp instanceof BitTrue)
      print("TRUE");
    else if (exp instanceof BitFalse)
      print("FALSE");
    else
      print(exp.getName());
    // return true;
  }
  
  public void visit(FunctionType exp)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    int n = infos.length;
    
    // special case of a predicate
    Type range = exp.getRange();
    if (range instanceof BooleanType)
    {
      print("PREDICATE(");
      print(infos, 0, n-1, ",", false);
      print(")");
      return;
    }
    
    // general case of a function
    if (n == 2)
      printLeftPar(infos[0], exp);
    else
    {
      print("(");
      print(infos, 0, n-1, ", ", false);
      print(")");
    }
    space(); print("->"); space();
    printRightPar(infos[n-1], exp);
    
    // we are done
    // return true;
  }
  
  public void visit(IfThenElseExpression exp)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    
    // print the expression
    print(IfThenElseExpression.ifKeyword); 
    space();
    lineMode = true;
    print(infos[0]);
    lineMode = mode;
    space();
    print(IfThenElseExpression.thenKeyword);
    changeLevel(true);
    newLine(); print(infos[1]);
    changeLevel(false);
    while (((IfThenElseExpression)info.tree).getElseBranch() 
        instanceof IfThenElseExpression)
    {
      info = infos[2];
      infos = info.getChildren();
      newLine(); print(IfThenElseExpression.elsifKeyword);
      space();
      lineMode = true;
      print(infos[0]);
      lineMode = mode;
      space();
      print(IfThenElseExpression.thenKeyword);
      changeLevel(true);
      newLine(); print(infos[1]);
      changeLevel(false);
    }
    newLine(); print(IfThenElseExpression.elseKeyword);
    changeLevel(true);
    newLine(); print(infos[2]);
    changeLevel(false);
    newLine(); print(IfThenElseExpression.endifKeyword);
    
    // we are done
    // return true;
  }
  
  public void visit(LetExpression exp)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    int n = infos.length;
    
    // print expression
    print(LetExpression.letKeyword);
    changeLevel(true);
    print(infos, 0, n-1, ", ", true);
    changeLevel(false);
    newLine(); print(LetExpression.inKeyword);
    changeLevel(true);
    newLine();
    printRightPar(infos[n-1], exp);
    changeLevel(false);
    
    // we are done
    // return true;
  }
  
  public void visit(PostfixApplication exp)
  { 
    // if (exp instanceof StateThrowsException)
    //   return visit((StateThrowsException)exp);
    
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print expression
    print(exp.getIdentifier().getString());
    print(PostfixApplication.operator);
    printRightPar(infos[0], exp);
    
    // we are done
    // return true;
  }
  
  public void visit(StateThrowsException exp)
  { 
    // if (exp instanceof StateThrowsException)
    //   return visit((StateThrowsException)exp);
    
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print expression
    print(exp.getIdentifier().getString());
    print("(");
    print(infos[1]);
    print(")");
    print(PostfixApplication.operator);
    printRightPar(infos[0], exp);
    
    // we are done
    // return true;
  }
  
  public void visit(ProgramVariableBase exp)
  {     
    boolean mode = lineMode;
    lineMode = true;
    print(exp.getKeyword());
    space();
    print(exp.getName());
    lineMode = mode;
    // return true;
  }
  
  public void visit(WritesOnlyFormula exp)
  { 
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // special case
    if (exp instanceof ReadsOnlyFormula)
    {
      print(ReadsOnlyFormula.keyword);
      return;
    }
    
    // general case
    boolean mode = lineMode;
    lineMode = true;
    print(WritesOnlyFormula.keyword);
    space();
    print(infos, 0, infos.length, ",", false);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(RecordTerm exp)
  { 
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    boolean mode = lineMode;
    lineMode = true;
    print(RecordTerm.beginToken);
    print(infos, 0, infos.length, ",", false);
    print(RecordTerm.endToken);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(RecordType exp)
  { 
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print expression
    boolean mode = lineMode;
    lineMode = true;
    print(RecordType.beginToken);
    print(infos, 0, infos.length, ",", false);
    print(RecordType.endToken);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(Reference exp)
  { 
    print(exp.getName());
    // return true;
  }
  
  // print expression
  public void visit(SelectionTerm exp)
  { 
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    boolean mode = lineMode;
    lineMode = true;
    printLeftPar(infos[0], exp);
    print(infos[1]);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(SelectorIdentifier exp)
  { 
    print(".");
    print(exp.getIdentifier().getString());
    // return true;
  }
  
  public void visit(SelectorIndex exp)
  { 
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print expression
    print("[");
    print(infos[0]);
    print("]");

    // we are done
    // return true;
  }
  
  public void visit(SelectorNumber exp)
  { 
    print(".");
    print(exp.getNumber().getName());
    // return true;
  }
  
  public void visit(StateType exp)
  { 
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print expression
    print("STATE");
    if (infos != null)
    {
      boolean mode = lineMode;
      lineMode = true;
      print("(");
      print(infos[0]);
      print(")");
      lineMode = mode;
    }
    
    // we are done
    // return true;
  }
  
  public void visit(SubrangeType exp)
  { 
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print expression
    boolean mode = lineMode;
    lineMode = true;
    print(SubrangeType.beginToken);
    print(infos[0]);
    print(SubrangeType.separatorToken);
    print(infos[1]);
    print(SubrangeType.endToken);
    lineMode = mode;
    
    // we are done
    // return true;
  } 
  
  public void visit(SubType exp)
  { 
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print expression
    boolean mode = lineMode;
    lineMode = true;
    print(SubType.keyword + "(");
    print(infos[0]);
    print(")");
    lineMode = mode;
    
    // we are done
    // return true;
  }  
    
  public void visit(TupleTerm exp)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print expression
    boolean mode = lineMode;
    lineMode = true;
    print(TupleTerm.beginToken);
    print(infos, 0, infos.length, ",", false);
    print(TupleTerm.endToken);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(TupleType exp)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print expression
    boolean mode = lineMode;
    lineMode = true;
    print(TupleTerm.beginToken);
    print(infos, 0, infos.length, ",", false);
    print(TupleTerm.endToken);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(TypedIdentifier exp)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print expression
    boolean mode = lineMode;
    lineMode = true;
    print(infos[0]);
    print(":"); space();
    print(infos[1]);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(UpdateTerm exp)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    int n = infos.length;
    
    // print expression
    boolean mode = lineMode;
    lineMode = true;
    printLeftPar(infos[0], exp);
    space();
    print(UpdateTerm.withKeyword);
    space();
    print(infos, 1, n-1, "", false);
    print(UpdateTerm.assignToken);
    printRightPar(infos[n-1], exp);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(ValueDeclarationInitialized def)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print declaration
    print(infos[0]);
    print(":"); space();
    print(infos[1]);
    space(); print("=");
    changeLevel(true);
    newLine(); 
    print(infos[2]);
    changeLevel(false);
    
    // we are done
    // return true;
  }
  
  public void visit(ValueDefinitionClass def)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print declaration
    print(infos[0]);
    space(); print("=");
    changeLevel(true);
    newLine(); 
    print(infos[1]);
    changeLevel(false);
    
    // we are done
    // return true;
  }
  
  public void visit(ValuedIdentifier exp)
  {
    // information for this node
    Info info = current;
    Info[] infos = info.getChildren();
    
    // print expression
    boolean mode = lineMode;
    lineMode = true;
    print(infos[0]);
    print(":=");
    print(infos[1]);
    lineMode = mode;
    
    // we are done
    // return true;
  }
}
