// ----------------------------------------------------------------------------
// SourceAnnotationPosition.java
// An annotation position in a source file with a position inside the annotation. 
// $Id: SourceAnnotationPosition.java,v 1.4 2010/03/18 17:15:59 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for identifieric Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import fmrisc.ProgramExplorer.Syntax.*;

/**
 * 
 * An annotation position in a source file with a position inside the annotation.
 *
 */
public final class SourceAnnotationPosition implements SourcePosition
{
  private SourceAnnotation annotation; // the annotation
  private int aline;                   // the line number in the annotation
  private int acolumn;                 // the column number in the annotation
  
  /***************************************************************************
   * Construct a source annotation position.
   * @param annotation the annotation
   * @param aline the line number in the annotation (base 0)
   * @param acolumn the column number in the annotation (base 0)
   **************************************************************************/
  public SourceAnnotationPosition(SourceAnnotation annotation,
    int aline, int acolumn)
  {
    this.annotation = annotation;
    this.aline = aline;
    this.acolumn = acolumn;
  }
  
  /***************************************************************************
   * Get the annotation of the position.
   * @return the annotation.
   **************************************************************************/
  public SourceAnnotation getAnnotation()
  {
    return annotation;
  }
  
  /***************************************************************************
   * Get the annotation  number of the position.
   * @return the line number.
   **************************************************************************/
  public int getAnnotationLine()
  {
    return aline;
  }
  
  /***************************************************************************
   * Get the annotation column number of the position.
   * @return the column number.
   **************************************************************************/
  public int getAnnotationColumn()
  {
    return acolumn;
  }
  
  /***************************************************************************
   * Get the source file.
   * @return the source file.
   **************************************************************************/
  public Source getSource()
  {
    return annotation.getPosition().getSource();
  }
  
  /***************************************************************************
   * Get the line number of the position.
   * @return the line number.
   **************************************************************************/
  public int getLine()
  {
    return annotation.getPosition().getLine();
  }
  
  /***************************************************************************
   * Get the column number of the position.
   * @return the column number.
   **************************************************************************/
  public int getColumn()
  {
    return annotation.getPosition().getColumn();
  }
  
  /****************************************************************************
   * Get offset of the position.
   * @return the number of characters up to (excluding) the position denoted
   * by the line and column index.
   ***************************************************************************/
  public int getOffset()
  {
    return annotation.getPosition().getOffset();
  }
  
  /**************************************************************************
   * Get number of characters from current position to end of line.
   * @return the number of characters.
   *************************************************************************/
  public int getLength()
  {
    return annotation.getPosition().getLength();
  }
  
  /**************************************************************************
   * Get printable text representation of position.
   * @return the text representation.
   *************************************************************************/
  public String toString()
  {
    return annotation.getPosition().toString() + 
      ":" + (aline+1) + ":" + (acolumn+1);
  }
  
  /**************************************************************************
   * Get source line corresponding to position.
   * @return the source line of the position
   *************************************************************************/
  public String getSourceLine()
  {
    return annotation.getPosition().getSourceLine();
  }
  
  /***************************************************************************
   * Register identifier with denoted position.
   * @param ident the identifier to be registered.
   **************************************************************************/
  public void setIdentifier(Identifier ident)
  {
    annotation.getPosition().setIdentifier(ident);
  }
  
  /***************************************************************************
   * Get identifier registered with denoted position.
   * @return the registered identifier (may be null).
   **************************************************************************/
  public Identifier getIdentifier()
  {
    return annotation.getPosition().getIdentifier();
  }
}
