// ----------------------------------------------------------------------------
// StringSourcePosition.java
// A position within a text represented by a string.
// $Id: StringSourcePosition.java,v 1.2 2010/12/06 17:06:43 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for identifieric Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import fmrisc.ProgramExplorer.Main;
import fmrisc.ProgramExplorer.Syntax.Identifier;

/**
 * 
 * A position within a text represented by a string.
 *
 */
public final class StringSourcePosition implements SourcePosition
{
  private String string;  // the string
  
  private int line;       // the line number
  private int column;     // the column number
  
  private Identifier ident;  // the registered identifier (may be null)
  
  /***************************************************************************
   * Construct a string source position.
   * @param string the string
   * @param line the line number.
   * @param column the column number.
   **************************************************************************/
  public StringSourcePosition(String string, int line, int column)
  {
    this.string = string;
    this.line = line;
    this.column = column;
  }
  
  /***************************************************************************
   * Get the line with the denoted number
   * @param line the line number.
   * @return the line corresponding to that number.
   **************************************************************************/
  private String getLine(int line)
  {
    int pos = 0;
    for (int i=0; i<line; i++)
    {
      int pos0 = string.indexOf(Source.NL, pos);
      if (pos0 == -1) Main.internalError("end of line not found");
      pos = pos0+Source.NL.length();
    }
    int pos0 = string.indexOf(Source.NL, pos);
    if (pos0 == -1) pos0 = string.length();
    return string.substring(pos, pos0);
  }
  
  /***************************************************************************
   * Get the source file.
   * @return null, i.e. there is no source file.
   **************************************************************************/
  public Source getSource()
  {
    // Main.internalError("source is requested from no source position");
    return null;
  }
  
  /***************************************************************************
   * Get the line number of the position.
   * @return the line number.
   **************************************************************************/
  public int getLine()
  {
    return line;
  }
  
  /***************************************************************************
   * Get the column number of the position.
   * @return the column number.
   **************************************************************************/
  public int getColumn()
  {
    return column;
  }
  
  /****************************************************************************
   * Get offset of denoted position
   * @return the number of characters up to (excluding) the position denoted
   * by the line and column index.
   ***************************************************************************/
  public int getOffset()
  {
    int offset = 0;
    for (int i=0; i<line; i++)
    {
      String line = getLine(i);
      offset = offset+line.length();
    }
    return offset+column;
  }
  
  /**************************************************************************
   * Get number of characters from current position to end of line.
   * @return the number of characters.
   *************************************************************************/
  public int getLength()
  {
    String line = getSourceLine();
    return line.length()-column;
  }
  
  /**************************************************************************
   * Get printable text representation of position.
   * @return the text representation.
   *************************************************************************/
  public String toString()
  {
    return "(no file):" + (line+1) + ":" + (column+1);
  }
  
  /**************************************************************************
   * Get source line corresponding to position.
   * @return the source line of the position
   *************************************************************************/
  public String getSourceLine()
  {
    return getLine(line);
  }
  
  /***************************************************************************
   * Register identifier with denoted position.
   * @param ident the identifier to be registered.
   **************************************************************************/
  public void setIdentifier(Identifier ident)
  {
    this.ident = ident;
  }
  
  /***************************************************************************
   * Get identifier registered with denoted position.
   * @return the registered identifier (may be null).
   **************************************************************************/
  public Identifier getIdentifier()
  {
    return ident;
  }
}
