// ----------------------------------------------------------------------------
// Normalization.java
// Normalize a specification formula with respect to the set of modified variables.
// $Id: Normalization.java,v 1.18 2012/10/18 16:12:58 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.ewuni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General private License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General private License for more details.
//
// You should have received a copy of the GNU General private License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Program.NamedType;
import fmrisc.ProgramExplorer.Syntax.Program.TypeExpression;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * 
 * Normalize a specification formula with respect to the set of modified variables.
 *
 */
public final class Normalization extends ASTCloner
{
  private final Collection<VariableSymbol> variables;
  
  /***************************************************************************
   * Create normalizer for a set of variables.
   * @param variables the variables that may be modified.
   **************************************************************************/
  private Normalization(Collection<VariableSymbol> variables)
  {
    this.variables = variables;
  }
  
  /**************************************************************************
   * Replace copy of formula with all poststate variables 
   * replaced by prestate variables.
   * Also add the "this" symbol to all plain object field references
   * and make references to class variables unique.
   * @param formula the formula to be normalized.
   * @return the copy of the formula.
   *************************************************************************/
  public static Formula perform(Formula formula)
  {
    Normalization cloner = new Normalization(new SetClass<VariableSymbol>());
    return (Formula)cloner.clone0(formula);
  }
  
  /**************************************************************************
   * Replace copy of term with all poststate variables 
   * replaced by prestate variables.
   * Also add the "this" symbol to all plain object field references
   * and make references to class variables unique.
   * @param term the term to be normalized.
   * @return the copy of the term.
   *************************************************************************/
  public static Term perform(Term term)
  {
    Normalization cloner = new Normalization(new SetClass<VariableSymbol>());
    return (Term)cloner.clone0(term);
  }
  
  /**************************************************************************
   * Replace copy of formula with all poststate variables that do not appear
   * in the set of modifiable variables replaced by prestate variables.
   * Also add the "this" symbol to all plain object field references
   * and make references to class variables unique.
   * @param formula the formula to be normalized.
   * @param variables the set of modifiable variables.
   * @return the copy of the formula.
   *************************************************************************/
  public static Formula perform(Formula formula, 
    Collection<VariableSymbol> variables)
  {
    Normalization cloner = new Normalization(variables);
    return (Formula)cloner.clone0(formula);
  }
  
  /**************************************************************************
   * Get normalized precondition of method (may be null).
   * @param method the method.
   * @param init add initialization precondition to constructors
   * @return the normalized precondition (for object method, "this." is
   * added in references to plain object variables and the given precondition 
   * is extended by "this /= null"). null is returned, if precondition 
   * is trivial.
   *************************************************************************/
  public static Formula preCondition(ParamSymbol method, boolean init)
  {
    // normalize precondition
    MethodSpec spec = method.getSpecification();
    Formula formula = null;
    if (spec != null) 
    {
      Name[] variables = spec.getVariables();
      Set<VariableSymbol> vars = new SetClass<VariableSymbol>();
      modifiableVariables(variables, vars);
      formula = spec.getPreCondition();
      if (formula != null) formula = perform(formula, vars);
    }
    
    // for class methods, we are done
    if (method instanceof ClassMethodSymbol) return formula;
    
    // the class of the method
    ClassSymbol csymbol = method.getClassSymbol();
     
    // for non-helper object methods (not constructors), we have to add class invariant
    if (method instanceof ObjectMethodSymbol && (spec == null || !spec.isHelper()))
    {
      ClassSpec cspec = csymbol.getDeclaration().getClassSpec();
      Formula invariant = null;
      if (cspec != null) invariant = cspec.getClassInvariant();
      if (invariant != null) invariant = perform(invariant);
      formula = Satisfies.andFormula(invariant, formula);
    }
    
    // for constructors, we have to add the initialization formula of the class
    if (init && method instanceof ConstructorSymbol)
      formula = Satisfies.andFormula(csymbol.getInitFormula(), formula);
   
    // for object methods/constructors, we have to add not null formula
    ThisSymbol tsymbol = csymbol.getThisSymbol();
    Formula notNull = ExpressionLogic.notNullFormula(
        Satisfies.preVariable(Name.construct(tsymbol.getName())), 
        BooleanType.construct());
    formula = Satisfies.andFormula(notNull, formula);
    
    // the normalized version of the precondition
    return formula;
  }
      
  /**************************************************************************
   * Get normalized postcondition of method (may be null).
   * @param method the method.
   * @return the normalized precondition (for object method, "this." is
   * added in references to plain object variables). null is returned,
   * if postcondition is trivial.
   *************************************************************************/
  public static Formula postCondition(ParamSymbol method)
  {
    // get specified postcondition
    MethodSpec spec = method.getSpecification();
    Formula formula = null;
    if (spec != null) formula = spec.getPostCondition();
    
    // for object constructors/methods add class invariant
    if (method instanceof ConstructorSymbol || 
        method instanceof ObjectMethodSymbol)
    {
      ClassSymbol csymbol = method.getClassSymbol();
      ClassSpec cspec = csymbol.getDeclaration().getClassSpec();
      Formula invariant = null;
      if (cspec != null) invariant = cspec.getClassInvariant();
      formula = Satisfies.andFormula(invariant, formula);
    }
    
    // normalize postcondition
    Set<VariableSymbol> vars = new SetClass<VariableSymbol>();
    Name[] variables = null;
    if (spec != null)
    {
      variables = spec.getVariables();
      modifiableVariables(variables, vars);
    }
    if (formula != null) formula = perform(formula, vars);
    
    // add variable constraints
    if (!vars.isEmpty())
    {
      Vector<Formula> equalities = new Vector<Formula>();;
      objectEqualities(variables, vars, equalities);
      for (Formula eq : equalities)
      {
        if (formula == null)
          formula = eq;
        else
          formula = Satisfies.andFormula(formula, eq);
      }
    }

    // the normalized version of the formula
    return formula;
  }
  
  /***************************************************************************
   * Derive the modifiable variables of a method.
   * @param names the names of the assignable clause of a method (may be null)
   * @return the symbols of those variables 
   *        that may be modified according to the listed names.
   **************************************************************************/
  public static Set<VariableSymbol> modifiableVariables(Name[] names)
  {
    Set<VariableSymbol> vars = new SetClass<VariableSymbol>();
    if (names == null) return vars;
    for (Name name : names) 
      vars.add(modifiableVariable(name));
    return vars;
  }
  
  /***************************************************************************
   * Derive the modifiable variables of a method.
   * @param names the names of the assignable clause of a method (may be null).
   * @param vars a set to which the symbols of those variables are added
   *        that may be modified according to the listed names.
   **************************************************************************/
  public static void modifiableVariables(Name[] names, Set<VariableSymbol> vars)
  {
    if (names == null) return;
    for (Name name : names) 
      vars.add(modifiableVariable(name));
  }
  
  /***************************************************************************
   * Derive the modifiable variable from a name.
   * @param name0 the expression denoting the variable.
   * @return the variable.
   **************************************************************************/
  public static VariableSymbol modifiableVariable(Name name0)
  {
    // make sure "this" is added on plain object symbols
    Name name = NameUtils.uniqueName(name0);
    int n = name.idents.length;
    for (int i=0; i<n; i++)
    {
      Symbol symbol0 = name.idents[i].getSymbol();
      if (symbol0 instanceof VariableSymbol)
        return(VariableSymbol)symbol0;
    }
    Main.internalError("name does not denote a variable");
    return null;
  }
  
  /***************************************************************************
   * Derive equalities from the assignable variables of a method.
   * @param names the variable names listed in the assignable clause.
   * @param variables the variables that may correspondingly be modified.
   * @param equs the list to which equalities for those object fields.
   *        are added that are not modified according to vars.
   **************************************************************************/
  private static void objectEqualities(Name[] names,
    Set<VariableSymbol> variables, Vector<Formula> equs)
  {
    // make names unique
    Collection<Name> names0 = new SetClass<Name>(names.length);
    for (Name name: names)
      names0.add(NameUtils.uniqueName(name));
    
    // process modifiable variables
    for (VariableSymbol var : variables)
    {
      Name name = NameUtils.uniqueName(var);
      Vector<Identifier> vector = new Vector<Identifier>();
      Identifier[] idents = name.idents;
      int n = idents.length;
      for (int i=0; i<n; i++) 
        vector.add(idents[i]);
      objectEqualities(vector, names0, equs);
    }
  }
  
  /***************************************************************************
   * Derive equalities from the assignable variables of a method.
   * @param idents a sequence of identifiers denoting an object field.
   * @param vars the names of the assignable variables of a method.
   * @param equs the list to which equalities for the field and its subfields
   *        are added that are not modified according to vars.
   **************************************************************************/
  private static void objectEqualities(Vector<Identifier> idents, 
    Collection<Name> vars, Vector<Formula> equs)
  {
    // field is the same as some of the variables
    if (isSame(idents, vars)) return;
    
    // field is not the prefix of any of the variables
    if (!isPrefix(idents, vars))
    {
      equs.add(
          EqualsFormula.construct(
              Satisfies.postVariable(Name.construct(idents)),
              Satisfies.preVariable(Name.construct(idents))));
      return;
    }
    
    // field is the prefix of some of the variables
    VariableSymbol osymbol = (VariableSymbol)idents.lastElement().getSymbol();
    TypeExpression type = osymbol.getType();
    
    // field does not denote an object
    if (!(type instanceof NamedType)) return;
    
    // field denotes an object, traverse its object fields
    NamedType type0 = (NamedType)type;
    ClassSymbol csymbol = type0.getSymbol();
    Collection<ObjectVariableSymbol> ovars = csymbol.getObjectVariableSet();
    int n = idents.size();
    for (ObjectVariableSymbol ovar : ovars)
    {
      idents.add(ovar.getName());
      objectEqualities(idents, vars, equs);
      idents.remove(n);
    }
  }
  
  /***************************************************************************
   * Is sequence of identifiers identical to some name.
   * @param idents a sequence of identifiers.
   * @param vars a set of variables.
   * @return true if sequence denotes some some variable in the list.
   **************************************************************************/
  private static boolean isSame(Vector<Identifier> idents, Collection<Name> vars)
  {
    for (Name var : vars)
    {
      if (isSame(idents, var)) return true;
    }
    return false;
  }
  
  /***************************************************************************
   * Is sequence of identifiers prefix of some name.
   * @param idents a sequence of identifiers.
   * @param var a variable.
   * @return true if sequence denotes the variable.
   **************************************************************************/
  private static boolean isSame(Vector<Identifier> idents, Name var)
  {
    Symbol symbol0 = idents.lastElement().getSymbol();
    Symbol symbol1 = var.getSymbol();
    return symbol0 == symbol1;
  }
  
  /***************************************************************************
   * Is sequence of identifiers prefix of some name.
   * @param idents a sequence of identifiers.
   * @param vars a set of variables.
   * @return true if sequence is prefix of some variable in the list.
   **************************************************************************/
  private static boolean isPrefix(Vector<Identifier> idents,  Collection<Name> vars)
  {
    for (Name var : vars)
    {
      if (isPrefix(idents, var)) return true;
    }
    return false;
  }
  
  /***************************************************************************
   * Is sequence of identifiers prefix of some name.
   * @param idents a sequence of identifiers.
   * @param var a variable.
   * @return true if sequence is prefix of the identifiers in the variable.
   **************************************************************************/
  private static boolean isPrefix(Vector<Identifier> idents, Name var)
  {
    Symbol symbol = idents.lastElement().getSymbol();
    Identifier[] idents0 = var.idents;
    int n = idents0.length;
    for (int i=0; i<n; i++)
    {
      if (symbol == idents0[i].getSymbol()) return true;
    }
    return false;
  }
  
  // ------------------------------------------------------------------------
  //
  // the visitor methods
  //
  // ------------------------------------------------------------------------
  
  /**************************************************************************
   * Replace poststate variable by prestate variable if it
   * does not appear in the set of modifiable variables.
   * Also add the "this" symbol to all plain object field references.
   * @param exp the variable to be normalized.
   *************************************************************************/
  public void visit(PostVariable exp)
  {
    Name name = NameUtils.uniqueName(exp.getName());
    int pos = NameUtils.getVariablePosition(name);
    Name name0 = name.getPrefix(pos+1);
    VariableSymbol symbol = (VariableSymbol)name0.getSymbol();
    Term base;
    if (variables.contains(symbol))
      base = PostVariable.construct(name0);
    else
      base = PreVariable.construct(name0);
    base.setType(symbol.getLogicType());
    this.result = makeSelectionTerm(base, name, pos+1);
  }
 
  /**************************************************************************
   * Replace poststate variable by prestate variable if it
   * does not appear in the set of modifiable variables.
   * Also add the "this" symbol to all plain object field references.
   * @param exp the variable to be normalized.
   *************************************************************************/
  public void visit(PreVariable exp)
  {
    Name name = NameUtils.uniqueName(exp.getName());
    int pos = NameUtils.getVariablePosition(name);
    Name name0 = name.getPrefix(pos+1);
    VariableSymbol symbol = (VariableSymbol)name0.getSymbol();
    Term base = PreVariable.construct(name0);
    base.setType(symbol.getLogicType());
    this.result = makeSelectionTerm(base, name, pos+1);
  }
  
  /***************************************************************************
   * Construct selection term from name with identifiers.
   * @param base the base term to which the selection is applied.
   * @param name the name holding the selection identifiers.
   * @param from the position where the construction starts.
   * @return the constructed selection term.
   **************************************************************************/
  private Term makeSelectionTerm(Term base, Name name, int from)
  {
    Term result = base;
    int n = name.idents.length;
    for (int i=from; i<n; i++)
    {
      Identifier ident = name.idents[i];
      VariableSymbol symbol = (VariableSymbol)ident.getSymbol();
      result = SelectionTerm.construct(result,
          SelectorIdentifier.construct(ident));
      result.setType(symbol.getLogicType());
    }
    return result;
  }
}