// ----------------------------------------------------------------------------
// TopWindow.java
// The top panel of the GUI.
// $Id: TopWindow.java,v 1.113 2023/06/22 08:34:01 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.SWT;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.Formula;
import fmrisc.ProgramExplorer.Syntax.Program.*;
import fmrisc.ProgramExplorer.Tasks.*;
import fmrisc.ProgramExplorer.Proving.*;
import fmrisc.ProgramExplorer.Judgements.*;

import java.io.*;
import java.util.*;

import org.eclipse.swt.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.browser.Browser;
import org.eclipse.swt.custom.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.*;
import org.eclipse.jface.text.*;
import org.eclipse.jface.text.source.*;
import org.eclipse.jface.text.rules.*;
import org.eclipse.jface.text.presentation.*;
import org.eclipse.jface.text.source.projection.*;
import org.eclipse.jface.text.reconciler.*;
import org.eclipse.core.runtime.*;
import org.eclipse.jface.action.*;

/**
 * 
 * The top panel of the GUI.
 *
 */
public final class TopWindow extends Composite
{
  // -------------------------------------------------------------------------
  //
  // text literals
  //
  // -------------------------------------------------------------------------
  
  public final static String aboutText = 
    "<html>\n" +
    "<body style=\"font-size:10pt;\">\n" +
    "<img align=right " +
    "src=\"http://www.risc.jku.at/people/schreine/risc/risc-logo-transparent.gif\"/>\n" +
    "<h1>RISC ProgramExplorer</h1>\n" +
    "<b>Version " + Main.getVersion() + "</b><br>\n" +
    "For the latest version of this program and further information, see " +
    "<a href=\"http://www.risc.jku.at/research/formal/software/ProgramExplorer\">" +
    "http://www.risc.jku.at/research/formal/software/ProgramExplorer</a>.<p>" +
    "Copyright (C) 2008-, <a href=\"http://www.risc.jku.at\">" +
    "Research Institute for Symbolic Computation (RISC)</a>,\n"+
    "Johannes Kepler University, Linz, Austria.<p>" +
    "This program is free software; you can redistribute it and/or modify " +
    "it under the terms of the GNU General Public License as published by " +
    "the Free Software Foundation, either version 3 of the License, or " +
    "(at your option) any later version.<p>" +
    "Author: Wolfgang Schreiner " +
    "&lt;<a href=\"mailto:Wolfgang.Schreiner@risc.jku.at\">" +
    "Wolfgang.Schreiner@risc.jku.at</a>&gt;\n<br>" +
    "</body>\n" +
    "</html>";

  // URL of help documents
  public final static String helpURL = 
    "http://www.risc.jku.at/research/formal/software/ProgramExplorer/manual";
  public final static String helpURLOld = 
      "http://www.risc.jku.at/research/formal/software/ProofNavigator/manual";
  
  // -------------------------------------------------------------------------
  //
  // Semantic information
  //
  // -------------------------------------------------------------------------
  
  // the top-level package
  private PackageSymbol topPackage = null;
  
  // -------------------------------------------------------------------------
  //
  // GUI Components
  //
  // -------------------------------------------------------------------------
  
  // image names
  // newImage is small, since otherwise images are clipped
  // seems related to https://bugs.eclipse.org/bugs/show_bug.cgi?id=183003
  private String runImageName = "go-next.png";
  private String newImageName = "document-new.png";
  private String openImageName = "document-open.png";
  private String saveImageName = "media-floppy.png";
  private String analyzeImageName = "edit-find.png";
  private String verifyImageName = "dialog-information.png";
  private String consoleImageName = "utilities-terminal.png";
  private String sourceImageName = "text-x-generic.png";
  private String okImageName = "dialog-ok.png";
  private String cancelImageName = "dialog-cancel.png";
  private String folderImageName = "folder.png";
  private String fileImageName = "text-x-script.png";
  private String refreshImageName = "view-refresh.png";
  private String deleteImageName = "edit-delete.png";
  private String packageImageName = "package-x-generic.png";
  private String classImageName = "media-playback-stop.png";
  private String theoryImageName = "media-playback-start.png";
  private String classVarImageName = "list-remove.png";
  private String objectVarImageName = "list-add.png";
  private String constructorImageName = "applications-development.png";
  private String classMethodImageName = "application-x-executable.png";
  private String objectMethodImageName = "applications-other.png";
  private String parameterImageName = "emblem-symbolic-link.png";
  private String errorImageName = "software-update-urgent.png";
  private String folderOpenImageName = "annotation-open.png";
  private String folderCloseImageName = "annotation-close.png";
  private String folderProcessImageName = "user-home.png";
  private String taskNewImageName = "weather-overcast.png";
  private String taskOpenImageName = "weather-few-clouds.png";
  private String taskDoneImageName = "weather-clear.png";
  private String taskFailedImageName = "weather-severe-alert.png";
  private String typeImageName = "preferences-system-windows.png";
  private String valueImageName = "office-calendar.png";
  private String formulaImageName = "help-browser.png";
  private String detailImageName = "system-search.png";
  
  // the images
  private Image runImage = null;        
  private Image newImage = null;      
  private Image openImage = null;   
  private Image saveImage = null;   
  private Image analyzeImage = null;
  private Image verifyImage = null;
  private Image consoleImage = null;
  private Image sourceImage = null;
  private Image okImage = null;
  private Image cancelImage = null;
  private Image folderImage = null;
  private Image fileImage = null;
  private Image refreshImage = null;
  private Image deleteImage = null;
  private Image packageImage = null;
  private Image classImage = null;
  private Image theoryImage = null;
  private Image classVarImage = null;
  private Image objectVarImage = null;
  private Image constructorImage = null;
  private Image classMethodImage = null;
  private Image objectMethodImage = null;
  private Image parameterImage = null;
  private Image errorImage = null;
  private Image folderOpenImage = null;
  private Image folderCloseImage = null;
  private Image folderProcessImage = null;
  private Image taskNewImage = null;
  private Image taskOpenImage = null;
  private Image taskDoneImage = null;
  private Image taskFailedImage = null;
  private Image typeImage = null;
  private Image formulaImage = null;
  private Image valueImage = null;
  private Image detailImage = null;
  
  // the colors
  private Color openColor;       // color for open tasks
  private Color closedColor;     // color for closed tasks
  private Color proofColor;      // color for open tasks with a closed proof
  private Color blackColor;      // black 
  private Color whiteColor;      // white
  
  // the buttons
  private ToolItem newButton = null;  
  private ToolItem openButton = null;  
  private ToolItem saveButton = null; 
  private ToolItem refreshButton = null;    
  private ToolItem runButton = null; 
  
  // the GUI of the window
  private MainSWT gui;
  
  // the shell of the window
  private Shell shell;
  
  // the top-level widgets
  private CTabFolder views = null;      // the folder holding the panels
  private ToolBar buttons = null;       // buttons associated to the folder
  private Menu explorerMenu = null;     // the menu for the explorer view
  
  // the views 
  private CTabItem analyzeView = null;  // the analyze view
  private CTabItem verifyView = null;   // the verification view
  private CTabItem detailView = null;   // the detail view
   
  // content of the analyze view
  private SashForm analyzeForm = null;   // the form holding the area
  private CTabFolder explorers = null;   // the explorer area
  private CTabItem fitem = null;         // the file explorer item
  private FileTree fileExplorer = null;  // the file explorer
  private CTabItem sitem = null;         // the symbol explorer item
  private SymbolTree symbolExplorer = null; // the symbol explorer
  private SashForm centerForm = null;    // the form holding the center area
  private CTabFolder editors = null;     // the editor area
  private CTabFolder displays = null;    // the display area
  private CTabFolder tasks = null;       // the tasks area
  private TaskTree taskExplorer = null;  // the task explorer
  private TaskList openExplorer = null;  // the open task explorer
  
  // the console 
  private Text console = null;
  private PrintWriter consoleOutput = null;
  
  // content of the verify view
  private fmrisc.ProofNavigator.SWT.Top verifyForm = null; 

  // content of the detail view
  private SashForm detailForm = null;    // the form holding the area
  private CTabFolder browsers = null;    // the browsers area
  
  // the menu items globally affected by operaions
  private MenuItem saveItem = null;
  private MenuItem undoItem = null;
  private MenuItem fontItem = null;
  private MenuItem closeItem = null;
  
  // the set of files being edited keyed by names and items
  private Map<String,FileInfo> fileNames = 
    new MapClass<String,FileInfo>();
  private Map<CTabItem,FileInfo> fileItems = 
    new MapClass<CTabItem,FileInfo>();
  
  // the names of the virtual directories that are browsed
  private Map<CTabItem,String> directoryNames = 
    new MapClass<CTabItem,String>();
  
  // the method symbols whose details are displayed
  private Map<ParamSymbol,CTabItem> methodItems =
    new MapClass<ParamSymbol,CTabItem>();
  private Map<CTabItem,ParamSymbol> displayedMethods =
    new MapClass<CTabItem,ParamSymbol>();
  
  // font handling
  private boolean smallFonts = true;
  private Font smallMono = null;
  private Font bigMono = null;
  private Font fontMono = null;
  
  // fonts and line widths for presenter
  private int smallFont = 10;
  private int bigFont = 12;
  private int smallWidth = 90;
  private int bigWidth = 80;
  
  // annotation handling
  private static final String UNDERLINE_SQUIGGLE = "fmrisc.squiggle";
  private static final String ERROR_TYPE = "fmrisc.error";
  private static final String MARKER_TYPE = "fmrisc.marker";
  
  // the various kinds of cursors
  private final Cursor arrowCursor =  
    new Cursor(getDisplay(), SWT.CURSOR_ARROW);
  private final Cursor ibeamCursor = 
    new Cursor(getDisplay(), SWT.CURSOR_IBEAM);
  private final Cursor waitCursor = 
    new Cursor(getDisplay(), SWT.CURSOR_WAIT);
  
  // -------------------------------------------------------------------------
  //
  // GUI construction
  //
  // -------------------------------------------------------------------------
  
  /****************************************************************************
   * Construct window connected to GUI.
   * @param gui the GUI.
   * @return a window connected to GUI.
   ***************************************************************************/
  public static TopWindow construct(MainSWT gui)
  {
    Display display = gui.getDisplay();
    Shell shell = new Shell(display, 
      SWT.CLOSE | SWT.MIN | SWT.MAX | SWT.RESIZE | SWT.TITLE | SWT.BORDER);
    shell.setText("RISC ProgramExplorer");
    Image riscImage = new Image(shell.getDisplay(),
        Main.class.getResourceAsStream("/" + MainSWT.riscImageName));
    shell.setImage(riscImage);
    return new TopWindow(gui, shell);
  }
  
  /****************************************************************************
   * Construct panel for denoted shell.
   * @param gui the GUI for the panel.
   * @param shell the shell to which the panel is attached.
   ***************************************************************************/
  private TopWindow(MainSWT gui, Shell shell)
  {
    // initialize window
    super(shell, SWT.NONE);
    this.gui = gui;
    this.shell = shell;
    
    // closing the shell raises the quit dialog
    shell.addShellListener(new ShellAdapter() { 
      public void shellClosed(ShellEvent e)
      {
        boolean open = checkOpenProof();
        if (open)
        {
          e.doit = false;
          return;
        }
        boolean confirm = ask("Quit Program", "Do you really want to quit?");
        if (!confirm)
        {
          e.doit = false;
          return;
        }
        // without this, java may trigger segfault on exit()
        displayAnalyzeView();
        boolean okay = closeAll();
        if (!okay) 
        {
          e.doit = false;
          return;
        }
        consoleOutput.println("cleaning task directory...");
        console.update();
        Main.exit(0);
      }
    });

    // create fonts
    smallMono = new Font(getDisplay(), "Monospace", smallFont, SWT.NORMAL);
    bigMono = new Font(getDisplay(), "Monospace", bigFont, SWT.NORMAL);
    fontMono = smallMono;
    
    // create colors
    openColor = new Color(getDisplay(), new RGB(255, 0, 0));
    closedColor = new Color(getDisplay(), new RGB(0, 0, 255));
    proofColor = new Color(getDisplay(), new RGB(128, 0, 255));
    blackColor = new Color(getDisplay(), new RGB(0, 0, 0));
    whiteColor = new Color(getDisplay(), new RGB(255, 255, 255));
     
    // create top structure
    createImages();
    createViews();
    createMenu();
    
    // create analyze view
    createAnalzeView();
    createExplorerArea();
    createCenterArea();
    createEditorArea();
    createTaskArea();
    createDisplayArea();
    createConsole();
    
    // create detail view
    createDetailView();
    createBrowserArea();
    
    // create verify view
    // must be created last to avoid its content displayed in all views ???
    // I have no idea what is going on here
    createVerifyView();
    
    // recreate detail view
    views.getItem(1).dispose();
    createDetailView();
    
    // layout everything
    layout();
    
    // display analyze view
    displayAnalyzeView();
  }
  
  
  /****************************************************************************
   * Open shell and perform initial processing of packages.
   ***************************************************************************/
  public void open()
  {
    java.awt.SplashScreen screen = java.awt.SplashScreen.getSplashScreen();
    if (screen != null) screen.close();
    shell.open();
    console.setFocus();
  }
  
  /****************************************************************************
   * Resize shell.
   ***************************************************************************/
  public void resize()
  {
    shell.setSize(1200, 900);
  }
  
  /****************************************************************************
   * Handle layout issues.
   ***************************************************************************/
  public void layout()
  {
    // set shell layout
    GridLayout shellLayout = new GridLayout();
    shell.setLayout(shellLayout);
    shellLayout.numColumns = 1;
    shellLayout.marginWidth = 0;
    shellLayout.marginHeight = 0;

    // set own layout
    setLayoutData(new GridData(GridData.FILL_BOTH));
    GridLayout topLayout =  new GridLayout();
    setLayout(topLayout);
    topLayout.numColumns = 1;
    topLayout.verticalSpacing = 3;
    topLayout.marginTop = 3;
    topLayout.marginHeight = 0;
    topLayout.marginWidth = 0;
    
    // set layout of views
    views.setLayoutData(new GridData(GridData.FILL_BOTH));
    GridLayout viewLayout =  new GridLayout();
    views.setLayout(viewLayout);
    viewLayout.numColumns = 1;
    viewLayout.verticalSpacing = 3;
    viewLayout.marginTop = 3;
    viewLayout.marginHeight = 0;
    viewLayout.marginWidth = 0;
    
    // set layout of analyze view
    analyzeForm.setLayoutData(new GridData(GridData.FILL_BOTH));
    GridLayout analyzeLayout =  new GridLayout();
    analyzeForm.setLayout(analyzeLayout);
    analyzeLayout.numColumns = 3;
    analyzeLayout.verticalSpacing = 0;
    analyzeLayout.marginTop = 0;
    analyzeLayout.marginHeight = 0;
    analyzeLayout.marginWidth = 0;  
    
    // set layout of center area
    centerForm.setLayoutData(new GridData(GridData.FILL_BOTH));
    GridLayout centerLayout =  new GridLayout();
    centerForm.setLayout(centerLayout);
    centerLayout.numColumns = 1;
    centerLayout.verticalSpacing = 0;
    centerLayout.marginTop = 0;
    centerLayout.marginHeight = 0;
    centerLayout.marginWidth = 0;  
    
    // set relative weights of components 
    analyzeForm.setWeights(new int[]{30, 100, 45});
    centerForm.setWeights(new int[]{65, 35});
  }
  
  /***************************************************************************
   * Open dialog to ask yes/no question.
   * @param title the dialog title
   * @param question the question asked
   * @return true iff question is answered positively.
   **************************************************************************/
  public boolean ask(String title, String question)
  {
    MessageBox box = 
      new MessageBox(shell, SWT.YES | SWT.NO | SWT.ICON_QUESTION
          | SWT.APPLICATION_MODAL);
    box.setText(title);
    box.setMessage(question);
    return box.open() == SWT.YES;
  }
  
  /***************************************************************************
   * Open dialog to ask yes/no/cancel question.
   * @param title the dialog title
   * @param question the question asked
   * @return 0 for "no", 1 for "yes", -1 for "cancel. 
   **************************************************************************/
  private int askCancel(String title, String question)
  {
    MessageBox box = 
      new MessageBox(shell, SWT.YES | SWT.NO | SWT.CANCEL | SWT.ICON_QUESTION
          | SWT.APPLICATION_MODAL);
    box.setText(title);
    box.setMessage(question);
    int result = box.open();
    if (result == SWT.YES) return 1;
    if (result == SWT.NO) return 0;
    return -1;
  }
  
  /***************************************************************************
   * Check whether proof is open (asking the user to close the proof).
   * @return true iff there is an open proof.
   **************************************************************************/
  public boolean checkOpenProof()
  {
    if (fmrisc.ProofNavigator.State.getProof() == null) return false;
    views.setSelection(verifyView);
    inform("Close Proof", "Close open proof first.");
    return true;
  }
  
  /***************************************************************************
   * Open information dialog.
   * @param title the dialog title
   * @param message the message text
   **************************************************************************/
  public void inform(String title, String message)
  {
    MessageBox box = 
      new MessageBox(shell, SWT.OK | SWT.ICON_INFORMATION
          | SWT.APPLICATION_MODAL);
    box.setText(title);
    box.setMessage(message);
    box.open();
  }
  
  /***************************************************************************
   * Create menu bar for explorer.
   **************************************************************************/
  private void createMenu()
  {
    explorerMenu = new Menu(shell, SWT.BAR);
    
    MenuItem fileTitle = new MenuItem(explorerMenu, SWT.CASCADE);
    fileTitle.setText("File");
    Menu fileMenu = new Menu(shell, SWT.DROP_DOWN);
    fileTitle.setMenu(fileMenu);
    MenuItem newItem = new MenuItem(fileMenu, SWT.NULL);
    newItem.setText("New File...");
    MenuItem openItem = new MenuItem(fileMenu, SWT.NULL);
    openItem.setText("Open File...");
    new MenuItem(fileMenu, SWT.SEPARATOR);
    closeItem = new MenuItem(fileMenu, SWT.NULL);
    closeItem.setText("Close File (Ctrl+W)");
    MenuItem closeAllItem = new MenuItem(fileMenu, SWT.NULL);
    closeAllItem.setText("Close All Files");
    new MenuItem(fileMenu, SWT.SEPARATOR);
    saveItem = new MenuItem(fileMenu, SWT.NULL);
    saveItem.setText("Save File (Ctrl+S)");
    new MenuItem(fileMenu, SWT.SEPARATOR);
    MenuItem workspaceItem = new MenuItem(fileMenu, SWT.NULL);
    workspaceItem.setText("Workspace...");
    MenuItem propertyItem = new MenuItem(fileMenu, SWT.NULL);
    propertyItem.setText("Properties...");
    new MenuItem(fileMenu, SWT.SEPARATOR);
    MenuItem quitItem = new MenuItem(fileMenu, SWT.NULL);
    quitItem.setText("Quit");

    MenuItem editTitle = new MenuItem(explorerMenu, SWT.CASCADE);
    editTitle.setText("Edit");
    Menu editMenu = new Menu(shell, SWT.DROP_DOWN);
    editTitle.setMenu(editMenu);
    undoItem = new MenuItem(editMenu, SWT.NULL);
    undoItem.setText("Undo (Ctrl+U)");
    new MenuItem(editMenu, SWT.SEPARATOR);
    fontItem = new MenuItem(editMenu, SWT.NULL);
    if (smallFonts)
      fontItem.setText("Bigger Font");
    else
      fontItem.setText("Smaller Font");
    
    MenuItem helpTitle = new MenuItem(explorerMenu, SWT.CASCADE);
    helpTitle.setText("Help");
    Menu helpMenu = new Menu(shell, SWT.DROP_DOWN);
    helpTitle.setMenu(helpMenu);
    MenuItem helpItem = new MenuItem(helpMenu, SWT.NULL);
    helpItem.setText("Online Manual");
    new MenuItem(helpMenu, SWT.SEPARATOR);
    MenuItem aboutItem = new MenuItem(helpMenu, SWT.NULL);
    aboutItem.setText("About RISC ProgramExplorer");
        
    // menu item "New File"
    newItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        newFile();
      }
    });
    
    // menu item "Open File"
    openItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        openFile();
      }
    });
    
    // menu item "Close File"
    closeItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        CTabItem item = editors.getSelection();
        if (item == null) return;
        closeFile(item);
      }
    });
    closeItem.setEnabled(false);
    closeItem.setAccelerator(SWT.CTRL+'W'); 
    
    // menu item "Close All Files"
    closeAllItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        closeAll();
      }
    });
    
    // menu item "Save File"
    saveItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        saveFile();
      }
    });
    saveItem.setEnabled(false);
    saveItem.setAccelerator(SWT.CTRL+'S'); 
    
    // menu item "Workspace": select workspace directories
    final TopWindow top = this;
    workspaceItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        DirectorySelectionWindow w = 
          DirectorySelectionWindow.construct(top, Main.getPath());
        openCentered(w.getShell());
      }
    });
    
    // menu item "Properties": configure system properties
    propertyItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        String[] properties = Main.getProperties();
        if (checkOpenProof()) return;
        PropertyConfigurationWindow w = 
          PropertyConfigurationWindow.construct(top, properties);
        openCentered(w.getShell());
      }
    });
    
    // menu item "Quit": quit program
    quitItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        shell.close();
      }
    });
    
    // menu item "Undo": undo operation in currently edited document
    undoItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        CTabItem item = editors.getSelection();
        if (item == null) return;
        FileInfo info = fileItems.get(item);
        SourceViewer viewer = info.viewer;
        if (viewer.canDoOperation(ITextOperationTarget.UNDO))
          viewer.doOperation(ITextOperationTarget.UNDO);
      }
    });
    undoItem.setEnabled(false);
    undoItem.setAccelerator(SWT.CTRL+'Z'); 
    
    // menu item "Bigger/Smaller Font": resize fonts
    fontItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        resizeFont();
      }
    });
    
    // menu item "Help": open help window
    helpItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        BrowserWindow b = BrowserWindow.construct(gui, "Help", 1000, 700);
        b.getBrowser().setUrl(helpURL);
        openCentered(b.getShell());
      }
    });
    
    // menu item "About": open about window
    aboutItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        BrowserWindow b = BrowserWindow.construct(gui, "About", 800, 420);
        b.getBrowser().setText(aboutText);
        openCentered(b.getShell());
      }
    });
  }
  
  /***************************************************************************
   * Resize fonts in display areas.
   **************************************************************************/
  public void resizeFont()
  {
    if (smallFonts)
    {
      fontMono = bigMono;
      fontItem.setText("Smaller Font");
    }
    else
    {
      fontMono = smallMono;
      fontItem.setText("Bigger Font");
    }
    smallFonts = !smallFonts;
    console.setFont(fontMono);
    CTabItem[] items = editors.getItems();
    for (CTabItem item : items)
    {
      FileInfo info = fileItems.get(item);
      info.viewer.getTextWidget().setFont(fontMono);
    }   
    createMenu();
    shell.setMenuBar(explorerMenu);
    fmrisc.ProofNavigator.SWT.MainSWT.resizeFont();
  }
  
  /***************************************************************************
   * Open window centered at top-level window
   * @param window the window to be opened.
   **************************************************************************/
  private void openCentered(Shell window)
  {
    Point p = shell.getLocation();
    Point s = shell.getSize();
    Point s0 = window.getSize();
    window.setLocation(p.x+s.x/2-s0.x/2,p.y+s.y/2-s0.y/2);
    window.open();
  }
  
  /****************************************************************************
   * Create image from denoted file name
   * @param name the name of the image file
   * @return the image
   ***************************************************************************/
  private Image newImage(String name)
  {
    return new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + name));
  }
  
  /****************************************************************************
   * Create images.
   **************************************************************************/
  private void createImages()
  {
    newImage = newImage(newImageName);
    openImage = newImage(openImageName);
    saveImage = newImage(saveImageName);
    runImage = newImage(runImageName);
    consoleImage = newImage(consoleImageName);
    sourceImage = newImage(sourceImageName);
    okImage = newImage(okImageName);
    cancelImage = newImage(cancelImageName);
    folderImage = newImage(folderImageName);
    fileImage = newImage(fileImageName);
    refreshImage = newImage(refreshImageName);
    deleteImage = newImage(deleteImageName);
    packageImage = newImage(packageImageName);
    classImage = newImage(classImageName);
    theoryImage = newImage(theoryImageName);
    classVarImage = newImage(classVarImageName);
    objectVarImage = newImage(objectVarImageName);
    constructorImage = newImage(constructorImageName);
    classMethodImage = newImage(classMethodImageName);
    objectMethodImage = newImage(objectMethodImageName);
    parameterImage = newImage(parameterImageName);
    errorImage = newImage(errorImageName);
    folderOpenImage = newImage(folderOpenImageName);
    folderCloseImage = newImage(folderCloseImageName);
    folderProcessImage = newImage(folderProcessImageName);
    taskNewImage = newImage(taskNewImageName);
    taskOpenImage = newImage(taskOpenImageName);
    taskDoneImage = newImage(taskDoneImageName);
    taskFailedImage = newImage(taskFailedImageName);
    typeImage = newImage(typeImageName);
    formulaImage = newImage(formulaImageName);
    valueImage = newImage(valueImageName);
    analyzeImage = newImage(analyzeImageName);
    verifyImage = newImage(verifyImageName);
    detailImage = newImage(detailImageName);
  }

  /***************************************************************************
   * Get image for okay buttons.
   * @return the image.
   **************************************************************************/
  public Image getOkayImage()
  {
    return okImage;
  }
  
  /***************************************************************************
   * Get image for cancel buttons.
   * @return the image.
   **************************************************************************/
  public Image getCancelImage()
  {
    return cancelImage;
  }
  
  /***************************************************************************
   * Get image for source files.
   * @return the image.
   **************************************************************************/
  public Image getSourceImage()
  {
    return sourceImage;
  }
  
  /***************************************************************************
   * Get image for folders.
   * @return the image.
   **************************************************************************/
  public Image getFolderImage()
  {
    return folderImage;
  }
  
  /***************************************************************************
   * Get image for generic files.
   * @return the image.
   **************************************************************************/
  public Image getFileImage()
  {
    return fileImage;
  }
  
  /***************************************************************************
   * Get analyze image for cancel buttons.
   * @return the image.
   **************************************************************************/
  public Image getAnalyzeImage()
  {
    return analyzeImage;
  }
  
  /***************************************************************************
   * Get run image.
   * @return the image.
   **************************************************************************/
  public Image getRunImage()
  {
    return runImage;
  }
  
  /***************************************************************************
   * Get refresh image.
   * @return the image.
   **************************************************************************/
  public Image getRefreshImage()
  {
    return refreshImage;
  }
  
  /***************************************************************************
   * Get delete image.
   * @return the image.
   **************************************************************************/
  public Image getDeleteImage()
  {
    return deleteImage;
  }
  
  /***************************************************************************
   * Get package image.
   * @return the image.
   **************************************************************************/
  public Image getPackageImage()
  {
    return packageImage;
  }
  
  /***************************************************************************
   * Get class image.
   * @return the image.
   **************************************************************************/
  public Image getClassImage()
  {
    return classImage;
  }
  
  /***************************************************************************
   * Get theory image.
   * @return the image.
   **************************************************************************/
  public Image getTheoryImage()
  {
    return theoryImage;
  }
  
  /***************************************************************************
   * Get class variable image.
   * @return the image.
   **************************************************************************/
  public Image getClassVarImage()
  {
    return classVarImage;
  }
  
  /***************************************************************************
   * Get object variable image.
   * @return the image.
   **************************************************************************/
  public Image getObjectVarImage()
  {
    return objectVarImage;
  }
  
  /***************************************************************************
   * Get constructor image.
   * @return the image.
   **************************************************************************/
  public Image getConstructorImage()
  {
    return constructorImage;
  }
  
  
  /***************************************************************************
   * Get class method image.
   * @return the image.
   **************************************************************************/
  public Image getClassMethodImage()
  {
    return classMethodImage;
  }
  
  /***************************************************************************
   * Get object method image.
   * @return the image.
   **************************************************************************/
  public Image getObjectMethodImage()
  {
    return objectMethodImage;
  }
  
  /***************************************************************************
   * Get parameter image.
   * @return the image.
   **************************************************************************/
  public Image getParameterImage()
  {
    return parameterImage;
  }
  
  /***************************************************************************
   * Get error image.
   * @return the image.
   **************************************************************************/
  public Image getErrorImage()
  {
    return errorImage;
  }
  
  /***************************************************************************
   * Get process folder image.
   * @return the image.
   **************************************************************************/
  public Image getFolderProcessImage()
  {
    return folderProcessImage;
  }
  
  /***************************************************************************
   * Get new task image.
   * @return the image.
   **************************************************************************/
  public Image getTaskNewImage()
  {
    return taskNewImage;
  }
  
  /***************************************************************************
   * Get open task image.
   * @return the image.
   **************************************************************************/
  public Image getTaskOpenImage()
  {
    return taskOpenImage;
  }
  
  /***************************************************************************
   * Get done task image.
   * @return the image.
   **************************************************************************/
  public Image getTaskDoneImage()
  {
    return taskDoneImage;
  }
  
  /***************************************************************************
   * Get failed task image.
   * @return the image.
   **************************************************************************/
  public Image getTaskFailedImage()
  {
    return taskFailedImage;
  }
  
  /***************************************************************************
   * Get type image.
   * @return the image.
   **************************************************************************/
  public Image getTypeImage()
  {
    return typeImage;
  }
  
  /***************************************************************************
   * Get formula image.
   * @return the image.
   **************************************************************************/
  public Image getFormulaImage()
  {
    return formulaImage;
  }
  
  /***************************************************************************
   * Get value image.
   * @return the image.
   **************************************************************************/
  public Image getValueImage()
  {
    return valueImage;
  }
  
  /***************************************************************************
   * Get detail image.
   * @return the image.
   **************************************************************************/
  public Image getDetailImage()
  {
    return detailImage;
  }
  
  /***************************************************************************
   * Get color for open tasks.
   * @return the color
   **************************************************************************/
  public Color getOpenColor()
  {
    return openColor;
  }
  
  /***************************************************************************
   * Get color for closed tasks.
   * @return the color
   **************************************************************************/
  public Color getClosedColor()
  {
    return closedColor;
  }
  
  /***************************************************************************
   * Get color for open tasks with a closed proof.
   * @return the color
   **************************************************************************/
  public Color getProofColor()
  {
    return proofColor;
  }
  
  /***************************************************************************
   * Get color black.
   * @return the color
   **************************************************************************/
  public Color getBlackColor()
  {
    return blackColor;
  }
  
  /***************************************************************************
   * Get color white.
   * @return the color
   **************************************************************************/
  public Color getWhiteColor()
  {
    return whiteColor;
  }
  
  /***************************************************************************
   * Create buttons for analyze view.
   **************************************************************************/
  public void createAnalyzeButtons()
  {
    if (buttons != null) buttons.setVisible(false);
    buttons = new ToolBar(views,
        SWT.HORIZONTAL | SWT.LEFT_TO_RIGHT | SWT.FLAT);
    
    newButton = new ToolItem(buttons, SWT.PUSH);
    newButton.setImage(newImage);
    newButton.setToolTipText("New File");
    newButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        newFile();
      }
    });
    newButton.setEnabled(true);
    
    openButton = new ToolItem(buttons, SWT.PUSH);
    openButton.setImage(openImage);
    openButton.setToolTipText("Open File");
    openButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        openFile();
      }
    });
    openButton.setEnabled(true);
    
    saveButton = new ToolItem(buttons, SWT.PUSH);
    saveButton.setImage(saveImage);
    saveButton.setToolTipText("Save File");
    saveButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        saveFile();
      }
    });
    CTabItem item = editors.getSelection();
    if (item == null)
      saveButton.setEnabled(false);
    else
    {
      FileInfo info = fileItems.get(item);
      saveButton.setEnabled(info.modified);
    }
    
    new ToolItem(buttons, SWT.SEPARATOR);

    refreshButton = new ToolItem(buttons, SWT.PUSH);
    refreshButton.setImage(refreshImage);
    refreshButton.setToolTipText("Refresh View");
    refreshButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        gui.execute(new Runnable() 
          {
            public void run() 
            {
              refresh();
              gui.terminate();
            }
          });
      }
    });
    refreshButton.setEnabled(true);
    
    // run button currently not required
    runButton = new ToolItem(buttons, SWT.PUSH);
    runButton.setImage(runImage);
    runButton.setToolTipText("Run Program");
    runButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        gui.execute(new Runnable() 
          {
            public void run() 
            {
              executeProgram();
              gui.terminate();
            }
          });
      }
    });
    runButton.setEnabled(true);
  }
  
  /***************************************************************************
   * Perform refresh action.
   **************************************************************************/
  public void refresh()
  {
    processPackages();
    closeTaskFolders();
  }
  
  /***************************************************************************
   * Create buttons for verify view.
   **************************************************************************/
  private void createVerifyButtons()
  {
    if (buttons != null) buttons.setVisible(false);
    buttons = new ToolBar(views, 
        SWT.HORIZONTAL | SWT.LEFT_TO_RIGHT | SWT.FLAT);
    newButton = new ToolItem(buttons, SWT.PUSH);
    newButton.setImage(newImage);
    newButton.setToolTipText("New File");
    newButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        newFile();
      }
    });
    newButton.setEnabled(true);
    
    openButton = new ToolItem(buttons, SWT.PUSH);
    openButton.setImage(openImage);
    openButton.setToolTipText("Open File");
    openButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        openFile();
      }
    });
    openButton.setEnabled(true);
  }
  
  /***************************************************************************
   * Create buttons for detail view.
   **************************************************************************/
  private void createDetailButtons()
  {
    if (buttons != null) buttons.setVisible(false);
    buttons = new ToolBar(views, 
        SWT.HORIZONTAL | SWT.LEFT_TO_RIGHT | SWT.FLAT);
    
    newButton = new ToolItem(buttons, SWT.PUSH);
    newButton.setImage(newImage);
    newButton.setToolTipText("New File");
    newButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        newFile();
      }
    });
    newButton.setEnabled(true);
    
    openButton = new ToolItem(buttons, SWT.PUSH);
    openButton.setImage(openImage);
    openButton.setToolTipText("Open File");
    openButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        openFile();
      }
    });
    openButton.setEnabled(true);
  }
  
  /***************************************************************************
   * Create views.
   **************************************************************************/
  public void createViews()
  {
    views = new CTabFolder(this, 
      SWT.TOP | SWT.RIGHT_TO_LEFT | SWT.MULTI | SWT.BORDER);
    views.setSimple(false);
    views.setSelectionBackground(views.getBackground());
    views.addSelectionListener(      
      new SelectionListener() {
        public void widgetSelected(SelectionEvent e) {
          CTabItem item = (CTabItem)e.item;
          if (item == analyzeView)
          {
            displayAnalyzeView();
          }
          else if (item == verifyView)
          {
            displayVerifyView();
          }
          else if (item == detailView)
          {
            if (directoryNames.isEmpty()) 
            {
              displayAnalyzeView();
              return;
            }
            displayDetailView();
          }
          else
            Main.internalError("unknown view");
        }
        public void widgetDefaultSelected(SelectionEvent e) {
          // widgetSelected(e);
        }
      });
  }
  
  /***************************************************************************
   * Create analyze view.
   **************************************************************************/
  private void createAnalzeView()
  {
    analyzeView = new CTabItem(views, SWT.NONE);
    analyzeView.setImage(analyzeImage);
    analyzeView.setText("Analysis");
    analyzeForm = new SashForm(views, 
      SWT.LEFT_TO_RIGHT | SWT.HORIZONTAL | SWT.SMOOTH);
    analyzeView.setControl(analyzeForm);
  }
  
  /***************************************************************************
   * Create verify view.
   **************************************************************************/
  private void createVerifyView()
  {
    verifyView = new CTabItem(views, SWT.NONE);
    verifyView.setImage(verifyImage);
    verifyView.setText("Verification");
    // verifyForm = new SashForm(views, 
    //   SWT.LEFT_TO_RIGHT | SWT.HORIZONTAL | SWT.SMOOTH);
    verifyForm = 
      new fmrisc.ProofNavigator.SWT.Top(this, views, getShell(),
          SWT.LEFT_TO_RIGHT | SWT.HORIZONTAL | SWT.SMOOTH);
    verifyView.setControl(verifyForm);
  }
    
  /***************************************************************************
   * Create detail view.
   **************************************************************************/
  private void createDetailView()
  {
    detailView = new CTabItem(views, SWT.NONE);
    detailView.setImage(detailImage);
    detailView.setText("Semantics");
    if (detailForm == null) // used for recreating the view
      detailForm = new SashForm(views, 
        SWT.LEFT_TO_RIGHT | SWT.HORIZONTAL | SWT.SMOOTH);
    detailView.setControl(detailForm);
  }
  
  /***************************************************************************
   * Display the analyze view from user thread (setting buttons and menu).
   **************************************************************************/
  public void displayAnalyzeView()
  {
    if (views.getSelection() != analyzeView) views.setSelection(analyzeView);
    shell.setMenuBar(explorerMenu);
    // if (buttons != null) buttons.dispose();
    createAnalyzeButtons();
    views.setTopRight(buttons);
  }
  
  /***************************************************************************
   * Display the analyze view from user thread (setting buttons and menu).
   **************************************************************************/
  public void displayVerifyView()
  {
    if (views.getSelection() != verifyView) views.setSelection(verifyView);
    shell.setMenuBar(verifyForm.getCurrentMenu());
    // if (buttons != null) buttons.dispose();
    createVerifyButtons();
    views.setTopRight(buttons);
  }
  
  /***************************************************************************
   * Display the detail view from user thread (setting buttons and menu).
   **************************************************************************/
  public void displayDetailView()
  {
    if (views.getSelection() != detailView) views.setSelection(detailView);
    shell.setMenuBar(explorerMenu);
    // if (buttons != null) buttons.dispose();
    createDetailButtons();
    views.setTopRight(buttons);
  }
  
  /***************************************************************************
   * Display the tasks associated to the symbol.
   * @param symbol a symbol
   **************************************************************************/
  public void displayTasks(Symbol symbol)
  {
    taskExplorer.display(symbol);
  }
  
  /***************************************************************************
   * Create explorer area.
   **************************************************************************/
  private void createExplorerArea()
  {
    explorers = new CTabFolder(analyzeForm, 
        SWT.TOP | SWT.LEFT_TO_RIGHT | SWT.MULTI | SWT.BORDER);
    explorers.setLayoutData(new GridData(GridData.FILL_BOTH));
    explorers.setSimple(false);
    explorers.setSelectionBackground(explorers.getBackground());
    
    // the file explorer
    fitem = new CTabItem(explorers, SWT.NULL);
    fitem.setText("Files");
    fitem.setImage(folderImage);
    fileExplorer = new FileTree(explorers, this);
    fitem.setControl(fileExplorer.getTree());
    redrawFileExplorer(Main.getPath());
    
    // the symbol explorer
    sitem = new CTabItem(explorers, SWT.NULL);
    sitem.setText("Symbols");
    sitem.setImage(packageImage);
    symbolExplorer = new SymbolTree(explorers, this);
    sitem.setControl(symbolExplorer.getTree());
    
    // the default selection
    explorers.setSelection(fitem);
  }
  
 
  /***************************************************************************
   * Redraw file explorer view for denoted path.
   * @param path the path to be used for redrawing.
   **************************************************************************/
  public void redrawFileExplorer(File[] path)
  {
    Main.setPath(path);
    fileExplorer.redraw(path);
  }
  
  /***************************************************************************
   * Redraw symbol explorer view for denoted package.
   * @param top the top-level package to be used for redrawing.
   **************************************************************************/
  public void redrawSymbolExplorer(final PackageSymbol top)
  {
    gui.update(new Runnable() {
      public void run() {
        symbolExplorer.redraw(top);
        explorers.setSelection(sitem);
      }});
  }
  
  /***************************************************************************
   * Redraw task explorer for denoted package.
   * @param top the top-level package to be used for redrawing.
   **************************************************************************/
  private void redrawTaskExplorer(final PackageSymbol top)
  {
    gui.update(new Runnable() {
      public void run() {
        TaskFolder folder = top.getTasks();
        taskExplorer.redraw(folder);
      }});
  }
  
  /***************************************************************************
   * Close in task tree folders that only contain closed tasks.
   **************************************************************************/
  public void closeTaskFolders()
  {
    gui.update(new Runnable() {
      public void run() {
        taskExplorer.close();
      }});
  }
  
  /***************************************************************************
   * Redraw task explorer.
   **************************************************************************/
  private void redrawTaskExplorer()
  {
    redrawTaskExplorer(topPackage);
  }
  
  /***************************************************************************
   * Redraw open task explorer for denoted package.
   * @param top the top-level package to be used for redrawing.
   **************************************************************************/
  private void redrawOpenExplorer(final PackageSymbol top)
  {
    gui.update(new Runnable() {
      public void run() {
        TaskFolder folder = top.getTasks();
        openExplorer.redraw(folder);
      }});
  }
  
  /***************************************************************************
   * Redraw open task explorer.
   **************************************************************************/
  public void redrawOpenExplorer()
  {
    redrawOpenExplorer(topPackage);
  }
  
  /***************************************************************************
   * Create center area.
   **************************************************************************/
  private void createCenterArea()
  {
    centerForm = new SashForm(analyzeForm, 
      SWT.VERTICAL | SWT.SMOOTH);
    centerForm.setLayout(new GridLayout());
  }
  
  /***************************************************************************
   * Create editor area.
   **************************************************************************/
  private void createEditorArea()
  {
    editors = new CTabFolder(centerForm, 
        SWT.TOP | SWT.LEFT_TO_RIGHT | SWT.MULTI | SWT.BORDER);
    editors.setLayoutData(new GridData(GridData.FILL_BOTH));
    editors.setSimple(false);
    editors.setSelectionBackground(editors.getBackground());
    editors.setMaximizeVisible(true);
    editors.setMinimizeVisible(true);
    editors.addCTabFolder2Listener(new CTabFolder2Adapter() {
      public void minimize(CTabFolderEvent event) {
        editors.setMinimized(true);
        displays.setMaximized(true);
        centerForm.setMaximizedControl(displays);
      }
      public void maximize(CTabFolderEvent event) {
        editors.setMaximized(true);
        displays.setMinimized(true);
        centerForm.setMaximizedControl(editors);
      }
      public void restore(CTabFolderEvent event) {
        editors.setMinimized(false);
        editors.setMaximized(false);
        displays.setMinimized(false);
        displays.setMaximized(false);
        centerForm.setMaximizedControl(null);
      }
      public void close(CTabFolderEvent event) {
        CTabItem item = (CTabItem)event.item;
        boolean okay = closeFile(item);
        if (!okay) event.doit = false;
      }
    });
    editors.addSelectionListener(new SelectionListener() {
      public void widgetSelected(SelectionEvent e) {
        CTabItem item = (CTabItem)e.item;
        FileInfo info = fileItems.get(item);
        saveStatus(info.modified);
      }
      public void widgetDefaultSelected(SelectionEvent e) {
        widgetSelected(e);
      }
    });
  }
  
  /****************************************************************************
   * Display source..
   * @param source the source to be displayed.
   ***************************************************************************/
  public void displaySource(Source source)
  {
    // create editor for source file or reuse existing one.
    File file = source.getFile();
    createEditor(file, true); 
    
    // display item
    FileInfo info = fileNames.get(file.getAbsolutePath());
    editors.setSelection(info.item);
  }
  
  /****************************************************************************
   * Display source at denoted position.
   * @param pos the source position to be displayed.
   * @param marker true if a marker is to be shown at that position.
   ***************************************************************************/
  public void displaySource(SourcePosition pos, boolean marker)
  {
    // create editor for source file or reuse existing one.
    Source source = pos.getSource();
    if (source == null) return;
    File file = source.getFile();
    createEditor(file, true); 
    
    // move to right position and display item
    FileInfo info = fileNames.get(file.getAbsolutePath());
    TextViewer viewer = info.viewer;
    viewer.setTopIndex(pos.getLine());
    editors.setSelection(info.item);
    
    // add marker at denoted position
    addMarker(pos, marker);
  }
  
  /***************************************************************************
   * Create editor for existing source file to be processed.
   * @param file the file to be loaded and processed.
   ***************************************************************************/
  public void createEditor(File file)
  {
    // create editor
    createEditor(file, true);
    
    // process class and display symbols
    FileInfo info = fileNames.get(file.getAbsolutePath());
    ClassSymbol csymbol = getClass(topPackage, info.source);
    if (csymbol != null) processClass(csymbol);
  }
  
  /****************************************************************************
   * Create editor for source file.
   * @param file the file.
   * @param load true iff file content is to be loaded.
   ***************************************************************************/
  // see the comment on "addStyledTextStrategy" below
  @SuppressWarnings("deprecation") 
  private void createEditor(File file, boolean load)
  {
    // make sure right view is displayed
    displayAnalyzeView();
    
    // make sure file is not edited twice 
    String path = file.getAbsolutePath();
    FileInfo info0 = fileNames.get(path);
    if (info0 != null)
    {
      editors.setSelection(info0.item);
      return;
    }
    
    // determine source file
    final Source source = Source.create(file, Main.getError());
    String text = null;
    if (load) text = source.getText();
    
    // create tab item
    CTabItem item = new CTabItem(editors, SWT.CLOSE);
    editors.setSelection(item);
    item.setText(file.getName());
    item.setImage(sourceImage);
    
    // annotation model and access
    AnnotationModel model = new AnnotationModel();
    SourceAnnotationAccess access = new SourceAnnotationAccess();

    // composite ruler on the left 
    CompositeRuler cruler = new CompositeRuler();
    cruler.setModel(model);
    
    // column 0: annotation ruler 
    AnnotationRulerColumn aruler = new AnnotationRulerColumn(model, 16, access);
    aruler.addAnnotationType(ERROR_TYPE);
    cruler.addDecorator(0, aruler);
    
    // column 1: line number ruler
    LineNumberRulerColumn lruler = new LineNumberRulerColumn();
    lruler.setForeground(getDisplay().getSystemColor(SWT.COLOR_DARK_GRAY));
    cruler.addDecorator(1, lruler);
    
    // overview ruler on the right
    SharedTextColors colors = new SharedTextColors();
    OverviewRuler oruler = new OverviewRuler(access, 12, colors);
    oruler.setModel(model);
    oruler.addAnnotationType(ERROR_TYPE);
    oruler.setAnnotationTypeLayer(ERROR_TYPE, 3);
    oruler.setAnnotationTypeColor(ERROR_TYPE, 
      getDisplay().getSystemColor(SWT.COLOR_RED));
    oruler.addHeaderAnnotationType(ERROR_TYPE);
    
    // create editor 
    final ProjectionViewer viewer = 
      new ProjectionViewer(editors, cruler, oruler, true,
        SWT.BORDER | SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL);
    
    // perform registrations
    final FileInfo info = new FileInfo(source, item, viewer, model);
    fileNames.put(source.getFile().getAbsolutePath(), info);
    fileItems.put(item, info);
    
    // configure editor
    if (Source.isTheorySource(file))
      viewer.configure(new TheoryEditorConfiguration(info));
    else
      viewer.configure(new ClassEditorConfiguration(info));
    
    // set document
    Document document = new Document(text);
    viewer.setDocument(document, model);
  
    // create projection support
    ProjectionSupport support = new ProjectionSupport(viewer, access, colors);
    support.install();
    viewer.doOperation(ProjectionViewer.TOGGLE);
    
    // set annotation painter
    // only works with deprecated drawing strategy, 
    // not the newer text strategy, seems to be known bug
    // https://bugs.eclipse.org/bugs/show_bug.cgi?id=206913
    AnnotationPainter painter = new AnnotationPainter(viewer, access);
    // painter.addTextStyleStrategy(UNDERLINE_SQUIGGLE, 
    //     new AnnotationPainter.UnderlineStrategy(SWT.UNDERLINE_SQUIGGLE));
    painter.addDrawingStrategy(UNDERLINE_SQUIGGLE, 
      new AnnotationPainter.SquigglesStrategy());
    painter.addAnnotationType(ERROR_TYPE, UNDERLINE_SQUIGGLE);
    painter.setAnnotationTypeColor(ERROR_TYPE, 
        getDisplay().getSystemColor(SWT.COLOR_RED));
    painter.addAnnotationType(MARKER_TYPE, UNDERLINE_SQUIGGLE);
    painter.setAnnotationTypeColor(MARKER_TYPE, 
        getDisplay().getSystemColor(SWT.COLOR_BLUE));
    viewer.addPainter(painter);
    
    // set hover manager
    AnnotationBarHoverManager manager = 
      new AnnotationBarHoverManager(cruler, viewer, new ErrorHover(model), 
        new ErrorInformationControlCreator());
    manager.install(aruler.getControl());

    // set document partitioner
    IDocumentPartitioner partitioner = 
      new FastPartitioner(new SourcePartitionScanner(), 
          SourcePartitionScanner.PARTITIONS);
    partitioner.connect(document);
    document.setDocumentPartitioner(partitioner);
    
    // set visual interface
    viewer.getTextWidget().setFont(fontMono);
    Control tcontrol = viewer.getControl();
    item.setControl(tcontrol);
    tcontrol.setFocus();
    
    // register listeners
    viewer.addTextListener(new ITextListener() {
      public void textChanged(TextEvent e) {
        // just the folding structure, not the text itself, may have changed
        if (e.getDocumentEvent() == null) return;
        info.modified = true;
        saveStatus(true);
      }
    });
    
    // set context menu for editor
    MenuManager menuManager = new MenuManager();
    final StyledText stext = viewer.getTextWidget();
    Menu menu = menuManager.createContextMenu(stext);
    stext.setMenu(menu);
    menuManager.setRemoveAllWhenShown(true);
    if (Source.isClassSource(source.getFile()))
    {
      menuManager.addMenuListener(new IMenuListener() {
        public void menuAboutToShow(IMenuManager manager) {      
          ClassSymbol csymbol = TopWindow.this.getClass(topPackage, source);
          if (csymbol == null) return;
          final ClassDeclaration cdecl = csymbol.getDeclaration();
          if (cdecl == null) return;
          manager.add(new Action("Print Command Specification") {
            public void run() {
              Point pos = stext.getSelection();
              ITextViewerExtension5 viewer0 = (ITextViewerExtension5)viewer;
              int selX = viewer0.widgetOffset2ModelOffset(pos.x);
              int selY = viewer0.widgetOffset2ModelOffset(pos.y);
              printStatSpecification(cdecl, selX, selY, source.getText());
            }});
          }
        });
      }
    
    // set global status
    saveStatus(false);
    closeItem.setEnabled(true);
  }
  
  /***************************************************************************
   * Print specification of statement in class denoted by selected position.
   * @param cdecl the declaration of the class.
   * @param selX the starting point of the selection (inclusive).
   * @param selY the end point of the selection (exclusive).
   * @param text the text of the source file.
   **************************************************************************/
  private void printStatSpecification(ClassDeclaration cdecl, 
    int selX, int selY, String text)
  {
    TopDeclaration[] decls = cdecl.decls;
    int n = decls.length;
    for (int i=0; i<n; i++)
    {
      TopDeclaration decl = decls[i];
      if (!(decl instanceof ParamDeclaration)) continue;
      ParamDeclaration decl0 = (ParamDeclaration)decl;
      Statement stat = decl0.getBody();
      int pos;
      if (i+1 < n)
        pos = decls[i+1].getPosition().getOffset();
      else
      {
        Source source = decl.getPosition().getSource();
        if (source == null) continue;
        pos = source.getSize();
      }
      boolean printed = printStatSpecification(stat, selX, selY, pos, text);
      if (printed) return;
    }
  }
  
  /***************************************************************************
   * Print specification of statement denoted by selected position.
   * @param stat the statement in which the selected statement may appear.
   * @param selX the starting point of the selection (inclusive).
   * @param selY the end point of the selection (exclusive).
   * @param next the offset of the next statement after the current one.
   * @param text the text of the source file.
   * @return true iff the selected statement was found and printed.
   **************************************************************************/
  private boolean printStatSpecification(Statement stat,
    int selX, int selY, int statY, String text)
  {
    // the statement has range [statX, statY0[
    // eat white space from the end of the statement
    int statX = stat.getPosition().getOffset();
    int statY0 = statY;
    while (statY0 >= statX && Character.isWhitespace(text.charAt(statY0-1)))
      statY0 = statY0-1;
    
    // the selection has range [selX, selY0[
    // a zero range is extended to the end of the statement
    int selY0 = selY; 
    if (selY0 == selX) selY0 = statY0;
    
    // the statement is excluded by the selection
    if (selY0 <= statX || selX >= statY0) return false;
    
    // the statement is fully captured by the selection
    if (selX <= statX && statY0 <= selY0)
    {
      ASTPrinter.print(stat, consoleOutput, ASTPrinter.PrintStatementSpec.TOP, true);
      consoleOutput.println();
      return true;
    }
    
    // a substatement may be captured by the selection
    if (stat instanceof BlockStatement)
    {
      BlockStatement stat0 = (BlockStatement)stat;
      Statement[] stats = stat0.stats;
      int n = stats.length;
      for (int i=0; i<n; i++)
      {
        Statement s = stats[i];
        int pos = statY;
        if (i+1 < n)
          pos = stats[i+1].getPosition().getOffset();
        boolean printed = printStatSpecification(s, selX, selY, pos, text);
        if (printed) return true;
      }
    }
    else if (stat instanceof ConditionalStatement)
    {
      ConditionalStatement stat0 = (ConditionalStatement)stat;
      int pos = statY;
      if (stat0.elsestat != null)
        pos = stat0.elsestat.getPosition().getOffset();
      boolean printed = printStatSpecification(stat0.thenstat, 
          selX, selY, pos, text);
      if (printed) return true;
      if (stat0.elsestat != null)
      {
        printed = printStatSpecification(stat0.thenstat, 
            selX, selY, statY, text);
        if (printed) return true;
      }
    }
    else if (stat instanceof WhileLoopStatement)
    {
      WhileLoopStatement stat0 = (WhileLoopStatement)stat;
      boolean printed = printStatSpecification(stat0.body, 
          selX, selY, statY, text);
      if (printed) return true;
    }
    else if (stat instanceof ForLoopStatement)
    {
      ForLoopStatement stat0 = (ForLoopStatement)stat;
      int pos = stat0.update.getPosition().getOffset();
      boolean printed = printStatSpecification(stat0.init, 
          selX, selY, pos, text);
      if (printed) return true;
      pos = stat0.body.getPosition().getOffset();
      printed = printStatSpecification(stat0.update, selX, selY, pos, text);
      if (printed) return true;
      printed = printStatSpecification(stat0.body, selX, selY, statY, text);
      if (printed) return true;
    }
    else if (stat instanceof TryCatchStatement)
    {
      TryCatchStatement stat0 = (TryCatchStatement)stat;
      int pos = stat0.handlers[0].getPosition().getOffset();
      boolean printed = printStatSpecification(stat0.body, 
          selX, selY, pos, text);
      if (printed) return true;
      Handler[] handlers = stat0.handlers;
      int n = handlers.length;
      for (int i=1; i<n; i++)
      {
        Handler handler = handlers[i];
        pos = statY;
        if (i+1 < n) pos = handlers[i+1].getPosition().getOffset();
        printed = printStatSpecification(handler.body, selX, selY, pos, text);
        if (printed) return true;
      }
    }
    
    // statement is at least partially captured by the selection
    ASTPrinter.print(stat, consoleOutput, ASTPrinter.PrintStatementSpec.TOP, true);
    consoleOutput.println();
    return true;
  }
  
  /****************************************************************************
   * Dispose editor.
   * @param item the tab item containing the editor to be disposed.
   ***************************************************************************/
  private void disposeEditor(CTabItem item)
  {
    FileInfo info = fileItems.get(item);
    fileNames.remove(info.source.getFile().getAbsolutePath());
    fileItems.remove(item);
    item.dispose();
  }
    
  /***************************************************************************
   * Create display area.
   **************************************************************************/
  private void createDisplayArea()
  {
    displays = new CTabFolder(centerForm, 
        SWT.TOP | SWT.LEFT_TO_RIGHT | SWT.MULTI | SWT.BORDER);
    displays.setLayoutData(new GridData(GridData.FILL_BOTH));
    displays.setSimple(false);
    displays.setSelectionBackground(displays.getBackground());
    displays.setMaximizeVisible(true);
    displays.setMinimizeVisible(true);
    displays.addCTabFolder2Listener(new CTabFolder2Adapter() {
      public void minimize(CTabFolderEvent event) {
        displays.setMinimized(true);
        editors.setMaximized(true);
        centerForm.setMaximizedControl(editors);
      }
      public void maximize(CTabFolderEvent event) {
        displays.setMaximized(true);
        editors.setMinimized(true);
        centerForm.setMaximizedControl(displays);
      }
      public void restore(CTabFolderEvent event) {
        displays.setMinimized(false);
        displays.setMaximized(false);
        editors.setMinimized(false);
        editors.setMaximized(false);
        centerForm.setMaximizedControl(null);
      }
    });
  }
  
  /***************************************************************************
   * Create display item.
   * @param header the text in the display header.
   * @param image the image in the display header.
   * @param control the control for the display.
   * @return the tab item for the display.
   **************************************************************************/
  private CTabItem createDisplayItem(String header, Image image, Control control)
  {
    CTabItem display = new CTabItem(displays, SWT.NONE);
    display.setText(header);
    display.setImage(image);
    display.setControl(control);
    return display;
  }
  
  /****************************************************************************
   * Create console.
   ***************************************************************************/
  private void createConsole()
  {
    console = new Text(displays, SWT.BORDER | SWT.MULTI | SWT.V_SCROLL | 
      SWT.WRAP | SWT.READ_ONLY);
    console.setFont(fontMono);
    consoleOutput = new PrintWriter(new TextOutputStream(console), true);
    Main.setOutput(consoleOutput);
    CTabItem item = 
      createDisplayItem("Console", consoleImage, console);
    displays.setSelection(item);
  }
  
  /***************************************************************************
   * Get console output writer.
   * @return the console output writer.
   **************************************************************************/
  public PrintWriter getConsole()
  {
    return consoleOutput;
  }
  
  /***************************************************************************
   * Create task area
   **************************************************************************/
  private void createTaskArea()
  {
    // create tab folder
    tasks = new CTabFolder(analyzeForm, 
        SWT.TOP | SWT.LEFT_TO_RIGHT | SWT.MULTI | SWT.BORDER);
    tasks.setLayoutData(new GridData(GridData.FILL_BOTH));
    tasks.setSimple(false);
    tasks.setSelectionBackground(tasks.getBackground());
    
    // create folder menu
    final Menu menu = new Menu(shell, SWT.POP_UP);
    tasks.setMenu(menu);
    menu.addListener (SWT.Show, new Listener () {
      public void handleEvent (Event event) {
        MenuItem [] menuItems = menu.getItems ();
        int n = menuItems.length;
        for (int i=0; i<n; i++) {
          menuItems [i].dispose ();
        }
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Execute All Tasks");
          menuItem.setImage(getRunImage());
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              TaskFolder tasks = topPackage.getTasks();
              tasks.resume();
              taskExplorer.redraw(tasks);
            }
          });
        }
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Show All Tasks");
          menuItem.setImage(taskOpenImage);
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              redrawTaskExplorer(topPackage);
            }
          });
        }
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Hide Completed Tasks");
          menuItem.setImage(taskDoneImage);
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              taskExplorer.close();
            }
          });
        }
      }});
    
    // the task explorer tab
    CTabItem titem = new CTabItem(tasks, SWT.NULL);
    titem.setText("All Tasks");
    titem.setImage(folderImage);
    taskExplorer = new TaskTree(tasks, this);
    titem.setControl(taskExplorer.getTree());
    
    // the open tasks tab
    CTabItem oitem = new CTabItem(tasks, SWT.NULL);
    oitem.setText("Open Tasks");
    oitem.setImage(taskOpenImage);
    openExplorer = new TaskList(tasks, this);
    oitem.setControl(openExplorer.getTree());
    
    // the default selection
    tasks.setSelection(titem);
  }
  
  /**************************************************************************
   * Redraw node to indicate status of task.
   * @param node the node.
   * @param task the task.
   * @param color true if text is to be colored.
   *************************************************************************/
  public void redraw(TreeItem node, Task task, boolean color)
  {
    Task.Status status = task.getStatus();
    node.setBackground(whiteColor);
    node.setForeground(blackColor);
      
    // set colors
    if (color && !(task instanceof TaskFolder))
    {
      if (status == Task.Status.DONE)
      {
        if (task instanceof TaskFolder)
          node.setForeground(blackColor);
        else
          node.setForeground(closedColor);
      }
      else if (task instanceof ProofNavigatorTask)
      {
        ProofNavigatorTask task0 = (ProofNavigatorTask)task;
        ProofNavigatorProblem problem = task0.proofNavigatorProblem();
        fmrisc.ProofNavigator.Proving.Proof proof = problem.getProof();
        if (proof != null && proof.isClosed() && proof.isTrusted() && proof.isAbsolute())
          node.setForeground(proofColor);
        else
          node.setForeground(openColor);
      }
      else
      {
        node.setForeground(openColor);
      }
    }
    
    // set images for folders
    if (task instanceof TaskFolder) 
    {
      if (status == Task.Status.FAILED)
        node.setImage(getTaskFailedImage());
      else if (status == Task.Status.DONE)
        node.setImage(getFolderImage());
      else
        node.setImage(getFolderProcessImage());
      return;
    }
    
    // set images for tasks
    if (status == Task.Status.NEW)
      node.setImage(getTaskNewImage());
    else if (status == Task.Status.OPEN || status == Task.Status.ALMOST)
      node.setImage(getTaskOpenImage());
    else if (status == Task.Status.DONE)
      node.setImage(getTaskDoneImage());
    else if (status == Task.Status.FAILED)
      node.setImage(getTaskFailedImage());
    else if (status == Task.Status.REVOKED)
      node.setImage((Image)null);
    else
      Main.internalError("unknown task status");
  }
  
  /***************************************************************************
   * Create editor area.
   **************************************************************************/
  private void createBrowserArea()
  {
    browsers = new CTabFolder(detailForm, 
        SWT.TOP | SWT.LEFT_TO_RIGHT | SWT.MULTI | SWT.BORDER);
    browsers.setLayoutData(new GridData(GridData.FILL_BOTH));
    browsers.setSimple(false);
    browsers.setSelectionBackground(editors.getBackground());
    //browsers.setMaximizeVisible(true);
    //browsers.setMinimizeVisible(true);
    browsers.addCTabFolder2Listener(new CTabFolder2Adapter() {
      public void close(CTabFolderEvent event) {
        CTabItem item = (CTabItem)event.item;
        String name = directoryNames.get(item);
        if (name != null) gui.deleteVirtualDirectory(name);    
        directoryNames.remove(item);
        if (directoryNames.isEmpty()) displayAnalyzeView();
        ParamSymbol symbol = displayedMethods.get(item);
        if (symbol != null)
        {
          displayedMethods.remove(item);
          methodItems.remove(symbol);
        }
      }
    });
  }
  
  /****************************************************************************
   * Create browser to display the content of the denoted virtual directory.
   * @param name the name of the virtual directory.
   * @param title the title to be used for the display.
   * @retrun the created item
   ***************************************************************************/
  private CTabItem createVirtualDirectoryBrowser(String name, String title)
  {
    CTabItem item = 
      createBrowser("http://localhost:" + gui.getVirtualDirectoryPort() + 
        "/" + name + "/" + VirtualDirectory.MAIN, title);
    directoryNames.put(item, name);
    return item;
  }
  
  /****************************************************************************
   * Create browser to display the content of the denoted URL.
   * @param url the URL.
   * @returns the tab item of the browser
   ***************************************************************************/
  private CTabItem createBrowser(String url, String title)
  {      
    // create tab item
    CTabItem item = new CTabItem(browsers, SWT.CLOSE);
    browsers.setSelection(item);
    item.setText(title);
    item.setImage(sourceImage);
   
    // create browser and register it
    Browser browser = new Browser(browsers, SWT.WEBKIT | SWT.BORDER);
    browser.setBackground(whiteColor);
    browser.setUrl(url);
    item.setControl(browser);

    // display browser
    displayDetailView();
    browser.setFocus();
    return item;
  }

  // --------------------------------------------------------------------------
  //
  // Application logic.
  //
  // --------------------------------------------------------------------------
    
  /***************************************************************************
   * Process packages in workspace.
   **************************************************************************/
  public void processPackages()
  {
    getDisplay().syncExec(new Runnable() {
      public void run() {
        if (checkOpenProof()) return;
        Main.resetState();
        signalRunning(true);
        taskExplorer.hide();
        topPackage = Main.processPackages();
        signalRunning(false);
        redrawExplorers();
        // closeTaskFolders();
      }});
  }
  
  /***************************************************************************
   * Get top package.
   * @return the top package.
   **************************************************************************/
  public PackageSymbol getTopPackage()
  {
    return topPackage;
  }
  
  /***************************************************************************
   * Process packages in workspace.
   **************************************************************************/
  public void redrawExplorers()
  {
    // redraw explorers
    redrawSymbolExplorer(topPackage);
    redrawTaskExplorer();
    redrawOpenExplorer();
    
    // attempt new tasks (activity is displayed)
    topPackage.getTasks().resume();
  }
  
  /***************************************************************************
   * Signal run status of background process.
   * @param running if we are waiting for background process to complete.
   **************************************************************************/
  private void signalRunning(final boolean running)
  {
    final TopWindow top = this;
    getDisplay().syncExec(new Runnable() {
      public void run() {
        top.setEnabled(!running);
        if (running)
        {
          setCursor(waitCursor);
          editors.setCursor(waitCursor);
          for (CTabItem editor : editors.getItems()) 
            fileItems.get(editor).viewer.getTextWidget().setCursor(waitCursor);
          console.setCursor(waitCursor);
        }
        else
        {
          setCursor(arrowCursor);
          editors.setCursor(arrowCursor);
          for (CTabItem editor : editors.getItems()) 
            fileItems.get(editor).viewer.getTextWidget().setCursor(ibeamCursor);
          console.setCursor(arrowCursor);
        }
      }});
  }
  
  /***************************************************************************
   * Process denoted class.
   * @param csymbol the class to be processed.
   **************************************************************************/
  public void processClass(ClassSymbol csymbol)
  {
    if (checkOpenProof()) return;
    Main.resetState();
    Set<ParamSymbol> methods = methodItems.keySet();
    Set<ParamSymbol> methods0 = new SetClass<ParamSymbol>();
    for (ParamSymbol method : methods)
    {
      // clone needed since we must not modify base collection while iterating
      if (csymbol == method.getClassSymbol()) 
        methods0.add(method);
    }
    for (ParamSymbol method : methods0)
    {
      CTabItem item = methodItems.get(method);
      item.dispose();
      methodItems.remove(method);
      directoryNames.remove(item);
      displayedMethods.remove(item);
    }
    TheorySymbol tsymbol = csymbol.getLocalTheory();
    if (tsymbol != null) symbolExplorer.remove(tsymbol);
    signalRunning(true);
    taskExplorer.hide();
    Main.processClass(csymbol);
    boolean error = !csymbol.isRead();
    markSource(csymbol.getSource(), error);
    if (error) 
      MethodTasks.clearTasks();
    else
      MethodTasks.generateTasks(consoleOutput);
    processDependencies(csymbol, error);
    signalRunning(false);
    // consoleOutput.print("still executing...");
    // consoleOutput.flush();
    redrawExplorers();
    consoleOutput.println("done.");
  }

  /***************************************************************************
   * Process denoted theory.
   * @param tsymbol the theory to be processed.
   **************************************************************************/
  public void processTheory(TheorySymbol tsymbol)
  {
    if (checkOpenProof()) return;
    Main.resetState();
    signalRunning(true);
    taskExplorer.hide();
    if (tsymbol.isFromClass())
      Main.processClass(tsymbol.getClassSymbol());
    else
      Main.processTheory(tsymbol);
    boolean error = !tsymbol.isRead();
    Source source = tsymbol.getSource();
    if (source != null) markSource(tsymbol.getSource(), error);
    if (error) 
      MethodTasks.clearTasks();
    else
      MethodTasks.generateTasks(consoleOutput);
    processDependencies(tsymbol, error);
    signalRunning(false);
    redrawExplorers();
  }
  
  /***************************************************************************
   * Process units that have been loaded and depend on the unit denoted by symbol.
   * @param symbol a symbol denoting a unit (class/theory)
   * @param serror true iff an error has occurred in the processing of symbol
   **************************************************************************/
  private void processDependencies(Symbol symbol, boolean serror)
  {
    Collection<Symbol> symbols = symbolExplorer.getUnitSymbols();
    Collection<Symbol> dep = ReferencedUnits.filterDependencies(symbol, symbols);
    dep.remove(symbol); // prevent duplicate processing
    
    // reset all symbols such that we do not have to process them in
    // topological order of dependencies
    for (Symbol s : dep)
    {
      if (s instanceof ClassSymbol)
      {
        ClassSymbol s0 = (ClassSymbol)s;
        TheorySymbol tsymbol = s0.getLocalTheory();
        if (tsymbol != null) symbolExplorer.remove(tsymbol);
        s0.reset();
      }
      else if (s instanceof TheorySymbol)
      {
        TheorySymbol s0 = (TheorySymbol)s;
        if (!s0.isFromClass()) s0.reset();
      }
      else
        Main.internalError("unknown symbol");
    }
    
    // process symbols
    ErrorStream error = Main.getError();
    for (Symbol s : dep)
    {
      if (s instanceof ClassSymbol)
      {
        ClassSymbol s0 = (ClassSymbol)s;
        if (serror)
        {
          error.newMessage("class " + s0.getLongName() + " depends on failed unit.");
          symbolExplorer.signalError(s0);
          markSource(s0.getSource(), true);
          continue;
        }
        error.startMessage("processing class " + s0.getLongName() + "...");
        Main.processClass(s0);
        boolean errors = !s0.isRead();
        markSource(s0.getSource(), errors);
        if (errors) 
          MethodTasks.clearTasks();
        else
          MethodTasks.generateTasks(consoleOutput);
        error.endMessage("done.");
        continue;
      }
      if (s instanceof TheorySymbol)
      {
        TheorySymbol s0 = (TheorySymbol)s;
        if (s0.isFromClass()) continue;
        if (serror)
        {
          error.newMessage("theory " + s0.getLongName() + " depends on failed unit.");
          symbolExplorer.signalError(s0);
          markSource(s0.getSource(), true);
          continue;
        }
        error.startMessage("processing theory " + s0.getLongName() + "...");
        Main.processTheory(s0);
        boolean errors = !s0.isRead();
        Source source = s0.getSource();
        if (source != null) markSource(source, errors);
        if (errors) 
          MethodTasks.clearTasks();
        else
          MethodTasks.generateTasks(consoleOutput);
        error.endMessage("done.");
        continue;
      }
      Main.internalError("unknown symbol");
    }
  }
  
  /***************************************************************************
   * Compile and execute program as specified in the "properties" dialog.
   **************************************************************************/
  private void executeProgram()
  {
    String[] properties = Main.getProperties();
    String compileCommand = properties[0];
    String executeCommand = properties[1];
    String workingDirectory = properties[2];
    String mainClass = properties[3];
    String command1 = compileCommand + " " + mainClass + ".java";
    String command2 = executeCommand + " " + mainClass;
    consoleOutput.println(command1);
    boolean okay = executeCommand(command1, workingDirectory);
    if (!okay) return; 
    consoleOutput.println(command2);
    okay = executeCommand(command2, workingDirectory);
    consoleOutput.println("done.");
  }
  
  /****************************************************************************
   * Execute command in denoted working directory.
   * @param command the command.
   * @param dir the working directory.
   * @return true iff everything went fine.
   ***************************************************************************/
  private boolean executeCommand(String command, String dir)
  {
    try
    {
      Process process = 
        Runtime.getRuntime().exec(command, 
            new String[] { }, new File(dir));
      PrintThread ethread = new PrintThread(process.getErrorStream());
      PrintThread othread = new PrintThread(process.getInputStream());
      ethread.start();
      othread.start();
      ethread.join();
      othread.join();
      return !ethread.error;
    }
    catch(IOException e)
    {
      consoleOutput.println(e.getMessage());
      return false;
    }
    catch(InterruptedException e)
    {
      consoleOutput.println(e.getMessage());
      return false;
    }
  }
  
  // thread that prints content of input stream on console window
  // sets flag errors if last line of output contains string "error"
  private final class PrintThread extends Thread
  {
    private InputStream input;
    public boolean error = false;
    public PrintThread(InputStream input)
    {
      this.input = input;
    }
    public void run()
    {
      try
      {
        BufferedReader reader = new BufferedReader(new InputStreamReader(input));
        while (true)
        {
          String line = reader.readLine();
          if (line == null) break;
          consoleOutput.println(line);
          error = line.indexOf("error") != -1;
        }
      }
      catch(IOException e)
      {
        consoleOutput.println(e.getMessage());
        error = true;
      }
    }
  }
  
  /***************************************************************************
   * Open a file in editor window.
   **************************************************************************/
  private void openFile()
  {
    FileDialog dialog = new FileDialog(shell, SWT.OPEN);
    dialog.setText("Open File");
    dialog.open();
    String fileName = dialog.getFileName();
    if (fileName == "") return;
    String filterPath = dialog.getFilterPath();
    File file = new File(filterPath, fileName);
    createEditor(file);
  }
  
  /***************************************************************************
   * Open a new file in editor window.
   **************************************************************************/
  private void newFile()
  {
    FileDialog dialog = new FileDialog(shell, SWT.SAVE);
    dialog.setText("New File");
    dialog.open();
    String fileName = dialog.getFileName();
    if (fileName == "") return;
    String filterPath = dialog.getFilterPath();
    File file = new File(filterPath, fileName);
    createEditor(file, file.exists());
  }
  
  /****************************************************************************
   * Save currently selected file.
   * @return true if file could be saved.
   ***************************************************************************/
  private boolean saveFile()
  {
    CTabItem item = editors.getSelection();
    FileInfo info = fileItems.get(item);
    boolean okay = saveFile(info);
    if (!okay) return false;
    info.modified = false;
    saveStatus(false);
    symbolExplorer.expand(info.source);
    return true;
  }
  
  /***************************************************************************
   * Set save status of current file.
   * @param on true if status is "unsaved", false, otherwise.
   *************************************************************************/
  private void saveStatus(boolean on)
  {
    if (views.getSelection() != analyzeView) return;
    saveButton.setEnabled(on);
    saveItem.setEnabled(on);
    undoItem.setEnabled(on);
  }
  
  /***************************************************************************
   * Save file.
   * @param info the information about the file.
   * @return true if file could be saved.
   **************************************************************************/
  private boolean saveFile(FileInfo info)
  {
    if (checkOpenProof()) return false;
    try
    {
      File file = info.source.getFile();
      PrintWriter writer = new PrintWriter(new FileWriter(file));
      writer.println(info.viewer.getDocument().get());
      writer.close();
      fileExplorer.update(file);
      if (topPackage != null)
      {
        if (Source.isClassSource(info.source.getFile()))
        {
          ClassSymbol csymbol = getClass(topPackage, info.source);
          if (csymbol == null)
          {
            processPackages();
            csymbol = getClass(topPackage, info.source);
            if (csymbol == null) return true;
          }
          processClass(csymbol);
        }
        else if (Source.isTheorySource(info.source.getFile()))
        {
          TheorySymbol tsymbol = getTheory(topPackage, info.source);
          if (tsymbol == null)
          {
            processPackages();
            tsymbol = getTheory(topPackage, info.source);
            if (tsymbol == null) return true;
          }
          processTheory(tsymbol);
        }
      }
    }
    catch(IOException e)
    {
      Main.getError().newMessage("Could not save file: " + e.getMessage());
      return false;
    }
    return true;
  }
  
  
  /***************************************************************************
   * Search recursively in package for class symbol denoted by source.
   * @param psymbol the package to search in recursively.
   * @param source the source of the class.
   * @return the class symbol (or null, if the class could not be found).
   **************************************************************************/
  private ClassSymbol getClass(PackageSymbol psymbol, Source source)
  {
    Collection<ClassSymbol> csymbols = psymbol.getClassSet();
    for (ClassSymbol symbol : csymbols)
    {
      Source csource = symbol.getSource();
      if (source == csource) return symbol;
    }
    Collection<PackageSymbol> psymbols = psymbol.getPackageSet();
    for (PackageSymbol symbol : psymbols)
    {
      ClassSymbol csymbol = getClass(symbol, source);
      if (csymbol != null) return csymbol;
    }
    return null;
  }

  /***************************************************************************
   * Search recursively in package for theory symbol denoted by source.
   * @param psymbol the package to search in recursively.
   * @param source the source of the theory.
   * @return the theory symbol (or null, if the theory could not be found).
   **************************************************************************/
  private TheorySymbol getTheory(PackageSymbol psymbol, Source source)
  {
    Collection<TheorySymbol> tsymbols = psymbol.getTheorySet();
    for (TheorySymbol symbol : tsymbols)
    {
      Source tsource = symbol.getSource();
      if (source == tsource) return symbol;
    }
    Collection<PackageSymbol> psymbols = psymbol.getPackageSet();
    for (PackageSymbol symbol : psymbols)
    {
      TheorySymbol tsymbol = getTheory(symbol, source);
      if (tsymbol != null) return tsymbol;
    }
    return null;
  }
  
  /***************************************************************************
   * Close file (modified files are saved upon user request).
   * @param item the editor tab denoting the file to be closed.
   * @return true if the file was closed.
   **************************************************************************/
  private boolean closeFile(CTabItem item)
  {
    FileInfo info = fileItems.get(item);
    if (info.modified)
    {
      int result = askCancel("Close File",
          "File \"" + info.source.getFile().getName() + 
          "\" has been modified. Save changes?");
      if (result == -1) return false;
      if (result == 1)
      {
        boolean okay = saveFile(info);
        if (!okay) return false;
      }
    }
    disposeEditor(item);
    item = editors.getSelection();
    if (item == null)
    {
      saveStatus(false);
      closeItem.setEnabled(false);
    }
    else
    {
      info = fileItems.get(item);
      saveStatus(info.modified);
    }
    return true;
  }
  
  /****************************************************************************
   * Close file possibly edited in window
   * @param file that is possibly edited
   * @return true if no editor window holds file any more.
   **************************************************************************/
  public boolean closeFile(File file)
  {
    FileInfo info = fileNames.get(file.getAbsolutePath());
    if (info == null) return true;
    boolean okay = closeFile(info.item);
    return okay;
  }
  
  /***************************************************************************
   * Close all open files (modified files after confirmation).
   * @return true if all files were closed.
   **************************************************************************/
  private boolean closeAll()
  {
    CTabItem[] items = editors.getItems();
    boolean closed = true;
    for (CTabItem item : items)
    {
      boolean okay = closeFile(item);
      if (!okay) closed = false;
    }
    return closed;
  }
  
  /****************************************************************************
   * Delete file/directory (recursively).
   * @param file file or directory to be deleted after confirmation
   ***************************************************************************/
  public void delete(File file)
  {
    String name = file.getName();
    String question;
    if (file.isDirectory())
      question = "Delete directory " + name + " and all its contents?";
    else
      question = "Delete file " + name + "?";
    boolean okay = ask("Delete Confirmation", question);
    if (!okay) return;
    deleteCore(file);
  }
  
  /***************************************************************************
   * Delete file/directory (recursively).
   * @param file the file/directory to be (recursively) deleted.
   * @return true if everything could be deleted.
   **************************************************************************/
  private boolean deleteCore(File file)
  {
    boolean okay = true;
    
    if (file.isDirectory())
    {
      // delete directory content
      File[] files = file.listFiles();
      for (File f : files)
      {
        boolean okay0 = deleteCore(f);
        okay = okay && okay0;
      }
    }
    else
    {
      // close possibly open editor window
      okay = closeFile(file);
    }
    if (!okay) return false;
    
    // delete file and remove it from explorer view
    okay = file.delete();
    if (okay) fileExplorer.delete(file);  
    
    // process packages again
    processPackages();
    return okay;
  }
  
  /***************************************************************************
   * Add annotation with given text at specified source position.
   * Has no effect, if source is not displayed in an editor.
   * @param pos the position.
   * @param text the text.
   **************************************************************************/
  public void addAnnotation(SourcePosition pos, String text)
  {
    Source source = pos.getSource();
    if (source == null) return;
    String name = source.getFile().getAbsolutePath();
    FileInfo info = fileNames.get(name);
    if (info == null) return;
    IAnnotationModel model = info.viewer.getAnnotationModel();
    ErrorAnnotation annotation = new ErrorAnnotation(text, pos.getLine());
    Position p = new Position(pos.getOffset(), pos.getLength());
    model.addAnnotation(annotation, p);
    info.item.setImage(errorImage);
  }
  
  /***************************************************************************
   * Mark source to indicate whether it has/has not errors.
   * Has no effect, if source is not displayed in an editor.
   * @param source the source (may be null, indicating no source).
   * @param error true iff source has error
   **************************************************************************/
  public void markSource(Source source, boolean error)
  {
    if (source == null) return;
    String name = source.getFile().getAbsolutePath();
    FileInfo info = fileNames.get(name);
    if (info == null) return;
    if (error)
      info.item.setImage(errorImage);
    else
      info.item.setImage(sourceImage);
  }
  
  /***************************************************************************
   * Remove all annotations from denoted source.
   * @param source the source from which to remove the annotations
   *        (may be null indicating no source)
   **************************************************************************/
  public void clearAnnotations(Source source)
  {
    if (source == null) return;
    String name = source.getFile().getAbsolutePath();
    FileInfo info = fileNames.get(name);
    if (info == null) return;
    IAnnotationModel model = info.viewer.getAnnotationModel();
    if (!(model instanceof AnnotationModel)) return;
    AnnotationModel model0 = (AnnotationModel)model;
    model0.removeAllAnnotations();
    info.item.setImage(sourceImage);
  }
  
  /***************************************************************************
   * Add marker at specified source position.
   * Has no effect, if source is not displayed in an editor.
   * @param pos the position.
   * @param marker iff false only previous marker is erased
   **************************************************************************/
  public void addMarker(SourcePosition pos, boolean marker)
  {
    Source source = pos.getSource();
    if (source == null) return;
    String name = source.getFile().getAbsolutePath();
    FileInfo info = fileNames.get(name);
    if (info == null) return;
    IAnnotationModel model = info.viewer.getAnnotationModel();
    if (info.marker != null) model.removeAnnotation(info.marker);
    info.marker = null;
    if (!marker) return;
    MarkerAnnotation annotation = new MarkerAnnotation();
    Position p = new Position(pos.getOffset(), pos.getLength());
    model.addAnnotation(annotation, p);
    info.marker = annotation;
  }
  
  /****************************************************************************
   * Print symbol on console.
   * @param symbol the symbol to be printed.
   ***************************************************************************/
  public void print(Symbol symbol)
  {
    PrintWriter console = getConsole();
    symbol.print(console);
    console.print("\n");
    console.flush();
  }
  
  /****************************************************************************
   * Print declaration/definition of symbol on console.
   * @param symbol the symbol whose declaration/definition is to be printed.
   ***************************************************************************/
  public void printDecl(Symbol symbol)
  {
    PrintWriter console = getConsole();
    Declaration decl = symbol.getDeclaration();
    if (decl == null)
      console.print("(no declaration)");
    else
      decl.print(console);
    console.print("\n");
    console.flush();
  }
  
  /****************************************************************************
   * Get current font size.
   * @return the font size.
   ***************************************************************************/
  public int getFontSize()
  {
    if (smallFonts) 
      return smallFont;
    else
      return bigFont;
  }
  
  /****************************************************************************
   * Show details of method body.
   * @param symbol the method whose details are to be shown.
   ***************************************************************************/
  public void showMethodDetail(ParamSymbol symbol)
  {
    int fontSize = getFontSize();
    int lineWidth = smallFonts ? smallWidth : bigWidth;
    VirtualDirectory directory = 
      Presenter.present(symbol, fontSize, lineWidth, gui.getCommandServerPort());
    String name = gui.enterVirtualDirectory(directory);
    CTabItem item = methodItems.get(symbol);
    if (item != null)
    {
      displayDetailView();
      browsers.setSelection(item);
      return;
    }
    item = createVirtualDirectoryBrowser(name, 
        symbol.getClassSymbol().getName().toString()+"."+symbol.getName().toString());
    methodItems.put(symbol, item);
    displayedMethods.put(item, symbol);
    /*
    StatJudgement judgement = symbol.getBodyJudgement();
    if (judgement != null)
    {
      PrintWriter out = getConsole();
      PrettyPrinter.print(judgement, 70, out);
      out.println();
    }
    */
  }
  
  
  /***************************************************************************
   * Propagate condition through method body.
   * @param method the method.
   * @param statement the statement to which the condition applies.
   * @param formula the formula representing the condition.
   * @param pre true if condition applies to prestate
   **************************************************************************/
  public void propagateCondition(ParamSymbol method,
    Statement statement, Formula formula, boolean pre)
  {
    // derive mapping of body statements to formulas
    Statement body = method.getDeclaration().getBody();
    Map<Statement, Formula> preMap = new MapClass<Statement, Formula>();
    Map<Statement, Formula> postMap = new MapClass<Statement, Formula>();
    Satisfies satisfies = method.getSatisfies();
    if (satisfies == null) Main.internalError("no judgement processor");
    Propagator.execute(body, statement, formula, pre, preMap, postMap, satisfies);
    
    // present the mapping in a document
    int fontSize = getFontSize();
    int lineWidth = smallFonts ? smallWidth : bigWidth;
    VirtualDirectory directory = 
      Presenter.present(method, statement, formula, pre,
          preMap, postMap,
          fontSize, lineWidth, gui.getCommandServerPort());
    final String name = gui.enterVirtualDirectory(directory);
    final ParamSymbol method0 = method;
    gui.update(new Runnable() 
    {
      public void run() 
      {
        createVirtualDirectoryBrowser(name, 
            method0.getClassSymbol().getName().toString() + "." + 
            method0.getName().toString() + ": State Conditions");
      }
    });
  }
  
  //-------------------------------------------------------------------------
  //
  // helper classes
  //
  // -------------------------------------------------------------------------
  
  //-------------------------------------------------------------------------
  // the information stored for each file being edited
  //-------------------------------------------------------------------------
  private class FileInfo
  {
    Source source;           // the source entry
    CTabItem item;           // the tab item for the file
    SourceViewer viewer;     // the editor for the file
    IAnnotationModel model;  // the annotation model
    MarkerAnnotation marker; // the marker set (may be null)
    
    boolean modified;       // the modification status
    
    public FileInfo(Source source, CTabItem item, SourceViewer viewer,
      IAnnotationModel model)
    {
      this.source = source;
      this.item = item;
      this.viewer = viewer;
      this.model = model;
      this.modified = false;
      this.marker = null;
    }
  }

  //-------------------------------------------------------------------------
  // the undo manager used for text viewers
  //-------------------------------------------------------------------------
  private class MyUndoManager extends TextViewerUndoManager
  {
    public MyUndoManager(int level)
    {
      super(level);
    }
    public void undo()
    {
      super.undo();
      if (!undoable())
      {
        saveStatus(false);
      }
    }
  }
    
  //-------------------------------------------------------------------------
  // my word detector (identifier-like)
  //-------------------------------------------------------------------------
  private class WordDetector implements IWordDetector 
  {
    public boolean isWordPart(char ch) 
    {
      return Character.isJavaIdentifierPart(ch);
    }
    public boolean isWordStart(char ch)  
    {
      return Character.isJavaIdentifierStart(ch);
    }       
   }
  
  //-------------------------------------------------------------------------
  // my word detector (extended version)
  //-------------------------------------------------------------------------
  private class ExtendedWordDetector implements IWordDetector 
  {
    public boolean isWordPart(char ch) 
    {
      return Character.isJavaIdentifierPart(ch)
          || ch == '<' || ch == '=' || ch == '>' ;
    }
    public boolean isWordStart(char ch)  
    {
      return Character.isJavaIdentifierStart(ch) ||
             ch == '<' || ch == '=';
    }       
   }
  
  //-------------------------------------------------------------------------
  // the source partition scanner: 
  // - comment partitions for appropriate handling of comments
  // - literal partitions for avoiding highlighting inside string literals
  // - specification literals as a special kind of comments
  //-------------------------------------------------------------------------
  private static class SourcePartitionScanner extends RuleBasedPartitionScanner
  {
    public final static String COMMENT = "comment";
    public final static String LITERAL = "literal";
    public final static String SPEC = "specification"; 
    public final static String[] PARTITIONS = { COMMENT, LITERAL, SPEC } ;
    
    public SourcePartitionScanner()
    {
      // single/multi-line comments
      IToken comment = new Token(COMMENT);
      EndOfLineRule scomments = new EndOfLineRule("//", comment);
      MultiLineRule mcomments = new MultiLineRule("/*", "*/", comment);
      
      // string/character literals
      IToken literal = new Token(LITERAL);
      SingleLineRule strings = new SingleLineRule("\"", "\"", literal, '\\');
      SingleLineRule chars = new SingleLineRule("'", "'", literal, '\\');
      
      // specification comments
      IToken scomment = new Token(SPEC);
      EndOfLineRule sscomments = new EndOfLineRule("//@", scomment);
      MultiLineRule mscomments = new MultiLineRule("/*@", "@*/", scomment);
      
      setPredicateRules(new IPredicateRule[] { 
        sscomments, mscomments, scomments, mcomments, strings, chars });
    }
  }
   
  // the comment scanner
  private class CommentScanner extends RuleBasedScanner
  {
    public CommentScanner()
    {
      // get current display
      Display display = Display.getCurrent();
      
      // define token
      Color commentColor = display.getSystemColor(SWT.COLOR_DARK_GREEN);
      IToken commentToken = new Token(new TextAttribute(commentColor));
 
      // set token
      this.setDefaultReturnToken(commentToken);
    }
  }
  
  // the string literal scanner
  private class LiteralScanner extends RuleBasedScanner
  {
    public LiteralScanner()
    {
      // get current display
      Display display = Display.getCurrent();
      
      // define token
      Color literalColor = display.getSystemColor(SWT.COLOR_DARK_RED);
      IToken literalToken = new Token(new TextAttribute(literalColor));
 
      // set token
      this.setDefaultReturnToken(literalToken);
    }
  }
  
  // the specification scanner
  private class SpecScanner extends RuleBasedScanner
  {
    public SpecScanner()
    {
      // get current display
      Display display = Display.getCurrent();
      
      // define token
      Color scommentColor = display.getSystemColor(SWT.COLOR_BLUE);
      IToken scommentToken = new Token(new TextAttribute(scommentColor));
 
      // set token
      setDefaultReturnToken(scommentToken);
    }
  }
  
  //-------------------------------------------------------------------------
  // the syntax scanner for classes
  //-------------------------------------------------------------------------
  private class ClassSyntaxScanner extends RuleBasedScanner
  {
    public ClassSyntaxScanner()
    {
      // get current display
      Display display = Display.getCurrent();
      
      // define tokens
      Color keywordColor = display.getSystemColor(SWT.COLOR_DARK_MAGENTA);
      IToken keywordToken = new Token(new TextAttribute(keywordColor));
      Color typeColor = display.getSystemColor(SWT.COLOR_DARK_MAGENTA);
      IToken typeToken = new Token(new TextAttribute(typeColor));
      
      // define literal constants as keywords
      WordRule consts = new WordRule(new WordDetector());
      consts.addWord("true", keywordToken);
      consts.addWord("false", keywordToken);
      consts.addWord("null", keywordToken);
      
      // define keywords
      WordRule keywords = new WordRule(new WordDetector());
      keywords.addWord("public", keywordToken);
      keywords.addWord("private", keywordToken);
      keywords.addWord("package", keywordToken);
      keywords.addWord("import", keywordToken);
      keywords.addWord("static", keywordToken);
      keywords.addWord("class", keywordToken);
      keywords.addWord("if", keywordToken);
      keywords.addWord("else", keywordToken);
      keywords.addWord("while", keywordToken);
      keywords.addWord("for", keywordToken);
      keywords.addWord("continue", keywordToken);
      keywords.addWord("break", keywordToken);
      keywords.addWord("return", keywordToken);
      keywords.addWord("throw", keywordToken);
      keywords.addWord("try", keywordToken);
      keywords.addWord("catch", keywordToken);
      
      // define builtin types
      WordRule types = new WordRule(new WordDetector());
      types.addWord("boolean", typeToken);
      types.addWord("int", typeToken);
      types.addWord("char", typeToken);
      types.addWord("String", typeToken);
      types.addWord("void", typeToken);
      
      // set rules
      IRule[] rules = 
        new IRule[] { consts, keywords, types } ;
      setRules(rules);
    }
  }
  
  //-------------------------------------------------------------------------
  // the syntax scanner for theories
  //-------------------------------------------------------------------------
  private class TheorySyntaxScanner extends RuleBasedScanner
  {
    public TheorySyntaxScanner()
    {
      // get current display
      Display display = Display.getCurrent();
      
      // define tokens
      Color keywordColor = display.getSystemColor(SWT.COLOR_DARK_MAGENTA);
      IToken keywordToken = new Token(new TextAttribute(keywordColor));
      Color typeColor = display.getSystemColor(SWT.COLOR_DARK_MAGENTA);
      IToken typeToken = new Token(new TextAttribute(typeColor));
      Color quantifierColor = display.getSystemColor(SWT.COLOR_RED);
      IToken quantifierToken = new Token(new TextAttribute(quantifierColor));
      Color connectiveColor = display.getSystemColor(SWT.COLOR_BLUE);
      IToken connectiveToken = new Token(new TextAttribute(connectiveColor));
      
      // define literal constants as keywords
      WordRule consts = new WordRule(new WordDetector());
      consts.addWord("TRUE", keywordToken);
      consts.addWord("FALSE", keywordToken);
      consts.addWord("NOW", keywordToken);
      consts.addWord("NEXT", keywordToken);
      consts.addWord("EXECUTES", keywordToken);
      consts.addWord("CONTINUES", keywordToken);
      consts.addWord("BREAKS", keywordToken);
      consts.addWord("RETURNS", keywordToken);
      consts.addWord("THROWS", keywordToken);
      consts.addWord("VALUE", keywordToken);
      consts.addWord("MESSAGE", keywordToken);
      consts.addWord("OLD", keywordToken);
      consts.addWord("VAR", keywordToken);
      
      // define keywords
      WordRule keywords = new WordRule(new WordDetector());
      keywords.addWord("public", keywordToken);
      keywords.addWord("package", keywordToken);
      keywords.addWord("import", keywordToken);
      keywords.addWord("theory", keywordToken);
      
      // define builtin types
      WordRule types = new WordRule(new WordDetector());
      types.addWord("TYPE", typeToken);
      types.addWord("FORMULA", typeToken);
      types.addWord("AXIOM", typeToken);
      types.addWord("BOOLEAN", typeToken);
      types.addWord("INT", typeToken);
      types.addWord("NAT", typeToken);
      types.addWord("REAL", typeToken);
      types.addWord("SUBTYPE", typeToken);
      types.addWord("PREDICATE", typeToken);
      
      // define quantifiers
      WordRule quantifiers = new WordRule(new WordDetector());
      quantifiers.addWord("FORALL", quantifierToken);
      quantifiers.addWord("EXISTS", quantifierToken);
      quantifiers.addWord("LAMBDA", quantifierToken);
      quantifiers.addWord("PRED", quantifierToken);
      quantifiers.addWord("ARRAY", quantifierToken);
      quantifiers.addWord("LET", quantifierToken);
      quantifiers.addWord("IN", quantifierToken);
      
      // define connectives
      WordRule connectives = new WordRule(new ExtendedWordDetector());
      connectives.addWord("AND", connectiveToken);
      connectives.addWord("OR", connectiveToken);
      connectives.addWord("NOT", connectiveToken);
      connectives.addWord("=>", connectiveToken);
      connectives.addWord("<=>", connectiveToken);
      connectives.addWord("XOR", connectiveToken);
      connectives.addWord("IF", connectiveToken);
      connectives.addWord("THEN", connectiveToken);
      connectives.addWord("ELSIF", connectiveToken);
      connectives.addWord("ELSE", connectiveToken);
      connectives.addWord("ENDIF", connectiveToken);
      
      // set rules
      IRule[] rules = 
        new IRule[] { consts, keywords, types, quantifiers, connectives } ;
      setRules(rules);
    }
  }
  
  // my version of the default damager/repairer for dbugging
  private class SourceDamagerRepairer extends DefaultDamagerRepairer
  {
    public SourceDamagerRepairer(ITokenScanner scanner)
    {
      super(scanner);
    }
  }
  
  //-------------------------------------------------------------------------
  // the viewer configuration for class files
  //-------------------------------------------------------------------------
  private class ClassEditorConfiguration extends SourceViewerConfiguration
  {
    private FileInfo info;
    
    public ClassEditorConfiguration(FileInfo info)
    {
      this.info = info;
    }
    
    public IPresentationReconciler 
    getPresentationReconciler(ISourceViewer viewer)
    {
      PresentationReconciler reconciler = new PresentationReconciler();

      // reconcile specifications
      RuleBasedScanner scscanner = new SpecScanner();
      SourceDamagerRepairer scdrepairer = new SourceDamagerRepairer(scscanner);
      reconciler.setDamager(scdrepairer, SourcePartitionScanner.SPEC);
      reconciler.setRepairer(scdrepairer, SourcePartitionScanner.SPEC);
      
      // reconcile comments
      RuleBasedScanner cscanner = new CommentScanner();
      SourceDamagerRepairer cdrepairer = new SourceDamagerRepairer(cscanner);
      reconciler.setDamager(cdrepairer, SourcePartitionScanner.COMMENT);
      reconciler.setRepairer(cdrepairer, SourcePartitionScanner.COMMENT);
      
      // reconcile literals
      RuleBasedScanner lscanner = new LiteralScanner();
      SourceDamagerRepairer ldrepairer = new SourceDamagerRepairer(lscanner);
      reconciler.setDamager(ldrepairer, SourcePartitionScanner.LITERAL);
      reconciler.setRepairer(ldrepairer, SourcePartitionScanner.LITERAL);
      
      // reconcile rest
      RuleBasedScanner sscanner = new ClassSyntaxScanner();
      SourceDamagerRepairer sdrepairer = new SourceDamagerRepairer(sscanner);
      reconciler.setDamager(sdrepairer, IDocument.DEFAULT_CONTENT_TYPE);
      reconciler.setRepairer(sdrepairer, IDocument.DEFAULT_CONTENT_TYPE);
      
      return reconciler;
    }
    
    public IUndoManager getUndoManager(ISourceViewer viewer)
    {
      return new MyUndoManager(99);
    }
    
    public IAutoEditStrategy[] 
    getAutoEditStrategies(ISourceViewer sourceViewer, String contentType)
    {
      return new IAutoEditStrategy[] { 
          new DefaultIndentLineAutoEditStrategy() };
    }
    
    public IAnnotationHover getAnnotationHover(ISourceViewer viewer)
    {
      return new ErrorHover(info.model);
    }
    
    public IReconciler getReconciler(ISourceViewer viewer)
    {
      ProjectionViewer pviewer = (ProjectionViewer)viewer;
      IReconcilingStrategy strategy = new SourceReconcilingStrategy(pviewer);
      MonoReconciler reconciler = new MonoReconciler(strategy,false);  
      return reconciler;
    }
    
    public ITextDoubleClickStrategy 
    getDoubleClickStrategy(ISourceViewer viewer, String contentType)
    {
      return new EditorDoubleClickStrategy(info);
    }
  }
  
  //-------------------------------------------------------------------------
  // the viewer configuration for theory files
  //-------------------------------------------------------------------------
  private class TheoryEditorConfiguration extends SourceViewerConfiguration
  {
    private FileInfo info;
    
    public TheoryEditorConfiguration(FileInfo info)
    {
      this.info = info;
    }
    
    public IPresentationReconciler 
    getPresentationReconciler(ISourceViewer viewer)
    {
      PresentationReconciler reconciler = new PresentationReconciler();
      
      // reconcile comments
      RuleBasedScanner cscanner = new CommentScanner();
      SourceDamagerRepairer cdrepairer = new SourceDamagerRepairer(cscanner);
      reconciler.setDamager(cdrepairer, SourcePartitionScanner.COMMENT);
      reconciler.setRepairer(cdrepairer, SourcePartitionScanner.COMMENT);
      
      // reconcile literals
      RuleBasedScanner lscanner = new LiteralScanner();
      SourceDamagerRepairer ldrepairer = new SourceDamagerRepairer(lscanner);
      reconciler.setDamager(ldrepairer, SourcePartitionScanner.LITERAL);
      reconciler.setRepairer(ldrepairer, SourcePartitionScanner.LITERAL);
      
      // reconcile rest
      RuleBasedScanner sscanner = new TheorySyntaxScanner();
      SourceDamagerRepairer sdrepairer = new SourceDamagerRepairer(sscanner);
      reconciler.setDamager(sdrepairer, IDocument.DEFAULT_CONTENT_TYPE);
      reconciler.setRepairer(sdrepairer, IDocument.DEFAULT_CONTENT_TYPE);
      
      return reconciler;
    }
    
    public IUndoManager getUndoManager(ISourceViewer viewer)
    {
      return new MyUndoManager(99);
    }
    
    public IAutoEditStrategy[] 
    getAutoEditStrategies(ISourceViewer sourceViewer, String contentType)
    {
      return new IAutoEditStrategy[] { 
          new DefaultIndentLineAutoEditStrategy() };
    }
    
    public IAnnotationHover getAnnotationHover(ISourceViewer viewer)
    {
      return new ErrorHover(info.model);
    }
    
    public IReconciler getReconciler(ISourceViewer viewer)
    {
      ProjectionViewer pviewer = (ProjectionViewer)viewer;
      IReconcilingStrategy strategy = new SourceReconcilingStrategy(pviewer);
      MonoReconciler reconciler = new MonoReconciler(strategy,false);  
      return reconciler;
    }
    
    public ITextDoubleClickStrategy 
    getDoubleClickStrategy(ISourceViewer viewer, String contentType)
    {
      return new EditorDoubleClickStrategy(info);
    }
  }
  
  //-------------------------------------------------------------------------
  // the error annotation
  //-------------------------------------------------------------------------
  private class ErrorAnnotation extends Annotation 
  {
    private int line;
    
    public ErrorAnnotation(String text, int line) 
    {
      super(ERROR_TYPE, true, text);
      this.line = line;
    }
    
    public int getLine()
    {
      return line;
    }
  }
  
  //-------------------------------------------------------------------------
  // the marker annotation
  //-------------------------------------------------------------------------
  private class MarkerAnnotation extends Annotation 
  {
    public MarkerAnnotation() 
    {
      super(MARKER_TYPE, false, "");
    }
  }
  
  // -------------------------------------------------------------------------
  // source annotation access: error markers and folding markers
  // -------------------------------------------------------------------------
  private class SourceAnnotationAccess 
  implements IAnnotationAccess, IAnnotationAccessExtension 
  {
    public Object getType(Annotation annotation) 
    { 
      return annotation.getType(); 
    }
    
    public boolean isMultiLine(Annotation annotation) 
    { 
      return false; 
    }

    public boolean isTemporary(Annotation annotation) 
    { 
      return !annotation.isPersistent(); 
    }

    public String getTypeLabel(Annotation annotation) 
    {
      return "Errors";
    }

    public int getLayer(Annotation annotation) 
    {
      return 3;
    }

    public void paint(Annotation annotation, GC gc, Canvas canvas, 
      Rectangle bounds) 
    {
      Image image = null;
      if (annotation instanceof ErrorAnnotation)
        image = errorImage;
      else if (annotation instanceof ProjectionAnnotation)
      {
        ProjectionAnnotation annotation0 = (ProjectionAnnotation)annotation;
        if (annotation0.isCollapsed())
          image = folderOpenImage;
        else
          image = folderCloseImage;
      }
      ImageUtilities.drawImage(image, gc, canvas, bounds, SWT.CENTER, SWT.TOP);
    }

    public boolean isPaintable(Annotation annotation) 
    {
      return true;
    }

    public boolean isSubtype(Object annotationType, Object potentialSupertype) 
    {
      return (annotationType.equals(potentialSupertype));
    }

    public Object[] getSupertypes(Object annotationType) 
    {
      return new Object[0];
    }
  }
  
  // -------------------------------------------------------------------------
  // error annotation access
  // -------------------------------------------------------------------------
  private class SharedTextColors implements ISharedTextColors
  {
    public Color getColor(RGB rgb) { return new Color(Display.getDefault(), rgb); }
    public void dispose() { }
  }
  
  // --------------------------------------------------------------------------
  // error annotation hover
  // --------------------------------------------------------------------------
  private class ErrorHover implements IAnnotationHover, ITextHover 
  {
    private IAnnotationModel model;
    
    public ErrorHover(IAnnotationModel model)
    {
      this.model = model;
    }
    
    public String getHoverInfo(ISourceViewer viewer, int line) 
    {
      @SuppressWarnings("rawtypes")
      Iterator iterator = model.getAnnotationIterator();
      while (iterator.hasNext()) 
      {
        Object info = iterator.next();
        if (!(info instanceof ErrorAnnotation)) continue;
        ErrorAnnotation annotation = (ErrorAnnotation)info;
        if (annotation.getLine() == line) return annotation.getText();
      }
      return null;
    }

    public String getHoverInfo(ITextViewer textViewer, IRegion hoverRegion) 
    {
      return null;
    }

    public IRegion getHoverRegion(ITextViewer textViewer, int offset) 
    {
      return null;
    }
  }

  // --------------------------------------------------------------------------
  // annotation configuration
  // --------------------------------------------------------------------------
  private class ErrorInformationControlCreator
  implements IInformationControlCreator {
    public IInformationControl createInformationControl(Shell shell) 
    {
      return new DefaultInformationControl(shell);
    }
  }
  
  // --------------------------------------------------------------------------
  // source reconciling strategy: set positions for folding
  // --------------------------------------------------------------------------
  private class SourceReconcilingStrategy 
  implements IReconcilingStrategy, IReconcilingStrategyExtension
  {
    private ProjectionViewer viewer;
    private IDocument document;
    
    // previously used annotations, their positions, and whether again used
    private Annotation[] oannotations;
    private Position[] opositions;
    private boolean[] oused;
    
    // newly constructed annotations and their positions
    private Vector<Annotation> nannotations;
    private Vector<Position> npositions;
    
    // reused annotations and their positions
    private Vector<Annotation> mannotations;
    private Vector<Position> mpositions;
    
    public SourceReconcilingStrategy(ProjectionViewer viewer)
    {
      this.viewer = viewer;
      this.oannotations = new Annotation[0];
      this.opositions = new Position[0];
      this.oused = new boolean[0];
      this.nannotations = new Vector<Annotation>();
      this.npositions = new Vector<Position>();
      this.mannotations = new Vector<Annotation>();
      this.mpositions = new Vector<Position>();
    }
    
    public void setDocument(IDocument document) 
    {
      this.document = document;
    }

    public void initialReconcile()
    {
      process();
      update(true);
    }
    
    public void reconcile(DirtyRegion dirtyRegion, IRegion subRegion) 
    {
      process();
      update(false);
    }

    public void reconcile(IRegion partition) 
    {
      process();
      update(false);
    }
    
    public void setProgressMonitor(IProgressMonitor monitor) 
    {
    }
    
    /*************************************************************************
     * Update the folding structure of the document.
     * @param collapsed true if folders are to be collapsed
     ************************************************************************/
    public void update(boolean collapse)
    { 
      // determine annotations to be deleted
      Vector<Annotation> dannotations = new Vector<Annotation>();
      int m = oannotations.length;
      for (int i=0; i<m; i++)
      {
        if (!oused[i]) dannotations.add(oannotations[i]);
      }
      int l = dannotations.size();
      Annotation[] dannotations0 = new Annotation[l];
      for (int i=0; i<l; i++) dannotations0[i] = dannotations.get(i);
      
      // create map of new annotations/positions
      MapClass<Annotation,Position> annotations = new MapClass<Annotation,Position>();
      int n = nannotations.size();
      for(int i=0; i<n; i++)
      {
        ProjectionAnnotation annotation = (ProjectionAnnotation)nannotations.get(i);
        if (collapse) annotation.markCollapsed();
        annotations.put(annotation,npositions.get(i));
      }
      
      // update model
      ProjectionAnnotationModel model = viewer.getProjectionAnnotationModel();
      if (model != null)
        model.modifyAnnotations(dannotations0,annotations,new Annotation[0]);
      
      // remember data
      int p = mannotations.size();
      int o = n+p;
      oannotations = new Annotation[o];
      opositions = new Position[o];
      oused = new boolean[o];
      for (int i=0; i<n; i++)
      {
        oannotations[i] = nannotations.get(i);
        opositions[i] = npositions.get(i);
        oused[i] = false;
      }
      for (int i=0; i<p; i++)
      {
        oannotations[i+n] = mannotations.get(i);
        opositions[i+n] = mpositions.get(i);
        oused[i+n] = false;
      }
      
      // reset status
      this.nannotations = new Vector<Annotation>();
      this.npositions = new Vector<Position>();
      this.mannotations = new Vector<Annotation>();
      this.mpositions = new Vector<Position>();
    }
    
    // current parse status
    private int length; // number of characters in document
    private int pos;    // index of next character to be read
    private int line;   // start index for current line
    private int begin;  // begin of folding position (-1, if none)
    
    /*************************************************************************
     * Initialize parser status.
     ************************************************************************/
    private void init()
    {
      length = document.getLength();
      pos = 0;
      line = 0;
      begin = -1;
    }
    
    /*************************************************************************
     * Read next character from document.
     * @return the character.
     ************************************************************************/
    private char readChar()
    {
      try
      {
        char ch = document.getChar(pos); pos = pos+1;
        if (ch == '\n') line = pos;
        return ch;
      }
      catch(BadLocationException e)
      {
        Main.internalError("bad location in document");
        return 0;
      }
    }
    
    /*************************************************************************
     * Go back n characters in input stream.
     * @param n the number of characters to go back.
     ************************************************************************/
    private void unreadChar(int n)
    {
      pos = pos-n;
    }
    
    /*************************************************************************
     * Is there still a character to be read?
     * @return true if there is a character.
     ************************************************************************/
    private boolean hasChar()
    {
      return pos < length;
    }
    
    /*************************************************************************
     * Folding starts at current position.
     ************************************************************************/
    private void startFolding()
    {
      begin = line;
    }
    
    /*************************************************************************
     * Folding of block specification ends at current position.
     ************************************************************************/
    private void endFoldingBlock()
    {
      while (hasChar())
      {
        char ch = readChar();
        if (ch == '\n') break;
      }
      createFolding(begin, pos-begin);
    }
    
    /*************************************************************************
     * Folding of lines ends at current position.
     ************************************************************************/
    private void endFoldingLines()
    {
      createFolding(begin, line-begin);
    }
    
    /*************************************************************************
     * Create folding for denoted region
     * @param offset the start position of the region
     * @param length its length
     ************************************************************************/
    private void createFolding(int offset, int length)
    {
      // folding is completed
      begin = -1;
      
      // search for given position in old pool
      int n = opositions.length;
      for (int i=0; i<n; i++)
      {
        Position p = opositions[i];
        if (p.offset == offset && p.length == length)
        {
          // take annotation from old pool
          mannotations.add(oannotations[i]);
          mpositions.add(opositions[i]);
          oused[i] = true;
          return;
        }
      }
      
      // create new annotation
      nannotations.add(new ProjectionAnnotation());
      npositions.add(new Position(offset, length));
    }
    
    /*************************************************************************
     * Process document and generate folding positions.
     ************************************************************************/
    private void process()
    {
      init();
      while (hasChar())
      {
        char ch = readChar();
        switch (ch)
        {
        case '\'':
        {
          parseCharLiteral();
          break;
        }
        case '\"':
        {
          parseStringLiteral();
          break;
        }
        case '/': 
        {
          if (!hasChar()) break;
          char ch0 = readChar();
          switch (ch0)
          {
          case '/':
          {
            if (!hasChar()) break;
            char ch1 = readChar();
            if (ch1 == '@')
              parseLineSpec();
            else
              parseLineComment();
            break;
          }
          case '*':
          {
            if (!hasChar()) break;
            char ch1 = readChar();
            if (ch1 == '@')
              parseBlockSpec();
            else
              parseBlockComment();
          }
          }
        }
        }
      }
    }
    
    /*************************************************************************
     * Various parsing functions.
     ************************************************************************/
    
    private void parseCharLiteral()
    {
      while (hasChar())
      {
        char ch = readChar();
        if (ch == '\\')
        {
          if (!hasChar()) return;
          readChar();
          continue;
        }
        if (ch == '\n' || ch == '\'') return;
      }
    } 
    
    private void parseStringLiteral()
    {
      while (hasChar())
      {
        char ch = readChar();
        if (ch == '\\')
        {
          if (!hasChar()) return;
          readChar();
          continue;
        }
        if (ch == '\n' || ch == '\"') return;
      }
    } 
    
    private void parseLineComment()
    {
      while (hasChar())
      {
        char ch = readChar();
        if (ch == '\n') return;
      }
    } 
   
    private void parseBlockComment()
    {
      while (hasChar())
      {
        char ch = readChar();
        if (ch == '*') 
        {
          if (!hasChar()) return;
          char ch0 = readChar();
          if (ch0 == '/') return;
        }
      }
    } 
    
    private void parseLineSpec()
    {
      startFolding();
      while (true)
      {
        while (hasChar())
        {
          char ch = readChar();
          if (ch == '\n') break;
        }
        while (hasChar())
        {
          char ch = readChar();
          if (ch != ' ' && ch != '\t' && ch != '\n')
          {
            unreadChar(1);
            break;
          }
        }
        if (!hasChar()) { endFoldingLines(); return; }
        char ch = readChar();
        if (ch != '/' || !hasChar() ) { unreadChar(1); endFoldingLines(); return; }
        char ch0 = readChar();
        if (ch0 != '/' || !hasChar() ) { unreadChar(2); endFoldingLines(); return; }
        char ch1 = readChar();
        if (ch1 != '@') { unreadChar(3); endFoldingLines(); return; }
      }
    }
    
    private void parseBlockSpec()
    {
      startFolding();
      while (hasChar())
      {
        char ch = readChar();
        if (ch == '@')
        {
          if (!hasChar()) return;
          char ch0 = readChar();
          if (ch0 == '*') 
          {
            if (!hasChar()) return;
            char ch1 = readChar();
            if (ch1 == '/') 
            {
              endFoldingBlock();
              return;
            }
          }
        }
      }
    }
  }
  
  /****************************************************************************
   * Strategy for double clicking in editor text: access referenced symbol.
   ***************************************************************************/
  private class EditorDoubleClickStrategy extends DefaultTextDoubleClickStrategy
  {
    private FileInfo info;
    
    public EditorDoubleClickStrategy(FileInfo info)
    {
      this.info = info;
    }
    
    public void doubleClicked(ITextViewer text)
    {
      super.doubleClicked(text);
      int offset = text.getSelectedRange().x;
      SourcePosition[] positions = info.source.getPositions();
      int n = positions.length;
      for (int i=0; i<n; i++)
      {
        SourcePosition p = positions[i];
        if (offset == p.getOffset()) 
        { 
          // same position may be registered multiple times
          boolean okay = process(p); 
          if (okay) return;
        }
      }
    }
    
    public boolean process(SourcePosition p)
    {
      Identifier ident = p.getIdentifier();
      if (ident == null) return false;
      if (ident.getString().equals("String")) 
      {
        TypeExpression stringType = topPackage.getTypeTable().getStringType();
        if (stringType instanceof StringType)
        {
          getConsole().print("type String");
          return true;
        }
        NamedType stringType0 = (NamedType)stringType;
        ident = stringType0.getSymbol().getName();
      }
      Symbol symbol = ident.getSymbol();
      if (symbol == null) return false;
      symbolExplorer.show(symbol);
      print(symbol);
      printDecl(symbol);
      return true;
    }
  }
}
