// ----------------------------------------------------------------------------
// Environment.java
// An environment giving meaning to identifiers.
// $Id: Environment.java,v 1.36 2022/12/05 07:10:56 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;
import fmrisc.ProgramExplorer.Main;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;
import fmrisc.ProgramExplorer.InOut.*;

/**
 * 
 * An environment giving meaning to identifiers.
 *
 */
public final class Environment
{   
  // the stream for printing error messages
  private ErrorStream error;
  
  // the level of the current context
  // the following declarations are only defined for the denoted minimum level
  private int level = 0;
  
  // indicates whether whether we are in class ("static") context
  private boolean isclass = true;
  
  // the name "java.lang"
  private Name javaLang = null;
  
  // --------------------------------------------------------------------------
  // level 0: top-level declarations
  // --------------------------------------------------------------------------
  
  // the package into which all "top-level" declarations are placed.
  private PackageSymbol top;
  
  // --------------------------------------------------------------------------
  // level 1: package-level declarations
  // --------------------------------------------------------------------------
  
  // null or the current package
  private PackageSymbol psymbol = null;
  
  // --------------------------------------------------------------------------
  // level 2: class-level declarations
  // --------------------------------------------------------------------------
  
  // null or the current class
  private ClassSymbol csymbol = null;
  
  // the environment for looking up classes from the context of a class
  private SymbolTable<ClassSymbol> classes = null;
  
  // the environment for looking up object variables during initialization
  private SymbolTable<ObjectVariableSymbol> ovars = null;
  
  // look up object variables in ovars?
  private boolean lookupOvars = false;
  
  // the environment for looking up class variables during initialization
  private SymbolTable<ClassVariableSymbol> cvars = null;
  
  // look up class variables in cvars?
  private boolean lookupCvars = false;
  
  // --------------------------------------------------------------------------
  // level 3: method-level declarations
  // --------------------------------------------------------------------------
  
  // null or the current constructor, class or object method
  private ParamSymbol msymbol = null;
  
  // --------------------------------------------------------------------------
  // level 4: body-level declarations
  // --------------------------------------------------------------------------
  
  // the stack of local variable tables for the current context
  private Stack<SymbolTable<BodySymbol>> locals = null;
  
  // --------------------------------------------------------------------------
  //
  // Environment creation
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Create a new environment at top-level.
   * @param top the package for holding "top-level" declarations.
   * @param error the stream for printing error messages.
   ***************************************************************************/
  public Environment(PackageSymbol top, ErrorStream error)
  {
    this.top = top;
    this.error = error;
    this.level = 0;
    
    // the name of the "java.lang" package
    javaLang = Name.construct(new Identifier[] {
        Identifier.construct("java"), Identifier.construct("lang") });
  }
  
  // --------------------------------------------------------------------------
  //
  // Entering/exiting scopes.
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Enter from top-level the package level.
   * @param psymbol the symbol of the package that is entered.
   ***************************************************************************/
  public void enterPackage(PackageSymbol psymbol)
  {
    if (level != 0) 
      Main.internalError("attempt to enter package from level " + level);
    level = 1;
    this.psymbol = psymbol;
  }
  
  /****************************************************************************
   * Return from package level to top-level.
   ***************************************************************************/
  public void exitPackage()
  {
    if (level != 1) 
      Main.internalError("attempt to exit package from level " + level);
    level = 0;
    this.psymbol = null;
  }
  
  /****************************************************************************
   * Enter from package level the class level.
   * @param csymbol the symbol of the class that is entered.
   * @return true iff everything went well (message is printed otherwise).
   ***************************************************************************/
  public boolean enterClass(ClassSymbol csymbol)
  {
    if (level != 1) 
      Main.internalError("attempt to enter class from level " + level);
    level = 2;
    this.csymbol = csymbol;
    
    // create object/class variables environment
    ovars = new SymbolTable<ObjectVariableSymbol>(error);
    cvars = new SymbolTable<ClassVariableSymbol>(error);
    
    // create class environment
    classes = new SymbolTable<ClassSymbol>(error);
    
    // classes in current package
    classes.add(psymbol.getClasses(), false, false);
    
    // classes in java.lang, skip classes already visible in current class
    PackageSymbol jsymbol = top.getPackage(javaLang, false);
    if (jsymbol != null) classes.add(jsymbol.getClasses(), false, true);
      
    // classes in package imports, skip classes already visible in current class
    Import[] imports = csymbol.getDeclaration().getImports();
    for (Import imp : imports)
    {
      if (!(imp instanceof PackageImport)) continue;
      PackageSymbol isymbol = (PackageSymbol)imp.getName().getSymbol();
      classes.add(isymbol.getClasses(), false, true);
    }
    
    // classes in class imports, override classes that are visible otherwise
    SymbolTable<ClassSymbol> isymbols = new SymbolTable<ClassSymbol>(error);
    for (Import imp : imports)
    {
      if (!(imp instanceof ClassImport)) continue;
      Name iname = imp.getName();
      Identifier base = iname.getBase();
      ClassSymbol isymbol = (ClassSymbol)base.getSymbol();
      boolean okay = isymbols.put(isymbol);
      if (!okay)
      {
        ClassSymbol isymbol0 = isymbols.get(base);
        error.newError("import of class " + iname + 
          " conflicts with already imported class " + isymbol0.getName());
        return false;
      }
    } 
    classes.add(isymbols, true, false);
    return true;
  }
  
  /****************************************************************************
   * Enter from class level the initialization of object variables.
   * @return true iff everything went well (message is printed otherwise).
   ***************************************************************************/
  public boolean enterObjectVariables()
  {
    if (level != 2 || lookupOvars) 
      Main.internalError("attempt to enter object variables from level " + level);
    lookupOvars = true;
    return true;
  }
    
  /****************************************************************************
   * Return from entering object variables.
   * @return always true.
   ***************************************************************************/
  public boolean exitObjectVariables()
  {
    if (level != 2 || !lookupOvars) 
      Main.internalError("attempt to exit object variables from level " + level);
    lookupOvars = false;
    return true;
  }
  
  /****************************************************************************
   * Enter in class level another object variable.
   * @param osymbol the symbol of the object variable that is entered.
   * @return true iff everything went well (message is printed otherwise).
   ***************************************************************************/
  public boolean enterObjectVariable(ObjectVariableSymbol osymbol)
  {
    if (level != 2) 
      Main.internalError("attempt to enter object variable in level " + level);
    ovars.put(osymbol);
    return true;
  }
    
  /****************************************************************************
   * Get currently visible object variables.
   * @return the set of object values (null if none).
   ***************************************************************************/
  public SymbolTable<ObjectVariableSymbol> getObjectVariables()
  {
    if (lookupOvars) return ovars;
    if (csymbol != null) return csymbol.getObjectVariables();
    return null;
  }
  
  /****************************************************************************
   * Enter from class level the initialization of class variables.
   * @return true iff everything went well (message is printed otherwise).
   ***************************************************************************/
  public boolean enterClassVariables()
  {
    if (level != 2 || lookupCvars) 
      Main.internalError("attempt to enter class variables from level " + level);
    lookupCvars = true;
    return true;
  }
    
  /****************************************************************************
   * Return from entering class variables.
   * @return always true.
   ***************************************************************************/
  public boolean exitClassVariables()
  {
    if (level != 2 || !lookupCvars) 
      Main.internalError("attempt to exit class variables from level " + level);
    lookupCvars = false;
    return true;
  }
  
  /****************************************************************************
   * Enter in class level another class variable.
   * @param csymbol the symbol of the class variable that is entered.
   * @return true iff everything went well (message is printed otherwise).
   ***************************************************************************/
  public boolean enterClassVariable(ClassVariableSymbol csymbol)
  {
    if (level != 2) 
      Main.internalError("attempt to enter class variable in level " + level);
    cvars.put(csymbol);
    return true;
  }
  
  /****************************************************************************
   * Get currently visible class variables.
   * @return the set of class values (null if none).
   ***************************************************************************/
  public SymbolTable<ClassVariableSymbol> getClassVariables()
  {
    if (lookupCvars) return cvars;
    if (csymbol != null) return csymbol.getClassVariables();
    return null;
  }
  
  /****************************************************************************
   * Return from class level to package level.
   ***************************************************************************/
  public void exitClass()
  {
    if (level != 2) 
      Main.internalError("attempt to exit class from level " + level);
    level = 1;
    this.csymbol = null;
    this.classes = null;
  }
  
  /****************************************************************************
   * Enter object ("non-static") context.
   ***************************************************************************/
  public void enterObject()
  {
    if (level != 2) 
      Main.internalError("attempt to enter object context from level " + level);
    if (!isclass)
      Main.internalError("attempt to enter object context twice");
    isclass = false;
  }
  
  /****************************************************************************
   * Leave object ("non-static") context.
   ***************************************************************************/
  public void exitObject()
  {
    if (level != 2) 
      Main.internalError("attempt to exit object context from level " + level);
    if (isclass)
      Main.internalError("attempt to exit object context twice");
    isclass = true;
  }
  
  /****************************************************************************
   * Enter from class level the constructor/method level.
   * @param msymbol the symbol of the method that is entered.
   ***************************************************************************/
  public void enterMethod(ParamSymbol msymbol)
  {
    if (level != 2) 
      Main.internalError("attempt to enter method from level " + level);
    level = 3;
    this.msymbol = msymbol;
  }
  
  /****************************************************************************
   * Return from method level to class level.
   ***************************************************************************/
  public void exitMethod()
  {
    if (level != 3) 
      Main.internalError("attempt to exit method from level " + level);
    level = 2;
    this.msymbol = null;
  }  
  
  /****************************************************************************
   * Get symbol of current method.
   * @return symbol of current method (null, if we are not inside a method).
   ***************************************************************************/
  public ParamSymbol getCurrentMethod()
  {
    return msymbol;
  }
  
  /****************************************************************************
   * Enter from constructor/method level the body level.
   ***************************************************************************/
  public void enterBody()
  {
    if (level != 3) 
      Main.internalError("attempt to enter body from level " + level);
    level = 4;
    this.locals = new Stack<SymbolTable<BodySymbol>>();
    this.locals.push(new SymbolTable<BodySymbol>(error));
  }
  
  /****************************************************************************
   * Return from body level to constructor/method level.
   ***************************************************************************/
  public void exitBody()
  {
    if (level != 4) 
      Main.internalError("attempt to exit body from level " + level);
    level = 3;
    this.locals = null;
  }
  
  /****************************************************************************
   * Enter a new local scope from (at least) body level.
   ***************************************************************************/
  public void enterScope()
  {
    if (level < 4)
      Main.internalError("attempt to enter local scope from level " + level);
    level = level+1;
    this.locals.push(new SymbolTable<BodySymbol>(error));
  }
  
  /****************************************************************************
   * Return from local scope to previous level.
   ***************************************************************************/
  public void exitScope()
  {
    if (level < 5) 
      Main.internalError("attempt to exit local scope from level " + level);
    level = level-1;
    this.locals.pop();
  }
  
  // --------------------------------------------------------------------------
  //
  // Enter symbols into corresponding tables.
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Enter new package symbol into environment.
   * @param symbol the package symbol.
   * @return true iff symbol could be successfully added.
   ***************************************************************************/
  public boolean put(PackageSymbol symbol)
  {
    if (level != 0)
      Main.internalError("attempt to put package symbol in level " + level);
    return top.getPackages().put(symbol);
  }
  
  /****************************************************************************
   * Enter new class symbol into environment.
   * @param symbol the class symbol.
   * @return true iff symbol could be successfully added.
   ***************************************************************************/
  public boolean put(ClassSymbol symbol)
  {
    if (level == 0)
      return top.getClasses().put(symbol);
    else if (level == 1)
      return psymbol.getClasses().put(symbol);
    else
      Main.internalError("attempt to put class symbol in level " + level);
    return false;
  }
      
  /****************************************************************************
   * Enter new constructor symbol into environment.
   * @param decl the constructor declaration giving rise to the symbol.
   * @return true iff symbol could be successfully added.
   ***************************************************************************/
  public boolean put(ConstructorDeclaration decl)
  {
    if (level != 2)
      Main.internalError("attempt to put constructor symbol in level " + level);
    TypeExpression[] types = getTypes(decl.params);
    if (types == null)
    {
      error.explainError("invalid parameter types");
      return false;
    }
    ConstructorSymbol symbol = ConstructorSymbol.construct(csymbol, 
        decl.name, decl.params, types, decl, error);
    if (symbol == null) 
    {
      error.explainError("constructor signature is invalid");
      return false;
    }
    boolean okay = csymbol.getConstructors().put(symbol);
    if (!okay) 
    {
      error.explainError("constructor is already declared in class");
      return false;
    }
    return true;
  }
  
  /****************************************************************************
   * Enter new class method into environment.
   * @param decl the declaration of the class method.
   * @return true iff symbol of the declaration could be successfully added.
   ***************************************************************************/
  public boolean put(ClassMethodDeclaration decl)
  {
    if (level != 2)
      Main.internalError("attempt to put class method symbol in level " + level);
    TypeExpression[] atypes = getTypes(decl.params);
    if (atypes == null)
    {
      error.explainError("invalid parameter types");
      return false;
    }
    TypeExpression rtype = getType(decl.type);
    if (rtype == null)
    {
      error.explainError("invalid return type");
      return false;
    }
    Identifier name = decl.getName();
    ClassMethodSymbol symbol = 
      ClassMethodSymbol.construct(csymbol, name, 
          decl.params, atypes, rtype, decl, error);
    if (symbol == null) 
    {
      error.explainError("signature of class method " + name + " is invalid");
      return false;
    }
    ObjectMethodSymbol symbol0 = 
      csymbol.getObjectMethods().get(name, atypes);
    if (symbol0 != null)
    {
      error.newError("there is an object method " + 
          Syntax.procHeader(name.toString(), atypes) +
          " declared in class");
      error.explainError("cannot declare a class method with same signature");
      return false;
    }
    boolean okay = csymbol.getClassMethods().put(symbol);
    if (!okay) 
    {
      error.explainError("there is already a class method " + 
          Syntax.procHeader(name.toString(), atypes) +
          " declared in class");
      error.explainError("cannot declare a class method twice with same signature");
      return false;
    }
    return true;
  }
  
  /****************************************************************************
   * Enter new object method symbol into environment.
   * @param decl the declaration givning rise to the symbol.
   * @return true iff symbol could be successfully added.
   ***************************************************************************/
  public boolean put(ObjectMethodDeclaration decl)
  {
    if (level != 2)
      Main.internalError("attempt to put objectmethod symbol in level " + level);
    TypeExpression[] atypes = getTypes(decl.params);
    if (atypes == null)
    {
      error.explainError("invalid parameter types");
      return false;
    }
    TypeExpression rtype = getType(decl.type);
    if (rtype == null)
    {
      error.explainError("invalid return type");
      return false;
    }
    Identifier name = decl.getName();
    ObjectMethodSymbol symbol = 
      ObjectMethodSymbol.construct(csymbol, name, 
          decl.params, atypes, rtype, decl, error);
    if (symbol == null) 
    {
      error.explainError("signature of object method " + name + " is invalid");
      return false;
    }
    ClassMethodSymbol symbol0 = 
      csymbol.getClassMethods().get(name, atypes);
    if (symbol0 != null)
    {
      error.newError("there is a class method " + 
          Syntax.procHeader(name.toString(), atypes) +
          " declared in class");
      error.explainError("cannot declare an object method with same signature");
      return false;
    }
    boolean okay = csymbol.getObjectMethods().put(symbol);
    if (!okay) 
    {
      error.explainError("there is already an object method " + 
          Syntax.procHeader(name.toString(), atypes) +
          " declared in class");
      error.explainError("cannot declare an object method twice with same signature");
      return false;
    }
    return true;
  }
  
  /****************************************************************************
   * Enter new class variable symbol into environment.
   * @param decl the declaration giving rise to the symbol.
   * @return true iff symbol could be successfully added.
   ***************************************************************************/
  public boolean put(ClassVariableDeclaration decl)
  {
    if (level != 2)
      Main.internalError("attempt to put class variable symbol in level " + level);
    Identifier name = decl.name;
    TypeExpression type = getType(decl.type);
    ClassVariableSymbol symbol = 
      ClassVariableSymbol.construct(csymbol, name, type, decl);
    if (symbol == null) 
    {
      error.explainError("class variable has invalid type");
      return false;
    }
    ObjectVariableSymbol symbol0 = getObjectVariables().get(name);
    if (symbol0 != null)
    {
      error.newError("there is an object variable " + name + 
          " declared in class");
      error.explainError("cannot declare a class variable with same name");
      return false;
    }
    boolean okay = csymbol.getClassVariables().put(symbol);
    if (!okay) 
    {
      error.explainError("class variable is already declared in class");
      return false;
    }
    return true;
  }
  
  /****************************************************************************
   * Enter new object variable symbol into environment.
   * @param decl the declaration giving rise to the symbol.
   * @return true iff symbol could be successfully added.
   ***************************************************************************/
  public boolean put(ObjectVariableDeclaration decl)
  {
    if (level != 2)
      Main.internalError("attempt to put object variable symbol in level " + level);
    Identifier name = decl.name;
    TypeExpression type = getType(decl.type);
    ObjectVariableSymbol symbol = 
      ObjectVariableSymbol.construct(csymbol, name, type, decl);
    if (symbol == null) 
    {
      error.explainError("object variable has invalid type");
      return false;
    }
    ClassVariableSymbol symbol0 = csymbol.getClassVariables().get(name);
    if (symbol0 != null)
    {
      error.newError("there is a class variable " + name + 
          " declared in class");
      error.explainError("cannot declare an object variable with same name");
      return false;
    }
    boolean okay = csymbol.getObjectVariables().put(symbol);
    if (!okay) 
    {
      error.explainError("object variable is already declared in class");
      return false;
    }
    return true;
  }
  
  /****************************************************************************
   * Enter new parameter symbol into environment.
   * @param symbol the parameter symbol.
   * @return true iff symbol could be successfully added.
   ***************************************************************************/
  public boolean put(ParameterSymbol symbol)
  {
    if (level == 3)
      return msymbol.getParams().put(symbol);
    else
      Main.internalError("attempt to put parameter symbol in level " + level);
    return false;
  }
  
  /****************************************************************************
   * Enter new local variable declaration into environment.
   * @param decl the declaration.
   * @return true iff symbol could be successfully added.
   ***************************************************************************/
  public boolean put(VariableDeclaration decl)
  {
    if (level < 4)
      Main.internalError("attempt to put body symbol in level " + level);
    Identifier name = decl.getName();
    TypeExpression type = getType(decl.getType());
    BodySymbol symbol = BodySymbol.construct(name, type, decl);
    if (symbol == null) 
    {
      error.explainError("local variable has invalid type");
      return false;
    }
    symbol.setLogicType(top.getTypeTranslator().logicType(type));
    // disallow redeclarations of variables declared at lower levels
    // because this may confuse later stages of the RISC ProgramExplorer
    // BodySymbol symbol0 = locals.peek().get(name);
    VariableSymbol symbol0 = getVariable(name);
    if (symbol0 != null)
    {
      error.newError("there is a variable " + name + 
          " declared in current scope");
      error.explainError("cannot declare another variable with same name");
      return false;
    }
    return put(symbol);
  }
  
  /****************************************************************************
   * Enter new body symbol into environment.
   * @param symbol the symbol.
   * @return true iff symbol could be successfully added.
   ***************************************************************************/
  public boolean put(BodySymbol symbol)
  {
    return locals.peek().put(symbol);
  }
  
  // --------------------------------------------------------------------------
  //
  // Lookup names.
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Return symbol for class denoted by name in current environment.
   * @param name a name which should denote a class.
   * @return the symbol or null if the name could not be looked up.
   *         in the first case, the symbol is fully populated with declarations.
   ***************************************************************************/
  public ClassSymbol getClass(Name name)
  {
    return getClass(name, true);
  }
  
  /****************************************************************************
   * Return symbol for class denoted by name in current environment.
   * @param name a name which should denote a class.
   * @param verbose true iff an error message is printed if the lookup fails.
   * @return the symbol or null if the name could not be looked up.
   *         in the first case, the symbol is fully populated with declarations.
   ***************************************************************************/
  private ClassSymbol getClass(Name name, boolean verbose)
  {
    Identifier[] idents = name.idents;
    int n = idents.length;
    if (n == 0) Main.internalError("empty name encountered");
    Identifier ident = idents[0];
    
    // name is plain identifier
    if (n == 1)
    {
      ClassSymbol csymbol;
      if (level == 0)
        csymbol = top.getClasses().get(ident);
      else if (level == 1)
        csymbol = psymbol.getClasses().get(ident);
      else
        csymbol = classes.get(ident);
      if (csymbol == null) 
      {
        if (verbose)
          error.newError("there is no class " + ident + " visible");
        return null;
      }
      boolean okay = csymbol.read(error);
      if (!okay) return null;
      return csymbol;
    }

    
    // name prefix must denote a package
    Name prefix = name.getPrefix();
    Identifier ident0 = name.getBase();
    PackageSymbol psymbol = top.getPackage(prefix, verbose);
    if (psymbol == null) return null;
    
    // lookup class in package
    ClassSymbol csymbol = psymbol.getClasses().get(ident0);
    if (csymbol != null) 
    {
      boolean okay = csymbol.read(error);
      if (!okay) return null;
      return csymbol;
    }
    
    // no symbol found in package
    if (verbose)
      error.newError("in package " + prefix + " there is no class " + ident0);
    return null;
  }
    
  /****************************************************************************
   * Return symbol for constructor denoted by name and argument types
   * in current environment.
   * @param stat the constructor call.
   * @param atypes the argument types of the constructor.
   * @return the symbol or null if the name could not be looked up
   ***************************************************************************/
  public ConstructorSymbol getConstructor(CallStatement stat, TypeExpression[] atypes)
  {
    Name name = stat.method;
    ClassSymbol csymbol = getClass(name);
    if (csymbol == null) return null;
    ParamSymbols<ConstructorSymbol> constructors = csymbol.getConstructors();
    ConstructorSymbol constructor = constructors.get(atypes);
    if (constructor == null) 
      error.newError("there is no constructor " + 
        Syntax.procHeader(name.toString(), atypes) );
    stat.setMethodSymbol(constructor);
    return constructor;
  }
  
  /****************************************************************************
   * Return symbol for a method called with arguments of denoted types
   * in current environment and annotate call with symbol.
   * @param call the method call.
   * @param atypes the argument types of the method.
   * @return the symbol or null if the name could not be looked up
   ***************************************************************************/
  public MethodSymbol getMethod(CallStatement call, TypeExpression[] atypes)
  { 
    // must be in body of a method
    if (level < 4) 
      Main.internalError("attempt to look up method in level " + level);
    
    // determine first identifier
    Name name = call.method;
    Identifier[] idents = name.idents;
    int n = idents.length;
    if (n == 0) Main.internalError("empty name encountered");
    Identifier ident = idents[0];

    // method is denoted by a simple identifier
    if (n == 1)
    {      
      // in object context, first look up object methods of current class
      if (msymbol instanceof ObjectMethodSymbol)
      {
        ObjectMethodSymbol symbol = 
          csymbol.getObjectMethods().get(ident, atypes);
        call.setMethodSymbol(symbol);
        if (symbol != null) return symbol;
      }
      
      // look up class methods of current class
      ClassMethodSymbol symbol = 
        csymbol.getClassMethods().get(ident, atypes);
      call.setMethodSymbol(symbol);
      if (symbol != null) return symbol;
      
      // no symbol found
      error.newError("there is no method " + 
          Syntax.procHeader(name.toString(), atypes) + " visible");
      return null;
    }
       
    // method name has a prefix
    Name prefix = name.getPrefix();
    Identifier base = name.getBase();
    
    // if first identifier is "this", we have an object method
    if (ident.toString().equals("this"))
    {
      if (!isclass)
      {
        ident.setSymbol(csymbol.getThisSymbol());
        return getObjectMethod(csymbol, name.getPostfix(), atypes);
      }
      error.newError("reference to \"this\" is only allowed in object context");
      return null;
    }
    
    // if prefix denotes a variable, we have an object method
    VariableSymbol vsymbol = getVariable(prefix, false);
    if (vsymbol != null)
    {
      TypeExpression vtype = vsymbol.getType();
      if (!(vtype instanceof NamedType))
      {
        error.newError("variable " + prefix + 
            " should denote an object but has type " + vtype);
        return null;
      }
      NamedType vtype0 = (NamedType)vtype;
      ClassSymbol csymbol = vtype0.getSymbol();
      ObjectMethodSymbol symbol = getObjectMethod(csymbol, base, atypes);
      call.setMethodSymbol(symbol);
      return symbol;
    }
    
    // otherwise we have a class method, prefix must denote a class
    ClassSymbol csymbol = getClass(prefix, false);
    if (csymbol == null) 
    {
      error.newError("there is no class method " + name);
      return null;
    }
    
    // lookup class method in class
    ClassMethodSymbol symbol = csymbol.getClassMethods().get(base, atypes);
    call.setMethodSymbol(symbol);
    if (symbol != null) return symbol;
    
    // no method found in class
    error.newError("in class " + prefix + " there is no class method " + 
      Syntax.procHeader(base.toString(), atypes));
    return null;
  }

  /****************************************************************************
   * Return object method denoted by name and argument types.
   * @param csymbol the symbol denoting the class where to look up the method.
   * @param name the name denoting the method.
   * @param atypes the argument types of the method.
   * @return the symbol denoting the method (null, if there is no such method.
   ***************************************************************************/
  private ObjectMethodSymbol getObjectMethod(ClassSymbol csymbol, 
    Name name, TypeExpression[] atypes)
  {
    Name prefix = name.getPrefix();
    ObjectVariableSymbol vsymbol = getObjectVariable(csymbol, prefix);
    if (vsymbol == null) return null;
    ClassSymbol csymbol0 = getClass(vsymbol);
    if (csymbol0 == null)
    {
      error.newError("Type " + vsymbol.getType() + " of variable " + 
        prefix + " does not denote a class");
      return null;
    }
    Identifier base = name.getBase();
    return getObjectMethod(csymbol0, base, atypes);
  }
  
  /****************************************************************************
   * Return object method denoted by name and argument types.
   * @param csymbol the symbol denoting the class where to look up the method.
   * @param name the name denoting the method.
   * @param atypes the argument types of the method.
   * @return the symbol denoting the method (null, if there is no such method.
   ***************************************************************************/
  private ObjectMethodSymbol getObjectMethod(ClassSymbol csymbol, 
    Identifier name, TypeExpression[] atypes)
  {
    ObjectMethodSymbol symbol = csymbol.getObjectMethods().get(name, atypes);
    if (symbol != null) return symbol;
    error.newError("in class " + csymbol.getLongName() + 
        " there is no object method " + 
        Syntax.procHeader(name.toString(), atypes));
    return null;
  }
  
  /****************************************************************************
   * Return object variable denoted by name.
   * @param csymbol the symbol denoting the class where to look up the variable.
   * @param name the name denoting the variable.
   * @return the symbol denoting the method (null, if there is no such method.
   ***************************************************************************/
  private ObjectVariableSymbol getObjectVariable(ClassSymbol csymbol, Name name)
  {
    ClassSymbol csymbol0 = csymbol;
    SymbolTable<ObjectVariableSymbol> ovars = getObjectVariables();
    ObjectVariableSymbol vsymbol = null;
    Identifier[] idents = name.idents;
    int n = idents.length;
    for (int i=0; i < n; i++)
    {
      Identifier ident = idents[i];
      vsymbol = ovars.get(ident);
      if (vsymbol == null)
      {
        error.newError("in class " + csymbol0.getLongName() + 
            " there is no object variable " + ident);
        return null;
      }
      if (i+1 < n)
      {
        csymbol0 = getClass(vsymbol);
        if (csymbol0 == null)
        {
          error.newError("type " + vsymbol.getType() + " of variable " + 
              vsymbol.getLongName() + " does not denote a class");
          return null;
        }
        ovars = csymbol0.getObjectVariables();
      }
    }
    return vsymbol;
  }
  
  /****************************************************************************
   * Return symbol for variable denoted by identifier in current environment.
   * @param ident a name which should denote a variable.
   * @return the symbol or null if the name could not be looked up
   *         (no message is generated in this case).
   ***************************************************************************/
  public VariableSymbol getVariable(Identifier ident)
  {
    return getVariable(ident, false);
  }
  
  /****************************************************************************
   * Return symbol for variable denoted by identifier in current environment.
   * @param name a name which should denote a variable.
   * @param verbose if true, an error message is printed on lookup failure.
   * @return the symbol or null if the name could not be looked up.
   ***************************************************************************/
  private VariableSymbol getVariable(Identifier ident, boolean verbose)
  {
    // the symbol for "this" object
    if (ident.toString().equals("this"))
    {
      if (!isclass) 
      {
        VariableSymbol symbol = csymbol.getThisSymbol();
        ident.setSymbol(symbol);
        return symbol;
      }
      if (verbose)
        error.newError("reference to \"this\" is only allowed in object context");
      return null;
    }
    
    // lookup body variables
    if (level >= 4)
    {
      int s = locals.size();
      for (int i=s-1; i>=0; i--)
      {
        SymbolTable<BodySymbol> table = locals.elementAt(i);
        VariableSymbol vsymbol = table.get(ident);
        if (vsymbol != null) return vsymbol;
      }
    }
    
    // lookup parameters
    if (level >= 3)
    {
      VariableSymbol vsymbol = msymbol.getParams().get(ident);
      if (vsymbol != null) return vsymbol;
    }
    
    // lookup object and class variables
    VariableSymbol vsymbol = getObjectVariables().get(ident);
    if (vsymbol != null) return vsymbol;
    vsymbol = getClassVariables().get(ident);
    return vsymbol;
  }
  
  /****************************************************************************
   * Return symbol for variable denoted by name in current environment.
   * @param name a name which should denote a variable.
   * @return the symbol or null if the name could not be looked up
   *         (in this case, an error message is printed)
   ***************************************************************************/
  public VariableSymbol getVariable(Name name)
  {
    return getVariable(name, true);
  }
  
  /****************************************************************************
   * Return symbol for variable denoted by name in current environment.
   * @param name a name which should denote a variable.
   * @param verbose true iff error messages shall be printed if the lookup fails.
   * @return the symbol or null if the name could not be looked up.
   ***************************************************************************/
  public VariableSymbol getVariable(Name name, boolean verbose)
  {
    // must be in a class
    // if (level < 2) 
    //  Main.internalError("attempt to look up variable in level " + level);
    
    // may be wrongly looked up from a specification
    if (level < 2) return null;
    
    // determine first identifier
    Identifier[] idents = name.idents;
    int n = idents.length;
    if (n == 0) Main.internalError("empty name encountered");
    Identifier ident = idents[0];

    // variable is denoted by a simple identifier
    if (n == 1)
    {      
      VariableSymbol vsymbol = getVariable(ident);
      if (vsymbol != null)
      {
        if (isclass && vsymbol instanceof ObjectVariableSymbol)
        {
          error.newError("object variable " + ident + " is not visible in class context");
          return null;
        }
        return vsymbol;
      }
      if (verbose)
        error.newError("no variable " + ident + " is visible in current context");
      return null;
    }
           
    // if first identifier is "this", we have an object method variable
    Identifier first = name.idents[0];
    if (first.toString().equals("this"))
    {
      if (!isclass) 
      {
        first.setSymbol(csymbol.getThisSymbol());
        return getObjectVariable(csymbol, name.getPostfix());
      }
      if (verbose)
        error.newError("reference to \"this\" is only allowed in object context");
      return null;
    }
    
    // method name has a prefix
    Name prefix = name.getPrefix();
    Identifier base = name.getBase();
    
    // if prefix denotes a variable, we have an object variable
    VariableSymbol vsymbol = getVariable(prefix, false);
    if (vsymbol != null)
    {
      TypeExpression vtype = vsymbol.getType();
      if (vtype instanceof ArrayType && base.toString().equals("length"))
      {
        ClassSymbol array = new ClassSymbol(top,
            Identifier.construct("Array"), null);
        TypeExpression intType = top.getTypeTable().intType;
        return ObjectVariableSymbol.construct(array, 
            base, intType, null);
      }
      if (!(vtype instanceof NamedType))
      {
        error.newError("variable " + prefix + 
            " should denote an object but has type " + vtype);
        return null;
      }
      NamedType vtype0 = (NamedType)vtype;
      ClassSymbol csymbol = vtype0.getSymbol();
      ObjectVariableSymbol symbol = csymbol.getObjectVariables().get(base);
      if (symbol == null)
      {
        error.newError("in class " + vtype + 
            " there is no object variable " + base);
        return null;
      }
      return symbol;
    }
    
    // otherwise we have a class method, prefix must denote a class
    ClassSymbol csymbol = getClass(prefix, false);
    if (csymbol == null) 
    {
      if (verbose)
        error.newError("there is no class variable " + name);
      return null;
    }
    
    // lookup class variable in class
    ClassVariableSymbol symbol = csymbol.getClassVariables().get(base);
    if (symbol != null) return symbol;
    
    // no variable found in class
    if (verbose)
      error.newError("in class " + prefix + 
          " there is no class variable " + base); 
    return null;
  }
  
  /****************************************************************************
   * Get class of value.
   * @param vsymbol the value
   * @return the class of the value
   *        (null, if the type of the value does not denote a class)
   ***************************************************************************/
  private static ClassSymbol getClass(VariableSymbol vsymbol)
  {
    TypeExpression type = vsymbol.getType();
    if (!(type instanceof NamedType)) return null;
    NamedType ntype = (NamedType)type;
    return ntype.getSymbol();
  }
  
  // -------------------------------------------------------------------------
  //
  // Auxiliaries.
  //
  // -------------------------------------------------------------------------
  
  /***************************************************************************
   * Get types from parameter list.
   * @param params the parameter list.
   * @return the unique sequence of parameter types (null for error)
   *************************************************************************/
  private TypeExpression[] getTypes(Parameter[] params)
  {
    TypeTable ttable = top.getTypeTable();
    int n= params.length;
    TypeExpression[] types = new TypeExpression[n];
    for (int i=0; i<n; i++)
    {
      types[i] = params[i].type;
    }
    return ttable.process(types);
  }
  
  /***************************************************************************
   * Get unique version of type
   * @param the type
   * @return the unique version of the type (null for error)
   **************************************************************************/
  private TypeExpression getType(TypeExpression type)
  {
    return top.getTypeTable().process(type);
  }
}
