// ----------------------------------------------------------------------------
// ReferencedUnits.java
// Compute the units (class/theory) referenced by a unit.
// $Id: ReferencedUnits.java,v 1.11 2012/10/15 15:16:48 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.StringType;
import fmrisc.ProgramExplorer.Syntax.Program.StringLiteral;

/**
 * 
 * Compute the units (theory/class) referenced by a unit.
 *
 */
public final class ReferencedUnits extends ASTVisitorBase
{
  /***************************************************************************
   * Compute referenced units.
   * @param tree the source of the unit.
   * @return a collection of the symbols of the referenced units. 
   **************************************************************************/
  public static Collection<Symbol> compute(AST tree)
  {
    ReferencedUnits ref = new ReferencedUnits();
    tree.accept(ref);
    return ref.symbols;
  }
  
  // the collection of symbols computed up to now
  private Collection<Symbol> symbols = new SetClass<Symbol>();
  
  /****************************************************************************
   * Process identifier.
   * @param ident an identifier.
   ***************************************************************************/
  public void visit(Identifier ident) 
  { 
    Symbol symbol = ident.getSymbol();
    if (symbol instanceof ClassSymbol || symbol instanceof TheorySymbol)
      symbols.add(symbol);
  }
  
  /****************************************************************************
   * Process string type.
   * @param type a string type expression.
   ***************************************************************************/
  public void visit(StringType type) 
  { 
    ClassSymbol stringClass = type.getStringClass();
    if (stringClass == null) return;
    visit(stringClass.getName());
  }
  
  /****************************************************************************
   * Process string literal
   * @param literal a string literal
   ***************************************************************************/
  public void visit(StringLiteral literal) 
  { 
    ClassSymbol stringClass = literal.getStringClass();
    if (stringClass == null) return;
    visit(stringClass.getName());
  }
  
  /****************************************************************************
   * Filter all those units that (directly or indirectly) depend on given unit.
   * Assumes that all units have their direct dependencies computed.
   * @param symbol the symbol denoting a unit
   * @param symbols a collection of units
   * @return the set of all elements of symbols depending directly or 
   *         indirectly on the unit denoted by symbol.
   ***************************************************************************/
  public static Collection<Symbol> filterDependencies(Symbol symbol, 
      Collection<Symbol> symbols)
  {
    // construct dependency matrix
    MapClass<Symbol,Collection<Symbol>> dep = 
      new MapClass<Symbol,Collection<Symbol>>();
    
    // initialize matrix with (clones of) direct dependencies
    for (Symbol s : symbols)
    {
      if (s instanceof ClassSymbol)
      {
        ClassSymbol s0 = (ClassSymbol)s;
        Collection<Symbol> rsymbols = s0.getReferencedUnits();
        if (rsymbols == null) continue; // class may not yet be read
        dep.put(s0, new SetClass<Symbol>(rsymbols));
        continue;
      }
      if (s instanceof TheorySymbol)
      {
        TheorySymbol s0 = (TheorySymbol)s;
        Collection<Symbol> rsymbols = s0.getReferencedUnits();
        if (rsymbols == null) continue; // theory may not yet be read
        dep.put(s0, new SetClass<Symbol>(rsymbols));
        continue;
      }
    }
    symbols = dep.keySet();
    int n = symbols.size();
    
    // n-1 multiplication steps to determine indirect dependencies
    for (int i=0; i<n-1; i++)
    {
      for (Symbol s : symbols)
      {
        Collection<Symbol> rsymbols = dep.get(s);
        Collection<Symbol> rsymbols0 = new SetClass<Symbol>();
        for (Symbol r : rsymbols)
        {
          Collection<Symbol> rsymbols1 = dep.get(r);
          // dependency to not yet encountered unit may have been established
          if (rsymbols1 != null) rsymbols0.addAll(rsymbols1);
        }
        rsymbols.addAll(rsymbols0);
      }
    }
    
    // determine result
    Collection<Symbol> result = new SetClass<Symbol>();
    for (Symbol s : symbols)
    {
      Collection<Symbol> rsymbols = dep.get(s);
      if (rsymbols.contains(symbol)) result.add(s);
    }
    return result; 
  }
  
  /***************************************************************************
   * Compute closure of unit dependencies.
   * @param in a collection of unit (class/theory) symbols.
   * @param out a vector of unit (class/theory) symbols different from in.
   *        to which the reflexive transitive closure of the unit (class/theory) 
   *        dependencies of in are added in dependence order.
   **************************************************************************/
  public static void computeClosure(Collection<Symbol> in, Vector<Symbol> out)
  {       
    // determine symbol set and inverse references
    Map<Symbol,Set<Symbol>> order = new MapClass<Symbol, Set<Symbol>>();
    Collection<Symbol> out0 = new SetClass<Symbol>();
    Collection<Symbol> set0 = new SetClass<Symbol>(in);
    while (!set0.isEmpty())
    {
      out0.addAll(set0);
      Collection<Symbol> set1 = new SetClass<Symbol>();
      for (Symbol symbol : set0)
      {
        Collection<Symbol> ref = null;
        if (symbol instanceof ClassSymbol)
          ref = ((ClassSymbol)symbol).getReferencedUnits();
        else if (symbol instanceof TheorySymbol)
          ref = ((TheorySymbol)symbol).getReferencedUnits();
        else
          Main.internalError("unknown symbol");
        if (ref == null) continue; // current class
        for (Symbol r : ref)
        {
          Set<Symbol> symbols = order.get(r);
          if (symbols == null)
          {
            symbols = new SetClass<Symbol>();
            order.put(r, symbols);
          }
          symbols.add(symbol);
        }
        set1.addAll(ref);
      }
      set0 = set1;
      set0.removeAll(out0);
    }
    
    // remove self-dependencies from order
    Set<Symbol> keys = order.keySet();
    for (Symbol key : keys)
    {
      Set<Symbol> symbols = order.get(key);
      symbols.remove(key);
    }
    
    // compute transitive closure of order
    boolean modified;
    do
    {
      modified = false;
      for (Symbol key : keys)
      {
        Set<Symbol> symbols = order.get(key);
        Set<Symbol> symbols0 = new SetClass<Symbol>();
        for (Symbol symbol : symbols)
        {
          Set<Symbol> symbols1 = order.get(symbol);
          if (symbols1 == null) continue;
          symbols0.addAll(symbols1);
        }
        symbols0.removeAll(symbols);
        if (!symbols0.isEmpty())
        {
          symbols.addAll(symbols0);
          modified = true;
        }
      }
    } while (modified);
    
    // add to output in denoted order    
    while (!out0.isEmpty())
    {
      Set<Symbol> domain = new SetClass<Symbol>();
      for (Symbol s : out0)
      {
        Set<Symbol> symbols = order.get(s);
        if (symbols == null) continue;
        domain.addAll(symbols);
      }
      Symbol symbol = null;
      for (Symbol s : out0)
      {
        if (!domain.contains(s))
        {
          symbol = s;
          break;
        }
      }
      if (symbol == null) Main.internalError("no minimum symbol");
      out0.remove(symbol);
      out.add(symbol);
      for (Symbol s : out0)
      {
        Set<Symbol> symbols = order.get(s);
        if (symbols == null) continue;
        symbols.remove(symbol);
      }
    }
  }
}
