// ----------------------------------------------------------------------------
// TCCGenerator.java
// A generator of type-checking conditions (tccs).
// $Id: TCCGenerator.java,v 1.18 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;
import java.io.*;

import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Tasks.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Judgements.Normalization;

/**
 * 
 * A generator of type-checking conditions (tccs).
 *
 */
public class TCCGenerator
{
  private TheorySymbol tsymbol;              // the theory of the generator
  private Vector<TheorySymbol> theories;     // the referenced theories
  private LogicEnvironment env;              // the referenced declarations
  private File parent;                       // a directory for writing info
  private ErrorStream out;                   // a stream for printing messages
  
  // a stack of assumptions
  private Stack<Vector<Formula>> astack = null; 
  private Vector<Formula> assumptions = null; // current top of stack
  
  // a stack of generated proving tasks
  private Stack<Vector<TypeCheckingTask>> tstack = null;
  private Vector<TypeCheckingTask> tasks = null; // current top of stack           
  
  /***************************************************************************
   * Construct a type checking condition generator.
   * @param tsymbol the theory in which the tcc generator operates.
   * @param theories the list of all theories referenced in dependence order.
   * @param env the environment holding the referenced declarations
   * @param parent the directory used to create subdirectories for tcc tasks
   * @param out a stream for printing messages
   **************************************************************************/
  public TCCGenerator(TheorySymbol tsymbol,
    Vector<TheorySymbol> theories, LogicEnvironment env,
    File parent, ErrorStream out)
  {
    this.tsymbol = tsymbol;
    this.theories = theories;
    this.env = env;
    this.parent = parent;
    this.out = out;
    reset();
  }
  
  /***************************************************************************
   * Reset to initial state.
   **************************************************************************/
  private void reset()
  {
    tstack = new Stack<Vector<TypeCheckingTask>>();
    tasks = null;
    astack = new Stack<Vector<Formula>>();
    assumptions = null;
    open();
  }
  
  /***************************************************************************
   * Set new parent directory for generating task directories.
   * @param parent the new parent directory.
   * @return the old directory.
   **************************************************************************/
  public File setDirectory(File parent)
  {
    File old = this.parent;
    this.parent = parent;
    return old;
  }
  
  /***************************************************************************
   * Add generated tasks to task folder and reset state.
   * @param folder the folder to add the tasks to.
   **************************************************************************/
  public void addTasks(TaskFolder folder)
  {
    for (Task task : tasks)
      folder.add(task);
    reset();
  }
  
  /**************************************************************************
   * Create an assumption.
   * @param formula the formula representing the assumption.
   *************************************************************************/
  public void assume(Formula formula)
  {
    assumptions.add(formula);
  }
  
  /***************************************************************************
   * Create a new task to prove the given formula.
   * The formula is also added to the current set of assumptions.
   * @param name the name of the task
   * @param position a related source position (may be null)
   * @param formula the formula to be proved
   * @param resultType the current result type (null, if none)
   * @param stateVars the current state vars (null, if none)
   * @param stateExcs the current state exceptions (null, if none)
   **************************************************************************/
  public void prove(String name, SourcePosition position, 
    Formula formula, StateType resultType, Name[] stateVars, Name[] stateExcs)
  {
    Declaration[][] decls = env.getDeclarations();
    Formula goal = createGoal(formula);
    TypeCheckingTask task = new TypeCheckingTask(tsymbol,
        name, position, parent,
        resultType, stateVars, stateExcs,
        theories, decls, goal, out);
    tasks.add(task);
    assume(formula);
  }
  
  /***************************************************************************
   * Create goal formula as the implication of the
   * - conjunction of the assumptions context and
   * - the given formula.
   * If there are no assumptions, then the goal is just the formula itself.
   **************************************************************************/
  private Formula createGoal(Formula formula)
  {    
    Formula hypothesis = null;
    int n = astack.size();
    for (int i=0; i<n; i++)
    { 
      Vector<Formula> formulas = astack.elementAt(i);
      int m = formulas.size();
      for (int j=0; j<m; j++)
      {
        Formula f = formulas.elementAt(j);
        if (hypothesis == null)
          hypothesis = f;
        else
          hypothesis = AndFormula.construct(hypothesis, f);
      }
    }
    if (hypothesis == null) return formula;
    return Normalization.perform(ImpliesFormula.construct(hypothesis, formula));
  }
  
  /****************************************************************************
   * Open a new subcontext. Subcontexts may be arbitrarily nested; 
   * every subcontext must be ultimately either aborted or committed.
   * If the subcontext is aborted, every task created in it is discarded.
   * If the subcontext is committed, every task created in it
   * is added to the parent context.
   **************************************************************************/
  public void open()
  {
    assumptions = new Vector<Formula>();
    astack.push(assumptions);
    tasks = new Vector<TypeCheckingTask>();
    tstack.push(tasks);
  }
  
  /***************************************************************************
   * Return to the parent context discarding all the assumptions and tasks
   * that were added to the subcontext since the last call of open().
   **************************************************************************/
  public void abort()
  {
    astack.pop();
    assumptions = astack.peek();
    tstack.pop();
    tasks = tstack.peek();
  }
  
  /***************************************************************************
   * Return to the parent context and add to it all the tasks that were
   * created since the last call of open(). All assumptions added since the 
   * last call of open() are discarded.
   **************************************************************************/
  public void commit()
  {
    astack.pop();
    assumptions = astack.peek();
    tstack.pop();
    Vector<TypeCheckingTask> tasks0 = tstack.peek();
    tasks0.addAll(tasks);
    tasks = tasks0;
  }
  
  /***************************************************************************
   * Return depth of context stack.
   * @return the depth.
   **************************************************************************/
  public int getDepth()
  {
    return astack.size();
  }
}
