// ----------------------------------------------------------------------------
// ASTPrinter.java
// A printer for abstract syntax trees.
// $Id: ASTPrinter.java,v 1.54 2012/10/18 16:12:57 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax;

import java.io.*;
import java.util.*;
import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Logic.ArrayType;
import fmrisc.ProgramExplorer.Syntax.Logic.BooleanType;
import fmrisc.ProgramExplorer.Judgements.*;

/**
 * 
 * A printer for abstract syntax trees.
 *
 */
public class ASTPrinter extends ASTVisitorBase
{
  /*
   * print command specification:
   * NONE: no printing.
   * TOP: print only specification of top-most command.
   * ALL: print specification for all commands (recursively).
   */
  public enum PrintStatementSpec { NONE, TOP, ALL };
  
  /****************************************************************************
   * Print tree on writer including specification commands.
   * @param tree the tree to be printed.
   * @param out the writer to be used for printing.
   * @param printStatSpec the flag to be used for printing command specifications.
   * @param printAnnotation true if annotations are to be printed.
   ***************************************************************************/
  public static void print(AST tree, PrintWriter out, 
    PrintStatementSpec printStatSpec, boolean printAnnotation)
  {
    tree.accept(new ASTPrinter(out, printStatSpec, printAnnotation));
    out.flush();
  }
  
  // --------------------------------------------------------------------------
  // Internals.
  // --------------------------------------------------------------------------
  
  private PrintStatementSpec printStatSpec; // print command specification
  private boolean printAnnotation;          // print annotations
  private int level;                // indentation level
  private final PrintWriter out;    // output writer
  
  // if true print parentheses around every expression
  private final boolean expPars = false;
  
  /****************************************************************************
   * Create printer.
   * @param out the writer to be used for printing
   * @param printStatSpec the flag to be used for printing command specifications.
   * @param printAnnotation true if annotations are to be printed.
   ***************************************************************************/
  protected ASTPrinter(PrintWriter out, PrintStatementSpec printStatSpec,
    boolean printAnnotation)
  {
    this.level = 0;
    this.out = out;
    this.printStatSpec = printStatSpec;
    this.printAnnotation = printAnnotation;
  }
  
  /****************************************************************************
   * Print string.
   * @param s the string.
   ***************************************************************************/
  protected void print(String s)
  {
    out.print(s);
  }
  
  /****************************************************************************
   * Print boolean.
   * @param b the boolean.
   ***************************************************************************/
  private void print(boolean b)
  {
    if (b) print("TRUE"); else print("FALSE");
  }
  
  /****************************************************************************
   * Print collection of class symbols.
   * @param c the collection.
   ***************************************************************************/
  private void printClasses(Collection<ClassSymbol> c)
  {
    print("{");
    if (c != null)
    {
      Iterator<ClassSymbol> iter = c.iterator();
      while (iter.hasNext())
      {
        ClassSymbol s = iter.next();
        print(s.getLongName());
        if (iter.hasNext()) print(",");
      }
    }
    print("}");
  }
  
  /****************************************************************************
   * Print collection of variable symbols
   * @param c the collection.
   ***************************************************************************/
  private void printVariables(Collection<VariableSymbol> c)
  {
    print("{");
    if (c != null)
    {
      Iterator<VariableSymbol> iter = c.iterator();
      while (iter.hasNext())
      {
        VariableSymbol s = iter.next();
        print(NameUtils.uniqueName(s));
        if (iter.hasNext()) print(",");
      }
    }
    print("}");
  }
  
  /****************************************************************************
   * Start a new line and indent appropriately.
   ***************************************************************************/
  protected void newLine()
  {
    out.println();
  }
  
  /****************************************************************************
   * Create indentation up to right level.
   ***************************************************************************/
  protected void tab()
  {
    for (int i=0; i < level; i++)
      print("  ");
  }
  
  /****************************************************************************
   * Increase/decrease indentation.
   * @param tab if true, increase indentation, else decrease.
   ***************************************************************************/
  private void indent(boolean tab)
  {
    if (tab)
      level = level+1;
    else
      level = level-1;
  }
  
  /****************************************************************************
   * Print an abstract syntax tree.
   * @param tree the abstract syntax tree.
   ***************************************************************************/
  private void print(AST tree)
  {
    tree.accept(this);
  }
  
  /****************************************************************************
   * Print an abstract syntax tree and start a new line with indentation.
   * @param tree the abstract syntax tree.
   ***************************************************************************/
  private void println(AST tree)
  {
    print(tree);
    newLine();
  }
   
  /****************************************************************************
   * Print a sequence of trees each on a new line and surrounded by "{ ... }".
   * @param trees the trees to be printed.
   * @param empty if an empty line is to be printed between elements
   ***************************************************************************/
  protected void printBlock(AST[] trees, boolean empty)
  {
    print("{");
    indent(true);
    int n = trees.length;
    for (int i=0; i<n; i++)
    {
      newLine();
      tab();
      print(trees[i]);
      if (empty && i<n-1) newLine();
    }
    indent(false);
    newLine();
    tab();
    print("}");
  }
  
  /****************************************************************************
   * Print body of an enclosing entity.
   * @param body the body to be printed.
   * @param block iff true body is to be printed as block with "{ ... }".
   ***************************************************************************/
  protected void printBody(Statement body, boolean block)
  {
    if (body instanceof BlockStatement)
    {
      newLine();
      tab();
      BlockStatement b = (BlockStatement)body;
      printBlock(b.stats, false);
    }
    else if (block)
    {
      print("{");
      indent(true);
      newLine();
      tab();
      print(body);
      indent(false);
      print("}");
    }
    else
    {
      indent(true);
      newLine();
      tab();
      print(body);
      indent(false);
    }
  }
  
  /****************************************************************************
   * Print sequence of trees separated by separator string.
   * @param trees the sequence of trees
   * @param sep the separator string
   * @param from the index of the first element of the sequence to be printed
   ***************************************************************************/
  private void printList(AST[] trees, String sep, int from)
  {
    int n = trees.length;
    if (n <= from) return;
    print(trees[from]);
    for (int i=from+1; i<n; i++)
    {
      print(sep);
      print(trees[i]);
    }   
  }
  
  /****************************************************************************
   * Print sequence of trees separated by separator string.
   * @param trees the sequence of trees
   * @param sep the separator string
   ***************************************************************************/
  private void printList(AST[] trees, String sep)
  {
    printList(trees, sep, 0);
  }
  
  // --------------------------------------------------------------------------
  // Classes.
  // --------------------------------------------------------------------------
  
  private void printUnitHeader(CompilationUnit c)
  {
    printAnnotation(c, true);
    Name pname = c.getPackage();
    if (pname != null)
    {
      print("package ");
      print(pname);
      print(";");
      newLine();
    }
    print(c.getImports());
  }
  
  private void print(Import[] imports)
  {
    int n = imports.length;
    for (int i=0; i<n; i++)
    {
      print(imports[i]);
      newLine();
    }
  }
  
  public void visit(Import imp)
  {
    Main.internalError("unknown import " + imp);
  }
  
  public void visit(PackageImport imp)
  {
    print("import ");
    print(imp.getName());
    print(".*");
    print(";");
    // return true;
  }
  
  public void visit(ClassImport imp)
  {
    print("import ");
    print(imp.getName());
    print(";");
    // return true;
  }
  
  public void visit(ClassDeclaration c)
  {
    printUnitHeader(c);
    print("public class ");
    println(c.name);
    printClassSpec(c);
    printBlock(c.decls, true);
    newLine();
    // return true;
  }
  
  private void printClassSpec(ClassDeclaration c)
  {
    ClassSpec spec = c.getClassSpec();
    if (spec == null) return;
    print(" /*@ ");
    indent(true);
    print(spec);
    indent(false);
    newLine(); tab();
    print("@*/");
    newLine(); tab();
  }
  
  public void visit(GlobalVariableDeclaration g)
  {
    print("public ");
    if (g instanceof ClassVariableDeclaration) print("static ");
    print(g.type);
    print(" ");
    print(g.name);
    if (g.value != null)
    {
      print(" = ");
      print(g.value);
    }
    print(";");
    // return true;
  }
  
  public void visit(ConstructorDeclaration g)
  {
    print("public ");
    print(g.name);
    print("(");
    printList(g.params, ", ");
    print(") ");
    printAnnotation(g, false);
    printBody(g.body, true);
    // return true;
  }
  
  public void visit(MethodDeclaration m)
  {
    print("public ");
    if (m instanceof ClassMethodDeclaration) print("static ");
    print(m.type);
    print(" ");
    print(m.name);
    print("(");
    printList(m.params, ", ");
    print(") ");
    printAnnotation(m, false);
    printBody(m.body, true);
    // return true;
  }
  
  private void printLineAnnotation(ProgramAnnotation annotation)
  {
    if (annotation == null) return;
    print("//@ ");
    print(annotation);
    newLine(); tab();
  }
  
  private void printLineAnnotation(ASTProgramAnnotated d)
  {
    printLineAnnotation(d.getAnnotation());
  }
  
  private void printAnnotation(ProgramAnnotation annotation, boolean newLine)
  {
    if (!printAnnotation || annotation == null) return;
    print("/*@ ");
    indent(true);
    print(annotation);
    indent(false);
    newLine(); tab();
    print("@*/");
    if (newLine) { newLine(); tab(); }
  }
  
  private void printAnnotation(ASTProgramAnnotated d, boolean newLine)
  {
    printAnnotation(d.getAnnotation(), newLine);
  }
  
  // --------------------------------------------------------------------------
  // Statements.
  // --------------------------------------------------------------------------
  
  public void visit(StatementBase tree) 
  { 
    PrintStatementSpec old = printStatSpec;
    if (printStatSpec == PrintStatementSpec.TOP)
      printStatSpec = PrintStatementSpec.NONE;
    if (tree instanceof AssertionStatement) 
      visitCore((AssertionStatement)tree);
    else if (tree instanceof AssignmentStatement) 
      visitCore((AssignmentStatement)tree);
    else if (tree instanceof BlockStatement) 
      visitCore((BlockStatement)tree);
    else if (tree instanceof BreakStatement) 
      visitCore((BreakStatement)tree);
    else if (tree instanceof CallStatement) 
      visitCore((CallStatement)tree);
    else if (tree instanceof ConditionalStatement) 
      visitCore((ConditionalStatement)tree);
    else if (tree instanceof ContinueStatement) 
      visitCore((ContinueStatement)tree);
    else if (tree instanceof EmptyStatement) 
      visitCore((EmptyStatement)tree);
    else if (tree instanceof ForLoopStatement) 
      visitCore((ForLoopStatement)tree);
    else if (tree instanceof ReturnStatement) 
      visitCore((ReturnStatement)tree);
    else if (tree instanceof TryCatchStatement) 
      visitCore((TryCatchStatement)tree);
    else if (tree instanceof WhileLoopStatement) 
      visitCore((WhileLoopStatement)tree);
    else
    {
      Main.internalError("unknown kind of statement");
    }
    if (old != PrintStatementSpec.NONE)
    {
      StatJudgement judgement = tree.getJudgement();
      if (judgement != null)
      {
        print(":");
        print(judgement);
      }
    }
    printStatSpec = old;
  }

  public void visit(fmrisc.ProgramExplorer.Judgements.StatJudgement spec)
  {
    if (spec == null) return;
    print("[");
    print(spec.sformula);
    print("]^");
    print(spec.continues);
    print(",");
    print(spec.breaks);
    print(",");
    print(spec.returns);
    print("_");
    printClasses(spec.exceptions);
    print(",");
    printVariables(spec.variables);
    // return true;
  }
  
  private void visitCore(EmptyStatement s)
  {
    printLineAnnotation(s);
    print(";");
    // return true;
  }
  
  private void visitCore(BlockStatement s)
  {
    printLineAnnotation(s);
    printBlock(s.stats, false);
    // return true;
  }
  
  private void printCore(AssignmentStatement s)
  {
    printLineAnnotation(s);
    print(s.lval);
    print(" = ");
    print(s.exp);
  }
  
  private void visitCore(AssignmentStatement s)
  {
    printLineAnnotation(s);
    printCore(s);
    print(";");
    // return true;
  }
  
  private void visitCore(CallStatement s)
  {
    printLineAnnotation(s);
    if (s instanceof AssignCallStatement)
    {
      AssignCallStatement a = (AssignCallStatement)s;
      print(a.lval);
      print(" = ");
      if (s instanceof AssignNewStatement) print("new ");
    }
    else if (s instanceof DeclCallStatement)
    {
      DeclCallStatement d = (DeclCallStatement)s;
      print(d.type);
      print(" ");
      print(d.name);
      print(" = ");
      if (s instanceof DeclNewStatement) print("new ");
    }
    else if (s instanceof ThrowStatement) 
    {
      visitCore((ThrowStatement)s);
      return;
    }
    print(s.method);
    print("(");
    printList(s.args, ", ");
    print(");");
    // return true;
  }
  
  public void visit(VariableStatement s)
  {
    printLineAnnotation(s);
    print(s.type);
    print(" ");
    print(s.name);
    if (s.exp != null)
    {
      print(" = ");
      print(s.exp);
    }
    print(";");
    if (printStatSpec != PrintStatementSpec.NONE)
    {
     StatJudgement judgement = s.getJudgement();
      if (judgement != null)
      {
        print(":");
        print(judgement);
      }
    }
    // return true;
  }
  
  private void visitCore(ConditionalStatement s)
  {
    printLineAnnotation(s);
    print("if (");
    print(s.cond);
    print(")");
    printBody(s.thenstat, false);
    if (s.elsestat != null)
    {
      newLine();
      tab();
      print("else");
      if (s.elsestat instanceof ConditionalStatement)
      {
        print(" ");
        print(s.elsestat);
      }
      else
        printBody(s.elsestat, false);
    }
    // return true;
  }
  
  private void visitCore(WhileLoopStatement s)
  {
    printLineAnnotation(s);
    print("while (");
    print(s.cond);
    print(")");
    printAnnotation(s.getLoopAnnotation(), false);
    printBody(s.body, false);
    // return true;
  }
  
  private void visitCore(ForLoopStatement s)
  {
    printLineAnnotation(s);
    print("for (");
    print(s.init);
    print(" ");
    print(s.cond);
    print("; ");
    if (s.update instanceof AssignmentStatement)
      printCore((AssignmentStatement)s.update);
    else
      Main.internalError("unknown kind of update statement");
    print(")");
    printAnnotation(s.getLoopAnnotation(), false);
    printBody(s.body, false);
    // return true;
  }
  
  private void visitCore(ContinueStatement s)
  {
    printLineAnnotation(s);
    print("continue;");
    // return true;
  }
  
  private void visitCore(BreakStatement s)
  {
    printLineAnnotation(s);
    print("break;");
    // return true;
  }
  
  private void visitCore(ReturnStatement s)
  {
    printLineAnnotation(s);
    print("return");
    if (s.value != null)
    {
      print(" ");
      print(s.value);
    }
    print(";");
    // return true;
  }
  
  private void visitCore(ThrowStatement s)
  {
    printLineAnnotation(s);
    print("throw new ");
    print(s.method);
    print("(");
    print(s.getValue());
    print(")");
    print(";");
    // return true;
  }
  
  private void visitCore(TryCatchStatement s)
  {
    printLineAnnotation(s);
    print("try");
    printBody(s.body, true);
    Handler[] handlers = s.handlers;
    int n = handlers.length;
    for (int i=0; i<n; i++)
    { 
      newLine();
      tab();
      print(handlers[i]);
    }
    // return true;
  }
  
  public void visit(Handler s)
  {
    print("catch (");
    print(s.param);
    print(")");
    printBody(s.body, true);
    // return true;
  }
  
  private void visitCore(AssertionStatement s)
  {
    printLineAnnotation(s);
    print("assert ");
    print(s.cond);
    print(";");
    // return true;
  }
  // --------------------------------------------------------------------------
  // Value expressions.
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Return operator of expression.
   * @param exp an expression
   * @return the operator of the expression (null, if it has none)
   ***************************************************************************/
  private String getOperator(ValueExpression exp)
  {
    if (exp instanceof BinaryValueExpression)
    {
      BinaryValueExpression e = (BinaryValueExpression)exp;
      return e.getOperator();
    }
    if (exp instanceof UnaryValueExpression)
    {
      UnaryValueExpression e = (UnaryValueExpression)exp;
      return e.getOperator();
    }
    return null;
  }
  
  /****************************************************************************
   * Print expression, possibly with parentheses.
   * @param precedence if expression has binary operator with precedence 
   *                   less than or equal this value, print with parentheses.
   * @param exp the expression to be printed. 
   * @param skip iff true and expression has operator op, then skip parentheses
   * @param op the operator for skipping               
   ***************************************************************************/
  private void printPar(int precedence, ValueExpression exp, 
    boolean skip, String op)
  {
    int p = exp.getPrecedence();
    boolean parens;
    if (exp instanceof VariableValue || 
        exp instanceof fmrisc.ProgramExplorer.Syntax.Program.IntLiteral || 
        exp instanceof fmrisc.ProgramExplorer.Syntax.Program.BooleanLiteral || 
        exp instanceof fmrisc.ProgramExplorer.Syntax.Program.StringLiteral ||
        exp instanceof CharLiteral)
      parens = false; 
    else if (expPars)
      parens = true;
    else if (p > precedence)
      parens = false;
    else if (p < precedence)
      parens = true;
    else if (!skip)
      parens = true;
    else
    {
      String op0 = getOperator(exp);
      parens = op0 == null || !op.equals(op0);
    }
    if (parens) print("(");
    print(exp);
    if (parens) print(")");
  }
  
  public void visit(BinaryValueExpression exp)
  {
    int p = exp.getPrecedence();
    String op = exp.getOperator();
    
    // associative operator?
    boolean assoc;
    if (exp instanceof AndExpression) assoc = true;
    else if (exp instanceof OrExpression) assoc = true;
    else if (exp instanceof PlusExpression) assoc = true;
    else if (exp instanceof TimesExpression) assoc = true;
    else assoc = false;
    
    // space around operator?
    boolean space;
    if (exp instanceof AndExpression) space = true;
    else if (exp instanceof OrExpression) space = true;
    else if (exp instanceof LessExpression) space = true;
    else if (exp instanceof LessEqualExpression) space = true;
    else if (exp instanceof GreaterExpression) space = true;
    else if (exp instanceof GreaterEqualExpression) space = true;
    else if (exp instanceof EqualsExpression) space = true;
    else if (exp instanceof NotEqualsExpression) space = true;
    else space = false;
    
    // print first expression 
    // if same operator, no parentheses
    printPar(p, exp.first, true, op);
    
    // print operator, possibly with surrounding spaces
    if (space) print(" ");
    print(exp.getOperator());
    if (space) print(" ");
    
    // print second expression
    // if same precedence and same associative operator, no parentheses
    printPar(p, exp.second, assoc, op);
    // return true;
  }
  
  public void visit(UnaryValueExpression exp)
  {
    int p = exp.getPrecedence();
    String op = exp.getOperator();
    print(op);
    printPar(p, exp.base, true, null);
    // return true;
  }
  
  public void visit(SelectorExpression exp)
  {
    printPar(exp.getPrecedence(), exp.base, false, null);
    fmrisc.ProgramExplorer.Syntax.Program.Selector[] sels = exp.sels;
    int n = sels.length;
    if (n == 0) Main.internalError("no selector");
    print(sels[0]);
    for (int i=1; i<n; i++)
    {
      print(sels[i]);
    }
    // return true;
  }
  
  public void visit(NewArrayExpression exp)
  {
    print("new ");
    print(exp.type);
    print("[");
    print(exp.length);
    print("]");
    // return true;
  }
  
  public void visit(VariableValue exp)
  {
    print(exp.ident);
    // return true;
  }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.IntLiteral exp)
  {
    print(exp.digits);
    // return true;
  }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.BooleanLiteral exp)
  {
    if (exp.value)
      print("true");
    else
      print("false");
    // return true;
  }
  
  public void visit(CharLiteral exp)
  {
    print("\'" + StringUtil.printable(exp.value) + "\'");
    // return true;
  }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.StringLiteral exp)
  {
    print("\"" + StringUtil.printable(exp.value) + "\"");
    // return true;
  }
  
  public void visit(NullLiteral exp)
  {
    print("null");
    // return true;
  }
  
  // --------------------------------------------------------------------------
  // Type expressions.
  // --------------------------------------------------------------------------
  
  public void visit(PrimitiveType type)
  {
    print(type.name);
    // return true;
  }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.ArrayType type)
  {
    print(type.base);
    print("[]");
    // return true;
  }
  
  public void visit(NamedType type)
  {
    print(type.name);
    // return true;
  }
  
  public void visit(NewType type)
  {
    print(type.base);
    // return true;
  }
  
  // --------------------------------------------------------------------------
  // Miscellaneous program constructs
  // --------------------------------------------------------------------------
  
  public void visit(LeftValue l)
  {
    print(l.name);
    printList(l.selectors, "");
    // return true;
  }
  
  public void visit(IndexSelector s)
  {
    print("[");
    print(s.index);
    print("]");
    // return true;
  }
  
  public void visit(IdentifierSelector s)
  {
    print(".");
    print(s.name);
    // return true;
  }
  
  public void visit(MessageSelector s)
  {
    print(".getMessage()");
    // return true;
  }
  
  public void visit(Parameter p)
  {
    print(p.type);
    print(" ");
    print(p.name);
    // return true;
  }
  
  /****************************************************************************
   * Print name (but do not print first identifier in name, if it is empty)
   * @param name the name to be printed.
   ***************************************************************************/
  public void visit(Name name)
  {
    Identifier[] idents = name.idents;
    int n = name.idents.length;
    if (n == 0) Main.internalError("empty name encountered");
    Identifier ident = idents[0];
    if (ident.getString().equals(""))
      printList(name.idents, ".", 1);
    else
      printList(name.idents, ".", 0);
    // return true;
  }
  
  public void visit(Identifier i)
  {
    print(i.getString());
    // return true;
  }
  
  // --------------------------------------------------------------------------
  // Program specifications.
  // --------------------------------------------------------------------------
 
  public void visit(TheoryDeclaration c)
  {
    printUnitHeader(c);
    print("theory ");
    print(c.name);
    Name[] theories = c.getTheories();
    if (theories.length > 0)
    {
      print(" uses ");
      printList(theories, ", ");
    }
    newLine(); tab();
    print("{");
    indent(true);
    print(c.getDeclarations());
    indent(false);
    newLine(); tab();
    print("}");
    // return true;
  }
  
  public void visit(UnitSpec spec)
  {
    print(spec.getImports());
    newLine(); tab();
    print("theory");
    newLine(); tab();
    print("{");
    indent(true);
    print(spec.getDeclarations());
    indent(false);
    newLine(); tab();
    print("}");
    // return true;
  }
  
  private void print(Declaration[] decls)
  {
    for (Declaration d : decls)
    {
      newLine(); tab();
      print(d);
      print(";");
    }
  }
  
  public void visit(ClassSpec spec)
  {
    Formula invariant = spec.getClassInvariant();
    if (invariant != null)
    {
      newLine(); tab();
      print(ClassSpec.invariantKeyword);
      print(" ");
      print(invariant);
      print(";");
    }
    // return true;
  }
  
  public void visit(MethodSpec spec)
  {
    // "requires" clause
    Formula pre = spec.getPreCondition();
    if (pre != null)
    {
      newLine(); tab();
      print(MethodSpec.preKeyword);
      print(" ");
      print(pre);
      print(";");
    }
    
    // "assignable" clause
    Name[] vars = spec.getVariables();
    if (vars != null)
    {
      newLine(); tab();
      print(MethodSpec.variablesKeyword);
      print(" ");
      printList(vars, ", ");
      print(";");
    }
    
    // "signals" clause
    Name[] exceptions = spec.getExceptions();
    if (exceptions != null)
    {
      newLine(); tab();
      print(MethodSpec.exceptionsKeyword);
      print(" ");
      printList(exceptions, ", ");
      print(";");
    }
    
    // "ensures" clause
    Formula post = spec.getPostCondition();
    if (post != null)
    {
      newLine(); tab();
      print(MethodSpec.postKeyword);
      print(" ");
      print(post);
      print(";");
    }
    
    // "diverges" clause
    Formula div = spec.getDivergesCondition();
    if (div != null)
    {
      newLine(); tab();
      print(MethodSpec.divergesKeyword);
      print(" ");
      print(div);
      print(";");
    }

    // "decreases" clause
    Term[] term = spec.getTerminationTerm();
    if (term != null)
    {
      newLine(); tab();
      print(MethodSpec.terminationKeyword);
      print(" ");
      int n = term.length;
      for (int i=0; i<n; i++)
      {
        print(term[i]);
        if (i+1 < n) print(", ");
      }
      print(";");
    }
    
    // return true;
  }
  
  public void visit(LoopSpec spec)
  {
    // "invariant" clause
    Formula invariant = spec.getLoopInvariant();
    if (invariant != null)
    {
      newLine(); tab();
      print(LoopSpec.invariantKeyword);
      print(" ");
      print(invariant);
      print(";");
    }
    
    // "decreases" clause
    Term[] term = spec.getTerminationTerm();
    if (term != null)
    {
      newLine(); tab();
      print(LoopSpec.terminationKeyword);
      print(" ");
      int n = term.length;
      for (int i=0; i<n; i++)
      {
        print(term[i]);
        if (i+1<n) print(", ");
      }
      print(";");
    }
    
    // return true;
  }
  
  public void visit(StatementSpec spec)
  {
    // "assert" clause
    Formula assertion = spec.getAssertion();
    if (assertion != null)
    {
      print(StatementSpec.assertionKeyword);
      print(" ");
      print(assertion);
      print(";");
    }
    
    // return true;
  }
  
  // --------------------------------------------------------------------------
  // Logical declarations/definitions
  // --------------------------------------------------------------------------
  
  public void visit(FormulaDefinition def)
  {
    print(def.getName());
    print(": ");
    if (def instanceof AxiomDefinition)
      print(AxiomDefinition.keyword);
    else
      print(FormulaDefinition.keyword);
    print(" ");
    print(def.getValue());
    // return true;
  }
  
  public void visit(ValueDefinitionClass def)
  {
    print(def.getName());
    print(" = ");
    print(def.getValue());
    // return true;
  }
  
  public void visit(TypeDeclaration decl)
  {
    print(decl.getName());
    print(": ");
    print(TypeDeclaration.keyword);
    if (decl instanceof TypeDefinition)
    {
      print(" = ");
      TypeDefinition def = (TypeDefinition)decl; 
      print(def.getValue());
    }
    // return true;
  }
 
  public void visit(ValueDeclarationClass decl)
  {
    print(decl.getName());
    print(": ");
    print(decl.getType());
    if (decl instanceof ValueDeclarationInitialized)
    {
      print(" = ");
      ValueDeclarationInitialized def = (ValueDeclarationInitialized)decl; 
      print(def.getValue());
    }
    if (decl instanceof ValueDeclarationFormula)
    {
      print(" <=> ");
      ValueDeclarationFormula def = (ValueDeclarationFormula)decl; 
      print(def.getFormula());
    }
    // return true;
  }
  
  // --------------------------------------------------------------------------
  // Logical selectors
  // --------------------------------------------------------------------------
  
  public void visit(SelectorIdentifier sel)
  {
    print(".");
    print(sel.getIdentifier());
    // return true;
  }
  
  public void visit(SelectorIndex sel)
  {
    print("[");
    print(sel.getIndex());
    print("]");
    // return true;
  }
  
  public void visit(SelectorNumber sel)
  {
    print(".");
    print(sel.getNumber());
    // return true;
  }
  
  // --------------------------------------------------------------------------
  // Typed/Valued Identifiers
  // --------------------------------------------------------------------------
 
  public void visit(TypedIdentifier ident)
  {
    print(ident.getName());
    print(": ");
    print(ident.getType());
    // return true;
  }
 
  public void visit(ValuedIdentifier ident)
  {
    print(ident.getIdentifier());
    print(":=");
    print(ident.getValue());
    // return true;
  }
  
  // --------------------------------------------------------------------------
  // Expressions
  // --------------------------------------------------------------------------
    
  private void printLeftPar(Expression exp, Expression outer)
  {
    if (!expPars &&
          (exp.getPrecedence() >= outer.getPrecedence() ||
             (exp.getClass() == outer.getClass() && exp.isLeftAssociative())))
    {
      print(exp);
      return;
    }
    print("(");
    print(exp);
    print(")");
  }
  
  private void printRightPar(Expression exp, Expression outer)
  {
    if (!expPars &&
          (exp.getPrecedence() >= outer.getPrecedence() ||
             (exp.getClass() == outer.getClass() && exp.isRightAssociative())))
    {
      print(exp);
      return;
    }
    print("(");
    print(exp);
    print(")");
  }
  
  public void visit(AtomicFormula exp)
  {
    print(exp.getFunction());
    print("(");
    printList(exp.getArguments(), ", ");
    print(")");
    // return true;
  }
  
  public void visit(FunctionApplication exp)
  {
    print(exp.getFunction());
    print("(");
    printList(exp.getArguments(), ", ");
    print(")");
    // return true;
  }
  
  public void visit(BinaryAtomicFormula exp)
  {
    print(exp.getFirst());
    print(" ");
    print(exp.getOperator());
    print(" ");
    print(exp.getSecond());
    // return true;
  }
  
  public void visit(BinaryExpression exp)
  {
    // if (exp instanceof BinaryAtomicFormula)
    //   return visit((BinaryAtomicFormula)exp);
    printLeftPar(exp.getFirst(), exp);
    if (exp instanceof Term)
      print(exp.getOperator());
    else
    {
      print(" ");
      print(exp.getOperator());
      print(" ");
    }
    printRightPar(exp.getSecond(), exp);
    // return true;
  }
  
  public void visit(Constant exp)
  {
    print(exp.getName());
    // return true;
  }
  
  public void visit(WritesOnlyFormula exp)
  {
    // if (exp instanceof ReadsOnlyFormula)
    // {
    //   print(ReadsOnlyFormula.keyword);
    //   return true;
    // }
    print(WritesOnlyFormula.keyword);
    print(" ");
    printList(exp.getVariables(), ", ");
    // return true;
  }
  
  public void visit(IfThenElseExpression exp)
  {
    print(IfThenElseExpression.ifKeyword);
    print(" ");
    print(exp.getCondition());
    print(" ");
    print(IfThenElseExpression.thenKeyword);
    print(" ");
    print(exp.getThenBranch());
    print(" ");
    if (IfThenElseExpression.elsifKeyword == null)
    {
      print(IfThenElseExpression.elseKeyword);
      print(" ");
      print(exp.getElseBranch());
      if (IfThenElseExpression.endifKeyword != null)
      {
        print(" ");
        print(IfThenElseExpression.endifKeyword);
        return;
      }
    }
    Expression elsebranch = exp.getElseBranch();
    while (elsebranch instanceof IfThenElseExpression)
    {
      IfThenElseExpression exp0 = (IfThenElseExpression)elsebranch;
      print(IfThenElseExpression.elsifKeyword);
      print(" ");
      print(exp0.getCondition());
      print(" ");
      print(IfThenElseExpression.thenKeyword);
      print(" ");
      print(exp0.getThenBranch());
      print(" ");
      elsebranch = exp0.getElseBranch();
    }
    print(IfThenElseExpression.elseKeyword);
    print(" ");
    print(exp.getElseBranch());
    if (IfThenElseExpression.endifKeyword != null)
    {
      print(" ");
      print(IfThenElseExpression.endifKeyword);
    }
    // return true;
  }
  
  public void visit(LetExpression exp)
  {
    print(LetExpression.letKeyword);
    print(" ");
    printList(exp.getDefinitions(), ", ");
    print(" ");
    print(LetExpression.inKeyword);
    print(" ");
    printRightPar(exp.getBody(), exp);
    // return true;
  }

  public void visit(PostfixApplication exp)
  {
    // if (exp instanceof StateThrowsException)
    //   return visit((StateThrowsException)exp);
    print(exp.getIdentifier());
    print(PostfixApplication.operator);
    printRightPar(exp.getArgument(), exp);
    // return true;
  }
 
  public void visit(StateThrowsException exp)
  {
    print(exp.getIdentifier());
    print("(");
    print(exp.getException());
    print(")");
    print(PostfixApplication.operator);
    printRightPar(exp.getArgument(), exp);
    // return true;
  }
  
  public void visit(QuantifiedExpression exp)
  {
    print(exp.getQuantifier());
    print("(");
    printList(exp.getVariables(), ", ");
    print("): ");
    printRightPar(exp.getBody(), exp);
    // return true;
  }
  
  public void visit(Reference exp)
  {
    print(exp.getName());
    // return true;
  }
  
  public void visit(ProgramVariableBase exp)
  {
    print(exp.getKeyword());
    print(" ");
    print(exp.getName());
    // return true;
  }
  
  public void visit(RecordTerm exp)
  {
    print(RecordTerm.beginToken);
    printList(exp.getBase(), ", ");
    print(RecordTerm.endToken);
    // return true;
  }
  
  public void visit(SelectionTerm exp)
  {
    printLeftPar(exp.getBase(), exp);
    print(exp.getSelector());
    // return true;
  }
  
  public void visit(TupleTerm exp)
  {
    print(TupleTerm.beginToken);
    printList(exp.getBase(), ", ");
    print(TupleTerm.endToken);
    // return true;
  }
  
  public void visit(UpdateTerm exp)
  {
    printLeftPar(exp.getBase(), exp);
    print(" ");
    print(UpdateTerm.withKeyword);
    print(" ");
    printList(exp.getSelectors(),"");
    print(UpdateTerm.assignToken);
    printRightPar(exp.getValue(), exp);
    // return true;
  }
  
  public void visit(ArrayType exp)
  {
    print(ArrayType.arrayKeyword);
    print(" ");
    print(exp.getIndex());
    print(" ");
    print(ArrayType.baseKeyword);
    print(" ");
    printRightPar(exp.getBase(), exp);
    // return true;
  }
  
  public void visit(AtomicType exp)
  {
    print(exp.getName());
    // return true;
  }
  
  public void visit(FunctionType exp)
  {
    Type[] domain = exp.getDomain();
    Type range = exp.getRange();
    if (range instanceof BooleanType)
    {
      print("PREDICATE(");
      printList(domain, ", ");
      print(")");
      return;
    }
    if (domain.length == 1)
      printLeftPar(domain[0], exp);
    else
    {
      print("(");
      printList(domain, ", ");
      print(")");
    }
    print(" -> ");
    printRightPar(range, exp);
    // return true;
  }

  public void visit(RecordType exp)
  {
    print(RecordType.beginToken);
    printList(exp.getBase(), ", ");
    print(RecordType.endToken);
    // return true;
  }
  
  public void visit(StateType exp)
  {
    print(StateType.stateKeyword);
    if (exp.getReturn() != null)
    {
      print("(");
      print(exp.getReturn());
      print(")");
    }
    // return true;
  }
  
  public void visit(SubrangeType exp)
  {
    print(SubrangeType.beginToken);
    print(exp.getFrom());
    print(SubrangeType.separatorToken);
    print(exp.getTo());
    print(SubrangeType.endToken);
    // return true;
  }
 
  public void visit(SubType exp)
  {
    print(SubType.keyword);
    print("(");
    print(exp.getPredicate());
    print(")");
    // return true;
  }
  
  public void visit(TupleType exp)
  {
    print(TupleType.beginToken);
    printList(exp.getBase(), ", ");
    print(TupleType.endToken);
    // return true;
  }
  
  public void visit(UnaryExpression exp)
  {
    String op = exp.getOperator();
    print(op);
    int n = op.length();
    char c = op.charAt(n-1);
    if (Character.isLetter(c)) print(" ");
    printRightPar(exp.getBase(), exp);
    // return true;
  }
}
