// ----------------------------------------------------------------------------
// NameUtils.java
// Utilities related to names.
// $Id: NameUtils.java,v 1.10 2011/06/17 15:18:00 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * 
 * Utilities related to names.
 *
 */
public class NameUtils
{
  /***************************************************************************
   * Get set of variable symbols denoted by name.
   * @param names a sequence of names denoting variables (may be null)
   * @return a set of the symbols denoted by the name.
   **************************************************************************/
  public static Set<VariableSymbol> getVariablePrefixSymbols(Name[] names)
  {
    Set<VariableSymbol> result = new SetClass<VariableSymbol>();
    if (names == null) return result;
    for (Name name : names)
      result.add(getVariablePrefixSymbol(name));
    return result;
  }
  
  /***************************************************************************
   * Get set of class symbols denoted by names.
   * @param names a sequence of names denoting classes (may be null).
   * @return a collection of the symbols denoted by the name.
   **************************************************************************/
  public static Set<ClassSymbol> getClassSymbols(Name[] names)
  {
    Set<ClassSymbol> result = new SetClass<ClassSymbol>();
    if (names == null) return result;
    for (Name name : names)
      result.add((ClassSymbol)name.getSymbol());
    return result;
  }
  
  /***************************************************************************
   * Get vector of theory symbols denoted by names.
   * @param names a sequence of names denoting theories (may be null).
   * @return a collection of the symbols denoted by the name.
   **************************************************************************/
  public static Vector<TheorySymbol> getTheorySymbols(Name[] names)
  {
    Vector<TheorySymbol> result = new Vector<TheorySymbol>();
    if (names == null) return result;
    for (Name name : names)
      TheorySymbol.add(result, (TheorySymbol)name.getSymbol());
    return result;
  }
  
  /***************************************************************************
   * Get names of symbols.
   * @param symbols a collection of symbols.
   * @return the set of names of the symbols.
   **************************************************************************/
  public static Set<Name> getNames(Collection<TheorySymbol> symbols)
  {
    Set<Name> names = new SetClass<Name>();
    for (GlobalSymbol symbol : symbols)
      names.add(symbol.getLongName());
    return names;
  }
  
  /***************************************************************************
   * Determine variable prefix of name.
   * @param name a name containing a variable prefix.
   * @return the variable prefix of the name
   *         (null, if there is none)
   **************************************************************************/
  public static Name getVariable(Name name)
  {
    int pos = getVariablePosition(name);
    if (pos == -1) return null;
    return name.getPrefix(pos+1);
  }
  
  /***************************************************************************
   * Determine variable prefix symbol of name.
   * @param name a name containing a variable prefix.
   * @return the symbol of the variable prefix of the name
   *         (null, if there is none)
   **************************************************************************/
  public static VariableSymbol getVariablePrefixSymbol(Name name)
  {
    Name name0 = uniqueName(name);
    int pos = getVariablePosition(name0);
    if (pos == -1) return null;
    return (VariableSymbol)name0.idents[pos].getSymbol();
  }
  
  /***************************************************************************
   * Determine variable prefix of name.
   * @param name a name containing a variable prefix.
   * @return the position of the variable in the name
   *         (-1, if there is no such variable)
   **************************************************************************/
  public static int getVariablePosition(Name name)
  {
    int n = name.idents.length;
    for (int i=0; i<n; i++)
    {
      Symbol symbol = name.idents[i].getSymbol();
      if (symbol instanceof VariableSymbol || symbol instanceof ThisSymbol) return i;
    }
    return -1;
  }
  
  /***************************************************************************
   * Get unique name by adding optional prefix.
   * @param name the name of a variable.
   * @return its unique name (add package/class/object prefix, if necessary).
   **************************************************************************/
  public static Name uniqueName(Name name)
  {
    Symbol symbol = name.idents[0].getSymbol();
    Identifier[] idents = null;
    if (symbol instanceof VariableSymbol)
    {
      VariableSymbol symbol0 = (VariableSymbol)symbol;
      idents = uniqueName(symbol0).idents;
    }
    else if (symbol instanceof ClassSymbol)
    {
      ClassSymbol symbol0 = (ClassSymbol)symbol;
      PackageSymbol psymbol = symbol0.getPackage();
      Name pname = psymbol.getLongName();
      int n = pname.idents.length;
      idents = new Identifier[n+1];
      for (int i=0; i<n; i++)
        idents[i] = pname.idents[i];
      idents[n] = name.idents[0];
    }
    else
      return name;
    int n0 = idents.length;
    int n1 = name.idents.length;
    Identifier[] idents0 = new Identifier[n0+n1-1];
    for (int i=0; i<n0; i++)
      idents0[i] = idents[i];
    for (int i=1; i<n1; i++)
      idents0[n0+i-1] = name.idents[i];
    return Name.construct(idents0);
  }
  
  /****************************************************************************
   * Get unique reference to program variable.
   * @param symbol the symbol of the program variable.
   * @return the name by which it can be referenced.
   ***************************************************************************/
  public static Name uniqueName(VariableSymbol symbol)
  {
    if (symbol instanceof ClassVariableSymbol)
    {
      // class variables are referenced by long name
      ClassVariableSymbol symbol0 = (ClassVariableSymbol)symbol;
      return symbol0.getLongName();
    }
    if (symbol instanceof ObjectVariableSymbol)
    {
      // object variables are referenced by "this.name"
      ObjectVariableSymbol symbol0 = (ObjectVariableSymbol)symbol;
      ThisSymbol thisSymbol = symbol0.getClassSymbol().getThisSymbol();
      return Name.construct(new Identifier[] { thisSymbol.getName(), symbol.getName() });
    }
    if (symbol instanceof LocalVariableSymbolBase)
    {
      // body symbols, parameters, this symbol referenced by short name
      return Name.construct(symbol.getName());
    }
    Main.internalError("unknown variable symbol");
    return null;
  }
}
