// ----------------------------------------------------------------------------
// DeclCallStatement.java
// A method call statement with a result value initializing a local variable.
// $Id: DeclCallStatement.java,v 1.5 2017/07/04 08:04:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Program;

import java.util.*;
import fmrisc.ProgramExplorer.Syntax.*;

/**
 * A method call statement with a result value initializing a local variable.
 */
public class DeclCallStatement 
extends CallStatement implements VariableDeclaration
{
  public final TypeExpression type;
  public final Identifier name;
    
  /****************************************************************************
   * Construct a call statement with a result value assigned to a variable.
   * @param type the type of the variable
   * @param var the name of the variable
   * @param method the method name
   * @param args the method arguments
   * @return the statement.
   ***************************************************************************/
  public static DeclCallStatement construct(TypeExpression type, 
    Identifier var, Name method, Vector<ValueExpression> args)
  {
    return new DeclCallStatement(type, var, method, 
        args.toArray(new ValueExpression[args.size()]));
  }
  
  /****************************************************************************
   * Construct a call statement with a result value assigned to a variable.
   * @param type the type of the variable
   * @param var the name of the variable
   * @param method the method name
   * @param args the method arguments
   * @return the statement.
   ***************************************************************************/
  public static DeclCallStatement construct(TypeExpression type, 
    Identifier var, Name method, ValueExpression[] args)
  {
    return new DeclCallStatement(type, var, method, args);
  }
  
  /****************************************************************************
   * Construct a call statement with a result value assigned to a variable.
   * @param type the type of the variable
   * @param name the name of the variable
   * @param method the method name
   * @param args the method arguments
   ***************************************************************************/
  protected DeclCallStatement(TypeExpression type, Identifier name, 
    Name method, ValueExpression[] args)
  {
    super(method, args);
    this.type = type;
    this.name = name;
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    type.accept(visitor);
    name.accept(visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    type.accept(visitor);
    name.accept(visitor);
  }
  
  /****************************************************************************
   * Get the name of the declared entity.
   * @return the entity's name.
   ***************************************************************************/
  public Identifier getName()
  {
    return name;
  }
  
  /****************************************************************************
   * Get the type of the declared entity.
   * @return the entity's type.
   ***************************************************************************/
  public TypeExpression getType()
  {
    return type;
  }
}
