// ----------------------------------------------------------------------------
// ForLoopStatement.java
// A for loop.
// $Id: ForLoopStatement.java,v 1.9 2017/07/04 08:04:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Program;

import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * A for loop.
 */
public final class ForLoopStatement extends LoopStatementBase
{
  // the for loop component
  public final Statement init;
  public final ValueExpression cond;
  public final Statement update;
  public final Statement body;
  
  // the while loop replacement
  public final BlockStatement whileReplacement;
  
  // the for loop precondition (lifted from while loop precondition)
  private Formula preCondition = null;
  
  /***************************************************************************
   * Construct a while loop.
   * @param init the loop variable initialization.
   * @param cond the loop condition.
   * @param update the loop variable update.
   * @param body the loop body.
   * @return the statement.
   **************************************************************************/
  public static ForLoopStatement construct(Statement init,
    ValueExpression cond, Statement update, Statement body)
  {
    return new ForLoopStatement(init, cond, update, body);
  }
  
  /***************************************************************************
   * Construct a while loop.
   * @param init the loop variable initialization.
   * @param cond the loop condition.
   * @param update the loop variable update.
   * @param body the loop body.
   **************************************************************************/
  private ForLoopStatement(Statement init,
    ValueExpression cond, Statement update, Statement body)
  {
    this.init = init;
    this.cond = cond;
    this.update = update;
    this.body = body;
    this.whileReplacement = 
      BlockStatement.construct(
          new Statement[] {
            init,
            WhileLoopStatement.construct(
                cond,
                BlockStatement.construct(
                  new Statement[] { body, update}))});
  }    
  
  /***************************************************************************
   * Set the loop annotation for this item.
   * @param annotation the annotation.
   **************************************************************************/
  public void setLoopAnnotation(ProgramAnnotation annotation)
  {
    super.setLoopAnnotation(annotation);
    WhileLoopStatement loop = (WhileLoopStatement)whileReplacement.stats[1];
    loop.setLoopAnnotation(annotation);
  }
  
  /***************************************************************************
   * Set the position of this node in the source code.
   * @param position the source code position.
   **************************************************************************/
  public void setPosition(SourcePosition position)
  {
    super.setPosition(position);
    WhileLoopStatement loop = (WhileLoopStatement)whileReplacement.stats[1];
    loop.setPosition(position);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    init.accept(visitor);
    cond.accept(visitor);
    update.accept(visitor);
    body.accept(visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    init.accept(visitor);
    cond.accept(visitor);
    update.accept(visitor);
    body.accept(visitor);
  }
  
  /****************************************************************************
   * Get type of statement.
   * @return a string identifying the type of statement.
   ***************************************************************************/
  public String getKind()
  {
    return "for loop";
  }
  
  /***************************************************************************
   * Set for-loop precondition.
   * @param pre the precondition.
   **************************************************************************/
  public void setPrecondition(Formula pre)
  {
    preCondition = pre;
  }
  
  /***************************************************************************
   * Get for-loop precondition.
   * @return the precondition (null, if not set).
   **************************************************************************/
  public Formula getPrecondition()
  {
    return preCondition;
  }
}
