// ----------------------------------------------------------------------------
// MethodTasks.java
// The delayed generation of method-related tasks.
// $Id: MethodTasks.java,v 1.12 2013/01/22 10:31:34 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.util.*;
import java.io.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Judgements.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Program.Statement;

/**
 * 
 * The delayed generation of method-related tasks.
 *
 */
public final class MethodTasks
{
  /**
   * 
   * The info necessary for delayed task creation.
   *
   */
  private final static class Info
  {
    public final ParamSymbol method;
    public final StateType stateType;
    public final Name[] stateVars;
    public final Name[] stateExcs;
    public final Vector<TheorySymbol> theories;
    public final Declaration[][] decls;
    public final ErrorStream out;
    
    public Info(ParamSymbol method,
      StateType stateType, Name[] stateVars, Name[] stateExcs,
      Vector<TheorySymbol> theories, Declaration[][] decls,
      ErrorStream out)
    {
      this.method = method;
      this.stateType = stateType;
      this.stateVars = stateVars;
      this.stateExcs = stateExcs;
      this.theories = theories;
      this.decls = decls;
      this.out = out;
    }
  }
  
  // the info registered for delayed task creation
  private static Vector<Info> infos = new Vector<Info>();
  
  /***************************************************************************
   * Register the construction of tasks for a given method.
   * @param method the symbol of a method.
   * @param stateType the type of the current state (null, if none)
   * @param stateVars the names of the modifiable variables (null, if none)
   * @param stateExcs the names of the throwable exceptions (null, if none)
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   **************************************************************************/
  public static void register(ParamSymbol method,
    StateType stateType, Name[] stateVars, Name[] stateExcs,
    Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream out)
  {
    infos.add(new Info(method, stateType, stateVars, stateExcs, theories, decls, out));
    TerminationTasks.register(method, stateType, stateVars, stateExcs, theories, decls, out);
  }
  
  /****************************************************************************
   * Clear the registered tasks.
   ***************************************************************************/
  public static void clearTasks()
  {
    infos = new Vector<Info>();
    ClassTasks.clearTasks();
    TerminationTasks.clearTasks();
  }
  
  /****************************************************************************
   * Determine whether there are tasks.
   * @return true iff there are tasks.
   ***************************************************************************/
  public static boolean exist()
  {
    return !infos.isEmpty();
  }
  
  /***************************************************************************
   * Generate those tasks for which the information has been previously registered.
   * @param out the stream to which messages are written.
   ***************************************************************************/
  public static void generateTasks(PrintWriter out)
  {
    if (!ClassTasks.exist() && !MethodTasks.exist() && !TerminationTasks.exist())
      return;
    out.print("executing tasks...");
    out.flush();
    ClassTasks.generateTasks(out);
    for (Info info : infos)
    {
      generate(info.method, info.stateType, info.stateVars, info.stateExcs,
          info.theories, info.decls, info.out);
      out.print(".");
      out.flush();
    }
    TerminationTasks.generateTasks(out);
    clearTasks();
  }
  
  /***************************************************************************
   * Construct the method tasks for a given method.
   * @param method the symbol of a method with a specification.
   * @param stateType the type of the current state (null, if none)
   * @param assignVars the names of the modifiable variables (null, if none)
   * @param throwExcepts the names of the throwable exceptions (null, if none)
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param error the stream to which to write error messages.
   **************************************************************************/
  private static void generate(ParamSymbol method,
    StateType stateType, Name[] assignVars, Name[] throwExcepts,
    Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream error)
  {
    // derive body judgement (generates invariant tasks)
    TaskFolder loops = method.getLoops();
    Satisfies satisfies = 
      new Satisfies(method, stateType, loops, theories, decls, error);
    Statement body = method.getDeclaration().getBody();
    Set<VariableSymbol> vars = Normalization.modifiableVariables(assignVars);
    StatJudgement S = satisfies.extendFrame(satisfies.derive(body), vars);
    method.setBodyJudgement(S);
    
    // generate frame task
    TaskFolder tasks = method.getObligations();
    Task frame = new FrameTask(method);
    tasks.add(frame);
    
    // propagate method precondition through body
    Map<Statement, Formula> preMap = new MapClass<Statement, Formula>();
    Map<Statement, Formula> postMap = new MapClass<Statement, Formula>();
    Formula pre = Normalization.preCondition(method, true);
    if (pre == null) pre = True.construct();
    Propagator.execute(body, body, pre, true, preMap, postMap, satisfies);
    method.setPreMap(preMap);
    
    // generate precondition tasks
    PreconditionTasks.generate(method, preMap,
        stateType, assignVars, throwExcepts, theories, decls, error);
    
    // the remaining tasks are only generated, if there is a postcondition
    if (Normalization.postCondition(method) == null) return;
    
    // generate optional tasks for validating specification
    TaskFolder optional = method.getOptional();
    Task satisfiability = new SpecTask(method, true,
        stateType, theories, decls, error);
    optional.add(satisfiability);
    Task nontriviality = new SpecTask(method, false,
        stateType, theories, decls, error);
    optional.add(nontriviality);
    
    // generate correctness task
    Task correctness = new CorrectnessTask(method, tasks.getDirectory(),
        stateType, assignVars, throwExcepts, theories, decls, error);
    tasks.add(correctness);
  }
}
