// ----------------------------------------------------------------------------
// TaskBase.java
// The base class of a verification task
// $Id: TaskBase.java,v 1.23 2011/05/30 14:08:13 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.io.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;

/**
 * 
 * The base class of a verification task
 *
 */
public class TaskBase implements Task
{
  private String name;             // name of the task
  private boolean optional;        // is it optional?
  private Status status;           // its status
  private SourcePosition position; // related source position (may be null)
  private TaskFolder parent;       // the parent folder (null, if none)
  private File pdir;               // the parent of dir
  private File dir;                // the directory for writing files
  private Strategy strategy;       // an applicable strategy (may be null)
  private String evidence;         // the evidence for CLOSED/FAILED tasks.
  private Object info;             // any additional information (may be null)
  private Observer observer;       // a task observer (may be null)
  
  /****************************************************************************
   * Construct task with denoted name and status NEW
   * @param name the name of the task
   * @param optional is the task optional?
   * @param position a related source position (may be null)
   * @param parent the parent of the directory created for writing 
   *        persistent information related to this task.
   * @param strategy the strategy for performing this task (may be null).
   ***************************************************************************/
  protected TaskBase(String name, boolean optional,
    SourcePosition position, File parent, Strategy strategy)
  {
    this.name = name;
    this.optional = optional;
    this.status = Status.NEW;
    this.position = position;
    this.pdir = parent;
    this.dir = null;
    this.strategy = strategy;
    this.evidence = null;
    this.observer = null;
  }
  
  /****************************************************************************
   * Get name of task.
   * @return the task name.
   ***************************************************************************/
  public String getName()
  {
    return name;
  }
  
  /****************************************************************************
   * Determine whether new task is to be resumed.
   * @return true if new task is to be resumed.
   ***************************************************************************/
  public boolean resumeOnNew()
  {
    return false;
  }
  
  /****************************************************************************
   * Determine whether task is optional.
   * @return true if task is optional.
   ***************************************************************************/
  public boolean isOptional()
  {
    return optional;
  }
  
  /****************************************************************************
   * Get status of task.
   * @return the task status.
   ***************************************************************************/
  public Status getStatus()
  {
    return status;
  }
  
  /****************************************************************************
   * Resume task.
   * @return the new status
   ***************************************************************************/
  public Task.Status resume()
  {
    if (strategy != null) strategy.apply(this);
    return status;
  }
  
  /****************************************************************************
   * Get source code position linked to this task (null, if none)
   ***************************************************************************/
  public SourcePosition getPosition()
  {
    return position;
  }
  
  /***************************************************************************
   * Set parent folder
   * @param parent the parent folder (may be null)
   **************************************************************************/
  public void setParent(TaskFolder parent)
  {
    this.parent = parent;
    if (parent != null) parent.computeStatus();
  }
  
  /***************************************************************************
   * Get parent folder
   * @return the parent (null, if root)
   **************************************************************************/
  public TaskFolder getParent()
  {
    return parent;
  }
  
  /***************************************************************************
   * Get parent directory
   * @return the parent of the directory of this task
   **************************************************************************/
  public File getParentDirectory()
  {
    return pdir;
  }
  
  /****************************************************************************
   * Get directory to be used for writing persistent data
   ***************************************************************************/
  public File getDirectory()
  {
    // create directory only on demand
    if (dir == null) createDirectory();
    return dir;
  }
  
  /****************************************************************************
   * Free directory that was used for writing persistent data
   * without clearing its contents.
   ***************************************************************************/
  private void freeDirectory()
  {
    if (dir == null) return;
    Directory.free(dir, false);
    dir = null;
  }
  
  /****************************************************************************
   * Set directory to be used for writing persistent data.
   * @param dir the directory.
   ***************************************************************************/
  public void setDirectory(File dir)
  {
    this.dir = dir;
  }
  
  /****************************************************************************
   * Create directory to be used for writing persistent data
   ***************************************************************************/
  protected void createDirectory()
  {
    setDirectory(Directory.create(pdir, name, false));
  }
  
  /****************************************************************************
   * Print linear representation of task.
   ***************************************************************************/
  public void print(PrintWriter out)
  {
    if (position instanceof SourceAnnotationPosition) 
    { 
      out.println("Task(" + position + "): " + name);
      out.println("Annotation line:");
      SourceAnnotationPosition pos0 = (SourceAnnotationPosition)position;
      out.println(pos0.getAnnotation().getLine(pos0.getAnnotationLine()));
      for (int i=0; i<pos0.getAnnotationColumn(); i++) out.print(" ");
      out.println("^");
    }
    else if (position != null)
      out.println("Task(" + position + "): " + name); 
    else
      out.println("Task: " + name); 
    out.print("Status: ");
    switch (getStatus())
    {
    case NEW: out.print("new"); break;
    case OPEN: out.print("open"); break;
    case ALMOST: out.print("almost done"); break;
    case DONE: out.print("done"); break;
    case FAILED: out.print("failed"); break;
    case REVOKED: out.print("failed"); break;
    default: Main.internalError("unknown status");
    }
    if (evidence != null)
      out.print(" (" + evidence + ")");
    out.println();
  }
  
  /****************************************************************************
   * Set task status.
   * @param evidence the evidence for the status (may be null).
   * @param info any additional information about the evidence (may be null).
   ***************************************************************************/
  public void setStatus(Status status, String evidence, Object info)
  {
    this.status = status;
    this.evidence = evidence;
    this.info = info;
    if (status == Status.REVOKED) freeDirectory();
    if (parent != null) parent.computeStatus();
    sendNotification();
  }
  
  /****************************************************************************
   * Get evidence for task status.
   * @return the evidence (may be null)
   ***************************************************************************/
  public String getEvidence()
  {
    return evidence;
  }
  
  /****************************************************************************
   * Get additional information for task status.
   * @return the information (may be null)
   ***************************************************************************/
  public Object getInfo()
  {
    return info;
  }
  
  /****************************************************************************
   * Determine whether task status is final.
   * @return true iff task status is DONE or FAILED.
   ***************************************************************************/
  public boolean isFinal()
  {
    return status == Status.DONE  || status == Status.FAILED;
  }
  
  /****************************************************************************
   * Set observer for this task.
   * @param observer a task observer
   ***************************************************************************/
  public void setObserver(Observer observer)
  {
    this.observer = observer;
  }
  
  /****************************************************************************
   * Notify observer about task status change.
   ***************************************************************************/
  private void sendNotification()
  {
    if (observer != null) observer.sendNotification(this);
  }
  
  /****************************************************************************
   * Reset the task to its initial (new) state.
   ***************************************************************************/
  public void reset()
  {
  }
  
  /****************************************************************************
   * Shorten text (sequence of lines) by cutting of too long/too many lines.
   * @param text a descriptive text.
   * @return the shortened version of the text.
   ***************************************************************************/
  protected String shortDescription(String text)
  {
    final int LENGTH = 60;
    final int NUMBER = 10;
    StringBuffer buffer = new StringBuffer();
    String[] lines = text.split("\n");
    int n = Math.min(lines.length, NUMBER);
    for (int i=0; i<n; i++)
    {
      String line = lines[i];
      if (line.length() > LENGTH)
      {
        line = line.substring(0, LENGTH);
        line = line + "...";
      }
      buffer.append(line);
      if (i+1 < n) buffer.append("\n");
    }
    if (NUMBER < lines.length) buffer.append("...");
    return buffer.toString();
  }
}
