// ----------------------------------------------------------------------------
// Case.java
// The "case" command.
// $Id: Case.java,v 1.12 2012/10/15 15:16:47 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;

import java.io.PrintWriter;
import java.util.*;
import org.w3c.dom.*;

/**
 * 
 * The "case" command: split state into multiple cases.
 *
 */
public final class Case extends ProofCommandBase
{
  public final static String Name = "case";
  private Expression[] formulas;
  
  /***************************************************************************
   * Create a "case" command.
   * @param formulas the (non-empty) sequence of formulas splitting the state.
   **************************************************************************/
  public Case(Expression[] formulas)
  {
    super(Name);
    this.formulas = formulas;
  }
  
  /****************************************************************************
   * Process command "case": split current proof state by formula
   * sequence F1, F2, ..., Fn (n>0) yielding at most 2n+1 children states.
   * A, ... |- B, ... =>
   *   ( A, ... |- TCC_F1 )
   *   A, ..., F1 |- B, ... 
   *   ( A, ..., ~F1 |- TCC_F2) 
   *   A, ..., ~F1, F2 |- B, ...
   *   ...
   *   ( A, ..., ~F1, ~F2, ... |- TCC_Fn)
   *   A, ..., ~F1, ~F2, ..., ~Fn |- B, ...
   * @return null or array of two or more proof states.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    Context context = State.getContext();
    Environment env = state.getEnvironment();
    Formula[] assumptions = state.getAssumptions();
    Formula[] goals = state.getGoals();
    int n = formulas.length;
    Vector<ProofState> children = new Vector<ProofState>();
    Vector<Expression> tccs = new Vector<Expression>();
    for (int i=0; i<n; i++)
    {
      context.open();
      if (!typeCheckFormula(env, formulas[i])) 
      {
        context.abort();
        return null;
      }
      Expression tcc = context.getFormula();
      context.abort();
      if (tcc != null)
      {
        ProofState state2 = new ProofState(state, env);
        children.add(state2);
        state2.setAssumptions(state2.copy(assumptions));
        state2.setGoals(new Formula[] { new Formula(state2, tcc, true) });
        tccs.add(tcc);
      }
      ProofState state0 = new ProofState(state, env);
      children.add(state0);
      Formula[] assumptions0 = state0.add(assumptions, 
        new Formula(state0, formulas[i], false));
      if (tccs.size() != 0) addAxioms(state0, tccs);
      state0.setAssumptions(assumptions0);
      state0.setGoals(state0.copy(goals));
      assumptions = state.add(assumptions,
        new Formula(state, ProofUtil.negate(formulas[i]), false));
    }
    ProofState state0 = new ProofState(state, env);
    children.add(state0);
    if (tccs.size() != 0) addAxioms(state0, tccs);
    state0.setAssumptions(state0.copy(assumptions));
    state0.setGoals(state0.copy(goals));
    return ProofUtil.toProofStateArray(children);
  }
  
  /***************************************************************************
   * Add to the axioms of the denoted state the formulas in the given vector.
   * @param state the state to which the axioms are to be added.
   * @param v a vector of Expression objects denoting the formulas to be added. 
   **************************************************************************/
  private void addAxioms(ProofState state, Vector<Expression> v)
  {
    int n = v.size();
    Formula[] formulas = new Formula[n];
    for (int i=0; i<n; i++)
      formulas[i] = new Formula(state, (Expression)(v.elementAt(i)), false);
    state.setAxioms(state.add(state.getAxioms(), formulas));
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    int n = formulas.length;
    for (int i=0; i<n; i++)
      node.appendChild(toNode(formulas[i]));
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    Expression[] formulas = new Expression[n];
    for (int i=0; i<n; i++)
    {
      formulas[i] = toExpression(children.item(i));
      if (formulas[i] == null) return null;
    }
    return new Case(formulas);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    out.print(" ");
    int n = formulas.length;
    for (int i=0; i<n; i++)
    {
      formulas[i].print(out);
      if (i+1 < n) out.print(", ");
    }
  }
}
