// ----------------------------------------------------------------------------
// CommandBase.java
// Base class of commands
// $Id: CommandBase.java,v 1.24 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.*;

import nl.tue.win.riaca.openmath.lang.*;
import org.w3c.dom.*;

import fmrisc.External.*;
import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Communication.*;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;

/**
 * 
 * Base class of commands.
 *
 */
public abstract class CommandBase implements Command
{
  //the name of the command
  private String name;              
  
  /***************************************************************************
   * construct command with denoted name
   * @param name the name of the command
   **************************************************************************/
  public CommandBase(String name)
  {
    this.name = name;
  }
  
  /***************************************************************************
   * get name of command
   * @return the command name
   **************************************************************************/
  public String getName()
  {
    return name;
  }

  /***************************************************************************
   * returns string representation of tree
   * @return the string representation
   **************************************************************************/
  public String toString()
  {
    StringWriter s = new StringWriter();
    printCore(new PrintWriter(s));
    return s.toString();
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination). 
   * @param out the stream on which the text is written
   **************************************************************************/
  public abstract void printCore(PrintWriter out);
  
  /***************************************************************************
   * the environment in which to process the command
   **************************************************************************/
  protected Environment getEnvironment()
  {
    Proof proof = State.getProof();
    if (proof == null)
      return State.getEnvironment();
    else
      return proof.getCurrent().getEnvironment();
  }
  
  /**************************************************************************
   * print error message on output stream
   * @param message the message to be printed
   *************************************************************************/
  protected static void error(String message)
  {
    State.setError(message);
  }
  
  // currently used document for node generation
  // set by toNode(document)
  private static Document document;
  private static NewOMDOMWriter writer;
  private static OMDoc omdoc;
  
  /***************************************************************************
   * Convert proof command to external representation as a DOM node.
   * @param  document the document to which the node belongs.
   * @return The representation of the proof command as a DOM node of the
   *         document. The name of the node is "fmrisc:command" with an
   *         attribute name="NAME" where NAME is the value returned 
   *         by getName().
   **************************************************************************/
  public Node toNode(Document document)
  {
    CommandBase.document = document;
    CommandBase.writer = new NewOMDOMWriter(document, 
      State.OpenMathURI, State.OpenMathPrefix);
    omdoc = new OMDoc(document);
    Element node = document.createElementNS(State.NamespaceURI, 
      State.NamespacePrefix + ":command");
    node.setAttribute("name", getName());
    setRepresentation(node);
    return node;
  }
    
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
  }
  
  /***************************************************************************
   * Convert AST object to DOM representation of OpenMath.
   * @param ast the object in AST representation.
   * @return the OpenMath representation as a DOM object
   *         (null, if an error occured)
   **************************************************************************/
  private static Node astToNode(AST ast)
  {
    OMObject om = OpenMath.toOpenMath(ast);
    if (om == null) return null;
    try
    {
      return writer.writeNode(om);
    }
    catch(IOException e)
    {
      return null;
    }
  }

  /***************************************************************************
   * Convert expression to DOM node.
   * @param exp the expression.
   * @return its DOM representation (null, if an error occured)
   **************************************************************************/
  protected static Node toNode(Expression exp)
  {
    return astToNode(exp);
  }
  
  /***************************************************************************
   * Convert type to DOM node.
   * @param type the type.
   * @return its DOM representation (null, if an error occured)
   **************************************************************************/
  protected static Node toNode(Type type)
  {
    return astToNode(type);
  }
 
  /***************************************************************************
   * Convert identifier to DOM node.
   * @param ident the identifier.
   * @return its DOM representation (null, if an error occured)
   **************************************************************************/
  protected static Node toNode(Identifier ident)
  {
    return astToNode(ident);
  }
  
  /***************************************************************************
   * Convert declaration to DOM node.
   * @param decl the declaration.
   * @return its DOM representation (null, if an error occured)
   **************************************************************************/
  protected static Node toNode(Declaration decl)
  {
    return omdoc.toNode(decl); 
  }
  
  /****************************************************************************
   * Mark identifier as referenced in current proof.
   * @param ident the identifier that is marked as referenced.
   **************************************************************************/
  protected static void referenced(Identifier ident)
  {
    if (!(ident instanceof Reference)) return;
    Proof proof = State.getProof();
    if (proof == null) return;
    proof.getReferenced().add(((Reference)ident).getSymbol());
  }
  
  /***************************************************************************
   * Create a new DOM node with given name.
   * @param name the name of the node
   * @return a DOM node with that name and namespace "fmrisc".
   **************************************************************************/
  protected Element newNode(String name)
  {
    return document.createElementNS(State.NamespaceURI,  
      State.NamespacePrefix + ":" + name);
  }
  
  /***************************************************************************
   * Create a new DOM text node with given text.
   * @param text the text of the node.
   * @return a DOM node with denoted text.
   **************************************************************************/
  protected Text newTextNode(String text)
  {
    return document.createTextNode(text);
  }
  
  /***************************************************************************
   * Convert node to expression.
   * @param node the DOM representation of the expression
   * @return the expression (null, in case of error).
   **************************************************************************/
  protected static Expression toExpression(Node node)
  {
    return Store.toExpression(node);
  }
  
  /***************************************************************************
   * Convert node to type.
   * @param node the DOM representation of the type
   * @return the type (null, in case of error).
   **************************************************************************/
  protected static Type toType(Node node)
  {
    return Store.toType(node);
  }
  
  /***************************************************************************
   * Convert node to identifier.
   * @param node the DOM representation of the identifier
   * @return the identifier (null, in case of error).
   **************************************************************************/
  protected static Identifier toIdentifier(Node node)
  {
    return Store.toIdentifier(node);
  }
  
  /***************************************************************************
   * Convert node to string.
   * @param node the DOM representation of the string
   * @return the string (null, in case of error).
   **************************************************************************/
  protected static String toString(Node node)
  {
    if (!(node instanceof Text))
    {
      State.setError("text node expected: " + node.toString());
      return null;
    }
    return ((Text)node).getData();
  }
  
  /***************************************************************************
   * Convert node to declaration.
   * @param node the DOM representation of the declaration
   * @return the type (null, in case of error).
   **************************************************************************/
  protected static Declaration toDeclaration(Node node)
  {
    return OMDoc.toDeclaration(node);
  }
  
  /***************************************************************************
   * Construct command from DOM representation.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command toCommand(Element node)
  {
    String name = node.getAttribute("name");
    if (name == null)
    {
      State.setError("Attribute 'name' expected: " + node.toString());
      return null;
    }
    if (name.equals(Assume.Name))
      return Assume.newCommand(node);
    if (name.equals(Auto.Name))
      return Auto.newCommand(node);
    if (name.equals(AutoStar.Name))
      return AutoStar.newCommand(node);
    if (name.equals(Case.Name))
      return Case.newCommand(node);
    if (name.equals(CounterExample.Name))
      return CounterExample.newCommand(node);
    if (name.equals(DeclarationC.Name))
      return DeclarationC.newCommand(node);
    if (name.equals(Decompose.Name))
      return Decompose.newCommand(node);
    if (name.equals(Empty.Name))
      return Empty.newCommand(node);
    if (name.equals(Expand.Name))
      return Expand.newCommand(node);
    if (name.equals(Flatten.Name))
      return Flatten.newCommand(node);
    if (name.equals(Flip.Name))
      return Flip.newCommand(node);
    if (name.equals(Goal.Name))
      return Goal.newCommand(node);
    if (name.equals(Induction.Name))
      return Induction.newCommand(node);
    if (name.equals(Instantiate.Name))
      return Instantiate.newCommand(node);
    if (name.equals(Lemma.Name))
      return Lemma.newCommand(node);
    if (name.equals(Option.Name))
      return Option.newCommand(node);
    if (name.equals(Proved.Name))
      return Proved.newCommand(node);
    if (name.equals(Scatter.Name))
      return Scatter.newCommand(node);
    if (name.equals(Simplify.Name))
      return Simplify.newCommand(node);
    if (name.equals(Skolemize.Name))
      return Skolemize.newCommand(node);
    if (name.equals(Split.Name))
      return Split.newCommand(node);
    if (name.equals(TypeAxiom.Name))
      return TypeAxiom.newCommand(node);
    State.setError("unknown command: " + node.toString());
    return null;
  }
}
