// ----------------------------------------------------------------------------
// PrettyPrinter.java
// Printing abstract syntax trees with appropriate line breaks.
// $Id: PrettyPrinter.java,v 1.13 2013/04/06 11:29:40 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Communication;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Syntax.*;
import fmrisc.ProofNavigator.Syntax.Number;

import java.io.*;
import java.util.*;

/**
 * 
 * Printing abstract syntax trees with appropriate line breaks.
 *
 */
public final class PrettyPrinter
{
  /****************************************************************************
   * The pretty printer depends on the current printing mode (LINE or BLOCK) 
   * and the current indentation level LEVEL.
   * 
   * Printing a tree starts in BLOCK mode, when a subtree is encountered 
   * that demands switching to LINE mode, the whole subtree is printed in 
   * LINE mode (no switching back to BLOCK mode within the subtree).
   * 
   * When printed in LINE mode, a syntax tree fills the current line as much 
   * as possible; if a line break has to be inserted, the remainder of the
   * tree is printed at increased level LEVEL+1 (not further increased).
   * 
   *  token token token token token ...      LEVEL 
   *    token token token token token token  LEVEL+1
   *    token token token token token token  LEVEL+1
   *    
   * The only trees that make use of the BLOCK mode are the ones listed below. 
   * If such a tree does not fit into the current line, parts of it are printed
   * at indentation level LEVEL+1 (which may be further increased within 
   * the parts).
   * 
   * IDENT:"TYPE"=                LEVEL
   *   TYPE                       LEVEL+1
   *   
   * IDENT:TYPE =                 LEVEL
   *   EXPRESSION;                LEVEL+1
   *   
   * IDENT:("FORMULA"|"AXIOM") =  LEVEL
   *   FORMULA                    LEVEL+1
   *   
   * QUANTIFIER(VARIABLES): LEVEL
   *   EXPRESSION           LEVEL+1 
   *   
   *   FORMULA          LEVEL+1
   * BINARY-CONNECTIVE  LEVEL
   *   FORMULA          LEVEL+1
   * BINARY-CONNECTIVE  LEVEL    
   *   FORMULA          LEVEL+1
   * 
   * IF FORMULA THEN     LEVEL    
   *   EXPRESSION        LEVEL+1
   * ELSIF FORMULA THEN  LEVEL
   *   EXPRESSION        LEVEL+1
   * ELSE                LEVEL
   *   EXPRESSION        LEVEL+1
   * ENDIF               LEVEL
   * 
   * LET             LEVEL
   *   DECLARATION,  LEVEL+1 
   *   DECLARATION,  LEVEL+1
   *   DECLARATION   LEVEL+1
   * IN              LEVEL
   *   EXPRESSION    LEVEL+1
   *   
   * The indentation level is not increased beyond MAXLEVEL; if this bound
   * is reached, the corresponding subtree is printed in LINE mode. Likewise,
   * if the number of character positions in a line drops below MINFREE, the
   * remainder of the tree is printed in LINE mode.
   ***************************************************************************/
  
  // the default width for printing
  public static final int WIDTH = 75;
  
  private final int MAXLEVEL = 7; // maximum indentation level
  private final int MINFREE = 20; // no further breaks if this is reached
  private final int INDENT = 2;   // amount of indentation per level
  
  private int width;       // the suggested line width
  private PrintWriter out; // the output stream to be used
  
  private boolean lineMode = false;  // we are in line mode
  private int level = 0;             // the current indentation level
  private int pos = 0;               // the current column position
  
  /****************************************************************************
   * Create pretty printer to print abstract syntax trees.
   * @param width the line width (suggestion, not a hard bound).
   * @param out the stream to print on.
   ***************************************************************************/
  public PrettyPrinter(int width, PrintWriter out)
  {
    this.width = width;
    this.out = out;
  }
  
  /****************************************************************************
   * Create pretty printer for default width WIDTH.
   * @param out the stream to print on.
   **************************************************************************/
  public PrettyPrinter(PrintWriter out)
  {
    this(WIDTH, out);
  }
  
  /****************************************************************************
   * Pretty-print declaration.
   * @param decl a declaration.
   ***************************************************************************/
  public void print(Declaration decl)
  {
    printCore(decl);
  }
  
  /****************************************************************************
   * Pretty-print expression.
   * @param exp an expression.
   ***************************************************************************/
  public void print(Expression exp)
  {
    printCore(exp);
  }
  
  /****************************************************************************
   * Pretty-print type.
   * @param type a type
   ***************************************************************************/
  public void print(Type type)
  {
    printCore(type);
  }
  
  /***************************************************************************
   * Print text.
   * @param text the text to be printed.
   **************************************************************************/
  public void print(String text)
  {
    if (width-pos < text.length()) 
    {
      resetLine();
      indent(true);
    }
    out.print(text);
    pos = pos+text.length();
  }
  
  /****************************************************************************
   * Change indentation level.
   * @param increase true if level is increased (decreased otherwise)
   ***************************************************************************/
  public void changeLevel(boolean increase)
  {
    if (lineMode) return;
    if (increase) 
    {
      level++;
      indent(false);
    }
    else
      level--;
  }
  
  /***************************************************************************
   * Start new (non-overflow)line and indent it appropriately.
   ***************************************************************************/
  public void newLine()
  {
    if (lineMode) 
      print(" ");
    else
    {
      resetLine();
      indent(false);
    }
  }
  
  /***************************************************************************
   * Insert a space (or a newline at the end of the line).
   **************************************************************************/
  public void space()
  {
    if (pos == width)
      newLine();
    else
      print(" ");
  }
  
  /***************************************************************************
   *
   * Auxiliary methods
   * 
   **************************************************************************/
  
  /***************************************************************************
   * Print abstract syntax tree.
   * @param tree an abstract syntax tree.
   **************************************************************************/
  private void printCore(AST tree)
  {
    print(BreakInfo.compute(tree, false));
  }
  
  /****************************************************************************
   * Print sequence of trees.
   * @param info the sequence to be printed.
   * @param separator the separator printed between elements.
   * @param nl true iff a new line is started before each element
   ***************************************************************************/
  private void print(BreakInfo[] info, String separator, boolean nl)
  {
    int n = info.length;
    
    // check whether sequence fits in single line
    int l = BreakInfo.length(info);
    if (n>1) l = l+(n-1)*(separator.length()+1);
    boolean single = (pos+l <= width);
    boolean mode = lineMode;
    if (single) 
    {
      // newLine();
      lineMode = true;
    }
    
    for (int i=0; i<n; i++)
    {
      if (!single && nl) newLine();
      print(info[i]);
      if (i<n-1) { print(separator); space(); }
    }
    
    // restore mode
    lineMode = mode;
  }
  
  /***************************************************************************
   * Reset to begin of line.
   *************************************************************************/
  private void resetLine()
  {
    out.println();
    pos = 0;
  }
  
  /***************************************************************************
   * If in block mode, indent to the apropriate position.
   * @param overflow true iff line is an overflow line
   **************************************************************************/
  private void indent(boolean overflow)
  {
    int indent = level*INDENT;
    if (overflow) indent = indent+2*INDENT;
    for (int i=pos; i<indent; i++)
      out.print(' ');
    pos = indent;
  }
  
  /***************************************************************************
   * Print tree with parentheses around it.
   * @param info the tree to be printed.
   * @param indent true iff tree is to be indented.
   **************************************************************************/
  private void printParens(BreakInfo info)
  {
    print("(");
    print(info);
    print(")");
  }
  
  /***************************************************************************
   * Print tree with parentheses around it, it its priority exceeds treshold.
   * @param info the tree to be printed.
   * @param treshold the reshold value for printing.
   **************************************************************************/
  private void printPriority(BreakInfo info, int treshold)
  {
    Expression exp = (Expression)info.getTree();
    int p = exp.getPriority();
    if (p > treshold) print("(");
    print(info);
    if (p > treshold) print(")");
  }
  
  /***************************************************************************
   * Print the tree denoted by info.
   * @param info the abstract syntax tree decorated by line break information.
   **************************************************************************/
  private void print(BreakInfo info)
  {
    AST tree = info.getTree();
    int len = info.getLength();
    int free = width-pos;
    
    // enough space left, we can print the simple way
    if (len <= free)
    {
      tree.print(out);
      pos = pos+len;
      return;
    }
    
    // we switch to line mode
    boolean oldMode = lineMode;
    if (level > MAXLEVEL || free < MINFREE)
      lineMode = true;
    
    // we have to break the linear representation
    if (tree instanceof Identifier) 
      printIdentifier(info);
    else if (tree instanceof Number)
      printNumber(info);
    else if (tree instanceof Expression)
      printExpression(info);
    else if (tree instanceof Type)
    {
      boolean mode = lineMode;
      lineMode = true;
      printType(info);
      lineMode = mode;
    }
    else if (tree instanceof Declaration)
      printDeclaration(info);
    else if (tree instanceof Selector)
      printSelector(info);
    else if (tree instanceof TypedIdentifier)
      printTypedIdentifier(info);
    else if (tree instanceof ValuedIdentifier)
      printValuedIdentifier(info);
    else
    {
      System.out.println("Unknown syntax tree to prettyprint: " + tree.toString());
      System.exit(-1);
    }
    
    // restore mode
    lineMode = oldMode;
  }
  
  /****************************************************************************
   * 
   * Printing the various kinds of abstract syntax trees.
   *
   ***************************************************************************/
  
  private void printIdentifier(BreakInfo info)
  {
    Identifier ident = (Identifier)info.getTree();
    print(ident.getName());
  }
  
  private void printNumber(BreakInfo info)
  {
    Number number = (Number)info.getTree();
    print(number.getDigits());
  }
  
  private void printType(BreakInfo info)
  {
    Type type = (Type)info.getTree();
    Vector<Object>  children = info.getChildren();
    if (type instanceof AtomicType)
    {
      AtomicType t = (AtomicType)type;
      String name = t.getName();
      if (State.booleanAsLogical())
      {
        if (name.equals("BOOLEAN"))
          print("LOGICAL");
        else if (name.equals("BIT"))
          print("BOOLEAN");
      }
      print(name);
      return;
    }
    if (type instanceof FunctionType)
    {
      FunctionType t = (FunctionType)type;
      Type[] domain = t.getDomain();
      int n = domain.length;
      BreakInfo[] dinfo = (BreakInfo[])children.elementAt(0);
      if (n == 1)
      {
        if (domain[0] instanceof FunctionType ||
            domain[0] instanceof ArrayType)
          printParens(dinfo[0]);
        else
          print(dinfo[0]);
      }
      else
      {
        print("(");
        for (int i=0; i<n; i++)
        {
          print(dinfo[i]);  
          if (i < n-1) print(", ");
        }
        print(")");
      }
      print("->");
      print((BreakInfo)children.elementAt(1));
      return;
    }
    if (type instanceof TupleType)
    {
      BreakInfo[] base = (BreakInfo[])children.elementAt(0);
      int n = base.length;
      print("[");
      for (int i=0; i<n-1; i++)
      {
        print(base[i]);
        print(", ");
      }
      print(base[n-1]);
      print("]");
      return;
    }
    if (type instanceof RecordType)
    {
      BreakInfo[] base = (BreakInfo[])children.elementAt(0);
      int n = base.length;
      print("[#");
      for (int i=0; i<n-1; i++)
      {
        print(base[i]);
        print(", ");
      }
      print(base[n-1]);
      print("#]");
      return;
    }
    if (type instanceof ArrayType)
    {
      print("ARRAY"); space();
      print((BreakInfo)children.elementAt(0));
      space(); print("OF"); space();
      print((BreakInfo)children.elementAt(1));
      return;
    }
    if (type instanceof SubrangeType)
    {
      print("[");
      print((BreakInfo)children.elementAt(0));
      print("..");
      print((BreakInfo)children.elementAt(1));
      print("]");
      return;
    }
    if (type instanceof SubType)
    {
      print("SUBTYPE");
      print("(");
      print((BreakInfo)children.elementAt(0));
      print(")");
      return;
    }
    if (type instanceof LetType)
    {
      print("LET"); space();
      BreakInfo[] declarations = (BreakInfo[])children.elementAt(0);
      int n = declarations.length;
      for (int i=0; i<n-1; i++)
      {
        print(declarations[i]);
        print(","); space();
      }
      print(declarations[n-1]);
      space(); print("IN"); space();
      print((BreakInfo)children.elementAt(1));   
      return;
    }
    System.out.println("Unknown type to pretty print: " + type.toString());
    System.exit(-1);
  }
  
  private void printDeclaration(BreakInfo info)
  {
    Declaration decl = (Declaration)info.getTree();
    Vector<Object>  children = info.getChildren();
    print((BreakInfo)children.elementAt(0));
    if (decl instanceof TypeDeclaration)
    {
      TypeDeclaration d = (TypeDeclaration)decl;
      Type t = d.getType();
      if (t == null)
      {
        print(":"); space(); print("TYPE");
      }
      else
      {
        print(":"); space(); print("TYPE"); space(); print("=");
        changeLevel(true);
        newLine();
        print((BreakInfo)children.elementAt(1));
        changeLevel(false);
      }
      return;
    }
    if (decl instanceof ValueDeclaration)
    {
      ValueDeclaration d = (ValueDeclaration)decl;
      Type t = d.getType();
      Expression e = d.getValue();
      int pos = 1;
      if (t != null)
      {
        print(":");
        print((BreakInfo)children.elementAt(pos));
        pos = pos+1;
      }
      if (e != null)
      {
        space(); print("="); 
        changeLevel(true);
        newLine(); 
        print((BreakInfo)children.elementAt(pos));
        changeLevel(false);
      }
      return;
    }
    if (decl instanceof FormulaDeclaration)
    {
      FormulaDeclaration d = (FormulaDeclaration)decl;
      if (d instanceof AxiomDeclaration)
      {
        print(":"); space(); print("AXIOM");
      }
      else
      {
        print(":"); space(); print("AXIOM");
      }
      changeLevel(true);
      newLine(); 
      print((BreakInfo)children.elementAt(1));
      changeLevel(false);
      return;
    }
    System.out.println("Unknown declaration to break: " + decl.toString());
    System.exit(-1);
  }
  
  private void printSelector(BreakInfo info)
  { 
    Selector sel = (Selector)info.getTree();
    Vector<Object>  children = info.getChildren();
    if (sel instanceof SelectorIdentifier)
    {
      print((BreakInfo)children.elementAt(0));
      print(".");
      return;
    }
    if (sel instanceof SelectorIndex)
    {
      print((BreakInfo)children.elementAt(0));
      print("[]");
      return;
    }
    if (sel instanceof SelectorNumber)
    {
      print((BreakInfo)children.elementAt(0));
      print(".");
      return;
    }
    System.out.println("Unknown selector to pretty print: " + sel.toString());
    System.exit(-1);
  }
  
  private void printTypedIdentifier(BreakInfo info)
  {
    boolean mode = lineMode;
    lineMode = true;
    Vector<Object>  children = info.getChildren();
    print((BreakInfo)children.get(0));
    print(":");
    print((BreakInfo)children.get(1));
    lineMode = mode;
  }
  
  private void printValuedIdentifier(BreakInfo info)
  {
    boolean mode = lineMode;
    lineMode = true;
    Vector<Object>  children = info.getChildren();
    print((BreakInfo)children.get(0));
    print(":=");
    print((BreakInfo)children.get(1));
    lineMode = mode;
  }
  
  /****************************************************************************
   * Print expression.
   * @param info the expression to be printed.
   ***************************************************************************/ 
  private void printExpression(BreakInfo info)
  {
    Expression exp = (Expression)info.getTree();
    if (exp instanceof ApplicationExpression)
      printApplicationExpression(info);
    else if (exp instanceof Logical)
      printLogical(info);
    else if (exp instanceof BitLogical)
      printBitLogical(info);
    else if (exp instanceof BinaryExpression)
      printBinaryExpression(info);
    else if (exp instanceof UnaryExpression)
      printUnaryExpression(info);
    else if (exp instanceof IfThenElseExpression)
      printIfThenElseExpression(info);
    else if (exp instanceof LetExpression)
      printLetExpression(info);
    else if (exp instanceof QuantifiedExpression)
      printQuantifiedExpression(info);
    else if (exp instanceof RecordTerm)
      printRecordTerm(info);
    else if (exp instanceof TupleTerm)
      printTupleTerm(info);
    else if (exp instanceof SelectionTerm)
      printSelectionTerm(info);
    else if (exp instanceof UpdateTerm)
      printUpdateTerm(info);
    else
    {
      System.out.println("Unknown expression to break: " + exp.toString());
      System.exit(-1);
    }
  }
  
  /****************************************************************************
   * 
   * Printing the various kinds of expressions
   * 
   ***************************************************************************/
  
  private void printApplicationExpression(BreakInfo info)
  {
    boolean mode = lineMode;
    lineMode = true;
    Vector<Object>  children = info.getChildren();
    ApplicationExpression exp = (ApplicationExpression)info.getTree();
    printPriority((BreakInfo)children.elementAt(0), exp.getPriority());
    BreakInfo[] args = (BreakInfo[])children.elementAt(1);
    int n = args.length;    
    print("(");
    for (int i=0; i<n-1; i++)
    {
      print(args[i]);
      print(",");
      space();
    }
    print(args[n-1]);
    print(")");
    lineMode = mode;
  }
  
  private void printLogical(BreakInfo info)
  {
    Logical exp = (Logical)info.getTree();
    if (State.booleanAsLogical())
    {
      if (exp.getValue())
        print("LOGICALTRUE");
      else
        print("LOGICALFALSE");
    }
    else
    {
      if (exp.getValue())
        print("TRUE");
      else
        print("FALSE");
    }
  }
  
  private void printBitLogical(BreakInfo info)
  {
    BitLogical exp = (BitLogical)info.getTree();
    if (State.booleanAsLogical())
    {
      if (exp.getValue())
        print("TRUE");
      else
        print("FALSE");
    }
    else
    {
      if (exp.getValue())
        print("BITTRUE");
      else
        print("BITFALSE");
    }
  }
  
  private void printBinaryExpression(BreakInfo info)
  {
    Vector<Object>  children = info.getChildren();
    BinaryExpression exp = (BinaryExpression)info.getTree();
    int p = exp.getPriority();
    Expression first = exp.getFirst();
    Expression second = exp.getSecond();
    BreakInfo firstInfo = (BreakInfo)children.elementAt(0);
    BreakInfo secondInfo = (BreakInfo)children.elementAt(1);
    
    // expressions printed in block mode
    if (exp instanceof EquivalentFormula || exp instanceof NotEquivalentFormula)
    {
      changeLevel(true);
      if (first instanceof EquivalentFormula ||
          first instanceof NotEquivalentFormula ||
          first instanceof ImpliesFormula)
        printParens(firstInfo);
      else
        printPriority(firstInfo, p);
      changeLevel(false);
      newLine(); 
      if (exp instanceof EquivalentFormula)
        print("<=>");
      else
        print("XOR");
      changeLevel(true);
      newLine();
      if (second instanceof EquivalentFormula ||
          second instanceof NotEquivalentFormula ||
          second instanceof ImpliesFormula)
        printParens(secondInfo);
      else
        printPriority(secondInfo, p);
      changeLevel(false);
      return;
    }
    if (exp instanceof ImpliesFormula)
    {
      changeLevel(true);
      while (true)
      {
        if (first instanceof EquivalentFormula ||
            first instanceof NotEquivalentFormula ||
            first instanceof ImpliesFormula)
          printParens(firstInfo);
        else
          printPriority(firstInfo, p);
        changeLevel(false);
        newLine(); 
        print("=>");
        if (!(exp.getSecond() instanceof ImpliesFormula)) break;
        if (secondInfo.getLength()+INDENT <= width-pos) break;
        changeLevel(true);
        newLine();
        info = secondInfo;
        children = info.getChildren();
        exp = (BinaryExpression)info.getTree();
        p = exp.getPriority();
        first = exp.getFirst();
        second = exp.getSecond();
        firstInfo = (BreakInfo)children.elementAt(0);
        secondInfo = (BreakInfo)children.elementAt(1);
      }
      changeLevel(true);
      newLine();
      if (second instanceof EquivalentFormula ||
          second instanceof NotEquivalentFormula)
        printParens(secondInfo);
      else
        printPriority(secondInfo, p);
      changeLevel(false);
      return;
    }
    if (exp instanceof AndFormula)
    {
      changeLevel(true);
      while (true)
      {
        while (true)
        {
          printPriority(firstInfo, p);
          if (!(exp.getSecond() instanceof AndFormula)) break;
          Vector<Object>  children0 = secondInfo.getChildren();
          BreakInfo thirdInfo = (BreakInfo)children0.elementAt(0);
          if (5+thirdInfo.getLength()+INDENT > width-pos) break;
          print(" AND ");
          exp = (BinaryExpression)exp.getSecond();
          firstInfo = thirdInfo;
          secondInfo = (BreakInfo)children0.elementAt(1);
        }
        changeLevel(false);
        newLine();
        print("AND");
        if (!(exp.getSecond() instanceof AndFormula)) break;
        if (secondInfo.getLength()+INDENT <= width-pos) break;
        changeLevel(true);
        newLine();
        info = secondInfo;
        children = info.getChildren();
        exp = (BinaryExpression)info.getTree();
        p = exp.getPriority();
        first = exp.getFirst();
        second = exp.getSecond();
        firstInfo = (BreakInfo)children.elementAt(0);
        secondInfo = (BreakInfo)children.elementAt(1);
      }
      changeLevel(true);
      newLine();
      printPriority(secondInfo, p);
      changeLevel(false);
      return;
    }
    if (exp instanceof OrFormula)
    {
      changeLevel(true);
      while (true)
      {
        while (true)
        {
          printPriority(firstInfo, p);
          if (!(exp.getSecond() instanceof OrFormula)) break;
          Vector<Object>  children0 = secondInfo.getChildren();
          BreakInfo thirdInfo = (BreakInfo)children0.elementAt(0);
          if (4+thirdInfo.getLength()+INDENT > width-pos) break;
          print(" OR ");
          exp = (BinaryExpression)exp.getSecond();
          firstInfo = thirdInfo;
          secondInfo = (BreakInfo)children0.elementAt(1);
        }
        changeLevel(false);
        newLine();
        print("OR");
        if (!(exp.getSecond() instanceof OrFormula)) break;
        if (secondInfo.getLength()+INDENT <= width-pos) 
          break;
        changeLevel(true);
        newLine();
        info = secondInfo;
        children = info.getChildren();
        exp = (BinaryExpression)info.getTree();
        p = exp.getPriority();
        first = exp.getFirst();
        second = exp.getSecond();
        firstInfo = (BreakInfo)children.elementAt(0);
        secondInfo = (BreakInfo)children.elementAt(1);
      }
      changeLevel(true);
      newLine();
      printPriority(secondInfo, p);
      changeLevel(false);
      return;
    }
    
    // expressions printed in line mode
    boolean mode = lineMode;
    lineMode = true;
    printPriority(firstInfo, p);
    if (exp instanceof PowerTerm)
    {
      print("^");
      if (second instanceof PowerTerm)
        printParens(secondInfo);
      else
        printPriority(secondInfo, p);
      lineMode = mode;
      return;
    }
    if (exp instanceof DividesTerm)
    {
      print("/");
      if (second instanceof TimesTerm || second instanceof DividesTerm)
        printParens(secondInfo);
      else
        printPriority(secondInfo, p);
      lineMode = mode;
      return;
    }
    else if (exp instanceof EqualsFormula)
    { space(); print("="); space(); }
    else if (exp instanceof GreaterEqualFormula)
    { space(); print(">="); space(); }
    else if (exp instanceof GreaterFormula)
    { space(); print(">"); space(); }
    else if (exp instanceof LessEqualFormula)
    { space(); print("<="); space(); }
    else if (exp instanceof LessFormula)
    { space(); print("<"); space(); }
    else if (exp instanceof MinusTerm)
    {
      print("-");
      if (second instanceof PlusTerm || second instanceof MinusTerm)
        printParens(secondInfo);
      else
        printPriority(secondInfo, p);
      lineMode = mode;
      return;
    }
    else if (exp instanceof NotEqualsFormula)
    { space(); print("/="); space(); }
    else if (exp instanceof PlusTerm)
      print("+");
    else if (exp instanceof BitOrTerm)
      print("|");
    else if (exp instanceof TimesTerm)
      print("*");
    else if (exp instanceof BitAndTerm)
      print("&");
    else
    {
      System.out.println("Unknown binary exp to prettyprint: " + exp.toString());
      System.exit(-1);
    }
    printPriority(secondInfo, p);
    lineMode = mode;
  }
  
  private void printUnaryExpression(BreakInfo info)
  {
    Vector<Object>  children = info.getChildren();
    UnaryExpression exp = (UnaryExpression)info.getTree();
    boolean mode = lineMode;
    if (exp instanceof NegationTerm)
    {
      lineMode = true;
      print("-");
    }
    else if (exp instanceof NotFormula)
    { print("NOT"); space(); }
    if (exp instanceof BitNotTerm)
    {
      lineMode = true;
      print("~");
    }
    else
    {
      System.out.println("Unknown unary exp to prettyprint: " + exp.toString());
      System.exit(-1);
    }
    printPriority((BreakInfo)children.elementAt(0), exp.getPriority());
    lineMode = mode;
  }
  
  private void printIfThenElseExpression(BreakInfo info)
  {
    Vector<Object>  children = info.getChildren();
    IfThenElseExpression exp = (IfThenElseExpression)info.getTree();
    print("IF"); space();
    boolean mode = lineMode;
    lineMode = true;
    print((BreakInfo)children.elementAt(0));
    lineMode = mode;
    space(); print("THEN");
    changeLevel(true);
    newLine(); 
    print((BreakInfo)children.elementAt(1));
    changeLevel(false);
    Expression elseBranch = exp.getElseBranch();
    while (elseBranch instanceof IfThenElseExpression)
    {
      info = (BreakInfo)children.elementAt(2);
      children = info.getChildren();
      exp = (IfThenElseExpression)elseBranch;
      newLine(); print("ELSIF"); space();
      lineMode = true;
      print((BreakInfo)children.elementAt(0));
      lineMode = mode;
      space(); print("THEN");
      changeLevel(true);
      newLine(); 
      print((BreakInfo)children.elementAt(1)); 
      changeLevel(false);
      elseBranch = exp.getElseBranch();
    }
    newLine(); print("ELSE");
    changeLevel(true);
    newLine(); 
    print((BreakInfo)children.elementAt(2));
    changeLevel(false);
    newLine(); print("ENDIF");
  }
  
  private void printQuantifiedExpression(BreakInfo info)
  {
    Vector<Object>  children = info.getChildren();
    QuantifiedExpression exp = (QuantifiedExpression)info.getTree();
    if (exp instanceof ForallFormula)
      print("FORALL");
    else if (exp instanceof ExistsFormula)
      print("EXISTS");
    else if (exp instanceof LambdaTerm)
      print("LAMBDA");
    else if (exp instanceof ArrayTerm)
      print("ARRAY");
    print("(");
    boolean mode = lineMode;
    lineMode = true;
    print((BreakInfo[])children.elementAt(0), ",", false);
    lineMode = mode;
    print(")");
    print(":");
    changeLevel(true);
    newLine();
    printPriority((BreakInfo)children.elementAt(1), exp.getPriority());
    changeLevel(false);
  }
  
  private void printLetExpression(BreakInfo info)
  {
    Vector<Object>  children = info.getChildren();
    print("LET");
    changeLevel(true);
    print((BreakInfo[])children.elementAt(0), ",", true);
    changeLevel(false);
    newLine(); print("IN");
    changeLevel(true);
    newLine(); 
    print((BreakInfo)children.elementAt(1));
    changeLevel(false);
  }
  
  private void printRecordTerm(BreakInfo info)
  {
    boolean mode = lineMode;
    lineMode = true;
    Vector<Object>  children = info.getChildren();
    print("(#");
    print((BreakInfo[])children.elementAt(0), ",", false);
    print("#)");
    lineMode = mode;
  }
  
  private void printTupleTerm(BreakInfo info)
  {
    boolean mode = lineMode;
    lineMode = true;
    Vector<Object>  children = info.getChildren();
    print("(");
    print((BreakInfo[])children.elementAt(0), ",", false);
    print(")");
    lineMode = mode;
  }
  
  private void printSelectionTerm(BreakInfo info)
  {
    boolean mode = lineMode;
    lineMode = true;
    Vector<Object>  children = info.getChildren();
    SelectionTerm exp = (SelectionTerm)info.getTree();
    printPriority((BreakInfo)children.elementAt(0), exp.getPriority());
    print((BreakInfo)children.elementAt(1));
    lineMode = mode;
  }
  
  private void printUpdateTerm(BreakInfo info)
  {
    boolean mode = lineMode;
    lineMode = true;
    Vector<Object>  children = info.getChildren();
    UpdateTerm exp = (UpdateTerm)info.getTree();
    printPriority((BreakInfo)children.elementAt(0), exp.getPriority());
    print((BreakInfo[])children.elementAt(1), "", false);
    print(":=");
    printPriority((BreakInfo)children.elementAt(2), exp.getPriority());
    lineMode = mode;
  }
}
