// ----------------------------------------------------------------------------
// CVCL.java
// CVC Lite Interface
// $Id: CVCL.java,v 1.55 2015/10/05 08:31:13 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Proving.CVCL;

import java.io.*;
import java.util.*;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;
import fmrisc.ProofNavigator.Syntax.Number;

/**
 * 
 * CVC Lite interface
 *
 */
public final class CVCL implements Prover
{
  /****************************************************************************
   * 
   * Static methods
   * 
   ***************************************************************************/
  
  /****************************************************************************
   * create CVCL identifier from given name
   * @param ident an identifier
   * @return an identifier usable for CVCL
   **************************************************************************/
  public static String getExternalName(ValueDeclIdentifier ident)
  {
    // CVCL has problems with identical identifiers in nested quantifiers
    int depth = ident.getSymbol().getDepth();
    if (depth == -1)
      // untranslated names
      return ident.getName();
    else if (depth == 0)
      // global names take their externally visible name
      return ident.getName() + "__";
    else
      // local names take their unique name
      return ident.getUniqueName() + "__";
  }
  
  /***************************************************************************
   * Get original name from external CVCL name.
   * @param name a CVCL name (may be returned by getExternalName())
   * @return the original name (may be name itself)
   **************************************************************************/
  public String getOriginalName(String name)
  {
    if (version == State.ProverVersion.CVC4)
    {
      if (name.startsWith("STRING_"))
        name = "STRING" + name.substring(7);
    }
    int pos = name.indexOf("__");
    if (pos == -1) return name;
    return name.substring(0, pos);
  }
  
  private String command;     // the invokation command
  private State.ProverVersion version; // the CVC version
  private String versionString; // the version as a string
  private PrintWriter log;    // logfile (may be null) 
  private Process process;    // the CVCL instance
  private KillThread killer;  // the killer thread for the process
  private PrintWriter out;    // its output stream
  private Reader in;          // its input stream
  private Reader err;         // its error stream
  private Reader reader;      // the current reader stream
  private String error;       // error message 
  private WatchDog watchDog;  // thread guarding the CVCL instance
  private Vector<String> context = new Vector<String>(); // the context set up so far for CVCL
  private boolean generate;   // counter examples to be generated
  private boolean useMask;      // true if mask is to be used
  private CVCLFormulaMask mask; // mask used for formulas
  private String prompt;      // prompt used by the checker
  private CVC4Transformer transformer; // transformation to reduced CVC4 language

  // special constants for CVCL4 and associated identifiers
  private final static String timesNat = "_timesnat";
  private final static String timesInt = "_timesint";
  private final static String timesReal = "_timesreal";
  private final static String powerNat = "_powernat";
  private final static String powerInt = "_powerint";
  private final static String powerReal = "_powerreal";
  private ValueDeclIdentifier timesNatId;
  private ValueDeclIdentifier timesIntId;
  private ValueDeclIdentifier timesRealId;
  private ValueDeclIdentifier powerNatId;
  private ValueDeclIdentifier powerIntId;
  private ValueDeclIdentifier powerRealId;
  
  /***************************************************************************
   * Create CVCL instance by call of command
   * @param command command line string to invoke CVCL
   * @param version the CVC version
   * @param log stream to log commands to (null, if none)
   * @param useMask true iff quantified formulas are to be masked from CVCL.
   **************************************************************************/
  public CVCL(String command, State.ProverVersion version,
    PrintWriter log, boolean useMask)
  {
    this.command = command;
    this.version = version;
    switch (version)
    {
    case CVCL: 
      prompt = "CVC> ";
      versionString = "CVCL";
      break;
    case CVC3:
      prompt = "CVC> ";
      versionString = "CVC3";
      break;
    case CVC4:
      prompt = "CVC4> ";
      versionString = "CVC4";
      transformer = new CVC4Transformer();
      timesNatId = new ValueDeclIdentifier(timesNat);
      timesNatId.getSymbol().setDepth(0);
      timesIntId = new ValueDeclIdentifier(timesInt);
      timesIntId.getSymbol().setDepth(0);
      timesRealId = new ValueDeclIdentifier(timesReal);
      timesRealId.getSymbol().setDepth(0);
      powerNatId = new ValueDeclIdentifier(powerNat);
      powerNatId.getSymbol().setDepth(0);
      powerIntId = new ValueDeclIdentifier(powerInt);
      powerIntId.getSymbol().setDepth(0);
      powerRealId = new ValueDeclIdentifier(powerReal);
      powerRealId.getSymbol().setDepth(0);
      break;
    default:
      setError("unknown checker version");
      return;
    }
    this.log = log;
    this.generate = false;
    this.useMask = useMask;
    this.mask = new CVCLFormulaMask();
    boolean okay = connect();
    if (!okay)
    {
      // wait a bit and try a second time
      try { Thread.sleep(3000); } catch(InterruptedException e) { }
      okay = connect();
      if (!okay)
      {
        setError("process could not be started.");
        return;
      }
    }
    watchDog = new WatchDog();
  }
  
  /***************************************************************************
   * 
   * State access methods
   * 
   **************************************************************************/
  
  /***************************************************************************
   * true iff CVCL instance is okay
   * @return true iff CVCL instance is okay
   **************************************************************************/
  public boolean isOkay()
  {
    return error == null;
  }
  
  /****************************************************************************
   * error message, if CVCL instance is not okay
   * @return error message, if !isOkay()
   **************************************************************************/
  public String getError()
  {
    return error;
  }
  
  /***************************************************************************
   * set error message
   * @param message the error message
   **************************************************************************/
  public void setError(String message)
  {
    writeLog("error: " + message);
    error = message;
  }
  
  /***************************************************************************
   * generate counterexamples, if possible.
   * @param generate true iff counterexamples are to be generated
   **************************************************************************/
  public void generateCounterExamples(boolean generate)
  {
    this.generate = generate;
  }
  
  /***************************************************************************
   * Try hard to simplify (potentially timeconsuming).
   * @param hard true iff prover is to try hard.
   **************************************************************************/
  public void tryHard(boolean hard)
  {
    this.useMask = !hard;
  }
  
  /***************************************************************************
   * Get version of checker.
   * @return its version.
   **************************************************************************/
  public State.ProverVersion getVersion()
  {
    return version;
  }
  
  /***************************************************************************
   *
   * Synchronized methods (to be invoked by multiple threads).
   * 
   **************************************************************************/
  
  /***************************************************************************
   * write message to log file
   * @param message to be written
   **************************************************************************/
  private synchronized void writeLog(String message)
  {
    if (log == null) return;
    log.println(new Date() + ": " + message);
    log.flush();
    // for debugging
    // System.out.println(message);
    // try
    // {
    //  Runtime.getRuntime().exec("/tmp/echox + " + message + "'");
    // }
    // catch(Exception e) {}
  }
  
  /***************************************************************************
   * Watchdog thread to guard process instance.
   * Thread is automatically started on creation.
   * If watchdog is not alive, process has died.
   **************************************************************************/
  private class WatchDog extends Thread
  {
    private boolean terminated = false;
    
    public WatchDog()
    {
      start();
    }
    
    public void setTerminated()
    {
      terminated = true;
    }
    
    public void run()
    {
      try
      {
        while (true)
        {
          int code = process.waitFor();
          writeLog("process terminated with result code " + code);
          if (terminated) return;
          // give readLine() some time to see end of input stream
          try { Thread.sleep(300); } catch(InterruptedException e) { }
          boolean okay = restart();
          if (!okay) return;
        }
      }
      catch(InterruptedException e)
      {
        setError("InterruptedException: " + e.toString());
      }
    }
  }
  
  /****************************************************************************
   * create and connect to a CVCL instance
   * @return true iff connection was successful
   **************************************************************************/
  private synchronized boolean connect()
  {
    try
    {
      writeLog("executing " + command);
      process = Runtime.getRuntime().exec(command);
      if (killer != null) killer.kill();
      killer = new KillThread(process);
      if (process == null) return false;
      out = new PrintWriter(process.getOutputStream(), true);
      in = new InputStreamReader(process.getInputStream());
      reader = in;
      err = new InputStreamReader(process.getErrorStream());
      String line;
      if (version == State.ProverVersion.CVC4)
      {
        line = readLineCVC4(false);
        if (!line.startsWith("cvc4"))
        {
          setError("CVC4 does not write \"cvc4\" but \"" + line + "\"");
          return false;
        }
        line = readLineCVC4(true);
      }
      else
      {
        line = readLineCVCL();
      }
      if (line == null)
      {
        setError("Process does not create output.");
        return false;
      }
      if (!line.equals(prompt))
      {
        setError("Process does not write \"" + prompt + "\" but \"" + 
          line + "\"");
        return false;
      }
      writeLog("process started");
      if (transformer != null)
      {
        addContext(getExternalName(timesNatId)  + ": (REAL,REAL)->INT;"); ignoreAnswer();
        addContext(getExternalName(timesIntId)  + ": (REAL,REAL)->INT;"); ignoreAnswer();
        addContext(getExternalName(timesRealId) + ": (REAL,REAL)->REAL;"); ignoreAnswer();
        addContext(getExternalName(powerNatId)  + ": (REAL,REAL)->INT;"); ignoreAnswer();
        addContext(getExternalName(powerIntId)  + ": (REAL,REAL)->INT;"); ignoreAnswer();
        addContext(getExternalName(powerRealId) + ": (REAL,REAL)->REAL;"); ignoreAnswer();
      }
      return true;
    }
    catch (IOException e)
    {
      setError("IOException(connect): " + e.toString());
      return false;
    }
  }
  
  /***************************************************************************
   * Restart prover process and send context.
   * @return true if restart was successful.
   **************************************************************************/
  private synchronized boolean restart()
  {
    writeLog("restarting process");
    boolean okay = connect();
    if (!okay) return false;
    int n = context.size();
    for (int i=0; i<n; i++)
    {
      send((String)context.elementAt(i));
      ignoreAnswer();
    }
    try
    {
      int code = process.exitValue();
      writeLog("process terminated with result code " + code);
      writeLog("giving up to restart process");
      return false;
    }
    catch(IllegalThreadStateException e)
    { }
    error = null;
    writeLog("process restarted");
    return true;
  }
  
  /***************************************************************************
   * Terminate prover instance.
   * @param wait true if to wait for termination.
   **************************************************************************/
  public synchronized void terminate(boolean wait)
  { 
    // signal watchdog not to restart process but to terminate
    watchDog.setTerminated(); 
    
    // terminate kill thread as well
    killer.kill();
    
    // terminate prover
    if (process != null) process.destroy();
    setError("process has been terminated by user");
    if (wait) try { process.waitFor(); } catch(InterruptedException e) { }
  }
  
  /***************************************************************************
   * Abort current prover activity. 
   * @return true iff new prover could be successfully started.
   **************************************************************************/
  public synchronized boolean abort()
  { 
    writeLog("aborting process");
    State.setAborted(true);
    terminate(true);
    int attempts = 2;
    while (true)
    {    
      try { Thread.sleep(1000); } catch(InterruptedException e) { }
      boolean okay = restart();
      if (okay) break;
      attempts = attempts-1;
      if (attempts == 0) return false;
      writeLog("Could not restart process, try again ...");
    }
    watchDog = new WatchDog();
    return true;
  }
  
  /****************************************************************************
   * Send message to CVCL instance.
   * @param message the message to be sent to the CVCL instance.
   ***************************************************************************/
  private synchronized void send(String message)
  {
    writeLog("send: " + message);
    
    // autoflushing enabled for stream
    out.println(message);

    // CVC4 writes input to stdout
    // if (version == State.ProverVersion.CVC4) readLine();
  }
  
  /***************************************************************************
   * add command to context
   * @param command the command to be added
   *************************************************************************/
  private synchronized void addContext(String command)
  {
    context.add(command);
    send(command);
  }
  
 /***************************************************************************
  *
  * Auxiliary methods (mostly private).
  * 
  **************************************************************************/
  
  /****************************************************************************
   * return CVCL representation of declaration
   * @param decl (value or type) declaration
   * @return the CVCL representation of decl
   ***************************************************************************/
  private String toCVCL(Declaration decl)
  {
    // ASTUtil.setPrintCVCL(true);
    // String result = decl.toString();
    // ASTUtil.setPrintCVCL(false);
    // return result;
    decl = (Declaration)transform(decl);
    return CVCLPrinter.toString(version, decl);
  }
  
  /****************************************************************************
   * return CVCL representation of expression
   * @param exp an expression (formula or term)
   * @return the CVCL representation of exp
   ***************************************************************************/
  private String toCVCL(Expression exp)
  {
    exp = (Expression)transform(exp);
    if (useMask) 
    {
      mask.resetSymbols();
      State.setFormulaMask(mask);
    }
    // do not rewrite, may put guard of TCC in wrong position
    // String result = CVCLPrinter.toString(Rewriting.rewrite(exp));
    String result = CVCLPrinter.toString(version, exp);
    if (useMask)
    {
      State.setFormulaMask(null);
      Iterator<ValueSymbol> iterator = mask.getSymbols().iterator();
      while (iterator.hasNext())
      {
        ValueSymbol symbol = iterator.next();
        addDeclaration(Construct.valueDeclaration(
          symbol.getValueDeclIdentifier(), 
          symbol.getType(),
          null));
      }
    }
    return result;
  }

  /****************************************************************************
   * read as much input as possible from process
   * but until end of line at most 
   * @return the text read from the process
   **************************************************************************/
  private String readLineCVCL()
  {
    try
    {
      StringBuffer buffer = new StringBuffer();
      int ch = reader.read();  // block until at least one character available
      if (ch == -1 && reader == in)
      {
        reader = err;
        ch = reader.read();
      }
      if (ch == -1) 
      {
        setError("End of input encountered (probably process aborted, see logfile).");
        return null;
      }
      do
      { 
        if (ch == '\n') break;   // new line terminates input
        if (ch != '\r') buffer.append((char)ch);
        if (!reader.ready()) break;
        ch = reader.read();
      }
      while (ch != -1);
      String string = buffer.toString();
      writeLog("received: " + string);
      return string;
    }
    catch(IOException e)
    {      
      setError("IOException(readline): " + e.toString());
      return null;
    }
  }

  /****************************************************************************
   * read line of input which may be terminated by prompt rather than 
   * a newline character.
   * @param prompt true if a prompt may terminate the line.
   * @return the text read from the process
   **************************************************************************/
  private String readLineCVC4(boolean prompt)
  {
    try
    {
      StringBuffer buffer = new StringBuffer();
      int ch = reader.read();  // block until at least one character available
      if (ch == -1 && reader == in)
      {
        reader = err;
        ch = reader.read();
      }
      if (ch == -1) 
      {
        setError("End of input encountered (probably process aborted, see logfile).");
        return null;
      }
      int n = CVCL.this.prompt.length();
      do
      {
        if (ch == '\n') break;   // new line terminates input
        buffer.append((char)ch);
        if (prompt)
        {
          int bn = buffer.length();
          int start = bn-n;
          if (start >= 0 && CVCL.this.prompt.charAt(0) == buffer.charAt(start))
          {
            String string0 = buffer.substring(start);
            if (string0.equals(CVCL.this.prompt)) break;
          }
        }
        ch = reader.read();
      }
      while (ch != -1);
      String string = buffer.toString();
      if (string.contains(" \r"))
        string = string.replace(" \r", ""); // CVC4 printing artifact
      writeLog("received: " + string);
      return string;
    }
    catch(IOException e)
    {      
      setError("IOException(readline): " + e.toString());
      return null;
    }
  }
  
 
  
  /****************************************************************************
   * get message from CVCL instance
   * @return message the message received from the CVCL instance
   *         (null, if an error occured)
   ***************************************************************************/
  private String receiveCVCL()
  {
    StringBuffer result = new StringBuffer();
    boolean empty = true;
    while (true)
    {
      // if (!watchDog.isAlive()) return watchDog.getStdError();
      String line = readLineCVCL();
      if (line == null) 
      {
        if (empty)
          return null;
        else
          return result.toString();
      }
      if (line.equals(prompt)) return result.toString();
      if (!empty) result.append('\n');
      empty = false;
      result.append(line);
    }
  }
  
  /****************************************************************************
   * get message from CVCL instance
   * @return message the message received from the CVCL instance
   *         (null, if an error occured)
   ***************************************************************************/
  private String receiveCVC4()
  {
    StringBuffer result = new StringBuffer();
    boolean empty = true;
    while (true)
    {
      // if (!watchDog.isAlive()) return watchDog.getStdError();
      String line = readLineCVC4(true);
      if (line == null) 
      {
        if (empty)
          return null;
        else 
          break;
      }
      if (line.equals(prompt)) break;
      if (line.startsWith(prompt)) continue;
      if (!empty) result.append('\n');
      empty = false;
      result.append(line);
    }
    String result0 = result.toString();
    try
    {
      // just keep last line of string since CVC4 may echo input
      BufferedReader reader = 
          new BufferedReader(new StringReader(result0));
      for (String line = reader.readLine(); 
          line != null; line = reader.readLine()) {
        result0 = line;
      }
      reader.close();
    }
    catch(IOException e) { }
    return result0;
  }
  
  /****************************************************************************
   * get message from CVC instance
   * @return message the message received from the CVCL instance
   *         (null, if an error occured)
   ***************************************************************************/
  private String receive()
  {
    if (version == State.ProverVersion.CVC4)
      return receiveCVC4();
    else
      return receiveCVCL();
  }
  
  //
  // A thread to kill a given process after some time
  //
  private class KillThread extends Thread
  {
    private Process process;
    private boolean doit;
    private int counter;
    private boolean killed;
    
    public KillThread(Process process)
    {
      if (process == null) return;
      this.process = process;
      this.doit = false;
      this.counter = 0;
      this.killed = false;
      start();
    }
    
    public void set(boolean doit)
    {
      this.doit = doit;
      if (doit) this.counter = this.counter+1;
    }
    
    private void kill()
    {
      this.killed = true;
    }
    
    public void run()
    {
      while (true)
      {             
        // wait for kill request
        while (true)
        {
          if (killed) return;
          if (doit) break;
          try { Thread.sleep(1000); } catch(InterruptedException e) {}
        }

        // remember request counter
        int c = counter;
        
        // wait 3 seconds for reset
        for (int i=0; i<3; i++)
        {
          if (killed) return;
          if (counter != c || !doit) break;
          try { Thread.sleep(1000); } catch(InterruptedException e) {}
        }
        
        // check whether request is still valid
        if (counter != c || !doit) continue;
        
        // finally kill it
        writeLog("process is terminated after timeout");
        watchDog.setTerminated();
        process.destroy();
        
        // unfortunately process.destroy() is not reliable
        try 
        {
          // give process time to terminate
          try { Thread.sleep(300); } catch(InterruptedException e) { }
          // if exception is thrown, process is not yet terminated
          process.exitValue();
        }
        catch(IllegalThreadStateException e)
        {
          try 
          {
            // brute force termination of all cvc* processes
            Runtime.getRuntime().exec(new String[]{"killall","-r","-9","cvc*"});
          }
          catch(IOException e0) { }
        }
        
        // now we are done
        process = null;
        return;
      }
    }
  }
  
  /***************************************************************************
   * Receive counterexample from CVCL interface.
   * This may fail depending on the theory used (which lets process terminate).
   **************************************************************************/
  private String receiveCounterExample()
  {
    // disable watchDog
    // WatchDog w = watchDog;
    // watchDog = null;
    
    // ask for counterexample
    send("COUNTEREXAMPLE;");
    String counterExample = receive();
    
    // give watchdog a chance to terminate (race condition!)
    // try { Thread.sleep(1); } catch(InterruptedException e) { }
    
    // counterexample generation failed, restart process and watchdog
    // if (w.isTerminated())
    // {
    //  boolean okay = restart();
    //  if (okay) watchDog = new WatchDog();
    //  return null;
    // }
    
    // counterexample generation successful, enable watchdog again
    // watchDog = w;
    
    return counterExample;
  }
  
  /***************************************************************************
   * receive answer from CVCL interface
   **************************************************************************/
  private Answer receiveAnswer()
  {
    String message = receive();
    if (message == null) 
      return null;
    message = message.toLowerCase(); // CVC4 has switched syntax
    if (message.startsWith("valid"))
      return new ValidAnswer(versionString);
    if (generate && message.startsWith("invalid"))
      return new InvalidAnswerCVCL(this, receiveCounterExample());
    return new UnknownAnswer(message);
  }
  
  /***************************************************************************
   * ignore answer from CVCL interface
   **************************************************************************/
  private void ignoreAnswer()
  {
    // for debugging
    // String result = 
    receive();
    // if (result != null && (result.startsWith("Parse")))
    //  System.out.println("Parse error: " + result);
  }
  
  /***************************************************************************
   * Parse CVCL String as an expression.
   * @param string the CVCL representation of an expression.
   * @return the corresponding expression (null, if an error occured)
   **************************************************************************/
  public Expression parseExpression(String string)
  {
    try
    {
      CVCLLexer lexer = new CVCLLexer(this, new StringReader(string));
      CVCLParser parser = new CVCLParser(this, lexer);
      if (useMask) State.setFormulaMask(mask);
      AST result = parser.main();
      if (useMask) State.setFormulaMask(null);
      if (!(result instanceof Expression)) 
      {
        State.setError("CVCL returned \"" + string +
          "\" which could not be parsed as an expression");
        return null;
      }
      return (Expression)result;
    }
    catch (antlr.ANTLRException e) 
    {
      setError("ANTLR exception: " + e);
      return null;
    }
  }
  
  /***************************************************************************
   * Parse CVCL String as a declaration.
   * @param string the CVCL representation of a declaration.
   * @return the corresponding declaration (null, if an error occured)
   **************************************************************************/
  public Declaration parseDeclaration(String string)
  {
    try
    {
      CVCLLexer lexer = new CVCLLexer(this, new StringReader(string));
      CVCLParser parser = new CVCLParser(this, lexer);
      if (useMask) State.setFormulaMask(mask);
      AST result = parser.declaration();
      if (useMask) State.setFormulaMask(null);
      if (!(result instanceof Declaration)) 
      {
        State.setError("CVCL returned \"" + string +
          "\" which could not be parsed as a declaration");
        return null;
      }
      return (Declaration)result;
    }
    catch (antlr.ANTLRException e) 
    {
      setError("ANTLR exception: " + e);
      return null;
    }
  }
  
  /***************************************************************************
   * receive expression from CVLC instance
   * @return the expression (null if an error occured)
   **************************************************************************/
  private Expression receiveExpression()
  {
    String message = receive();
    if (message == null) return null;
    return parseExpression(message);
  }
  
 /***************************************************************************
  *
  * Public unsychronized access methods.
  * 
  **************************************************************************/
  
  /***************************************************************************
   * Add declaration.
   * @param decl declaration that is added to prover state.
   *************************************************************************/
  public void addDeclaration(Declaration decl)
  {
    addContext(toCVCL(decl) + ";");
    ignoreAnswer();
    if (decl instanceof ValueDeclaration)
    {
      ValueDeclaration vdecl = (ValueDeclaration)decl;
      TypeExpression te = 
        vdecl.getValueDeclIdentifier().getSymbol().getClearType();
      if (te != null)
      {
        Expression formula = te.getExpression();
        if (formula != null) assertAxiom(formula);
      }
    }
  }
  
  /***************************************************************************
   * Start a new subcontext for formula assertions.
   *************************************************************************/
  public void pushContext()
  {
    addContext("PUSH;");
    ignoreAnswer();
  }
  
  /***************************************************************************
   * End subcontext for formula assertions. All assertions since the last
   * pushContext() are undone.
   *************************************************************************/
  public void popContext()
  {
    addContext("POP;");
    ignoreAnswer();
  }
  
  /***************************************************************************
   * assert axiom about previously asserted declarations.
   * @param axiom formula whose truth is asserted;
   *        declaration symbols must be only used in context of axiom.
   *************************************************************************/
  public void assertAxiom(Expression axiom)
  {
    // expose CVCL to unmasked formula
    // boolean use = useMask;
    // useMask = false;
    assertFormula(axiom);
    // useMask = use;
  }
  
  /***************************************************************************
   * Assert truth of formula.
   * @param formula formula whose truth is asserted.
   *************************************************************************/
  public void assertFormula(Expression formula)
  {
    addContext("ASSERT " + toCVCL(formula) + ";");
    ignoreAnswer();
  }
  
  /***************************************************************************
   * Query truth of formula.
   * @param formula formula whose truth is queried.
   * @return the prover answer (null, if a problem occured).
   *************************************************************************/
  public Answer queryFormula(Expression formula)
  {
    send("PUSH;");
    ignoreAnswer();
    send("QUERY " + toCVCL(formula) + ";");
    killer.set(true);;
    Answer answer=receiveAnswer();
    killer.set(false);
    send("POP;");
    ignoreAnswer();
    return answer;
  }

  /***************************************************************************
   * Simplify expression.
   * @param exp expression to be simplified.
   * @return the simplified expression (null, if a problem occured).
   *************************************************************************/
  public Expression simplifyExpression(Expression exp)
  {
    String exp0 = toCVCL(exp);
    send("TRANSFORM " + exp0 + ";");
    Expression result = receiveExpression();
    if (result == null) return Rewriting.rewrite(exp);
    Checking.setWeakerTypes(true);
    Type type = Checking.processTermOrFormula(result);
    Checking.setWeakerTypes(false);
    if (type == null) 
    {
      // do not give warnings because type checking may simply fail because 
      // exp0 contains weaker type declarations than exp
      // State.getSystemOut().println("Warning: CVCL returns mistyped simplification result.");
      // State.getSystemOut().println("Type checking message: " + State.getError());
      // State.resetError();
      // State.getSystemOut().println("CVCL result is ignored.");
      State.resetError();
      result = exp;
    }
    return Rewriting.rewrite(result);
  }
  
  /****************************************************************************
   * 
   * Transformer to reduced CVC4 language:
   * 
   * 1) replace array terms by constants collecting constant declarations 
   *    and associated axioms
   *    (the type checker has already annotated every array term with
   *     the necessary information)
   *     
   * 2) transform updates with multiple selectors to nested updates
   *    e WITH .x.y.z := v ---> 
   *    e WITH .x := (e.x WITH .y := (e.x.y WITH .z := v)
   *  
   ***************************************************************************/
  
  /****************************************************************************
   * Transform abstract syntax tree and possibly emit context.
   * @param tree the tree to be transformed.
   * @return the transformed version of the tree.
   ***************************************************************************/
  private AST transform(AST tree)
  {
    if (transformer == null) return tree;
    CVC4Transformer transformer0 = transformer;
    transformer = null;
    AST tree0 = transformer0.transform(tree);
    for (ValueDeclaration d : transformer0.getDeclarations())
      addDeclaration(d);
    for (AxiomDeclaration d : transformer0.getAxioms())
    {
      // speedup CVC4 (quantified axioms are not of much use anyway)
      // boolean use = useMask;
      // useMask = false;
      assertAxiom(d.getFormula());
      // useMask = use;
    }
    transformer = transformer0;
    return tree0;
  }
  
  /***************************************************************************
   * The auxiliary class for performing the transformation.
   **************************************************************************/
  private final class CVC4Transformer extends ASTVisitorBase
  {
    // the declarations generated for the constants
    private Vector<ValueDeclaration> decls = new Vector<ValueDeclaration>();
   
    // the axioms generated for the constants
    private Vector<AxiomDeclaration> axioms = new Vector<AxiomDeclaration>();

    // the maps of replacement names to symbols
    // and of symbols to original and new ASTs
    private Map<String, Symbol> symbols = new HashMap<String, Symbol>();
    private Map<Symbol, ArrayTerm> original = new HashMap<Symbol, ArrayTerm>();
    private Map<Symbol, Expression> replaced = new HashMap<Symbol, Expression>();
    
    /**************************************************************************
     * Replace array terms by constants and add axiom.
     * @param tree the term in which the replacements are to be performed.
     * @return the replacement term
     *************************************************************************/
    public AST transform(AST tree)
    {
      return tree.accept(this);
    }
    
    /**************************************************************************
     * Get declarations of constants generated by replacements.
     * @return the declarations 
     *************************************************************************/
    public Vector<ValueDeclaration> getDeclarations()
    {
      return decls;
    }
    
    /**************************************************************************
     * Get declarations of axioms generated by replacements.
     * @return the declarations
     *************************************************************************/
    public Vector<AxiomDeclaration> getAxioms()
    {
      return axioms;
    }
    
    /**************************************************************************
     * Get original value of reference.
     * @param exp the reference
     * @return its replacement (or the original, if none).
     *************************************************************************/
    public Expression untransformReference(Reference exp)
    {
      String name = exp.getName();
      ValueSymbol symbol = (ValueSymbol)symbols.get(name);
      if (symbol == null) return exp;
      Type type = symbol.getType();
      if (type instanceof FunctionType) return exp;
      // must prevent sharing of result expression
      return (Expression)original.get(symbol).instantiate();
    }
    
    /**************************************************************************
     * Get original value of function application.
     * @param exp the application.
     * @return its replacment (or the original, if none).
     *************************************************************************/
    public Expression untransformApplication(ApplicationExpression exp)
    {
      Expression fun = exp.getFunction();
      if (!(fun instanceof Reference)) return exp;
      String name = ((Reference)fun).getName();
      ValueSymbol symbol = (ValueSymbol)symbols.get(name);
      if (symbol == null) 
      {
        boolean times = false;
        boolean power = false;
        Type type = null;
        switch (getOriginalName(name))
        {
        case timesNat: times = true; type = Checking.getNatType(); break;
        case timesInt: times = true; type = Checking.getIntType(); break;
        case timesReal: times = true; type = Checking.getRealType(); break;
        case powerNat: power = true; type = Checking.getNatType(); break;
        case powerInt: power = true; type = Checking.getIntType(); break;
        case powerReal: power = true; type = Checking.getRealType(); break;
        default: return exp;
        }
        Expression[] args = exp.getArguments();
        if (times)
        {
          TimesTerm result = Construct.timesTerm(args[0], args[1]);
          result.setType(type);
          return result;
        }
        else if (power)
        {
          PowerTerm result = Construct.powerTerm(args[0], args[1]);
          result.setType(type);
          return result;          
        }
        return exp;
      }
      ArrayTerm oexp = original.get(symbol);
      Expression rexp = (Expression)replaced.get(symbol);
      // must prevent sharing of result expression
      if (exp.toString().equals(rexp.toString())) return oexp.instantiate();
      ValueDeclIdentifier[] free = oexp.getFree();
      int n = free.length;
      TypedIdentifier[] tid = new TypedIdentifier[n];
      for (int i=0; i<n; i++)
      {
        ValueDeclIdentifier id = free[i];
        tid[i] = Construct.typedIdentifier(id, id.getSymbol().getType());
      }
      return Construct.applicationExpression(
          Construct.lambdaTerm(tid, oexp),
          exp.getArguments());
    }
    
    /**************************************************************************
     * Signal whether tree is to be cloned.
     * @return true iff tree is to be cloned, if visit returns null.
     *************************************************************************/
    public boolean isCloned()
    {
      return true;
    }
    
    /**************************************************************************
     * Replace non-linear multiplication term by function application.
     * @param tree the array term to be replaced.
     * @return the replacment.
     *************************************************************************/
    protected AST visit(TimesTerm tree) 
    {
      Expression first = tree.getFirst();
      Expression second = tree.getSecond();
      if (isConst(first) || isConst(second)) return super.visit(tree);
      Type type = tree.getType();
      ValueDeclIdentifier fun = null;
      if (Checking.isNat(type)) fun = timesNatId;
      else if (Checking.isInt(type)) fun = timesIntId;
      else if (Checking.isReal(type)) fun = timesRealId;
      if (fun == null) return super.visit(tree);
      Expression first0 = (Expression)first.accept(this);
      Expression second0 = (Expression)second.accept(this);
      return Construct.applicationExpression(
          Construct.reference(fun), new Expression[] { first0, second0 });
    }
    
    /**************************************************************************
     * returns true if expression is a numeric constant (potentially with sign)
     * @param exp the expression
     * @return true if exp is a numeric constant
     *************************************************************************/
    private boolean isConst(Expression exp)
    {
      if (exp instanceof Number) return true;
      if (exp instanceof NegationTerm)
      {
        NegationTerm exp0 = (NegationTerm)exp;
        if (exp0.getBase() instanceof Number) return true;
      }
      return false;
    }
    
    /**************************************************************************
     * Replace power term by function application.
     * @param tree the array term to be replaced.
     * @return the replacment.
     *************************************************************************/
    protected AST visit(PowerTerm tree) 
    {
      Expression first = tree.getFirst();
      Expression second = tree.getSecond();
      Type type = tree.getType();
      ValueDeclIdentifier fun = null;
      if (Checking.isNat(type)) fun = powerNatId;
      else if (Checking.isInt(type)) fun = powerIntId;
      else if (Checking.isReal(type)) fun = powerRealId;
      if (fun == null) return super.visit(tree);
      Expression first0 = (Expression)first.accept(this);
      Expression second0 = (Expression)second.accept(this);
      return Construct.applicationExpression(
          Construct.reference(fun), new Expression[] { first0, second0 });
    }
    
    /**************************************************************************
     * Replace selection of array term by application of lambda application.
     * @param term the selection term
     * @return the replacement
     *************************************************************************/
    protected AST visit(SelectionTerm term)
    {
      Expression base = term.getBase();
      if (!(base instanceof ArrayTerm)) return super.visit(term);
      ArrayTerm base0 = (ArrayTerm)base;
      TypedIdentifier[] vars = base0.getVariables();
      Expression base1 = (Expression)base0.getTerm().accept(this);
      Expression base2 = Construct.lambdaTerm(vars,base1).instantiate();
      SelectorIndex sel = (SelectorIndex)term.getSelector();
      Expression index = (Expression)sel.getIndex().accept(this);
      return Construct.applicationExpression(base2, new Expression[]{index});
    }
    
    /**************************************************************************
     * Replace array term by constant and add axiom.
     * @param tree the array term to be replaced.
     * @return the constant replacing the array term.
     *************************************************************************/
    protected AST visit(ArrayTerm tree) 
    { 
      // get declaration and axiom and replacement term
      // (the term has to be instantiated to perform substitution)
      ValueDeclaration decl = tree.getConstant();
      AxiomDeclaration axiom = tree.getAxiom();
      
      // removal in declaration and axiom (axiom is always unmasked)
      Expression result = (Expression)tree.getTerm().accept(this);
      // speedup CVC4 (axioms are not useful anyway)
      // boolean use = useMask;
      // useMask = false;
      AxiomDeclaration axiom0 = (AxiomDeclaration)axiom.accept(this);
      // useMask = use;
      
      // add declarations and axiom
      // (may be already contained by axiom replacement)
      Symbol symbol = decl.getSymbol();
      if (!original.containsKey(symbol))
      {
        decls.add(decl);
        axioms.add(axiom0);
        symbols.put(symbol.getIdentifier().getName(), symbol);
        original.put(symbol, tree);
        replaced.put(symbol, result);
      }
      
      // the replacement term
      return result;
    }
    
    // do not transform masked formula
    protected AST visit(ForallFormula tree)
    {
      if (useMask) return tree.instantiate(); // prevent sharing
      return super.visit(tree);
    }
    
    // do not transform masked formula
    protected AST visit(ExistsFormula tree)
    {
      if (useMask) return tree.instantiate(); // prevent sharing
      return super.visit(tree);
    }
    
    /**************************************************************************
     * Transform updates with multiple selectors to nested updates
     *   e WITH .x.y.z := v ---> 
     *   e WITH .x := (e.x WITH .y := (e.x.y WITH .z := v)
     * @param tree the update term to be transformed
     * @return the transformed term
     *************************************************************************/
    protected AST visit(UpdateTerm tree)
    {
      Selector[] selectors = tree.getSelectors();
      int n = selectors.length;
      if (n == 1) return null;
      Expression base0 = (Expression)tree.getBase().accept(this);
      Selector[] selectors0 = new Selector[n];
      for (int i=0; i<n; i++)
        selectors0[i] = (Selector)selectors[i].accept(this); 
      Expression update0 = (Expression)tree.getUpdate().accept(this);
      return transformUpdate(base0, selectors0, 0, update0);
    }
    
    /**************************************************************************
     * Transform update expression with multiple selectors.
     * @param base the base expression.
     * @param selectors the sequence of selectors.
     * @param start the position of the first selector to be considered.
     * @param update the update term.
     * @return the replacement term.
     *************************************************************************/
    private Expression transformUpdate(Expression base, 
      Selector[] selectors, int start, Expression update)
    {
      Selector selector = selectors[start];
      if (start == selectors.length-1) 
        return Construct.updateTerm(base, new Selector[]{selector}, update);
      SelectionTerm base0 = Construct.selectionTerm(base, selector);
      Expression update0 = transformUpdate(base0, selectors, start+1, update);
      return Construct.updateTerm(base, new Selector[]{selector}, update0);
    }
  }
  
  /***************************************************************************
   * Perform backward transformation performed by CVCLTransformer.
   * @param exp the transformed expression.
   * @return the untransformed expression
   **************************************************************************/
  public Expression untransform(Expression exp)
  {
    // not needed because masked formulas are not transformed
    return exp;
    // if (transformer == null) return exp;
    // return (Expression)CVC4Untransformer.untransform(exp, transformer);
  }
  
  /**************************************************************************
   * Get original value of reference.
   * @param exp the reference
   * @return its replacement (or the original, if none).
   *************************************************************************/
  public Expression untransformReference(Reference exp)
  {
    if (transformer == null) return exp;
    return transformer.untransformReference(exp);
  }
  
  /**************************************************************************
   * Get original value of function application.
   * @param exp the application.
   * @return its replacement (or the original, if none).
   *************************************************************************/
  public Expression untransformApplication(ApplicationExpression exp)
  {
    if (transformer == null) return exp;
    return transformer.untransformApplication(exp);
  }
  
  /***************************************************************************
   * The auxiliary class for the backward transformation.
   **************************************************************************/
  /*
  private final static class CVC4Untransformer extends ASTVisitorBase
  {
    private CVC4Transformer transformer;
    
    public static AST untransform(AST tree, CVC4Transformer transformer)
    {
      CVC4Untransformer untransformer = new CVC4Untransformer(transformer);
      return tree.accept(untransformer);
    }
    
    private CVC4Untransformer(CVC4Transformer transformer)
    {
      this.transformer = transformer;
    }
    
    public boolean isCloned()
    {
      return true;
    }
    
    public AST visit(Reference tree)
    {
      return transformer.untransformReference(tree);
    }
    
    public AST visit(ApplicationExpression tree)
    {
      return transformer.untransformApplication(tree);
    }
  }
  */
}
