// ----------------------------------------------------------------------------
// GroundExpressions.java
// Compute ground (sub)expressions.
// $Id: GroundExpressions.java,v 1.8 2012/10/23 15:57:44 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Proving;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;

import java.util.*;

/**
 * 
 * Compute ground (sub)expressions.
 *
 */
public final class GroundExpressions extends ASTVisitorBase
{
  /***************************************************************************
   * Compute collection of all ground subexpressions of an expression.
   * @param exp an expression.
   * @return a collection of Expression objects representing all ground
   * subexpressions of exp (including exp itself), i.e. all expressions 
   * that do not contain free variables.
   **************************************************************************/
  public static GroundExpressions compute(Expression exp)
  {
    GroundExpressions ground = new GroundExpressions();
    exp.accept(ground);
    return ground;
  }
  
  /***************************************************************************
   * Compute collection of ground subexpressions of a proof state.
   * @param state a proof state.
   * @return a collection of Expression objects representing all ground
   * subexpressions of the formulas (assumptions and goals) in state, i.e. all 
   * expressions that do not contain free variables.
   **************************************************************************/
  public static GroundExpressions compute(ProofState state)
  {
    GroundExpressions ground = new GroundExpressions();
    Formula[] formulas0 = state.getAssumptions();
    int n0 = formulas0.length;
    for (int i=0; i<n0; i++)
      formulas0[i].getExpression().accept(ground);
    Formula[] formulas1 = state.getGoals();
    int n1 = formulas1.length;
    for (int i=0; i<n1; i++)
      formulas1[i].getExpression().accept(ground);
    return ground;
  }
  
  // collection of (String key)->(Expression value) mappings
  // representing the set of ground terms collected so far
  // where 'key' is the textual representation of 'value'
  private MapClass<String,Expression> map = new MapClass<String,Expression>();
  
  /***************************************************************************
   * Get the collection of ground expressions collected in this set.
   * @return a Collection of Expression objects representing the 
   * expressions collected in this set (every expression occurs exactly 
   * once in the collection).
   *************************************************************************/
  public Collection<Expression> getExpressions()
  {
    return map.values();
  }
  
  /***************************************************************************
   * Check whether collection contains expression.
   * @param exp an expression.
   * @return true iff collection contains exp.
   **************************************************************************/
  public boolean has(Expression exp)
  {
    return map.containsKey(exp.toString());
  }
  
  /****************************************************************************
   * Add ground expression to this set.
   * @param exp an expression; if this expression is a ground expression,
   *        it is added to this set.
   * @return null (signalling that no subexpressions remain to be visited).
   ***************************************************************************/
  protected AST visit(Expression exp) 
  { 
    if (!Checking.validTerm(exp)) return null;
    Vector<ValueSymbol> fv = FreeVariables.compute(exp, true);
    int n = fv.size();
    for (int i=0; i<n; i++)
    {
      ValueSymbol symbol = fv.elementAt(i);
      if (symbol.getDepth() != 0) return null;
    }
    map.put(exp.toString(), exp);
    return null;
  }
  
  /****************************************************************************
   * Add ground expression to this set.
   * @param ident an identifier; if this identifier is not free,
   *        it is added to this set.
   * @return null (signalling that no subexpressions remain to be visited).
   ***************************************************************************/
  protected AST visit(Reference ident) 
  { 
    Symbol symbol = ident.getSymbol();
    if (!(symbol instanceof ValueSymbol)) return null;
    ValueSymbol symbol0 = (ValueSymbol)symbol;
    if (symbol0.getDepth() != 0) return null;
    if (!Checking.validTermType(symbol0.getType())) return null;
    map.put(ident.toString(), ident);
    return null;
  }
}
