// ----------------------------------------------------------------------------
// BoundVariables.java
// Compute the variables bound within an expression.
// $Id: BoundVariables.java,v 1.3 2012/10/15 15:16:47 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Semantics;

import fmrisc.ProofNavigator.Syntax.*;

import java.util.*;

/**
 * 
 * Compute the variables bound within an expression.
 *
 */
public final class BoundVariables extends ASTVisitorBase
{
  /***************************************************************************
   * Compute variables bound some quantifier within an expression.
   * @param exp an expression.
   * @return a vector of Strings representing the names of the variables
   *         bound by some quantifier within the expression (each string
   *         occurs once in the vector).
   **************************************************************************/
  public static Vector<String> compute(Expression exp)
  {
    BoundVariables bv = new BoundVariables();
    exp.accept(bv);
    return bv.getVariables();
  }
  
  // the vector holding the names of the bound variables computed up to now
  private Vector<String> result = new Vector<String>();
  
  /****************************************************************************
   * Get bound variables computed by the expression visitor.
   * @return a vector of String objects holding the names of the bound vars.
   ***************************************************************************/
  private Vector<String> getVariables()
  {
    return result;
  }
  
  /****************************************************************************
   * Process quantified expression adding the bound variables to the result.
   * @param exp the quantified expression to be processed.
   * @return null (signalling that subtree remains to be visited).
   ***************************************************************************/
  protected AST visit(QuantifiedExpression exp)
  {
    TypedIdentifier[] variables = exp.getVariables();
    int n = variables.length;
    for (int i=0; i<n; i++)
    {
      String name = variables[i].getIdentifier().getName();
      if (!result.contains(name)) result.add(name);
    }
    return null;
  }
  
  /****************************************************************************
   * Process let expression adding the bound variables to the result.
   * @param exp the quantified expression to be processed.
   * @return null (signalling that subtree remains to be visited).
   ***************************************************************************/
  protected AST visit(LetExpression exp)
  {
    ValueDeclaration[] decls = exp.getDeclarations();
    int n = decls.length;
    for (int i=0; i<n; i++)
    {
      String name = decls[i].getIdentifier().getName();
      if (!result.contains(name)) result.add(name);
    }
    return null;
  }
}
