// ----------------------------------------------------------------------------
// ApplicationExpression.java
// Handling of function and predicate applications.
// $Id: ApplicationExpression.java,v 1.16 2008/10/13 09:32:01 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;

import java.io.*;

/**
 * Handling of function and predicate applications.
 */
public final class ApplicationExpression extends ExpressionBase
{
  // the application components
  private Expression fun;    // the function/predicate
  private Expression[] args; // the sequence of arguments
 
  /****************************************************************************
   * construct application of fun to args
   * @param fun the function/predicate
   * @param args the sequence of arguments
   ***************************************************************************/
  public ApplicationExpression(Expression fun, Expression[] args)
  {
    this.fun = fun;
    this.args = args;
  }
  
  /***************************************************************************
   * get binding level for printing (lower numbers mean greater binding power)
   * @return the priority level
   **************************************************************************/
  public int getPriority()
  {
    return 3;
  }
  
  /***************************************************************************
   * get function/predicate
   * @return the function/predicate
   **************************************************************************/
  public Expression getFunction()
  {
    return fun;
  }
  
  /***************************************************************************
   * get arguments
   * @return the argument sequence
   **************************************************************************/
  public Expression[] getArguments()
  {
    return args;
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Expression instantiate()
  {
    Expression fun = getFunction().instantiate();
    Expression[] args = getArguments();
    int n = args.length;
    Expression[] args0 = new Expression[n];
    for (int i=0; i<n; i++)
      args0[i] = args[i].instantiate();
    return Construct.applicationExpression(fun, args0);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    fun.printPriority(getPriority(), out);
    int n = args.length;    
    out.print('(');
    for (int i=0; i<n-1; i++)
    {
      args[i].print(out);
      out.print(", ");
    }
    args[n-1].print(out);
    out.print(')');
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * Overrides ASTBase.print such that parentheses are never printed
   * around the application.
   * @param out the stream on which the text is written
   **************************************************************************/
  public void print(PrintWriter out)
  {
    printCore(out);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null).
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    AST fun0 = fun.accept(visitor);
    int n = args.length;
    Expression[] args0 = new Expression[n];
    for (int i=0; i<n; i++)
      args0[i] = (Expression)args[i].accept(visitor);
    if (!visitor.isCloned()) return null;
    return Construct.applicationExpression((Expression)fun0, args0);
  }
}
