// ----------------------------------------------------------------------------
// ArrayTerm.java
// Handling of array terms.
// $Id: ArrayTerm.java,v 1.8 2015/09/17 09:31:04 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;

/**
 * Handling of array terms.
 */
public final class ArrayTerm extends QuantifiedExpression
{  
  // needed for replacing array terms by constants
  private ValueDeclaration decl;    // the constant declaration
  private AxiomDeclaration axiom;   // the associated axiom
  private Expression term;          // the replacement term
  private ValueDeclIdentifier[] free; // the free variables
  
  /****************************************************************************
   * construct array term of variables and base term
   * @param variables the quantified variables
   * @param base the base term
   ***************************************************************************/
  public ArrayTerm(TypedIdentifier[] variables, Expression base)
  {
    super("ARRAY", variables, base);
  }
  
  /***************************************************************************
   * Construct quantified expression.
   * @param variables typed variables.
   * @param base an expression.
   * @return a quantified expression of the same kind as this expression
   *         with the denoted variables and base expression.
   **************************************************************************/
  public QuantifiedExpression construct(
    TypedIdentifier[] variables, Expression base)
  {
    return Construct.arrayTerm(variables, base);
  }
  
  /****************************************************************************
   * Set declaration of constant replacing expression
   * @param d the declaration
   ***************************************************************************/
  public void setConstant(ValueDeclaration d)
  {
    decl = d;
  }
  
  /****************************************************************************
   * Get declaration of constant replacing expression
   * @return the declaration
   ***************************************************************************/
  public ValueDeclaration getConstant()
  {
    return decl;
  }
  
  /****************************************************************************
   * Set declaration of axiom replacing expression
   * @param d the declaration
   ***************************************************************************/
  public void setAxiom(AxiomDeclaration d)
  {
    axiom = d;
  }
  
  /****************************************************************************
   * Get declaration of axiom replacing expression
   * @return the declaration
   ***************************************************************************/
  public AxiomDeclaration getAxiom()
  {
    return axiom;
  }
  
  /****************************************************************************
   * Set replacement term
   * @param t the term
   ***************************************************************************/
  public void setTerm(Expression t)
  {
    term = t;
  }
  
  /****************************************************************************
   * Get replacement term
   * @return the term
   ***************************************************************************/
  public Expression getTerm()
  {
    return term;
  }
  
  /****************************************************************************
   * Set free variables
   * @param v the variables
   ***************************************************************************/
  public void setFree(ValueDeclIdentifier[] v)
  {
    free = v;
  }
  
  /****************************************************************************
   * Get free variables
   * @return the variables
   ***************************************************************************/
  public ValueDeclIdentifier[] getFree()
  {
    return free;
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Expression instantiate()
  {
    ArrayTerm result = (ArrayTerm)super.instantiate();

    // we leave these annotations unchanged since they do not involve
    // free variables 
    result.setConstant(decl);
    result.setAxiom(axiom);
    
    // here we perform substitutions, since they involve free variables
    result.setTerm(term.instantiate());
    int n = free.length;
    ValueDeclIdentifier[] free0 = new ValueDeclIdentifier[n];
    for (int i=0; i<n; i++)
    {
      ValueDeclIdentifier id = free[i];
      ValueDeclIdentifier id0 = id.getSubstitution();
      if (id0 == null) id0 = id;
      free0[i] = id0;
    }
    result.setFree(free0);
    return result;
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null).
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    int n = variables.length;
    TypedIdentifier[] variables0 = new TypedIdentifier[n];
    for (int i=0; i<n; i++)
    {
      variables0[i] = (TypedIdentifier)variables[i].accept(visitor);
      if (variables0[i] != null)
        variables[i].setSubstitution(variables0[i]);
    }
    Expression base0 = (Expression)base.accept(visitor);
    for (int i=0; i<n; i++)
      variables[i].resetSubstitution();
    if (!visitor.isCloned()) return null; 
    ArrayTerm result0 = (ArrayTerm)construct(variables0, base0);

    // we leave these annotations unchanged since they do not involve
    // free variables 
    result0.setConstant(decl);
    result0.setAxiom(axiom);
    
    // here we perform substitutions, since they involve free variables
    result0.setTerm((Expression)term.accept(visitor));
    int n0 = free.length;
    ValueDeclIdentifier[] free0 = new ValueDeclIdentifier[n0];
    for (int i=0; i<n0; i++)
    {
      free0[i] = (ValueDeclIdentifier)free[i].accept(visitor);
    }
    result0.setFree(free0);
    return result0;
  }
}
