// ----------------------------------------------------------------------------
// FunctionType.java
// Handling of function type expressions
// $Id: FunctionType.java,v 1.18 2008/10/13 09:32:01 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;
import java.io.*;

/**
 * Handling of function type expressions
 */
public final class FunctionType extends TypeBase
{
  private Type[] domain; // the domain types
  private Type   range;  // the range type

  /****************************************************************************
   * construct function type from domain and range.
   * @param domain the sequence of types representing the domain of the type.
   * @param range the range of the type
   ***************************************************************************/
  public FunctionType(Type[] domain, Type range)
  {
    this.domain = domain;
    this.range = range;
  }
  
  /****************************************************************************
   * returns domain of type
   * @return the domain
   ***************************************************************************/
  public Type[] getDomain()
  {
    return domain;
  }
  
  /****************************************************************************
   * returns range of type
   * @return the range
   ***************************************************************************/
  public Type getRange()
  {
    return range;
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Type instantiateType()
  {
    int n = domain.length;
    Type[] domain0 = new Type[n];
    for (int i=0; i<n; i++)
      domain0[i] = domain[i].instantiateType();
    Type range0 = range.instantiateType();
    return Construct.functionType(domain0, range0);
  }
  
  /***************************************************************************
   * Prints text representation of type on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    int n = domain.length;
    if (n == 1)
    {
      if (domain[0] instanceof FunctionType ||
          domain[0] instanceof ArrayType)
        domain[0].printParens(out);
      else
        domain[0].print(out);
    }
    else
    {
      out.print("(");
      for (int i=0; i<n; i++)
      {
        domain[i].print(out);
        if (i<n-1) out.print(", ");
      }
      out.print(")");
    }
    out.print("->");
    range.print(out);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null).
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    int n = domain.length;
    Type[] domain0 = new Type[n];
    for (int i=0; i<n; i++)
      domain0[i] = (Type)domain[i].accept(visitor);
    Type range0 = (Type)range.accept(visitor);
    if (!visitor.isCloned()) return null;
    return Construct.functionType(domain0, range0);
  }
}
