// ----------------------------------------------------------------------------
// LetExpression.java
// Handling of expressions with local bindings.
// $Id: LetExpression.java,v 1.20 2010/03/23 17:08:01 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;

import fmrisc.ProofNavigator.Semantics.*;

import java.io.*;

/**
 * Handling of terms with local bindings.
 */
public final class LetExpression extends ExpressionBase
{
  private ValueDeclaration[] declarations; // the local bindings
  private Expression base;                 // the base expression

  /****************************************************************************
   * construct let expressions of declarations and base expressions
   * @param declarations the local bindings in the expressions
   * @param base the base expressions
   ***************************************************************************/
  public LetExpression(ValueDeclaration[] declarations, Expression base)
  {
    this.declarations = declarations;
    this.base = base;
  }
  
  /***************************************************************************
   * get binding level for printing (lower numbers mean greater binding power)
   * @return the priority level
   **************************************************************************/
  public int getPriority()
  {
    return 90;
  }
  
  /****************************************************************************
   * get declarations
   * @return the declaration sequence
   ***************************************************************************/
  public ValueDeclaration[] getDeclarations()
  {
    return declarations;
  }
  
  /****************************************************************************
   * get base expression
   * @return the base expression
   ***************************************************************************/
  public Expression getBase()
  {
    return base;
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Expression instantiate()
  {
    ValueDeclaration[] decls = getDeclarations();
    int n = decls.length;
    ValueDeclaration[] decls0 = new ValueDeclaration[n];
    for (int i=0; i<n; i++)
    {
      decls0[i] = decls[i].instantiate();
      decls[i].getValueDeclIdentifier().setSubstitution(
        decls0[i].getValueDeclIdentifier());
    }
    Expression base0 = getBase().instantiate();
    for (int i=0; i<n; i++)
      decls[i].getValueDeclIdentifier().resetSubstitution();
    return Construct.letExpression(decls0, base0);
  }
  
  /***************************************************************************
   * Prints text representation of tree on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    int ov = ASTUtil.getVarNumber(); 
    out.print("LET ");
    int n = declarations.length;
    for (int i=0; i<n-1; i++)
    {
      int v = ASTUtil.incVarNumber();
      ValueDeclIdentifier ident = declarations[i].getValueDeclIdentifier();
      ValueSymbol symbol = ident.getSymbol();
      if (symbol != null) symbol.setVarNumber(v);
      declarations[i].print(out);
      out.print(", ");
    }
    if (n > 0)
    {
      int v = ASTUtil.incVarNumber();
      ValueDeclIdentifier ident = declarations[n-1].getValueDeclIdentifier();
      ValueSymbol symbol = ident.getSymbol();
      if (symbol != null) symbol.setVarNumber(v);
      declarations[n-1].print(out);
    }
    out.print(" IN ");
    base.printPriority(getPriority(), out);
    ASTUtil.setVarNumber(ov);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null)
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    int n = declarations.length;
    ValueDeclaration[] declarations0 = new ValueDeclaration[n];
    for (int i=0; i<n; i++)
    {
      declarations0[i] = (ValueDeclaration)declarations[i].accept(visitor);
      if (declarations0[i] != null)
        declarations[i].getValueDeclIdentifier().setSubstitution(
          declarations0[i].getValueDeclIdentifier());
    }
    AST base0 = base.accept(visitor);
    for (int i=0; i<n; i++)
      declarations[i].getValueDeclIdentifier().resetSubstitution();
    if (!visitor.isCloned()) return null;
    return Construct.letExpression(declarations0, (Expression)base0);
  }
}
