// ----------------------------------------------------------------------------
// Reference.java
// Handling of identifiers used in value references.
// $Id: Reference.java,v 1.14 2008/10/13 09:32:01 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;

import fmrisc.ProofNavigator.Semantics.*;

import java.io.*;

/**
 * Handling of identifiers used in declarations
 */
public final class Reference extends Identifier implements Type, Expression
{
  Symbol symbol; // corresponding symbol

  /****************************************************************************
   * construct reference from name (symbol is set to null)
   * @param name the text representation of the reference
   ***************************************************************************/
  public Reference(String name)
  {
    super(name);
  }
  
  /****************************************************************************
   * construct reference from name and symbol
   * @param name the text representation of the reference
   * @param symbol the symbol associated to the reference
   ***************************************************************************/
  public Reference(String name, Symbol symbol)
  {
    super(name);
    this.symbol = symbol;
  }
   
  /****************************************************************************
   * returns symbol associated to reference
   * @return the symbol
   ***************************************************************************/
  public Symbol getSymbol()
  {
    return symbol;
  }
  
  /****************************************************************************
   * sets symbol associated to reference
   * @param symbol the symbol
   ***************************************************************************/
  public void setSymbol(Symbol symbol)
  {
    this.symbol = symbol;
  }
  
  /***************************************************************************
   * get binding level for printing (lower numbers mean greater binding power)
   * @return the priority level
   **************************************************************************/
  public int getPriority()
  {
    return 0;
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Expression instantiate()
  {
    ValueDeclIdentifier did = ((ValueSymbol)symbol).getValueDeclIdentifier();
    ValueDeclIdentifier substitution = did.getSubstitution();
    if (substitution == null)
      return this;
    else
      return Construct.reference(substitution);
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Type instantiateType()
  {
    TypeDeclIdentifier did = ((TypeSymbol)symbol).getTypeDeclIdentifier();
    TypeDeclIdentifier substitution = did.getSubstitution();
    if (substitution == null)
      return this;
    else
      return Construct.reference(substitution);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    // use the symbol table identifier, if available
    if (symbol != null)
      symbol.getIdentifier().print(out);
    else
      super.printCore(out);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * Overrides ASTBase.print such that parentheses are never printed
   * around the number.
   * @param out the stream on which the text is written
   **************************************************************************/
  public void print(PrintWriter out)
  {
    super.print(out);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null).
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this); 
    if (result != null) return result;
    if (!visitor.isCloned()) return null;
    if (symbol == null) return this;
    if (symbol instanceof ValueSymbol)
      return this.instantiate();
    else
      return this.instantiateType();
  }
}
