// ----------------------------------------------------------------------------
// TimesTerm.java
// Handling of binary products
// $Id: TimesTerm.java,v 1.13 2015/09/17 09:31:04 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;

/**
 * Handling of conjunctions
 */
public final class TimesTerm extends BinaryExpression
{
  // type of expression
  private Type type;
  
  /****************************************************************************
   * construct binary product with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   ***************************************************************************/
  public TimesTerm(Expression base1, Expression base2)
  {
    super ("*", base1, base2);
  }
  
  /****************************************************************************
   * get result type
   * @return the type
   ***************************************************************************/
  public Type getType()
  {
    return type;
  }
  
  /****************************************************************************
   * set result type
   * @param t the type
   ***************************************************************************/
  public void setType(Type t)
  {
    type = t;
  }
  
  /***************************************************************************
   * get binding level for printing (lower numbers mean greater binding power)
   * @return the priority level
   **************************************************************************/
  public int getPriority()
  {
    return 11; // slightly bigger than "/"
  }
  
  /***************************************************************************
   * Construct binary expression of the same kind as this expression.
   * @param first an expression.
   * @param second an expression.
   * @return a binary expression of the same kind as this expression
   *         with subexpressions first and second.
   **************************************************************************/
  public BinaryExpression construct(Expression first, Expression second)
  {
    TimesTerm result = Construct.timesTerm(first, second);
    result.setType(type);
    return result;
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Expression instantiate()
  {
    TimesTerm result = (TimesTerm)
        construct(getFirst().instantiate(), getSecond().instantiate());
    result.setType(type);
    return result;
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null).
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    AST first0 = getFirst().accept(visitor);
    AST second0 = getSecond().accept(visitor);
    if (!visitor.isCloned()) return null;
    TimesTerm result0 = (TimesTerm)
        construct((Expression)first0, (Expression)second0);
    result0.setType(type);
    return result0;
  }
}

